/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//----------------------------------------------------------------------------
template <class Expression>
float
 FunctionExpr<Expression> :: operator() (float* (&d),int dx, int dy, int dz) const
   {
     return static_cast<const Expression&>(*this).template eval<float>(d,dx,dy,dz);
   }
//----------------------------------------------------------------------------
template <class Expression>
double
 FunctionExpr<Expression> :: operator() (double* (&d),int dx, int dy, int dz) const
   {
     return static_cast<const Expression&>(*this).template eval<double>(d,dx,dy,dz);
   }
//----------------------------------------------------------------------------
template <class Expression>
std::complex<double>
 FunctionExpr<Expression> :: operator() ( std::complex<double>* (&d) ,int dx, int dy, int dz) const
   {
     return static_cast<const Expression&>(*this).template eval<std::complex<double> >(d,dx,dy,dz);
   }
//----------------------------------------------------------------------------
template <typename eTYPE>
template <typename DTYPE>
DTYPE
 MonomX<eTYPE>::eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
   {
     if (ddy + ddz == 0)
        {
          switch (ddx)
            {
               case 0:
                 return std::pow(data[0],exponent);
                 break;
               case 1:
                 return std::pow(data[0],exponent-1) * static_cast<DTYPE>(exponent);
                 break;
#ifdef SECOND_DERIVATIVES
               case 2:
                 return std::pow(data[0],exponent-2) * static_cast<DTYPE>(exponent*(exponent-1));
                 break;
#endif
               default:
                 assert (!"This should never happen");
                 break;
             }
        }
     return 0;
   }
//----------------------------------------------------------------------------
template <typename eTYPE>
template <typename DTYPE>
DTYPE
 MonomY<eTYPE>::eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
   {
     if (ddx + ddz == 0)
        {
          switch (ddy)
            {
               case 0:
                 return std::pow(data[1],exponent);
                 break;
               case 1:
                 return std::pow(data[1],exponent-1) * static_cast<DTYPE>(exponent);
                 break;
#ifdef SECOND_DERIVATIVES
               case 2:
                 return std::pow(data[1],exponent-2) * static_cast<DTYPE>(exponent * (exponent-1));
                 break;
#endif
               default:
                 assert (!"This should never happen");
                 break;
             }
        }
     return 0;
   }
//----------------------------------------------------------------------------
template <typename eTYPE>
template <typename DTYPE>
DTYPE
 MonomZ<eTYPE>::eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
   {
     if (ddx + ddy == 0)
        {
          switch (ddz)
            {
               case 0:
                 return std::pow(data[2],exponent);
                 break;
               case 1:
                 return std::pow(data[2],exponent-1) * static_cast<DTYPE>(exponent);
                 break;
#ifdef SECOND_DERIVATIVES
               case 2:
                 return std::pow(data[2],exponent-2) * static_cast<DTYPE>(exponent * (exponent-1));
                 break;
#endif
               default:
                 assert (!"This should never happen");
                 break;
             }
        }
     return 0;
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionDDX<A>
 d_dx (const FunctionExpr<A>& a)
   {
     return FunctionDDX<A>( static_cast<const A&>(a) );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionDDY<A>
 d_dy (const FunctionExpr<A>& a)
   {
     return FunctionDDY<A>( static_cast<const A&>(a) );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionDDZ<A>
 d_dz (const FunctionExpr<A>& a)
   {
     return FunctionDDZ<A>( static_cast<const A&>(a) );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionMinus<A>
 operator- (const FunctionExpr<A>& a)
   {
     return FunctionMinus<A>( static_cast<const A&>(a) );
   }
//----------------------------------------------------------------------------
template <class A, class B>
inline FunctionAdd<A,B>
 operator+ (const FunctionExpr<A>& a, const FunctionExpr<B>& b)
   {
     return FunctionAdd<A,B>( static_cast<const A&>(a), static_cast<const B&>(b) );
   }
//----------------------------------------------------------------------------
template <class A, class B>
inline FunctionSub<A,B>
 operator- (const FunctionExpr<A>& a, const FunctionExpr<B>& b)
   {
     return FunctionSub<A,B>( static_cast<const A&>(a), static_cast<const B&>(b) );
   }
//----------------------------------------------------------------------------
template <class A, class B>
inline FunctionMult<A,B>
 operator* (const FunctionExpr<A>& a, const FunctionExpr<B>& b)
   {
     return FunctionMult<A,B>( static_cast<const A&>(a), static_cast<const B&>(b) );
   }
//----------------------------------------------------------------------------
template <class A, class B>
inline FunctionDiv<A,B>
 operator/ (const FunctionExpr<A>& a, const FunctionExpr<B>& b)
   {
     return FunctionDiv<A,B>( static_cast<const A&>(a), static_cast<const B&>(b) );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,int>
 operator+ (const FunctionExpr<A>& a, int b)
   {
     return FunctionCAdd<A,int>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,float>
 operator+ (const FunctionExpr<A>& a, float b)
   {
     return FunctionCAdd<A,float>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,double>
 operator+ (const FunctionExpr<A>& a, double b)
   {
     return FunctionCAdd<A,double>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,std::complex<double> > 
 operator+ (const FunctionExpr<A>& a, std::complex<double> b)
   {
     return FunctionCAdd<A,std::complex<double> >( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,int>
 operator+ (int b, const FunctionExpr<A>& a)
   {
     return FunctionCAdd<A,int>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,float>
 operator+ (float b, const FunctionExpr<A>& a)
   {
     return FunctionCAdd<A,float>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,double>
 operator+ (double b, const FunctionExpr<A>& a)
   {
     return FunctionCAdd<A,double>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCAdd<A,std::complex<double> > 
 operator+ (std::complex<double> b, const FunctionExpr<A>& a)
   {
     return FunctionCAdd<A,std::complex<double> >( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECSub<A,int>
 operator- (const FunctionExpr<A>& a, int b)
   {
     return FunctionECSub<A,int>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECSub<A,float>
 operator- (const FunctionExpr<A>& a, float b)
   {
     return FunctionECSub<A,float>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECSub<A,double>
 operator- (const FunctionExpr<A>& a, double b)
   {
     return FunctionECSub<A,double>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECSub<A,std::complex<double> > 
 operator- (const FunctionExpr<A>& a, std::complex<double> b)
   {
     return FunctionECSub<A,std::complex<double> >( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCESub<int,A>
 operator- (int b, const FunctionExpr<A>& a)
   {
     return FunctionCESub<int,A>(b,static_cast<const A&>(a));
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCESub<float,A> 
 operator- (float b, const FunctionExpr<A>& a)
   {
     return FunctionCESub<float,A>(b,static_cast<const A&>(a));
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCESub<double,A>
 operator- (double b, const FunctionExpr<A>& a)
   {
     return FunctionCESub<double,A>(b,static_cast<const A&>(a));
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCESub<std::complex<double>,A> 
 operator- (std::complex<double> b, const FunctionExpr<A>& a)
   {
     return FunctionCESub<std::complex<double>,A>(b, static_cast<const A&>(a));
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,int>
 operator* (const FunctionExpr<A>& a, int b)
   {
     return FunctionCMult<A,int>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,float>
 operator* (const FunctionExpr<A>& a, float b)
   {
     return FunctionCMult<A,float>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,double>
 operator* (const FunctionExpr<A>& a, double b)
   {
     return FunctionCMult<A,double>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,std::complex<double> >
 operator* (const FunctionExpr<A>& a, std::complex<double> b)
   {
     return FunctionCMult<A,std::complex<double> >( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,int>
 operator* (int b, const FunctionExpr<A>& a)
   {
     return FunctionCMult<A,int>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,float>
 operator* (float b, const FunctionExpr<A>& a)
   {
     return FunctionCMult<A,float>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,double>
 operator* (double b, const FunctionExpr<A>& a)
   {
     return FunctionCMult<A,double>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionCMult<A,std::complex<double> >
 operator* (std::complex<double> b, const FunctionExpr<A>& a)
   {
     return FunctionCMult<A,std::complex<double> >( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECDiv<A,int>
 operator/ (const FunctionExpr<A>& a, int b)
   {
     return FunctionECDiv<A,int>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECDiv<A,float>
 operator/ (const FunctionExpr<A>& a, float b)
   {
     return FunctionECDiv<A,float>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECDiv<A,double>
 operator/ (const FunctionExpr<A>& a, double b)
   {
     return FunctionECDiv<A,double>( static_cast<const A&>(a), b );
   }
//----------------------------------------------------------------------------
template <class A>
inline FunctionECDiv<A,std::complex<double> >
 operator/ (const FunctionExpr<A>& a, std::complex<double> b)
   {
     return FunctionECDiv<A,std::complex<double> >( static_cast<const A&>(a), b );
   }
//-----------------------------------------------------------------------------
template<class A>
inline FunctionExp<A>
 Exp(const FunctionExpr<A>& a)
   {
     return FunctionExp<A>(static_cast<const A&>(a));
   }
//-----------------------------------------------------------------------------
template<class A>
inline FunctionSin<A>
 Sin(const FunctionExpr<A>& a)
   {
     return FunctionSin<A>(static_cast<const A&>(a));
   }
//-----------------------------------------------------------------------------
template<class A>
inline FunctionCos<A>
 Cos(const FunctionExpr<A>& a)
   {
     return FunctionCos<A>(static_cast<const A&>(a));
   }
//----------------------------------------------------------------------------
template <typename eTYPE>
inline MonomX<eTYPE>
 X_(eTYPE exponent)
   {
     return MonomX<eTYPE>(exponent);
   }
//----------------------------------------------------------------------------
template <typename eTYPE>
inline MonomY<eTYPE>
 Y_(eTYPE exponent)
   {
     return MonomY<eTYPE>(exponent);
   }
//----------------------------------------------------------------------------
template <typename eTYPE>
inline MonomZ<eTYPE>
 Z_(eTYPE exponent)
   {
     return MonomZ<eTYPE>(exponent);
   }
//----------------------------------------------------------------------------
inline MonomX<int>
 X_()
   {
     return MonomX<int>(1);
   }
//----------------------------------------------------------------------------
inline MonomY<int>
 Y_()
   {
     return MonomY<int>(1);
   }
//----------------------------------------------------------------------------
inline MonomZ<int>
 Z_()
   {
     return MonomZ<int>(1);
   }
//==============================================================================
   
