/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: Polynom.h,v 1.24 2006/06/28 06:18:49 jochen Exp $
//
//==============================================================================
struct Base {
  virtual float operator() (float* (&d), int dx = 0, int dy = 0, int dz = 0) const  =  0;
  virtual double operator() (double* (&d), int dx = 0, int dy = 0, int dz = 0) const  =  0;
  virtual std::complex<double> operator()(std::complex<double>* (&d), int dx = 0, int dy = 0, int dz = 0) const = 0;
  virtual ~Base(){};
};
//----------------------------------------------------------------------------
template <class Expression>
struct FunctionExpr : public Base
   {
     virtual float operator() (float* (&d), int dx = 0, int dy = 0, int dz = 0) const ;
     virtual double operator() (double* (&d), int dx = 0, int dy = 0, int dz = 0) const ;
     virtual std::complex<double> operator()(std::complex<double>* (&d), int dx = 0, int dy = 0, int dz = 0) const;
   };
//----------------------------------------------------------------------------
class Null_ : public FunctionExpr<Null_>
   {
      public:  
     template <typename DTYPE>
     DTYPE 
      eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return (DTYPE)(0);
        } 
   };
//----------------------------------------------------------------------------
template <typename eTYPE>
class MonomX : public FunctionExpr <MonomX<eTYPE> >
  {
    private:
     eTYPE exponent;
    public:
     MonomX (eTYPE exponent_ = 1) : exponent(exponent_) {}
     template <typename DTYPE>
     DTYPE 
      eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const; 
  };
//----------------------------------------------------------------------------
template <typename eTYPE>
class MonomY : public FunctionExpr <MonomY<eTYPE> >
  {
    private:
     eTYPE exponent;
    public:
     MonomY (eTYPE exponent_ = 1) : exponent(exponent_) {}
     template <typename DTYPE>
     DTYPE 
      eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const; 
  };
//----------------------------------------------------------------------------
template <typename eTYPE>
class MonomZ : public FunctionExpr <MonomZ<eTYPE> >
  {
    private:
     eTYPE exponent;
    public:
     MonomZ (eTYPE exponent_ = 1) : exponent(exponent_) {}
     template <typename DTYPE>
     DTYPE 
      eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const; 
  };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionDDX : public FunctionExpr<FunctionDDX<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionDDX(const Expr1& expr1_): expr1(expr1_) {}

      template <typename DTYPE>
      DTYPE eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx+1,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionDDY : public FunctionExpr<FunctionDDY<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionDDY(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy+1,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionDDZ : public FunctionExpr<FunctionDDZ<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionDDZ(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy,ddz+1);
        }
   };
#ifdef SECOND_DERIVATIVES
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DX2 : public FunctionExpr<FunctionD2DX2<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DX2(const Expr1& expr1_): expr1(expr1_) {}

      template <typename DTYPE>
      DTYPE eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx+2,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DXY : public FunctionExpr<FunctionD2DXY<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DXY(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx+1,ddy+1,ddz);
        }
   };
#if 0
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DYX : public FunctionExpr<FunctionD2DYX<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DYX(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx+1,ddy+1,ddz);
        }
   };
#endif
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DXZ : public FunctionExpr<FunctionD2DXZ<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DXZ(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx+1,ddy,ddz+1);
        }
   };
#if 0
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DZX : public FunctionExpr<FunctionD2DZX<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DZX(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx+1,ddy,ddz+1);
        }
   };
#endif
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DY2 : public FunctionExpr<FunctionD2DY2<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DY2(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy+2,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DYZ : public FunctionExpr<FunctionD2DYZ<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DYZ(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy+1,ddz+1);
        }
   };
#if 0
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DZY : public FunctionExpr<FunctionD2DZY<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DZY(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy+1,ddz+1);
        }
   };
#endif
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionD2DZ2 : public FunctionExpr<FunctionD2DZ2<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionD2DZ2(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy,ddz+2);
        }
   };
#endif
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionMinus : public FunctionExpr<FunctionMinus<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
      FunctionMinus(const Expr1& expr1_): expr1(expr1_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return -expr1.eval(data,ddx,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, typename TYPE>
class FunctionCAdd : public FunctionExpr<FunctionCAdd<Expr1,TYPE> >
   {
     private:
      Expr1 expr1;
      TYPE constant;
     public:
      FunctionCAdd(const Expr1& expr1_, TYPE constant_)
        : expr1(expr1_) , constant(constant_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
        {
          if ( 0 < ddx+ddy+ddz )
             {
               return expr1.eval(data,ddx,ddy,ddz);
             }
          return expr1.eval(data,ddx,ddy,ddz) + (DTYPE)(constant);
        }
   };
//----------------------------------------------------------------------------
template <typename TYPE, class Expr1>
class FunctionCESub : public FunctionExpr<FunctionCESub<TYPE,Expr1> >
   {
     private:
      Expr1 expr1;
      TYPE constant;
     public:
      FunctionCESub(TYPE constant_, const Expr1& expr1_)
        : expr1(expr1_) , constant(constant_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
        {
          if ( 0 < ddx+ddy+ddz )
             {
               return -expr1.eval(data,ddx,ddy,ddz);
             }
          return (DTYPE)(constant) - expr1.eval(data,ddx,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, typename TYPE>
class FunctionECSub : public FunctionExpr<FunctionECSub<Expr1,TYPE> >
   {
     private:
      Expr1 expr1;
      TYPE constant;
     public:
      FunctionECSub(const Expr1& expr1_, TYPE constant_)
        : expr1(expr1_) , constant(constant_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
        {
          if ( 0 < ddx+ddy+ddz )
             {
               return expr1.eval(data,ddx,ddy,ddz);
             }
          return expr1.eval(data,ddx,ddy,ddz) - (DTYPE)(constant);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, typename TYPE>
class FunctionCMult : public FunctionExpr<FunctionCMult<Expr1,TYPE> >
   {
     private: 
      Expr1 expr1;
      TYPE constant;
     public:
      FunctionCMult(const Expr1& expr1_, TYPE constant_)
        : expr1(expr1_) , constant(constant_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy,ddz) * (DTYPE)(constant);
        }
   };
//----------------------------------------------------------------------------
template <typename TYPE, class Expr1>
class FunctionCEDiv : public FunctionExpr<FunctionCEDiv<TYPE,Expr1> >
   {
     private:
      Expr1 expr1;
      TYPE constant;
     public:
      FunctionCEDiv( TYPE constant_, const Expr1& expr1_)
        : expr1(expr1_) , constant(constant_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return (DTYPE)(constant) / expr1.eval(data,ddx,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, typename TYPE>
class FunctionECDiv : public FunctionExpr<FunctionECDiv<Expr1,TYPE> >
   {
     private:
      Expr1 expr1;
      TYPE constant;
     public:
      FunctionECDiv(const Expr1& expr1_, TYPE constant_)
        : expr1(expr1_) , constant(constant_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy,ddz) / (DTYPE)(constant);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, class Expr2>
class FunctionAdd : public FunctionExpr<FunctionAdd<Expr1,Expr2> >
   {
     private:
      Expr1 expr1;
      Expr2 expr2;
     public:
      FunctionAdd(const Expr1& expr1_, const Expr2& expr2_)
        : expr1(expr1_), expr2(expr2_) {}

      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy,ddz) + expr2.eval(data,ddx,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, class Expr2>
class FunctionSub : public FunctionExpr<FunctionSub<Expr1,Expr2> >
   {
     private:
      Expr1 expr1;
      Expr2 expr2;
     public:
      FunctionSub(const Expr1& expr1_, const Expr2& expr2_)
        : expr1(expr1_), expr2(expr2_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          return expr1.eval(data,ddx,ddy,ddz) - expr2.eval(data,ddx,ddy,ddz);
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, class Expr2>
class FunctionMult : public FunctionExpr<FunctionMult<Expr1,Expr2> >
   {
     private:
      Expr1 expr1;
      Expr2 expr2;
     public:
      FunctionMult(const Expr1& expr1_, const Expr2& expr2_)
        : expr1(expr1_), expr2(expr2_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
#ifdef SECOND_DERIVATEIVES
          int dx1 = ((0 < ddx)?1:0), dy1 = (((0<ddy)&&(ddx==0))?1:0), 
              dz1 = (((0<ddz)&&(ddx==0)&&(ddy==0))?1:0);
#endif
          switch( ddx+ddy+ddz )
             {
               case 0:
                 return expr1.eval(data,ddx,ddy,ddz) * expr2.eval(data,ddx,ddy,ddz);
                 break;
               case 1:
                 return expr1.eval(data,0,0,0) * expr2.eval(data,ddx,ddy,ddz) +
                        expr2.eval(data,0,0,0) * expr1.eval(data,ddx,ddy,ddz) ;
                 break;
#ifdef SECOND_DERIVATEIVES
               case 2:
                  return expr1.eval(data,ddx,ddy,ddz) * expr2.eval(data,0,0,0) +
                         expr1.eval(data,ddx-dx1,ddy-dy1,ddz-dz1) * expr2.eval(data,dx1,dy1,dz1) +
                         expr1.eval(data,dx1,dy1,dz1) * expr2.eval(data,ddx-dx1,ddy-dy1,ddz-dz1) +
                         expr1.eval(data,0,0,0) * expr2.eval(data,ddx,ddy,ddz) ;
#endif
                 break;
               default:
                 assert (!"This case should never happen");
                 break;
             }
          return 0.;
        }
   };
//----------------------------------------------------------------------------
template <class Expr1, class Expr2>
class FunctionDiv : public FunctionExpr<FunctionDiv<Expr1,Expr2> >
   {
     private: 
      Expr1 expr1;
      Expr2 expr2;
     public:
      FunctionDiv(const Expr1& expr1_, const Expr2& expr2_)
        : expr1(expr1_), expr2(expr2_) {}
      template <typename DTYPE>
      DTYPE
       eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          switch( ddx+ddy+ddz )
             {
               case 0:
                 return expr1.eval(data,ddx,ddy,ddz) / expr2.eval(data,ddx,ddy,ddz);
                 break;
               case 1:
                 return ( expr2.eval(data,0,0,0) * expr1.eval(data,ddx,ddy,ddz) -
                          expr1.eval(data,0,0,0) * expr2.eval(data,ddx,ddy,ddz) ) /
                          std::pow(expr2.eval(data,ddx,ddy,ddz),2)  ;
                 break;
#ifdef SECOND_DERIVATEIVES
               case 2:
#endif
               default:
                 assert (!"This case should never happen");
                 break;
             }
          return 0.;
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionExp : public FunctionExpr<FunctionExp<Expr1> >
   {
     private:
      Expr1 expr1;
     public:
     FunctionExp(const Expr1& expr1_): expr1(expr1_) {}
     template <typename DTYPE>
     DTYPE
      eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
        {
          if ( ddx+ddy+ddz == 0 )
             {
               return std::exp(expr1.eval(data,ddx,ddy,ddz));
             }
          if ( ddx+ddy+ddz == 1 )
             {
               return expr1.eval(data,ddx,ddy,ddz)* std::exp(expr1.eval(data,0,0,0));
             }
          assert (!"This case should never happen");
          return 0.;
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionSin : public FunctionExpr<FunctionSin<Expr1> >
   {
     private: 
      Expr1 expr1;
     public:
     FunctionSin(const Expr1& expr1_): expr1(expr1_) {}

     template <typename DTYPE>
     DTYPE eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const
        {
          if ( ddx+ddy+ddz == 0 )
             {
               return std::sin(expr1.eval(data,ddx,ddy,ddz));
             }
          if ( ddx+ddy+ddz == 1 )
             {
               return expr1.eval(data,ddx,ddy,ddz)* std::cos(expr1.eval(data,0,0,0));
             }
          assert (!"This case should never happen");
          return 0.;
        }
   };
//----------------------------------------------------------------------------
template <class Expr1>
class FunctionCos : public FunctionExpr<FunctionCos<Expr1> >
   {
     private: 
      Expr1 expr1;
     public:
     FunctionCos(const Expr1& expr1_): expr1(expr1_) {}

     template <typename DTYPE>
     DTYPE eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz) const 
        {
          if ( ddx+ddy+ddz == 0 )
             {
               return std::cos(expr1.eval(data,ddx,ddy,ddz));
             }
          if ( ddx+ddy+ddz == 1 )
             {
               return - expr1.eval(data,ddx,ddy,ddz)* std::sin(expr1.eval(data,0,0,0));
             }
          assert (!"This case should never happen");
          return 0.;
        }
   };
//----------------------------------------------------------------------------
template <class Expr>
struct VecFuncExpr {};
//----------------------------------------------------------------------------
template <class Expr1, class Expr2>
class FunctionExprVector2D : public VecFuncExpr<FunctionExprVector2D<Expr1,Expr2> >
                              // : public FunctionExpr <FunctionExprVector2D<Expr1,Expr2> >
   {
     private: 
      Expr1 expr1;
      Expr2 expr2;
     public:
      enum {Dimension = 2};
      FunctionExprVector2D(const Expr1& expr1_, const Expr2& expr2_): expr1(expr1_), expr2(expr2_) {}

      const Expr1 getExpr1() const
         {
           return expr1;
         }
      const Expr2 getExpr2() const
         {
           return expr2;
         }
#if 0 // not needed ... 
      template <typename DTYPE>
      DTYPE eval(DTYPE*(&data), const int ddx, const int ddy, const int ddz, SpaceDirection dir) const 
        {
          switch (dir)
             {
               case dirX:
                 return expr1.eval(data,ddx,ddy,ddz);
                 break;
               case dirY:
                 return expr2.eval(data,ddx,ddy,ddz);
                 break;
               default:
                 assert (!"This case should never happen");
                 break;
             }
          return 0.;
        }
#endif
   };

//----------------------------------------------------------------------------
template <class Expr1, class Expr2, class Expr3>
class FunctionExprVector3D : public VecFuncExpr<FunctionExprVector3D<Expr1,Expr2,Expr3> >
// : public FunctionExpr <FunctionExprVector3D<Expr1,Expr2,Expr3> >
   {
     private: 
      Expr1 expr1;
      Expr2 expr2;
      Expr3 expr3;
     public:
      enum {Dimension = 3};
      FunctionExprVector3D(const Expr1& expr1_, const Expr2& expr2_, const Expr3& expr3_)
         : expr1(expr1_), expr2(expr2_), expr3(expr3_) {}

      const Expr1 getExpr1() const
         {
           return expr1;
         }
      const Expr2 getExpr2() const
         {
           return expr2;
         }
      const Expr3 getExpr3() const
         {
           return expr3;
         }
   };
//----------------------------------------------------------------------------
template <class Expr1, class Expr2>
inline FunctionExprVector2D<Expr1,Expr2> fVec(const FunctionExpr<Expr1>& expr1,
                                              const FunctionExpr<Expr2>& expr2)
  {
     return FunctionExprVector2D<Expr1,Expr2>(static_cast<const Expr1&>(expr1),
                                              static_cast<const Expr2&>(expr2));
  }
//----------------------------------------------------------------------------
template <class Expr1, class Expr2, class Expr3>
inline FunctionExprVector3D<Expr1,Expr2,Expr3> fVec(const FunctionExpr<Expr1>& expr1,
                                                    const FunctionExpr<Expr2>& expr2,
                                                    const FunctionExpr<Expr3>& expr3)
  {
     return FunctionExprVector3D<Expr1,Expr2,Expr3>(static_cast<const Expr1&>(expr1),
                                                    static_cast<const Expr2&>(expr2),
                                                    static_cast<const Expr3&>(expr3));
  }
//==============================================================================
