/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: Stencil.h,v 1.28 2006/06/30 09:27:35 jochen Exp $
//
//==============================================================================
//-----------------------------------------------------------------------------
template <class TYPE> 
struct FunctionInterfaces
   {
     typedef TYPE (*functionInterfaceOne)(TYPE x);
     typedef TYPE (*functionInterfaceTwo)(TYPE x, TYPE y);
     typedef TYPE (*functionInterfaceThree)(TYPE x, TYPE y, TYPE z);
   };

template <>
struct FunctionInterfaces< std::complex<double> >
   {
     typedef std::complex<double> (*functionInterfaceOne)(double x);
     typedef std::complex<double> (*functionInterfaceTwo)(double x, double y);
     typedef std::complex<double> (*functionInterfaceThree)(double x, double y, double z);
   };
//-----------------------------------------------------------------------------
template <class A> 
struct BasisFunctionExpr
   {
     enum{length = 1, size = 1};
     template<class Element, unsigned int iterLength> 
     static inline const typename Element::TYPE
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <int uniqueFloatId>
class Float_  : public BasisFunctionExpr<Float_<uniqueFloatId> >
   {
    private:
     static float constantValue;
    public:
     enum {vContainment = 0, wContainment =0, pContainment = 0} ;
     inline static void 
      print(std::ostream & os); 
     inline Float_(){ } 
     inline Float_(float constantValue_); 
     inline Float_<uniqueFloatId> 
      operator = (const float& constantValue_); 
     inline Float_<uniqueFloatId> 
      operator ()(float constantValue_); 
     template<class Element, unsigned int iterLength> 
     static inline const typename Element::TYPE
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <int uniqueDoubleId>
class Double_  : public BasisFunctionExpr<Double_<uniqueDoubleId> >
   {
    private:
     static double constantValue;
    public:
     enum {vContainment = 0, wContainment =0, pContainment = 0} ;
     inline static void 
      print(std::ostream & os); 
     inline Double_(){ } 
     inline Double_(double constantValue_); 
     inline Double_<uniqueDoubleId> 
      operator = (const double& constantValue_); 
     inline Double_<uniqueDoubleId> 
      operator ()(double constantValue_); 
     template<class Element, unsigned int iterLength> 
     static inline const typename Element::TYPE
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <int uniqueComplexId>
struct Complex_  : public BasisFunctionExpr<Complex_<uniqueComplexId> >
   {
     private:
      static std::complex<double> constantValue;
     public:
      enum {vContainment = 0, wContainment =0, pContainment = 0} ;
      inline static void 
       print(std::ostream & os);
      inline Complex_() {} 
      inline Complex_(const std::complex<double>& constantValue_); 
      inline Complex_<uniqueComplexId> 
       operator = (const std::complex<double>& constantValue_); 
      inline Complex_<uniqueComplexId>
       operator ()(const std::complex<double>& constantValue_); 
      template<class Element, unsigned int iterLength> 
      static inline const typename Element::TYPE&
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <BasisFunctionSet number, int functionId>
struct BasisFunction 
  : public BasisFunctionExpr<BasisFunction<number,functionId> >
    {
      enum { vContainment = ((functionId == 0 && number == BasisSet_1)?1:0), 
             wContainment = ((functionId == 1 && number == BasisSet_1)?1:0), 
             pContainment = ((functionId == 0 && number == BasisSet_2)?1:0)};
      inline static void print(std::ostream &os); 
      template<class Element, unsigned int iterLength> 
      static inline const typename Element::TYPE
       eval(const Element& element,const int (&iterator)[iterLength]); 
      template<what typ, class Element, unsigned int iterLength> 
      static inline const typename Element::TYPE
       evaluateDerivative(const Element& element,const int (&iterator)[iterLength]); 
};
typedef BasisFunction<BasisSet_1,0> Ansatz_Function;
typedef BasisFunction<BasisSet_1,1> Testing_Function;
typedef BasisFunction<BasisSet_2,0> Mixed_Function;
//-----------------------------------------------------------------------------
template <int uniqueFunctionId, class T>
struct FUNC1_ : public BasisFunctionExpr<FUNC1_<uniqueFunctionId,T> >
   {
     enum {vContainment = 0, wContainment =0, pContainment = 0};
     static typename FunctionInterfaces<T>::functionInterfaceOne func_;
     inline static void 
      print(std::ostream &os); 
     inline FUNC1_(const typename FunctionInterfaces<T>::functionInterfaceOne& f); 
     template<class Element, unsigned int iterLength> 
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <int uniqueFunctionId, class T>
struct FUNC2_ : public BasisFunctionExpr<FUNC2_<uniqueFunctionId,T> >
   {
     enum {vContainment = 0, wContainment =0, pContainment = 0};
     static typename FunctionInterfaces<T>::functionInterfaceTwo func_;
     inline static void 
      print(std::ostream &os); 
     inline FUNC2_(const typename FunctionInterfaces<T>::functionInterfaceTwo& f); 
     template<class Element, unsigned int iterLength> 
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <int uniqueFunctionId, class T>
struct FUNC3_ : public BasisFunctionExpr<FUNC3_<uniqueFunctionId,T> >
   {
     enum {vContainment = 0, wContainment =0, pContainment = 0};
     static typename FunctionInterfaces<T>::functionInterfaceThree func_;
     inline static void 
      print(std::ostream &os); 
     inline FUNC3_(const typename FunctionInterfaces<T>::functionInterfaceThree& f); 
     template<class Element, unsigned int iterLength> 
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <int ex, SpaceDirection direction>
struct BasisMonom
  : public BasisFunctionExpr<BasisMonom<ex,direction> >
    {
      enum {vContainment = 0, wContainment = 0, pContainment = 0};
      inline static void 
       print(std::ostream &os); 
      template<class Element, unsigned int iterLength> 
      static inline typename Element::TYPE 
       eval(const Element& element,const int (&iterator)[iterLength]); 
    };
//-----------------------------------------------------------------------------
namespace Interfaces 
   {
     template <int p>
     inline BasisMonom<p,dirX> x_(); 
     template <int p>
     inline BasisMonom<p,dirY> y_(); 
     template <int p>
     inline BasisMonom<p,dirZ> z_(); 
   }
//-----------------------------------------------------------------------------
template <int position>
struct NormalComponent 
  : public BasisFunctionExpr< NormalComponent<position> >
    {
      enum {vContainment = 0, wContainment =0, pContainment = 0};
      inline static void 
       print(std::ostream &os); 
      template<class Element, unsigned int iterLength>
      static inline typename Element::TYPE 
       eval(const Element& element,const int (&iterator)[iterLength]); 
    };
//-----------------------------------------------------------------------------
template <int position>
struct UnitNormalComponent 
  : public BasisFunctionExpr< UnitNormalComponent<position> >
    {
      enum {vContainment = 0, wContainment =0, pContainment = 0};
      inline static void 
       print(std::ostream &os); 
      template<class Element, unsigned int iterLength> 
      static inline typename Element::TYPE 
       eval(const Element& element,const int (&iterator)[iterLength]); 
    };
//-----------------------------------------------------------------------------
template <int uniqueVertexVectorId, class T>
struct VertexVector : public BasisFunctionExpr<VertexVector<uniqueVertexVectorId,T> >
   {
     static T data;
     enum {vContainment = 1, wContainment =0, pContainment = 0};
     inline static void print(std::ostream &os); 
     inline VertexVector(const T& d); 
     template<class Element, unsigned int iterLength> 
     static inline typename Element::TYPE 
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <class A> 
struct Conjugate : public BasisFunctionExpr<Conjugate<A> >
   {
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
    inline static void 
     print(std::ostream &os); 
    template<class Element, unsigned int iterLength>
    static inline typename Element::TYPE 
     eval(const Element& element,const int (&iterator)[iterLength]);
   };
//-----------------------------------------------------------------------------
template<class A> 
struct Min_ : public BasisFunctionExpr<Min_<A> >
   {
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline static void 
      print(std::ostream &os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template<class A, class B> 
struct Add_ : public BasisFunctionExpr<Add_<A,B> >
   {
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
    inline static void 
     print(std::ostream & os); 
    template<class Element, unsigned int iterLength>
    static inline typename Element::TYPE
     eval(const Element& element,const int (&iterator)[iterLength]);
};
//-----------------------------------------------------------------------------
template<class A, class B>
struct Sub_ : public BasisFunctionExpr<Sub_<A,B> >
   {
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     inline static void
      print(std::ostream& os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
};
//-----------------------------------------------------------------------------
template<class A, class B> 
struct Mult_: public BasisFunctionExpr<Mult_<A,B> >
   {
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     inline static void 
      print(std::ostream &os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template<class A, class B> 
struct Div_: public BasisFunctionExpr<Div_<A,B> >
   {
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     inline static void 
      print(std::ostream &os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE 
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template<class A> 
struct SQRT_: public BasisFunctionExpr<SQRT_<A> >
   {
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline static void 
      print(std::ostream &os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE 
      eval(const Element& element,const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template<int number, class A, typename eTYPE> 
struct POW_: public BasisFunctionExpr<POW_<number,A,eTYPE> >
   {
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     static eTYPE exponent;
     inline POW_ (eTYPE e);
     inline static void 
      print(std::ostream &os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template<class A> 
struct EXP_: public BasisFunctionExpr<EXP_<A> >
   {
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline static void 
      print(std::ostream &os); 
     template<class Element, unsigned int iterLength>
     static inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//-----------------------------------------------------------------------------
template <class A, what typ> 
struct Deri_ : public BasisFunctionExpr<Deri_<A,typ> >
   {
     inline static void 
      print(std::ostream &os); 
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     template<class Element, unsigned int iterLength>
     static inline const typename Element::TYPE
      eval(const Element& element, const int (&iterator)[iterLength]); 
   };
//==============================================================================

