// ------------------------------------------------------------
// main.cc
//
// ------------------------------------------------------------

#include "source/ugblock.h"


double sinus (double x){
  return sin(x);
}

double cosinus (double x){
  return cos(x);
}

double sinush (double x){
  return sinh(x);
}



template <typename DTyp>
void Gauss_Seidel(Variable<DTyp>& u, Variable<DTyp>& f, int iter,
		  Local_stiffness_matrix<DTyp>&  lsm,
		  Marker& marker) {

  for(int i=0;i<iter;++i) {
    u = u - (lsm(u) -f) / lsm.diag() | marker;
  }
}

template <typename DTyp>
void MG( MG_array<Variable<DTyp> > u, 
	 MG_array<Variable<DTyp> > f,
	 MG_array<Variable<DTyp> > r,
	 int level, int max_level, int iter,
	 MG_array<Local_stiffness_matrix<DTyp> >  lsm,
	 MG_array<Multigrid_operator > MGop,
	 Marker& marker) {
  if(level > 0) {
    // pre-smoothing
    Gauss_Seidel(u[level],f[level],iter,lsm[level],marker);

    // restriction
    r[level] = f[level] - lsm[level](u[level]) | marker;


    f[level-1] = MGop[level-1].Restrict(r[level]);

    
    level = level-1;

    u[level] = 0.0;

    // coarse-grid correction
    MG(u,f,r,level,max_level,iter,lsm,MGop,marker);

    // prolongation
    level = level+1;

    r[level] = MGop[level-1].Prolongate(u[level-1]);

    u[level] = u[level] + r[level];

    // post-smoothing
    Gauss_Seidel(u[level],f[level],iter,lsm[level],marker);
  }
  else {
    // smoothing on coarsest grid
    Gauss_Seidel(u[level],f[level],iter,lsm[level],marker);
  }
}



using namespace ::_COLSAMM_;

int main(int argc, char** argv)
{
  cout.precision(10);
  cout.setf(std::ios::fixed,std::ios::floatfield);

  std::ifstream PARAMETER;
  int iteration, i, n, m, m_max, iter_relax;
  int level;

  double normi;  

  ofstream DATEI;

  PARAMETER.open("para.dat",std::ios::in);
  PARAMETER >> n >> iteration >> iter_relax;
  PARAMETER.close();

  cout << "\n Solving Poisson's equation on a cube \n";
  cout << " with Dirichlet boundary conditions by multigrid: ";
  cout << "\n Depth of grid: " << n;
  cout << "\n Maximal number of iterations: " << iteration;
  cout << "\n Number of smoothing steps: " 
       << iter_relax << endl;     

  Hexahedron hexa(1.0, 1.0, 1.0);

  Boundary_Marker boundary(&hexa);
  Unstructured_grid_Marker inner_points(&hexa);
  inner_points.complement(&boundary);

  DATEI.open("hex.dx",std::ios :: out);
  hexa.Print_Dx(&DATEI);
  DATEI.close();

  cout << " Die Anzahl der Freiheitsgrade sind "
       << hexa.degree_of_freedom() << endl;

  //  block_grids
  ////////////////

  MG_array<Blockgrid> blockgrids(n);

  m_max = 2;
  for(i=1;i<n;++i) m_max = m_max * 2;

  m = 2; 
  for(i=0;i<n;++i) {
    blockgrids(i) = new Blockgrid(&hexa,m,m,m_max);
    m = m * 2;
  }
  m = m/2;

  cout << " Blockgrids_hexa erstellt! " << endl;


  
  //  variables
  ////////////////
  MG_array<Variable<double> > u(n);
  MG_array<Variable<double> > r(n);
  MG_array<Variable<double> > f(n);


  for(i=0;i<n;++i) {
    u(i) = new Variable<double>(blockgrids[i]);
    r(i) = new Variable<double>(blockgrids[i]);
    f(i) = new Variable<double>(blockgrids[i]);

    u[i] = 0.0;
    f[i] = 0.0;
    r[i] = 0.0;
  }

  
  Variable<double> U_exakt(blockgrids[n-1]);


  X_coordinate X(blockgrids[n-1]);
  Y_coordinate Y(blockgrids[n-1]);
  Z_coordinate Z(blockgrids[n-1]);

  Function1<double,double> Sin(sinus);
  Function1<double,double> Cos(cosinus);
  Function1<double,double> Sinh(sinush);


  //  multigrid
  ////////////////
  MG_array<Multigrid_operator> MGop(n-1);

  for(i=0;i<n-1;++i) {
    MGop(i) = new Multigrid_operator(blockgrids[i],
				     blockgrids[i+1]);
  }


  //  local stiffness matrices
  ////////////////
  MG_array<Local_stiffness_matrix<double> > laplace(n);

  for(i=0;i<n;++i)
    laplace(i) = new Local_stiffness_matrix<double>(blockgrids[i]);

  laplace[n-1].Calculate(grad(v_())*grad(w_()));

  for(i=n-2;i>=0;--i)
    laplace[i]=MGop[i].Restrict(laplace[i+1]);

  Local_stiffness_matrix<double>  Helm(blockgrids[n-1]);
  Helm.Calculate(v_()*w_());

  //  start
  ////////////////
  level = n-1;

  u[level] = 0.0;
  u[level] = 1.0 | inner_points;
  normi = product(u[level],u[level]);


  cout << "\n Solving equation! \n";

  U_exakt = Cos(X*M_PI)*Cos(Y*M_PI)*Cos(Z*M_PI);

  u[level] = U_exakt | boundary;
  u[level] = 0.0     | inner_points;

  r[level] = 0.0;
  r[level] = Cos(X*M_PI)*Cos(Y*M_PI)*Cos(Z*M_PI) 
    * 3.0 * M_PI * M_PI;                                // right hand side
  
  f[level] = Helm(r[level])    | inner_points;
  r[level] = 0.0;


  
  cout << "\n MG: " << endl;
  cout << "-------------------- " << endl;

  for(i=1;i<=iteration;++i) {
    MG(u,f,r,level,level,iter_relax,laplace,MGop,inner_points);
    r[level] = u[level]-U_exakt;
    cout << "Iteration: " << i << "  Error: Max:" 
	 << L_infty(r[level],inner_points);
    cout << ",  Error L2: " << sqrt(product(r[level],r[level])/normi) << endl;

  }



  DATEI.open("u_exakt.dx",std::ios :: out);
  U_exakt.Print_Dx(&DATEI);
  DATEI.close();

  DATEI.open("u.dx",std::ios :: out);
  u[level].Print_Dx(&DATEI);
  DATEI.close();


  cout << " Variable ausgegeben! " << endl;

}


