/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "../../../linAlgExptemp/source/linAlg.h"

#include "../../../common_source/extemp_type.h"

#ifndef EXTEMPALG_H
#define EXTEMPALG_H

template <class A> struct ExprAlg {
	inline operator const A&() const {
		return *static_cast<const A*> ( this );
	}
};

//----------------------------------------------------------------------------
template <class A, class B>
class AddAlg : public ExprAlg<AddAlg<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline AddAlg ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		void startI_S(int max_size) const {
		     a_.startI_S(max_size);
		     b_.startI_S(max_size);
		};
		void startJ(int max_size) const {
		     a_.startJ(max_size);
		     b_.startJ(max_size);
		};

		
		typedef _TypeOf2_ ( A, B ) Result;

		inline Result getValueS(int Sglobal) const { 
		    return a_.getValueS(Sglobal) + b_.getValueS(Sglobal);  
		}
		inline Result getValueI(int Sglobal) const { 
		    return a_.getValueI(Sglobal) + b_.getValueI(Sglobal);  
		}

		inline Result getValueII(int Iglobal, int Jglobal) const { 
		    return a_.getValueII(Iglobal,Jglobal) + b_.getValueII(Iglobal,Jglobal);  
		}
		inline Result getValueSI(int Is, int Jglobal) const { 
		    return a_.getValueSI(Is,Jglobal) + b_.getValueSI(Is,Jglobal);  
		}
		inline Result getValueIS(int Iglobal, int Js) const { 
		    return a_.getValueIS(Iglobal,Js) + b_.getValueIS(Iglobal,Js);  
		}
		inline Result getValueSS(int Is, int Js) const { 
		    return a_.getValueSS(Is,Js) + b_.getValueSS(Is,Js);  
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline AddAlg<A, B> operator+ ( const ExprAlg<A>& a, const ExprAlg<B>& b ) {
	return AddAlg<A, B> ( a, b );
}


//----------------------------------------------------------------------------
template <class A, typename TYPE>
class CAddAlg : public ExprAlg<CAddAlg<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline CAddAlg(const A& a, TYPE b ) : a_ ( a ), b_ ( b ) {}

		inline void startI_S(int max_size) const { 
		  a_.startI_S(max_size);   
		}		
		inline void startJ(int max_size) const { 
		  a_.startJ(max_size);   
		}		
		
		typedef _TypeOf1_ ( A, TYPE ) Result;

		inline Result getValue(int Sglobal) const { 
		     return a_.getValue(Sglobal) + b_;  
		}
		inline Result getValue(int Iglobal, int Jglobal) const { 
		     return a_.getValue(Iglobal, Jglobal) + b_;  
		}
};

//----------------------------------------------------------------------------
template <class A>
inline CAddAlg<A, double>
operator+(ExprAlg<A>& a, double b ) {
	return CAddAlg<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddAlg<A, double>
operator+(double b, ExprAlg<A>& a ) {
	return CAddAlg<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddAlg<A, double>
operator-(ExprAlg<A>& a, double b ) {
	return CAddAlg<A, double> ( a, -b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddAlg<A, std::complex<double> >
operator+(ExprAlg<A>& a, std::complex<double>  b ) {
	return CAddAlg<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddAlg<A, std::complex<double> >
operator+ ( std::complex<double>  b, ExprAlg<A>& a ) {
	return CAddAlg<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddAlg<A, std::complex<double> >
operator-(ExprAlg<A>& a, std::complex<double>  b ) {
	return CAddAlg<A, std::complex<double> > ( a, -b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class SubAlg : public ExprAlg<SubAlg<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline SubAlg ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		void startI_S(int max_size) const {
		     a_.startI_S(max_size);
		     b_.startI_S(max_size);
		};
		void startJ(int max_size) const {
		     a_.startJ(max_size);
		     b_.startJ(max_size);
		};
		
		typedef _TypeOf2_ ( A, B ) Result;

		 inline Result getValueS(int Ss) const { 
		    return a_.getValueS(Ss) - b_.getValueS(Ss);  
		 }
		 inline Result getValueI(int Sglobal) const { 
		    return a_.getValueI(Sglobal) - b_.getValueI(Sglobal);  
		 }

		 inline Result getValueII(int Iglobal, int Jglobal) const { 
		    return a_.getValueII(Iglobal, Jglobal) - b_.getValueII(Iglobal, Jglobal);  
		 }
		 inline Result getValueSI(int Iglobal, int Jglobal) const { 
		    return a_.getValueSI(Iglobal, Jglobal) - b_.getValueSI(Iglobal, Jglobal);  
		 }
		 inline Result getValueIS(int Iglobal, int Jglobal) const { 
		    return a_.getValueIS(Iglobal, Jglobal) - b_.getValueIS(Iglobal, Jglobal);  
		 }
		 inline Result getValueSS(int Iglobal, int Jglobal) const { 
		    return a_.getValueSS(Iglobal, Jglobal) - b_.getValueSS(Iglobal, Jglobal);  
		 }
};

//----------------------------------------------------------------------------
template <class A, class B>
inline SubAlg<A, B> operator- ( const ExprAlg<A>& a, const ExprAlg<B>& b ) {
	return SubAlg<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <typename TYPE, class B>
class CSubAlg : public ExprAlg<CSubAlg<TYPE, B> > {
		TYPE a_;
		const B& b_;
		
	public:
		inline CSubAlg ( TYPE a, const B& b ) : a_ ( a ), b_ ( b ) {}

		inline void startI_S(int max_size) const { 
		  b_.startI_S(max_size);   
		}		
		inline void startJ(int max_size) const { 
		  b_.startJ(max_size);   
		}		
		
		typedef _TypeOf1_ ( B, TYPE ) Result;

		 inline Result getValueS(int Sglobal) const { 
		    return a_ - b_.getValueS(Sglobal);  
		 }
		 inline Result getValueI(int Sglobal) const { 
		    return a_ - b_.getValueI(Sglobal);  
		 }

		 inline Result getValueSS(int Iglobal, int Jglobal) const { 
		    return a_ - b_.getValueSS(Iglobal, Jglobal);  
		 }
		 inline Result getValueIS(int Iglobal, int Jglobal) const { 
		    return a_ - b_.getValueIS(Iglobal, Jglobal);  
		 }
		 inline Result getValueSI(int Iglobal, int Jglobal) const { 
		    return a_ - b_.getValueSI(Iglobal, Jglobal);  
		 }
		 inline Result getValueII(int Iglobal, int Jglobal) const { 
		    return a_ - b_.getValueII(Iglobal, Jglobal);  
		 }
};

//----------------------------------------------------------------------------
template <class B>
inline CSubAlg<double, B> operator- ( double a, const ExprAlg<B>& b ) {
	return CSubAlg<double, B> ( a, b );
}

template <class B>
inline CSubAlg<std::complex<double>, B> operator- ( std::complex<double> a, const ExprAlg<B>& b ) {
	return CSubAlg<std::complex<double>, B> ( a, b );
}


//----------------------------------------------------------------------------
template <class A>

class MinusAlg : public ExprAlg<MinusAlg<A> > {
		const A& a_;
		
	public:
		inline MinusAlg ( const A& a ) : a_ ( a ) {}
		
		inline void startI_S(int max_size) const { 
		  a_.startI_S(max_size);   
		}		
		inline void startJ(int max_size) const { 
		  a_.startJ(max_size);   
		}		
				
		typedef typename A::Result Result;
		
		 inline Result getValueS(int Sglobal) const { 
		    return - a_.getValueS(Sglobal);  
		 }
		 inline Result getValueI(int Sglobal) const { 
		    return - a_.getValueI(Sglobal);  
		 }

		 inline Result getValueII(int Iglobal, int Jglobal) const { 
		    return - a_.getValueII(Iglobal, Jglobal);  
		 }
		 inline Result getValueSI(int Iglobal, int Jglobal) const { 
		    return - a_.getValueSI(Iglobal, Jglobal);  
		 }		
		 inline Result getValueIS(int Iglobal, int Jglobal) const { 
		    return - a_.getValueIS(Iglobal, Jglobal);  
		 }		
		 inline Result getValueSS(int Iglobal, int Jglobal) const { 
		    return - a_.getValueSS(Iglobal, Jglobal);  
		 }
};

//----------------------------------------------------------------------------
template <class A>
inline MinusAlg<A> operator- ( const ExprAlg<A>& a ) {
	return MinusAlg<A> ( a );
}


//----------------------------------------------------------------------------
template <class A, typename TYPE>

class TimesAlg : public ExprAlg<TimesAlg<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline TimesAlg ( const A& a, const TYPE& b ) : a_ ( a ), b_ ( b ) {}
		
		void startI_S(int max_size) const {
		     a_.startI_S(max_size);
		};
		void startJ(int max_size) const {
		     a_.startJ(max_size);
		};

		typedef _TypeOf1_ ( A, TYPE ) Result;

		inline Result getValueS(int Sglobal) const { 
		    return a_.getValueS(Sglobal) * b_;  
		}
		inline Result getValueI(int Sglobal) const { 
		    return a_.getValueI(Sglobal) * b_;  
		}

		inline Result getValueII(int Iglobal, int Jglobal) const { 
		    return a_.getValueII(Iglobal, Jglobal) * b_;  
		}
		inline Result getValueIS(int Iglobal, int Jglobal) const { 
		    return a_.getValueIS(Iglobal, Jglobal) * b_;  
		}				
		inline Result getValueSI(int Iglobal, int Jglobal) const { 
		    return a_.getValueSI(Iglobal, Jglobal) * b_;  
		}				
		inline Result getValueSS(int Iglobal, int Jglobal) const { 
		    return a_.getValueSS(Iglobal, Jglobal) * b_;  
		}
};


//----------------------------------------------------------------------------
template <class A>
inline TimesAlg<A, double>
operator* ( const ExprAlg<A>& a, double b ) {
	return TimesAlg<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesAlg<A, double>
operator* ( double b, const ExprAlg<A>& a ) {
	return TimesAlg<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesAlg<A, std::complex<double> >
operator* ( const ExprAlg<A>& a, std::complex<double> b ) {
	return TimesAlg<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesAlg<A, std::complex<double> >
operator* ( std::complex<double> b, const ExprAlg<A>& a ) {
	return TimesAlg<A, std::complex<double> > ( a, b );
}


//----------------------------------------------------------------------------
template <typename TYPE, class B>

class DivAlg : public ExprAlg<DivAlg<TYPE, B> > {
		const TYPE a_;
		const B& b_;
		
	public:
		inline DivAlg ( const TYPE& a, const B& b ) : a_ ( a ), b_ ( b ) {}

		void startI_S(int max_size) const {
		     b_.startI_S(max_size);
		};
		void startJ(int max_size) const {
		     b_.startJ(max_size);
		};

		typedef _TypeOf1_ ( B, TYPE ) Result;

		inline Result getValueS(int Sglobal) const { 
		    return a_ / b_.getValueS(Sglobal);  
		}
		inline Result getValueI(int Sglobal) const { 
		    return a_ / b_.getValueI(Sglobal);  
		}
             // keine Division fuer Matrizen
};

//----------------------------------------------------------------------------
template <class A>
inline TimesAlg<A, double> operator/ ( const ExprAlg<A>& a, double b ) {
	return TimesAlg<A, double> ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesAlg<A, std::complex<double> > operator/ ( const ExprAlg<A>& a, std::complex<double> b ) {
	return TimesAlg<A, std::complex<double> > ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class B>
inline DivAlg<double, B> operator/ ( double a, const ExprAlg<B>& b ) {
	return DivAlg<double, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class B>
inline DivAlg<std::complex<double>, B> operator/ ( std::complex<double> a, const ExprAlg<B>& b ) {
	return DivAlg<std::complex<double>, B> ( a, b );
}


//----------------------------------------------------------------------------
template <class A, class B>

class VTimesAlg : public ExprAlg<VTimesAlg<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VTimesAlg ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		void startI_S(int max_size) const {
		     a_.startI_S(max_size);
		     b_.startI_S(max_size);
		};
		void startJ(int max_size) const {
		     a_.startJ(max_size);
		     b_.startJ(max_size);
		};
		
		typedef _TypeOf2_ ( A, B ) Result;

		 inline Result getValueI(int Sglobal) const { 
		    return a_.getValueI(Sglobal) * b_.getValueI(Sglobal);  
		 }
		 inline Result getValueS(int Sglobal) const { 
		    return a_.getValueS(Sglobal) * b_.getValueS(Sglobal);  
		 }				
		 
		 inline Result getValueII(int Iglobal, int Jglobal) const { 
		    Result sum = Result(0.0);
		    int sizeS = a_.getSizeJ();
		    		    
		    for(int s=0;s<sizeS;++s) {
		       sum = sum + (a_.getValueIS(Iglobal, s) * b_.getValueSI(s, Jglobal));
// test
//Result va = a_.getValueIS(Iglobal, s);
//Result vb = b_.getValueSI(s, Jglobal);
//cout << " Iglobal: " << Iglobal << " Jglobal: " << Jglobal << " va: " << va << " vb: " << vb << endl; 
//sum = sum + va * vb;
		    }
		    return sum;  
		 }
		 inline Result getValueSI(int Is, int Jglobal) const { 
		    Result sum = 0.0;
		    int sizeS = a_.getSizeJ();
		    for(int s=0;s<sizeS;++s)
		       sum = sum + a_.getValueIS(Is, s) * b_.getValueSI(s, Jglobal); 
		    return sum;  
		 }
		 inline Result getValueIS(int Iglobal, int Js) const { 
		    Result sum = 0.0;
		    int sizeS = a_.getSizeJ();
		    for(int s=0;s<sizeS;++s)
		       sum = sum + a_.getValueIS(Iglobal, s) * b_.getValueSS(s, Js); 
		    return sum;  
		 }		
		 inline Result getValueSS(int Is, int Js) const { 
		    Result sum = 0.0;
		    int sizeS = a_.getSizeJ();
		    for(int s=0;s<sizeS;++s)
		       sum = sum + a_.getValueSS(Is, s) * b_.getValueSS(s, Js); 
		    return sum;
		 }		
};


//----------------------------------------------------------------------------
template <class A, class B>
inline VTimesAlg<A, B> operator* ( const ExprAlg<A>& a, const ExprAlg<B>& b ) {
	return VTimesAlg<A, B> ( a, b );
}


//----------------------------------------------------------------------------
template <class A, class B>

class VDivAlg : public ExprAlg<VDivAlg<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VDivAlg ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}

		void startI_S(int max_size) const {
		     a_.startI_S(max_size);
		     b_.startI_S(max_size);
		};
		void start(int max_size) const {
		     a_.startJ(max_size);
		     b_.startJ(max_size);
		};
		
		typedef _TypeOf2_ ( A, B ) Result;

		 inline Result getValueS(int Sglobal) const { 
		    return a_.getValueS(Sglobal) / b_.getValueS(Sglobal);  
		 }				
		 inline Result getValueI(int Sglobal) const { 
		    return a_.getValueI(Sglobal) / b_.getValueI(Sglobal);  
		 }
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VDivAlg<A, B> operator/ ( const ExprAlg<A>& a, const ExprAlg<B>& b ) {
	return VDivAlg<A, B> ( a, b );
}

//----------------------------------------------------------------------------

#endif // EXTEMP_H
