/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// vector.h
//
// ------------------------------------------------------------

#ifndef VECTOR_ALG_H
#define VECTOR_ALG_H


///////////////////////////
// 1. Simple Index class
///////////////////////////

class SimpleIndex {  
  public:
    SimpleIndex(int l) { 
      size = l;
      In = new int[l]; 
      for(int i=0;i<size;++i) In[i] = i;
    }
    int getSize()     const  { return size; } 
    int* getIndices() const  { return In; }

  private:
    int size;
    int* In;
};

/////////////////////////////////////////////
//  Restriction 
/////////////////////////////////////////////

template <class A>
class ExprAlg_Restriction_Index {
   public:
     template <class IndexClass>
     ExprAlg_Restriction_Index(const ExprAlg<A>& a, const IndexClass& index) 
           : a_(a), size(index.getSize()), Sn(index.getIndices()) { }

     typedef typename A::Result Result;

     Result getValueI(int Sglobal) const {  return a_.getValueI(Sglobal);  }
     void startI_S(int max_size)   const {  a_.startI_S(max_size);	  }

   public:
     const A& a_;
     // Speicher Indizes
     const int size;
     const int *Sn;	
};

template <class A, class IndexClass>
inline ExprAlg_Restriction_Index<A> operator| ( const ExprAlg<A>& a, const IndexClass& index ) {
	return ExprAlg_Restriction_Index<A>(a,index);
}

///////////////////////////
// 2. vector
///////////////////////////



template <class DTyp>
class VectorAlg : public ExprAlg<VectorAlg< DTyp > > {		
	public:			
		VectorAlg(int size_) { size = size_; data = new DTyp[size]; }
		typedef DTyp Result;
		
		template <class A>
		void operator= ( const ExprAlg<A>& a );
		void operator=(const VectorAlg<DTyp>& v);
		void operator= ( DTyp x );
		
		void startI_S(int max_size)   const { assert(size<=max_size); }
		DTyp getValueS(int Ss) const { return data[Ss]; }
		int getSize() const { return size; }
			
		void Print();	
	private:		
		// own data
		DTyp* data;     
		int size;
};





template <class DTyp>
class VectorIndex : public ExprAlg<VectorIndex< DTyp > > {		
	public:			
	        template <class IndexClass>
		VectorIndex(const IndexClass& index) {
		  size = index.getSize(); 
		  Sn   = index.getIndices(); 
		  data = new DTyp[size];
		  s    = new int;
		  Smy  = new int;
		}
		typedef DTyp Result;
		
		template <class A>
		void operator=(const ExprAlg<A>& a );
		void operator=(const VectorIndex<DTyp>& v);

		template <class A>
		void operator=(const ExprAlg_Restriction_Index<A>& algRestriction);
		void operator= ( DTyp x );
		
		void startI_S(int max_size) const  {
		     (*s) = 0; 
		     if(size>0) (*Smy) = Sn[(*s)];
		}
		
		void setValueS(int Ss, DTyp x) { assert(Ss<size); data[Ss] = x; }	
		DTyp getValueS(int Ss) const { return data[Ss]; }		
		DTyp getValueI(int Sglobal) const ; 
		int getSize() const { return size; }
		
		void Print();	
		void Test();
	private:		
		// own data
		DTyp* data;     
		
		// Speicher Indizes
		int size;
		int *Sn;
		
		// für die Auswertung
		int *Smy;
		int *s;
};



///////////////////////////
// 3. Other functions, max, ...
///////////////////////////


/*
template <class A, class B>
_TypeOf2_ ( A, B ) product ( ExprAlg<A>& a, ExprAlg<B>& b, Marker2D& marker );
*/
template <class A, class B>
_TypeOf2_ ( A, B ) product(const ExprAlg<A>& a, const ExprAlg<B>& b );



template <class A>
double L_infty ( const ExprAlg<A>& a );



template <class A>
double Maximum ( const ExprAlg<A>& a );


template <class A>
double Minimum ( const ExprAlg<A>& a );



//////////////////////////////////////////
// and it's implementation



//////////////////////////////////////////
// 4. Implementation of inline functions
//////////////////////////////////////////

		
		
		
template<class DTyp>
DTyp VectorIndex<DTyp>::getValueI(int Sglobal) const {
   while(Sglobal > (*Smy) && (*s) < size) {
      ++(*s);
      (*Smy) = Sn[(*s)];
   }
   if((*Smy) > Sglobal || (*s)>=size) return DTyp(0);
   
   return data[(*s)];
}




template <class DTyp>
template <class A>
void VectorIndex<DTyp>::operator=(const ExprAlg<A>& a) {
	const A& ao ( a );
 	
	ao.startI_S(size);
	for(int ss = 0;ss < size;++ss ) {
            data[ss] =  ao.getValueI(Sn[ss]);
	}  
}

template <class DTyp>
void VectorIndex<DTyp>::operator=(const VectorIndex<DTyp>& v) {
	v.startI_S(size);
	for(int ss = 0;ss < size;++ss ) {
            data[ss] =  v.getValueI(Sn[ss]);
	}  
}

template <class DTyp>
template <class A>
void VectorIndex<DTyp>::operator=(const ExprAlg_Restriction_Index<A>& algRestriction) {
    int sizeIndex = algRestriction.size;
    algRestriction.startI_S(sizeIndex);
    startI_S(sizeIndex);
    for(int ss = 0;ss < sizeIndex;++ss ) {
        int Sglobal = algRestriction.Sn[ss];
        
	while(Sglobal > (*Smy) && (*s) < size) {
           ++(*s);
           (*Smy) = Sn[(*s)];
        }
        if((*Smy) == Sglobal && (*s)<size) {
	   data[(*s)] = algRestriction.getValueI(Sglobal);
	}
    }
}



template <class DTyp>
void VectorIndex<DTyp>::operator=(DTyp x) {
	for(int ss = 0;ss < size;++ss ) {
            data[ss] =  x;
	}  
}

template <class DTyp>
void VectorIndex<DTyp>::Test() {
  cout << " Test: Index Vector of size: " << size << endl;
  for(int ss = 0;ss < size;++ss ) {
      cout << "i: " << ss << " index: " << Sn[ss] << endl;
  }  
}


template <class DTyp>
void VectorIndex<DTyp>::Print() {
  cout << " Index Vector of size: " << size << endl;
  for(int ss = 0;ss < size;++ss ) {
      cout << "S[" << ss << "]=" << Sn[ss] << " value: " << data[ss] << endl;
  }  
}

template <class DTyp>
void VectorAlg<DTyp>::Print() {
  cout << " Algebra Vector of size: " << size << endl;
  for(int ss = 0;ss < size;++ss ) {
      cout << "i: " << ss << " value: " << data[ss] << endl;
  }  
}


template <class DTyp>
template <class A>
void VectorAlg<DTyp>::operator=(const ExprAlg<A>& a) {
	const A& ao ( a );
	ao.startI_S(size);
	for(int ss = 0;ss < size;++ss ) {
            data[ss] =  ao.getValueS(ss);
	}  
}


template <class DTyp>
void VectorAlg<DTyp>::operator=(DTyp x) {  
	for(int ss = 0;ss < size;++ss ) {
            data[ss] =  x;
	}  
}

#endif // VARIABLE_H
