
#pragma once

// Compiler identification, COMPILER_*

/// COMPILER_GNUC set to 1 for all compilers compatible with GCC
#ifdef __GNUC__
#define COMPILER_GNUC 1
#define GNUC_AT_LEAST( x, y ) ( ( __GNUC__ == x && __GNUC_MINOR__ >= y ) || __GNUC__ > x )
#define GNUC_AT_MOST( x, y ) ( ( __GNUC__ == x && __GNUC_MINOR__ <= y ) || __GNUC__ < x )
#define GNUC_AT( x, y ) ( __GNUC__ == x && __GNUC_MINOR__ == y )
#else
#define COMPILER_GNUC 0

#define GNUC_AT_LEAST( x, y ) 0
#define GNUC_AT_MOST( x, y ) 0
#define GNUC_AT( x, y ) 0
#endif

/// COMPILER_CLANG set to major+minor version (e.g., 307 for clang 3.7) if the compiler is clang
#if defined( __clang__ )
#define COMPILER_CLANG ( __clang_major__ * 100 + __clang_minor__ )
#else
#define COMPILER_CLANG 0
#endif

/// COMPILER_LLVM set to 1 if the compiler backend is llvm
#if defined( __llvm__ )
#define COMPILER_LLVM 1
#else
#define COMPILER_LLVM 0
#endif

/// COMPILER_ICC set to __INTEL_COMPILER if the compiler is Intel compiler, 0 otherwise
#if defined( __INTEL_COMPILER )
#define COMPILER_ICC __INTEL_COMPILER
#else
#define COMPILER_ICC 0
#endif

/// COMPILER_MSVC set to _MSC_VER if the compiler is Microsoft Visual C++, 0 otherwise.
#if defined( _MSC_VER )
#define COMPILER_MSVC _MSC_VER
#else
#define COMPILER_MSVC 0
#endif

// STRONG_INLINE is a stronger version of the ALWAYS_INLINE, using __forceinline on MSVC,
// but it still doesn't use GCC's always_inline. This is useful in (common) situations where MSVC
// needs forceALWAYS_INLINE
// but GCC is still doing fine with just ALWAYS_INLINE.
#if COMPILER_MSVC || COMPILER_ICC
#define STRONG_INLINE __forceinline
#else
#define STRONG_INLINE inline
#endif

// inline is the stronget, it has the effect of making the function ALWAYS_INLINE and adding every
// possible
// attribute to maximize inlining. This should only be used when really necessary: in particular,
// it uses __attribute__((always_inline)) on GCC, which most of the time is useless and can severely
// harm compile times.
// FIXME with the always_inline attribute,
// gcc 3.4.x reports the following compilation error:
//   Eval.h:91: sorry, unimplemented: inlining failed in call to 'const Eigen::Eval<Derived>
//   Eigen::MatrixBase<Scalar, Derived>::eval() const'
//    : function body not available
#if GNUC_AT_LEAST( 4, 0 )
#define ALWAYS_INLINE __attribute__( ( always_inline ) ) inline
#else
#define ALWAYS_INLINE STRONG_INLINE
#endif

#if COMPILER_GNUC
#define DONT_INLINE __attribute__( ( noinline ) )
#elif COMPILER_MSVC
#define DONT_INLINE __declspec( noinline )
#else
#define DONT_INLINE
#endif
