/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//#include "../../../linAlgExptemp/source/linAlg.h"
#include "../../../common_source/extemp_type.h"


#ifndef EXTEMP_H
#define EXTEMP_H


template <class A> struct Expr {
	inline operator const A&() const {
		return *static_cast<const A*> ( this );
	}
};

//----------------------------------------------------------------------------
template <class A, class B>
class Add : public Expr<Add<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline Add ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			return b_.Give_stencil_typ();
		};

		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};

		int getTotalNumberData() const {  return a_.getTotalNumberData(); }
		
		control_typ Give_control_typ() const {
			if ( a_.Give_control_typ() == not_thread_save ) return not_thread_save;
			
			return b_.Give_control_typ();
		};
		
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out ) +
			       b_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params_out_sten ) +
			       b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
				
		/*
		template <elementTyp TYP_EL>
		  inline Result Give_matrix(params_in_loc_matrix) const
		  {
		    return a_.template Give_matrix<TYP_EL>(params_out_loc_matrix) +
		  b_.template Give_matrix<TYP_EL>(params_out_loc_matrix);
		  }
		*/
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) +
			       b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_.Give_cell_hexahedra ( params_out_cell ) +
			       b_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_.Give_fromTotal(i) +
			       b_.Give_fromTotal(i);
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline Add<A, B> operator+ ( const Expr<A>& a, const Expr<B>& b ) {
	return Add<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, typename TYPE>
class CAdd : public Expr<CAdd<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline CAdd ( const A& a, TYPE b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			return a_.Give_blockgrid();
		}
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			return a_.totalCalcNotPossible();
		};		

		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const {
			return a_.Give_control_typ();
		};
		
		typedef _TypeOf1_ ( A, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template  Give_data<TYP_EL> ( params_out ) + b_;
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_.Give_data_sten<TYP_EL> ( params_out_sten ) + b_;
		}
		
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) + b_;
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_.Give_cell_hexahedra ( params_out_cell ) + b_;
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_.Give_fromTotal(i) + b_;
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline CAdd<A, double>
operator+ ( const Expr<A>& a, double b ) {
	return CAdd<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd<A, double>
operator+ ( double b, const Expr<A>& a ) {
	return CAdd<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd<A, double>
operator- ( const Expr<A>& a, double b ) {
	return CAdd<A, double> ( a, -b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd<A, std::complex<double> >
operator+ ( const Expr<A>& a, std::complex<double>  b ) {
	return CAdd<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd<A, std::complex<double> >
operator+ ( std::complex<double>  b, const Expr<A>& a ) {
	return CAdd<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd<A, std::complex<double> >
operator- ( const Expr<A>& a, std::complex<double>  b ) {
	return CAdd<A, std::complex<double> > ( a, -b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class Sub : public Expr<Sub<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline Sub ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			return b_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};
		
		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const {
			if ( a_.Give_control_typ() == not_thread_save ) return not_thread_save;
			
			return b_.Give_control_typ();
		};
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out ) -
			       b_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params_out_sten ) -
			       b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) -
			       b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_.Give_cell_hexahedra ( params_out_cell ) -
			       b_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_.Give_fromTotal(i) -
			       b_.Give_fromTotal(i);
		}
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline Sub<A, B> operator- ( const Expr<A>& a, const Expr<B>& b ) {
	return Sub<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <typename TYPE, class B>

class CSub : public Expr<CSub<TYPE, B> > {
		TYPE a_;
		const B& b_;
		
	public:
		inline CSub ( TYPE a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			return b_.Give_blockgrid();
		}
		
		stencil_typ Give_stencil_typ() const {
			return b_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			return b_.totalCalcNotPossible();
		};
		
		int getTotalNumberData() const { return b_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const { 
			return b_.Give_control_typ(); 
		}
		
		typedef _TypeOf1_ ( B, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_ - b_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_ - b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_ - b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_ - b_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_ - b_.Give_fromTotal(i);
		}
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class B>
inline CSub<double, B> operator- ( double a, const Expr<B>& b ) {
	return CSub<double, B> ( a, b );
}

template <class B>
inline CSub<std::complex<double>, B> operator- ( std::complex<double> a, const Expr<B>& b ) {
	return CSub<std::complex<double>, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>

class Minus : public Expr<Minus<A> > {
		const A& a_;
		
	public:
		inline Minus ( const A& a ) : a_ ( a ) {}
		
		Blockgrid* Give_blockgrid() const {
			return a_.Give_blockgrid();
		}
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			return a_.totalCalcNotPossible();
		};
		
		int getTotalNumberData() const {  return a_.getTotalNumberData(); }
		
		control_typ Give_control_typ() const { 
			return a_.Give_control_typ(); 
		}
		
		typedef typename A::Result Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return - a_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return - a_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return - a_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return - a_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return - a_.Give_fromTotal(i);
		}
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline Minus<A> operator- ( const Expr<A>& a ) {
	return Minus<A> ( a );
}

//----------------------------------------------------------------------------
template <class A, typename TYPE>

class Times : public Expr<Times<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline Times ( const A& a, const TYPE& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			return a_.totalCalcNotPossible();
		};

		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const { 
			return a_.Give_control_typ();
		}
		
		typedef _TypeOf1_ ( A, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out ) * b_;
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params_out_sten ) * b_;
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) * b_;
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_.Give_cell_hexahedra ( params_out_cell ) * b_;
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_.Give_fromTotal(i) * b_;
		}
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline Times<A, double>
operator* ( const Expr<A>& a, double b ) {
	return Times<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times<A, double>
operator* ( double b, const Expr<A>& a ) {
	return Times<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times<A, std::complex<double> >
operator* ( const Expr<A>& a, std::complex<double> b ) {
	return Times<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times<A, std::complex<double> >
operator* ( std::complex<double> b, const Expr<A>& a ) {
	return Times<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <typename TYPE, class B>

class Div : public Expr<Div<TYPE, B> > {
		const TYPE a_;
		const B& b_;
		
	public:
		inline Div ( const TYPE& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			return b_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			return b_.totalCalcNotPossible();
		};
		
		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const { 
			return b_.Give_control_typ();
		}
		
		typedef _TypeOf1_ ( B, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_ / b_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_ / b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_ / b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_ / b_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_ / b_.Give_fromTotal(i);
		}
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline Times<A, double> operator/ ( const Expr<A>& a, double b ) {
	return Times<A, double> ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times<A, std::complex<double> > operator/ ( const Expr<A>& a, std::complex<double> b ) {
	return Times<A, std::complex<double> > ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class B>
inline Div<double, B> operator/ ( double a, const Expr<B>& b ) {
	return Div<double, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class B>
inline Div<std::complex<double>, B> operator/ ( std::complex<double> a, const Expr<B>& b ) {
	return Div<std::complex<double>, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class VTimes : public Expr<VTimes<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VTimes ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil) return yes_stencil;
			return b_.Give_stencil_typ();
		}
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};
		
		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const { 
			if ( a_.Give_control_typ() == not_thread_save) {
				return not_thread_save;
			}
			return b_.Give_control_typ();
		}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out ) *
			       b_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params_out_sten ) *
			       b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) *
			       b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_.Give_cell_hexahedra ( params_out_cell ) *
			       b_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_.Give_fromTotal(i) *
			       b_.Give_fromTotal(i);
		}		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VTimes<A, B> operator* ( const Expr<A>& a, const Expr<B>& b ) {
	return VTimes<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class VDiv : public Expr<VDiv<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VDiv ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil) return yes_stencil;
			return b_.Give_stencil_typ();
		}

		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};
				
		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const { 
			if ( a_.Give_control_typ() == not_thread_save) {
				return not_thread_save;
			}
			return b_.Give_control_typ();
		}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out ) /
			       b_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params_out_sten ) /
			       b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
			return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) /
			       b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
			return a_.Give_cell_hexahedra ( params_out_cell ) /
			       b_.Give_cell_hexahedra ( params_out_cell );
		}
		
		inline Result Give_fromTotal(int i) const {
			return a_.Give_fromTotal(i) /
			       b_.Give_fromTotal(i);
		}
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VDiv<A, B> operator/ ( const Expr<A>& a, const Expr<B>& b ) {
	return VDiv<A, B> ( a, b );
}

//----------------------------------------------------------------------------

/** \addtogroup ExpressionTemplates **/
/* @{ */  
/**
 * epsilin Parameter for save division
 **/
struct saveDivision {
    static double eps;	
};
/* @} */



template <class A, class B>
class SaveDiv : public Expr<SaveDiv<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline SaveDiv ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil) return yes_stencil;
			return b_.Give_stencil_typ();
		}

		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};
				
		int getTotalNumberData() const {  return a_.getTotalNumberData(); }		
		
		control_typ Give_control_typ() const { 
			if ( a_.Give_control_typ() == not_thread_save) {
				return not_thread_save;
			}
			return b_.Give_control_typ();
		}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
		       Result valueB = b_.template Give_data<TYP_EL> ( params_out );
		       if(myABS(valueB) < saveDivision::eps) return 0.0;
                       return a_.template Give_data<TYP_EL> ( params_out ) / valueB;		  
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params_in_sten ) const {
		       Result valueB = b_.template Give_data_sten<TYP_EL> ( params_out_sten );
		       if(myABS(valueB) < saveDivision::eps) return 0.0;
                       return a_.template Give_data_sten<TYP_EL> ( params_out_sten ) / valueB;		  
		}
		
		inline Result Give_matrix_hexahedra ( params_in_loc_matrix ) const {
		       Result valueB = b_.Give_matrix_hexahedra ( params_out_loc_matrix );
		       if(myABS(valueB) < saveDivision::eps) return 0.0;
                       return a_.Give_matrix_hexahedra ( params_out_loc_matrix ) / valueB;			  
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const {
		       Result valueB = b_.Give_cell_hexahedra ( params_out_cell );
		       if(myABS(valueB) < saveDivision::eps) return 0.0;
                       return a_.Give_cell_hexahedra ( params_out_cell ) / valueB;		  
		}
		
		inline Result Give_fromTotal(int i) const {
		       Result valueB = b_.Give_fromTotal(i);
		       if(myABS(valueB) < saveDivision::eps) return 0.0;
                       return a_.Give_fromTotal(i) / valueB;
		}		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}	
		
   private:
     static double myABS(double a) {  return a < 0.0 ? -a : a; }
     static std::complex<double> myABS(std::complex<double> c) {  
         return (c.real() < 0.0 ? -c.real() : c.real()) + (c.imag() < 0.0 ? -c.imag() : c.imag());
    }
};


//----------------------------------------------------------------------------

/** \addtogroup ExpressionTemplates **/
/* @{ */  
/**
 * performs save division. This means if divisor is smaller than saveDivision::eps, then result is 0.0
 **/
template <class A, class B>
inline SaveDiv<A, B> saveDiv ( const Expr<A>& a, const Expr<B>& b ) {
	return SaveDiv<A, B> ( a, b );
}
/* @} */



//----------------------------------------------------------------------------
#endif // EXTEMP_H
