/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 *              Department Informatik Lehrstuhl 10 - Systemsimulation
 *              Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// functor.h
//
// ------------------------------------------------------------

#ifndef FUNCTOR_H
#define FUNCTOR_H

/*************************************************************
 * Defines functor expression templates. Use
 * functor objects instead of function pointers.
 *
 * A functor class has to define a evaluate
 * function. No other requirements are imposed on
 * the functor class implementation.
 * Example for 3 arguments
 * (DTyp_Result = float,DTyp = double):
 *
 * class MyFunctor3
 * {
 * public:
 *      float evaluate(double x, double y, double z) const;
 * };
 * 
 * Furthermore, vector of variables is allowed in
 *     class FunctorVector
 * 
************************************************************/

 
template <class A, class DTyp_Result, class DTyp, class Func>

class Exp_Functor1 : public Expr<Exp_Functor1<A, DTyp_Result, DTyp, Func> > {
                const A& a_;
                Func* functor_;
                control_typ funCtrTyp_;

        public:
                inline Exp_Functor1 ( const A& a, Func* functor, control_typ funCtrTyp )
                                : a_ ( a ), functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {}

                stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ();     }
		bool totalCalcNotPossible()    const { return a_.totalCalcNotPossible(); }                
                control_typ Give_control_typ() const { return funCtrTyp_; }


                typedef DTyp_Result Result;

                template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params_in ) const;

                inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;

		inline Result Give_fromTotal(int i) const;			
		
                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                }
};


template <class A, class DTyp_Result, class DTyp, class Func>
template <elementTyp TYP_EL>
inline DTyp_Result
Exp_Functor1<A, DTyp_Result, DTyp, Func>::Give_data ( params_in ) const {
        return functor_->evaluate ( a_.template Give_data<TYP_EL> ( params_out ) );
}

template <class A, class DTyp_Result, class DTyp, class Func>
inline DTyp_Result
Exp_Functor1<A, DTyp_Result, DTyp, Func>::Give_cell_hexahedra ( params_in_cell ) const {
        return functor_->evaluate ( a_.Give_cell_hexahedra ( params_out_cell ) );
}

template <class A, class DTyp_Result, class DTyp, class Func>
inline DTyp_Result
Exp_Functor1<A, DTyp_Result, DTyp, Func>::Give_fromTotal(int i) const {
        return functor_->evaluate ( a_.Give_fromTotal(i) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
/**
 * The class Func must contain a member function
 *    DTyp Func::evalute(DTyp x);
 * Then, one apply Functor1 as follows
 *   \verbatim
 *       class Func test{
 *          public: 
 *              Func(double f) : mf(f) {};
 *              double evaluate(double x) { return x * mf; }
 *          private:
 *              double mf;
 *       };
 *      .....
 * 
 *       Func multiplyFive(5.0);
 *       Functor1 multiplyFiveFunc(multiplyFive);
 *      .....
 * 
 *       a = multiplyFiveFunc(b);
 *   \endverbatim
 **/
template <class DTyp_Result, class DTyp, class Func>
class Functor1 {
        public:
                Functor1 ( Func& functor,  control_typ funCtrTyp = thread_save ) : functor_ ( &functor ), funCtrTyp_ ( funCtrTyp ) {};
                Functor1 ( Func* functor,  control_typ funCtrTyp = thread_save ) : functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {};

                template <class A>
                inline Exp_Functor1<A, DTyp_Result, DTyp, Func>
                operator() ( const Expr<A>& a ) const
                { return Exp_Functor1<A, DTyp_Result, DTyp, Func> ( a, functor_, funCtrTyp_ ); }

        private:
                Func* functor_;
                control_typ funCtrTyp_;
};
/* @} */

/**
*  2 arguments
**/
template <class A, class B, class DTyp_Result, class DTyp, class Func>
class Exp_Functor2
                        : public Expr<Exp_Functor2<A, B, DTyp_Result, DTyp, Func> > {
                const A& a_;
                const B& b_;
                Func* functor_;
                control_typ funCtrTyp_;

        public:
                inline Exp_Functor2 ( const A& a, const B& b, Func* functor, control_typ funCtrTyp ) :
                                a_ ( a ), b_ ( b ), functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {}

		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			return b_.Give_stencil_typ();
		};		
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};	
                control_typ Give_control_typ() const { return funCtrTyp_; }

                typedef DTyp_Result Result;

                template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params_in ) const;

                inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;

		inline Result Give_fromTotal(int i) const;			
		
                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                }
};

template <class A, class B, class DTyp_Result, class DTyp, class Func>
template <elementTyp TYP_EL>
inline DTyp_Result
Exp_Functor2<A, B, DTyp_Result, DTyp, Func>::Give_data ( params_in ) const {
        return functor_->evaluate ( a_.template Give_data<TYP_EL> ( params_out ),
                                    b_.template Give_data<TYP_EL> ( params_out ) );
}

template <class A, class B, class DTyp_Result, class DTyp, class Func>
inline DTyp_Result
Exp_Functor2<A, B, DTyp_Result, DTyp, Func>::Give_cell_hexahedra ( params_in_cell ) const {
        return functor_->evalute ( a_.Give_cell_hexahedra ( params_out_cell ),
                                   b_.Give_cell_hexahedra ( params_out_cell ) );
}


template <class A, class B, class DTyp_Result, class DTyp, class Func>
inline DTyp_Result
Exp_Functor2<A, B, DTyp_Result, DTyp, Func>::Give_fromTotal(int i) const {
        return functor_->evalute ( a_.Give_fromTotal(i),
                                   b_.Give_fromTotal(i) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
/**
 * The class Func must contain a member function
 *    DTyp Func::evalute(DTyp x, DTyp y);
 * see Functor1
 **/
template <class DTyp_Result, class DTyp, class Func>
class Functor2 {

        public:
                Functor2 ( Func& functor, control_typ funCtrTyp = thread_save) : functor_ ( &functor ), funCtrTyp_ ( funCtrTyp )  {};
                Functor2 ( Func* functor, control_typ funCtrTyp = thread_save) : functor_ ( functor ), funCtrTyp_ ( funCtrTyp )  {};

                template <class A, class B>
                inline Exp_Functor2<A, B, DTyp_Result, DTyp, Func>
                operator() ( const Expr<A>& a, const Expr<B>& b ) const
                { return Exp_Functor2<A, B, DTyp_Result, DTyp, Func> ( a, b, functor_, funCtrTyp_ ); }

        private:
                Func* functor_;
                control_typ funCtrTyp_;
};
/* @} */

// 3 arguments

template < class A, class B, class C,

class DTyp_Result, class DTyp, class Func >

class Exp_Functor3
                        : public Expr<Exp_Functor3<A, B, C, DTyp_Result, DTyp, Func> > {
                const A& a_;
                const B& b_;
                const C& c_;
                Func* functor_;
                control_typ funCtrTyp_;

        public:
                inline Exp_Functor3 ( const A& a,
                                      const B& b,
                                      const C& c,
                                      Func* functor,
                                      control_typ funCtrTyp ) :
                                a_ ( a ), b_ ( b ), c_ ( c ), functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {}

		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			if(b_.Give_stencil_typ() == yes_stencil ) return yes_stencil;			
			return c_.Give_stencil_typ();
		};		
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			if(b_.totalCalcNotPossible()) return true;			
			return c_.totalCalcNotPossible();
		};
                control_typ Give_control_typ() const { return funCtrTyp_; }

                typedef DTyp_Result Result;

                template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params_in ) const;

                inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;

		inline Result Give_fromTotal(int i) const;	
		
                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template Update<TYP_EL> ( id );
                }
};


template < class A, class B, class C,

class DTyp_Result, class DTyp, class Func >
template <elementTyp TYP_EL>
inline DTyp_Result
Exp_Functor3<A, B, C, DTyp_Result, DTyp, Func>::Give_data ( params_in ) const {
        return functor_->evaluate ( a_.template Give_data<TYP_EL> ( params_out ),
                                    b_.template Give_data<TYP_EL> ( params_out ),
                                    c_.template Give_data<TYP_EL> ( params_out ) );
}

template < class A, class B, class C,
class DTyp_Result, class DTyp, class Func >
inline DTyp_Result
Exp_Functor3<A, B, C, DTyp_Result, DTyp, Func>::Give_cell_hexahedra ( params_in_cell ) const {
        return functor_->evaluate ( a_.Give_cell_hexahedra ( params_out_cell ),
                                    b_.Give_cell_hexahedra ( params_out_cell ),
                                    c_.Give_cell_hexahedra ( params_out_cell ) );

}


template < class A, class B, class C,
class DTyp_Result, class DTyp, class Func >
inline DTyp_Result
Exp_Functor3<A, B, C, DTyp_Result, DTyp, Func>::Give_fromTotal(int i) const {
        return functor_->evaluate ( a_.Give_fromTotal(i),
                                    b_.Give_fromTotal(i),
                                    c_.Give_fromTotal(i) );

}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
/**
 * The class Func must contain a member function
 *    DTyp Func::evalute(DTyp x, DTyp y, DTyp z);
 * see Functor1
 **/
template <class DTyp_Result, class DTyp, class Func>
class Functor3 {

        public:
                Functor3 ( Func& functor, control_typ funCtrTyp = thread_save) : functor_ ( &functor ), funCtrTyp_ ( funCtrTyp ) {};
                Functor3 ( Func* functor, control_typ funCtrTyp = thread_save) : functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {};

                template <class A, class B, class C>
                inline Exp_Functor3<A, B, C, DTyp_Result, DTyp, Func>
                operator() ( const Expr<A>& a,
                             const Expr<B>& b,
                             const Expr<C>& c ) const
                { return Exp_Functor3<A, B, C, DTyp_Result, DTyp, Func> ( a, b, c, functor_, funCtrTyp_ ); }


        private:
                Func* functor_;
                control_typ funCtrTyp_;
};
/* @} */

// 4 arguments

template <class A, class B, class C, class D, class DTyp_Result, class DTyp, class Func>

class Exp_Functor4
                        : public Expr<Exp_Functor4<A, B, C, D, DTyp_Result, DTyp, Func> > {
                const A& a_;
                const B& b_;
                const C& c_;
                const D& d_;
                const control_typ funCtrTyp_;
                Func* functor_;

        public:
                inline Exp_Functor4 ( const A& a,
                                      const B& b,
                                      const C& c,
                                      const D& d,
                                      Func* functor,
                                      control_typ funCtrTyp ) :
                                a_ ( a ), b_ ( b ), c_ ( c ), d_ ( d ), functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {}

                stencil_typ Give_stencil_typ() const { return no_stencil; }

                control_typ Give_control_typ() const { return funCtrTyp_; }

                typedef DTyp_Result Result;

                template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params_in ) const;

                inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;


                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                }
};


template < class A, class B, class C, class D,

class DTyp_Result, class DTyp, class Func >
template <elementTyp TYP_EL>
inline DTyp_Result
Exp_Functor4<A, B, C, D, DTyp_Result, DTyp, Func>::Give_data ( params_in ) const {
        return functor_->evaluate ( a_.template Give_data<TYP_EL> ( params_out ),
                                    b_.template Give_data<TYP_EL> ( params_out ),
                                    c_.template Give_data<TYP_EL> ( params_out ),
                                    d_.template Give_data<TYP_EL> ( params_out ) );
}

template < class A, class B, class C, class D,

class DTyp_Result, class DTyp, class Func >
inline DTyp_Result
Exp_Functor4<A, B, C, D, DTyp_Result, DTyp, Func>::Give_cell_hexahedra ( params_in_cell ) const {
        return functor_->evaluate ( a_.template Give_cell_hexahedra ( params_out_cell ),
                                    b_.template Give_cell_hexahedra ( params_out_cell ),
                                    c_.template Give_cell_hexahedra ( params_out_cell ),
                                    d_.template Give_cell_hexahedra ( params_out_cell ) );
}



/** \addtogroup ExpressionTemplates **/
/* @{ */  
/**
 * The class Func must contain a member function
 *    DTyp Func::evalute(DTyp x, DTyp y);
 * see Functor1
 **/
template <class DTyp_Result, class DTyp, class Func>
class Functor4 {

        public:
                Functor4 ( Func* functor, control_typ funCtrTyp = thread_save ) : functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {

                };

                Functor4 ( Func& functor, control_typ funCtrTyp = thread_save ) : functor_ ( &functor ), funCtrTyp_ ( funCtrTyp ) {

                };

                template <class A, class B, class C, class D>
                inline Exp_Functor4<A, B, C, D, DTyp_Result, DTyp, Func>
                operator() ( const Expr<A>& a,
                             const Expr<B>& b,
                             const Expr<C>& c,
                             const Expr<D>& d ) const {
                        return Exp_Functor4<A, B, C, D, DTyp_Result, DTyp, Func> ( a, b, c, d, functor_ ,
                                funCtrTyp_ );
                }


        private:
                Func* functor_;
                control_typ funCtrTyp_;
};
/* @} */

////////////////////////////////////////////////////////////////////////////////////////////////////
// list of arguments
////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////
//  Anwendung
//  1. FunctorVector definieren:
//          FunctorVector  func(funcF);
//  2. Als Expression anwenden:
//        a = func()
////////////////


/*
template <class DTyp_Result, class DTyp, class Func>
class Exp_Functor_Vector : public Expr<Exp_Functor_Vector<DTyp_Result, DTyp, Func> > {
                int sizeVector;
		Func* functor;
		Variable<DTyp>* vecVar;
		const control_typ funCtrTyp;
	        DTyp* vectorCopy;
	public:
		inline Exp_Functor_Vector(int sizeVector_, 
					  const Variable<DTyp>* vecVar_, 
			                  Func* functor_, 
			                  const control_typ funCtrTyp_,
			                  DTyp* vectorCopy_) :	
		      sizeVector(sizeVector_), functor(functor_), 
		      vecVar(vecVar_),         funCtrTyp(funCtrTyp_), 
		      vectorCopy(vectorCopy_)  {	
		}
				
		stencil_typ Give_stencil_typ() const { return no_stencil; }		
		control_typ Give_control_typ() const { return funCtrTyp; }
		
		typedef DTyp_Result Result;
		
		
		template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params_in ) const;
				
		inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;
		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			for(int i=0;i<sizeVector;++i) vecVar[i].template  Update<TYP_EL> ( id );
		}
		
};


template <class DTyp_Result, class DTyp, class Func>
template <elementTyp TYP_EL>
inline DTyp_Result
Exp_Functor_Vector<DTyp_Result, DTyp, Func>::Give_data ( params_in ) const {
        for(int i=0;i<sizeVector;++i) vectorCopy[i] = vecVar.template Give_data<TYP_EL>(params_out);
	return functor->evaluate( vectorCopy );
}

template <class DTyp_Result, class DTyp, class Func>
inline DTyp_Result
Exp_Functor_Vector<DTyp_Result, DTyp, Func>::Give_cell_hexahedra ( params_in_cell ) const {
       for(int i=0;i<sizeVector;++i) vectorCopy[i] = vecVar.Give_cell_hexahedra(params_out_cell);
       return functor->evaluate( vectorCopy );  
}




template <class DTyp_Result, class DTyp, class Func>
class FunctorVector {
  public:
     FunctorVector(Func* functor, int sizeVector_, control_typ funCtrTyp = thread_save ) 
        : functor_ ( functor ), funCtrTyp_ ( funCtrTyp ) {
	  sizeVector = sizeVector_;
	  vectorCopy = new DTyp[sizeVector];
     };
     ~FunctorVector() { delete[] vectorCopy; }
		
     inline Exp_Functor_Vector<DTyp_Result, DTyp, Func>
        operator() (const Variable<DTyp>* vecVar) const {
		    return Exp_Functor_Vector<DTyp_Result, DTyp, Func>(sizeVector, vecVar, functor_ , funCtrTyp_, vectorCopy);
	}

  private:
    Func* functor_;
    control_typ funCtrTyp_;
    DTyp* vectorCopy;
    int sizeVector;
};

*/



#endif /* FUNCTOR_H */
