/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "extemp.h"
#include "parallel.h"
#include "variable.h"

template <>
void Variable<double>::Print_AVS(ofstream *Datei) {
  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei->precision(3);
  Datei->setf(std::ios::fixed,std::ios::floatfield);

  for(id=0;id<ug->Give_number_hexahedra();++id) {
    Update<hexahedronEl>(id);
    num_total += blockgrid->Give_N_total_hexahedron(id);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id))
                        *(blockgrid->Give_Nz_hexahedron(id));
  }

  // Teil 0: Write information
  if(my_rank==0) {
    *Datei << "# File created by 3DUGBlock " << endl;
    if(ug->partitioning_is_done())
      *Datei << "# parallel with  " 
	     << ug->number_of_used_processors() << " used processors " << endl;
      *Datei << "# UCD file format for AVS " << endl;
      *Datei << "# variable " << endl;

    // Teil 1: Kopfzeile schreiben
    *Datei << num_total << " "
           << cell_total << " 1 0 0" << endl;
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  id = 0;
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {

    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      *Datei << id
		     << " " << v.x
		     << " " << v.y
		     << " " << v.z
		     << "\n";
	      ++id; 
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei << id
	           <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	    ++id;
	  }
    }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    int id_start;
    id_start = 0;
    id = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    *Datei << id << "  1  hex  "
		   << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
		   << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
		   << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
		   << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
		   << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
		   << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
		   << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
		   << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
		   << "\n";
 	    ++id; 
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }

    // Teil 4:
    *Datei << "1 1" << "\n";

    // Teil 5:
    *Datei << "variable " << "\n";
  }

    // Teil 6: Ausgabe der Werte der Variable

  id = 0;
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {

    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      *Datei << id
		     << "  " << data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]
		     << "\n";
	      ++id;
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei << id
                   << "  " << data_coord[i + (Nx+1) * (j + (Ny+1) * k )]
                   << "\n";
	    ++id;
	  }
    }
  }
}

template <>
void Variable<std::complex<double> >::Print_AVS(ofstream *Datei) {

  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei->precision(3);
  Datei->setf(std::ios::fixed,std::ios::floatfield);

  for(id=0;id<ug->Give_number_hexahedra();++id) {
    Update<hexahedronEl>(id);
    num_total += blockgrid->Give_N_total_hexahedron(id);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id))
                        *(blockgrid->Give_Nz_hexahedron(id));
  }

  // Teil 0: Write information
  if(my_rank==0) {
    *Datei << "# File created by 3DUGBlock " << endl;
    if(ug->partitioning_is_done())
      *Datei << "# parallel with  " 
	     << ug->number_of_used_processors() << " used processors " << endl;
      *Datei << "# UCD file format for AVS " << endl;
      *Datei << "# variable " << endl;

    // Teil 1: Kopfzeile schreiben
    *Datei << num_total << " "
           << cell_total << " 1 0 0" << endl;
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  id = 0;
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {

    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      *Datei << id
		     << " " << v.x
		     << " " << v.y
		     << " " << v.z
		     << "\n";
	      ++id; 
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei << id
	           <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	    ++id;
	  }
    }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    int id_start;
    id_start = 0;
    id = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    *Datei << id << "  1  hex  "
		   << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
		   << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
		   << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
		   << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
		   << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
		   << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
		   << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
		   << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
		   << "\n";
 	    ++id; 
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }

    // Teil 4:
    *Datei << "1 1" << "\n";

    // Teil 5:
    *Datei << "variable " << "\n";
  }

    // Teil 6: Ausgabe der Werte der Variable

  id = 0;
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {

    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      *Datei << id
		     << "  " << abs(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)])
		     << "\n";
	      ++id;
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = abs(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei << id
                   << "  " << data_coord[i + (Nx+1) * (j + (Ny+1) * k )]
                   << "\n";
	    ++id;
	  }
    }
  }
}
