/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "extemp.h"
#include "parallel.h"
#include "variable.h"

template <>
void Variable<double>::Print_Dx(ofstream *Datei) {
  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;
  int id_WSD, id_ESD, id_WND, id_END;
  int id_WST, id_EST, id_WNT, id_ENT;

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei->precision(6);
  Datei->setf(std::ios::fixed,std::ios::floatfield);

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Update<hexahedronEl>(id_hex);
    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
  }

  // Teil 0: Write information
  if(my_rank==0) {
    *Datei << "# File created by 3DUGBlock " << endl;
    if(ug->partitioning_is_done())
      *Datei << "# parallel with  " 
	     << ug->number_of_used_processors() << " used processors " << endl;
    *Datei << "# dx file format for Opendx " << endl;
    *Datei << "# variable " << endl;
    
    // Teil 1: Kopfzeile schreiben
    *Datei << "object 1 class array type float rank 1 shape 3 items "
           << num_total << " data follows" << endl;
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      *Datei << " " << v.x
		     << " " << v.y
		     << " " << v.z
		     << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
  }
  
  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    *Datei << "object 2 class array type int rank 1 shape 4 items "
           << 6*cell_total << " data follows" << endl;
    
    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
	for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    id_WNT = id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) );
	    id_WST = id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) );
	    id_EST = id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) );
	    id_ENT = id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) );
	    id_WND = id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) );
	    id_WSD = id_start + i   + (Nx+1)*(j  +  k * (Ny+1) );
	    id_ESD = id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) );
	    id_END = id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) );
	    
	    // 0. Tet: WND, WNT, WST, EST
	    *Datei << id_WNT  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << "\n";
	    
	    // 1. Tet: EST, WND, WST, ESD
	    *Datei << id_EST  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << "\n";
	    
	    // 2. Tet: WND, WSD, WST, ESD
	    *Datei << id_WND  << "  ";
	    *Datei << id_WSD  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << "\n";
	    
	    // 3. Tet: EST, WND, ESD, END
	    *Datei << id_EST  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	    
	    // 4. Tet: ENT, WNT, EST, END
	    *Datei << id_ENT  << "  ";
	    *Datei << id_WNT  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	    
	    // 5. Tet: WNT, WND, EST, END
	    *Datei << id_WNT  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }
    
    // Teil 4:
    *Datei << "attribute " << '"' << "element type"
	   << '"' << " string " << '"' << "tetrahedra" 
	   << '"' << "" << endl;
    *Datei << "attribute " << '"' << "ref" << '"'
	   << " string " << '"' << "positions" 
	   << '"' << "" << endl;
    
    // Teil 5: Ausgabe der id der Points
    *Datei << "object 3 class array type float rank 0 items "
           << num_total << " data follows" << endl;
  }

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      *Datei << "  " << data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]
		     << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei <<	"  " << data_coord[i + (Nx+1) * (j + (Ny+1) * k )] << "\n";
	  }
    }
  }

  if(my_rank==0) {
    // Teil 6: end variable
    *Datei << "attribute "
	   << '"' << "dep"
	   << '"' << " string " << '"' << "positions" << '"' << "" << endl;
    
    // Teil 7: end remarks
    *Datei << "object " << '"'
	   << "irregular positions irregular connections"
	   << '"' << " class field" << endl;
    *Datei << "component " << '"' << "positions" << '"' << " value 1" << endl;
    *Datei << "component "
	   << '"' << "connections" << '"' << " value 2" << endl;
    *Datei << "component " << '"' << "data" << '"' << " value 3" << endl;
    *Datei << "end" << endl;
  }
}

template <>
void Variable<std::complex<double> >::Print_Dx(ofstream *Datei) {

  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;
  int id_WSD, id_ESD, id_WND, id_END;
  int id_WST, id_EST, id_WNT, id_ENT;

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei->precision(6);
  Datei->setf(std::ios::fixed,std::ios::floatfield);

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Update<hexahedronEl>(id_hex);
    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
  }

  // Teil 0: Write information
  if(my_rank==0) {
    *Datei << "# File created by 3DUGBlock " << endl;
    if(ug->partitioning_is_done())
      *Datei << "# parallel with  " 
	     << ug->number_of_used_processors() << " used processors " << endl;
    *Datei << "# dx file format for Opendx " << endl;
    *Datei << "# variable " << endl;
    
    // Teil 1: Kopfzeile schreiben
    *Datei << "object 1 class array type float rank 1 shape 3 items "
           << num_total << " data follows" << endl;
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      *Datei << " " << v.x
		     << " " << v.y
		     << " " << v.z
		     << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
  }
  
  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    *Datei << "object 2 class array type int rank 1 shape 4 items "
           << 6*cell_total << " data follows" << endl;
    
    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
	for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    id_WNT = id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) );
	    id_WST = id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) );
	    id_EST = id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) );
	    id_ENT = id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) );
	    id_WND = id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) );
	    id_WSD = id_start + i   + (Nx+1)*(j  +  k * (Ny+1) );
	    id_ESD = id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) );
	    id_END = id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) );
	    
	    // 0. Tet: WND, WNT, WST, EST
	    *Datei << id_WNT  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << "\n";
	    
	    // 1. Tet: EST, WND, WST, ESD
	    *Datei << id_EST  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << "\n";
	    
	    // 2. Tet: WND, WSD, WST, ESD
	    *Datei << id_WND  << "  ";
	    *Datei << id_WSD  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << "\n";
	    
	    // 3. Tet: EST, WND, ESD, END
	    *Datei << id_EST  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	    
	    // 4. Tet: ENT, WNT, EST, END
	    *Datei << id_ENT  << "  ";
	    *Datei << id_WNT  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	    
	    // 5. Tet: WNT, WND, EST, END
	    *Datei << id_WNT  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }
    
    // Teil 4:
    *Datei << "attribute " << '"' << "element type"
	   << '"' << " string " << '"' << "tetrahedra" 
	   << '"' << "" << endl;
    *Datei << "attribute " << '"' << "ref" << '"'
	   << " string " << '"' << "positions" 
	   << '"' << "" << endl;
    
    // Teil 5: Ausgabe der id der Points
    *Datei << "object 3 class array type float rank 0 items "
           << num_total << " data follows" << endl;
  }

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      *Datei << "  " << abs(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)])
		     << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = abs(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei <<	"  " << data_coord[i + (Nx+1) * (j + (Ny+1) * k )] << "\n";
	  }
    }
  }

  if(my_rank==0) {
    // Teil 6: end variable
    *Datei << "attribute "
	   << '"' << "dep"
	   << '"' << " string " << '"' << "positions" << '"' << "" << endl;
    
    // Teil 7: end remarks
    *Datei << "object " << '"'
	   << "irregular positions irregular connections"
	   << '"' << " class field" << endl;
    *Datei << "component " << '"' << "positions" << '"' << " value 1" << endl;
    *Datei << "component "
	   << '"' << "connections" << '"' << " value 2" << endl;
    *Datei << "component " << '"' << "data" << '"' << " value 3" << endl;
    *Datei << "end" << endl;
  }
}

template <>
void Variable<double>::Print_Dx( Variable<double>& ux,
				 Variable<double>& uy,
				 Variable<double>& uz, std::ofstream *Datei) {

  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;
  int id_WSD, id_ESD, id_WND, id_END;
  int id_WST, id_EST, id_WNT, id_ENT;

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei->precision(6);
  Datei->setf(std::ios::fixed,std::ios::floatfield);

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Update<hexahedronEl>(id_hex);

    ux.Update<hexahedronEl>(id_hex);
    uy.Update<hexahedronEl>(id_hex);
    uz.Update<hexahedronEl>(id_hex);

    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
  }

  // Teil 0: Write information
  if(my_rank==0) {
    *Datei << "# File created by 3DUGBlock " << endl;
    if(ug->partitioning_is_done())
      *Datei << "# parallel with  " 
	     << ug->number_of_used_processors() << " used processors " << endl;
    *Datei << "# dx file format for Opendx " << endl;
    *Datei << "# variable with deformation  " << endl;
    
    // Teil 1: Kopfzeile schreiben
    *Datei << "object 1 class array type float rank 1 shape 3 items "
           << num_total << " data follows" << endl;
  }

  // Teil 2: Koordinaten der Punkte ausgeben

  double** data_hexahedra_ux = ux.Give_pointer_data_hexahedra();
  double** data_hexahedra_uy = uy.Give_pointer_data_hexahedra();
  double** data_hexahedra_uz = uz.Give_pointer_data_hexahedra();

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      *Datei << " " << v.x + data_hexahedra_ux[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]
		     << " " << v.y + data_hexahedra_uy[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]
		     << " " << v.z + data_hexahedra_uz[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]
		     << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x +
                 data_hexahedra_ux[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y +
                 data_hexahedra_uy[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z +
                 data_hexahedra_uz[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    *Datei << "object 2 class array type int rank 1 shape 4 items "
           << 6*cell_total << " data follows" << endl;
    
    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
	for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    id_WNT = id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) );
	    id_WST = id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) );
	    id_EST = id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) );
	    id_ENT = id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) );
	    id_WND = id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) );
	    id_WSD = id_start + i   + (Nx+1)*(j  +  k * (Ny+1) );
	    id_ESD = id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) );
	    id_END = id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) );
	    
	    // 0. Tet: WND, WNT, WST, EST
	    *Datei << id_WNT  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << "\n";
	    
	    // 1. Tet: EST, WND, WST, ESD
	    *Datei << id_EST  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << "\n";
	    
	    // 2. Tet: WND, WSD, WST, ESD
	    *Datei << id_WND  << "  ";
	    *Datei << id_WSD  << "  ";
	    *Datei << id_WST  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << "\n";
	    
	    // 3. Tet: EST, WND, ESD, END
	    *Datei << id_EST  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_ESD  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	    
	    // 4. Tet: ENT, WNT, EST, END
	    *Datei << id_ENT  << "  ";
	    *Datei << id_WNT  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	    
	    // 5. Tet: WNT, WND, EST, END
	    *Datei << id_WNT  << "  ";
	    *Datei << id_WND  << "  ";
	    *Datei << id_EST  << "  ";
	    *Datei << id_END  << "  ";
	    *Datei << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }


    // Teil 4:
    *Datei << "attribute " << '"' << "element type"
	   << '"' << " string " << '"' << "tetrahedra" 
	   << '"' << "" << endl;
    *Datei << "attribute " << '"' << "ref" << '"'
	   << " string " << '"' << "positions" 
	   << '"' << "" << endl;
    
    // Teil 5: Ausgabe der id der Points
    *Datei << "object 3 class array type float rank 0 items "
           << num_total << " data follows" << endl;
  }

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      *Datei << "  " << data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]
		     << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    *Datei <<	"  " << data_coord[i + (Nx+1) * (j + (Ny+1) * k )] << "\n";
	  }
    }
  }

  if(my_rank==0) {
    // Teil 6: end variable
    *Datei << "attribute "
	   << '"' << "dep"
	   << '"' << " string " << '"' << "positions" << '"' << "" << endl;
    
    // Teil 7: end remarks
    *Datei << "object " << '"'
	   << "irregular positions irregular connections"
	   << '"' << " class field" << endl;
    *Datei << "component " << '"' << "positions" << '"' << " value 1" << endl;
    *Datei << "component "
	   << '"' << "connections" << '"' << " value 2" << endl;
    *Datei << "component " << '"' << "data" << '"' << " value 3" << endl;
    *Datei << "end" << endl;
  }
}
