/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// vari_array_eqn.h
//
// ------------------------------------------------------------

#ifndef VARI_ARRAY_EQN_H
#define VARI_ARRAY_EQN_H
#ifdef _OPENMP
#include <omp.h>
#endif




//////////////////////////////////////////////////////////////
// 1. class equation of array
// 2. 
//////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////
// 1. class equation of array
//////////////////////////////////////////////////////////////

template <class A, class DTyp>
class Expr_Equation_of_Array {
  const Variable<DTyp>& varleft;
  const A& a_;
  public:
  Expr_Equation_of_Array(const Variable<DTyp>& v_,
			 const Expr<A>& a) : varleft(v_), a_(a) {}

  Blockgrid* Give_blockgrid() const { return a_.Give_blockgrid(); };
  stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ(); }

  DTyp** Give_left_v_data_hexahedra()   const { return varleft.data_hexahedra; }
  DTyp** Give_left_v_data_quadrangle() const { return varleft.data_quadrangles; }
  DTyp** Give_left_v_data_edges()       const { return varleft.data_edges; }
  DTyp** Give_left_v_data_points()      const { return varleft.data_points; }

  typedef typename A::Result Result;

  template <elementTyp TYP_EL>
  inline Result Give_data(params_in) const {
    return a_.Give_data<TYP_EL>(params_out); }

  template <elementTyp TYP_EL>
  void Update(int id) const
     {
       a_.Update<TYP_EL>(id);
     }
};

template <class A, class DTyp>
  inline Expr_Equation_of_Array<A,DTyp> operator== (const Variable<DTyp>& v,
					     const Expr<A>& a){
  return Expr_Equation_of_Array<A,DTyp>(v,a);
}


//////////////////////////////////////////////////////////////
// 2. class equation of array
//////////////////////////////////////////////////////////////

template <class A, class ADTyp, class B, class BDTyp>
void Iterate_system(const Expr_Equation_of_Array<A,ADTyp>& eqnA,
	       const Expr_Equation_of_Array<B,BDTyp>& eqnB,
	       const Marker& sg) {
  int Nx, Ny, Nz, i,j,k;

  stencil_typ sten_typ;

  Blockgrid* blockgrid  = eqnA.Give_blockgrid();
  if(eqnB.Give_blockgrid()!=blockgrid) 
    cout << "error 1 in Iterate_system" << endl;
  Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();

  //  Variable<ADTyp>* vAleft = eqnA.Give_left_v();
  //  Variable<BDTyp>* vBleft = eqnB.Give_left_v();

  if(eqnA.Give_stencil_typ()== yes_stencil ||
     eqnB.Give_stencil_typ()== yes_stencil)
    sten_typ = yes_stencil;
  else
    sten_typ = no_stencil;

  // hexahedra
  ADTyp** data_hexahedraA = eqnA.Give_left_v_data_hexahedra();
  BDTyp** data_hexahedraB = eqnB.Give_left_v_data_hexahedra(); 

  for(int id=0;id<ug->Give_number_hexahedra();++id) {
    if(sten_typ == yes_stencil){
      eqnA.template Update<hexahedronEl>(id);
      eqnB.template Update<hexahedronEl>(id);
    }
    if(ug->Give_hexahedron(id)->my_object(my_rank) &&
       sg.Give_marker<hexahedronEl>(id)==yes_mark) {
      Nx = blockgrid->Give_Nx_hexahedron(id);
      Ny = blockgrid->Give_Ny_hexahedron(id);
      Nz = blockgrid->Give_Nz_hexahedron(id);


     #pragma omp parallel for private(j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
     for(k=1;k<Nz;++k) {
	for(j=1;j<Ny;++j)
	  for(i=1;i<Nx;++i) {
	    // equation A:
	    data_hexahedraA[id][i+(Nx+1)*(j+(Ny+1)*k)] =
	      eqnA.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	    // equation B:
	    data_hexahedraB[id][i+(Nx+1)*(j+(Ny+1)*k)] =
	      eqnB.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	  }
      }
    }
  }

  
  // quadrangles
  ADTyp** data_quadrangleA = eqnA.Give_left_v_data_quadrangle();
  BDTyp** data_quadrangleB = eqnB.Give_left_v_data_quadrangle();

  for(int id=0;id<ug->Give_number_quadrangles();++id) {
    if(sten_typ == yes_stencil){
      eqnA.template Update<quadrangleEl>(id);
      eqnB.template Update<quadrangleEl>(id);
    }
    if(ug->Give_quadrangle(id)->my_object(my_rank) &&
       sg.Give_marker<quadrangleEl>(id)==yes_mark) {
      Nx = blockgrid->Give_Nx_quadrangle(id);
      Ny = blockgrid->Give_Ny_quadrangle(id);

      for(j=1;j<Ny;++j)
	for(i=1;i<Nx;++i) {
	  // equation A:
	  data_quadrangleA[id][i+(Nx+1)*j] =
	    eqnA.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);

	  // equation B:
	  data_quadrangleB[id][i+(Nx+1)*j] =
	    eqnB.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	}
    }
  }

  // edges
  ADTyp** data_edgesA = eqnA.Give_left_v_data_edges(); 
  BDTyp** data_edgesB = eqnB.Give_left_v_data_edges();  

  for(int id=0;id<ug->Give_number_edges();++id) {
    if(sten_typ == yes_stencil){
      eqnA.template Update<edgeEl>(id);
      eqnB.template Update<edgeEl>(id);
    }
    if(ug->Give_edge(id)->my_object(my_rank) &&
       sg.Give_marker<edgeEl>(id)==yes_mark) {
      Nx = blockgrid->Give_Nx_edge(id);

      for(i=1;i<Nx;++i) {
	// equation A:
	data_edgesA[id][i] =
	  eqnA.template Give_data<edgeEl>(id,i,0,0,0,0);

	// equation B:
	data_edgesB[id][i] =
	  eqnB.template Give_data<edgeEl>(id,i,0,0,0,0);
      }
    }
  }

  // points
  ADTyp** data_pointsA = eqnA.Give_left_v_data_points(); 
  BDTyp** data_pointsB = eqnB.Give_left_v_data_points(); 

  for(int id=0;id<ug->Give_number_points();++id) {
    if(sten_typ == yes_stencil){
      eqnA.template Update<pointEl>(id);
      eqnB.template Update<pointEl>(id);
    }
    if(ug->Give_point(id)->my_object(my_rank) &&
       sg.Give_marker<pointEl>(id)==yes_mark) {
      // equation A:
      data_pointsA[id][0] =
	eqnA.template Give_data<pointEl>(id,0,0,0,0,0);

      // equation B:
      data_pointsB[id][0] =
	eqnB.template Give_data<pointEl>(id,0,0,0,0,0);

    }
  }
}



#endif // VARI_ARRAY_EQN_H

