/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// vari_zordering.h
//
// ------------------------------------------------------------

#ifndef VARI_ZORDERING_H
#define VARI_ZORDERING_H
#ifdef _OPENMP
#include <omp.h>
#endif
#include <stdexcept>


//////////////////////////////////////////////////////////////
// 1. Maximum at z-point
// 2. Integral at z-point
// 3. Plane_z_relaxation
//////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////
// 1. Maximum at z-point
//////////////////////////////////////////////////////////////

template <class A>
double L_infty_at_z_point(const Expr<A>& ao, double z) {
  int Nx, Ny, Nz;

  double infty = 0.0;
  double infty_total;
  double value;

  const A& a ( ao );

  Blockgrid* blockgrid  = a.Give_blockgrid();
  Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();
  MPI_Comm comm = ug->Give_MPI_comm();

  if(a.Give_stencil_typ()==yes_stencil) {
    throw(std::runtime_error("error in L_infty_at_z_point! do not apply to stencil! "));
    return 0.0;
  }

  double z_left;
  double z_right;

  for(int id=0; id<ug->Give_number_hexahedra(); ++id) {
    if(ug->Give_hexahedron(id)->my_object(my_rank)) {
      Nx = a.Give_blockgrid()->Give_Nx_hexahedron(id);
      Ny = a.Give_blockgrid()->Give_Ny_hexahedron(id);
      Nz = a.Give_blockgrid()->Give_Nz_hexahedron(id);
      
      z_left  = blockgrid->Give_coord_hexahedron(id, 0  ,0  ,0 ).z;
      z_right = blockgrid->Give_coord_hexahedron(id, 0  ,0  ,Nz).z; 
      
      if(z_left <= z && z <= z_right) {
	a.template Update<hexahedronEl>(id);
	bool weiter=true;
	for(int k=0;k<Nz && weiter;++k) {
	  z_left  = blockgrid->Give_coord_hexahedron(id,0,0, k  ).z;
	  z_right = blockgrid->Give_coord_hexahedron(id,0,0, k+1).z; 
	  if(z_left <= z && z <= z_right) {
	    weiter=false;
	    for(int j=0;j<=Ny;++j)
	      for(int i=0;i<=Nx;++i) {
		value = 
		  ABS((a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny)
		       *(z_right-z) +
		       a.template Give_data<hexahedronEl>(id,i,j,k+1,Nx,Ny)
		       *(z-z_left))
		      / (z_right-z_left));
		
		if(infty < value)	infty = value;
	      }
	  }
	}
      }
    }
  }

  MPI_Allreduce(&infty,&infty_total,1,MPI_DOUBLE,MPI_MAX,comm);

  return infty_total;
}

//////////////////////////////////////////////////////////////
// 2. Integral at z-point
//////////////////////////////////////////////////////////////

template <class A>
_TypeOf2_(A,A) Integral_at_z_point(const Expr<A>& ao, double z) {
  int Nx, Ny, Nz;

  typedef typename A::Result Result;

  Result sum = 0.0;
  Result sum_total = 0.0;
  Result value;

  const A& aa ( ao );

  Blockgrid* blockgrid  = aa.Give_blockgrid();
  Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();
  MPI_Comm comm = ug->Give_MPI_comm();

  if(aa.Give_stencil_typ()==yes_stencil) {
    std::cout << " error in Integral_at_z_point! do not apply to stencil! " << std::endl;
    return 0.0;
  }

  double z_left;
  double z_right;

  for(int id=0; id<ug->Give_number_hexahedra(); ++id) {
    if(ug->Give_hexahedron(id)->my_object(my_rank)) {
      Nx = aa.Give_blockgrid()->Give_Nx_hexahedron(id);
      Ny = aa.Give_blockgrid()->Give_Ny_hexahedron(id);
      Nz = aa.Give_blockgrid()->Give_Nz_hexahedron(id);
      
      z_left  = blockgrid->Give_coord_hexahedron(id, 0  ,0  ,0 ).z;
      z_right = blockgrid->Give_coord_hexahedron(id, 0  ,0  ,Nz).z; 
      
      if(z_left <= z && z <= z_right) {
	aa.template Update<hexahedronEl>(id);
	bool weiter=true;
	for(int k=0;k<Nz && weiter;++k) {
	  z_left  = blockgrid->Give_coord_hexahedron(id,0,0, k  ).z;
	  z_right = blockgrid->Give_coord_hexahedron(id,0,0, k+1).z; 
	  if(z_left <= z && z <= z_right) {
	    weiter=false;
	    for(int j=0;j<Ny;++j)
	      for(int i=0;i<Nx;++i) {
		value = 
		  ((aa.template Give_data<hexahedronEl>(id,i  ,j  ,k,Nx,Ny) +
		    aa.template Give_data<hexahedronEl>(id,i+1,j  ,k,Nx,Ny) +
		    aa.template Give_data<hexahedronEl>(id,i  ,j+1,k,Nx,Ny) +
		    aa.template Give_data<hexahedronEl>(id,i+1,j+1,k,Nx,Ny))
		   *(z_right-z) +
		   (aa.template Give_data<hexahedronEl>(id,i  ,j  ,k+1,Nx,Ny) +
		    aa.template Give_data<hexahedronEl>(id,i+1,j  ,k+1,Nx,Ny) +
		    aa.template Give_data<hexahedronEl>(id,i  ,j+1,k+1,Nx,Ny) +
		    aa.template Give_data<hexahedronEl>(id,i+1,j+1,k+1,Nx,Ny))
		   *(z-z_left))
		  / (z_right-z_left);
		
		
		// viereck:
                //   (NWx,NWy)   c   (NEx,NEy)
                //   d                       b
                //   (SWx,SWy)   a   (SEx,SEy)
		/////////////////
		  // Formel von Brahmagupta

		double a,b,c,d,s, area;
		
 
		double SWx = blockgrid->Give_coord_hexahedron(id,i  ,j  , k  ).x;
		double SWy = blockgrid->Give_coord_hexahedron(id,i  ,j  , k  ).y;
		double SEx = blockgrid->Give_coord_hexahedron(id,i+1,j  , k  ).x;
		double SEy = blockgrid->Give_coord_hexahedron(id,i+1,j  , k  ).y;
		double NWx = blockgrid->Give_coord_hexahedron(id,i  ,j+1, k  ).x;
		double NWy = blockgrid->Give_coord_hexahedron(id,i  ,j+1, k  ).y;
		double NEx = blockgrid->Give_coord_hexahedron(id,i+1,j+1, k  ).x;
		double NEy = blockgrid->Give_coord_hexahedron(id,i+1,j+1, k  ).y;

		d = sqrt( (SWx-NWx)*(SWx-NWx) + (SWy-NWy)*(SWy-NWy));
		a = sqrt( (SWx-SEx)*(SWx-SEx) + (SWy-SEy)*(SWy-SEy));
		c = sqrt( (NWx-NEx)*(NWx-NEx) + (NWy-NEy)*(NWy-NEy));
		b = sqrt( (NEx-SEx)*(NEx-SEx) + (NEy-SEy)*(NEy-SEy));

                s = (a+b+c+d)*0.5;

		area = sqrt((s-a)*(s-b)*(s-c)*(s-d));

		sum = sum + value * area;
	      }
	  }
	}
      }
    }
  }

  
  sum_total = Make_MPI_All_Sum<Result>::Do(sum,comm);

  return sum_total;
}



//////////////////////////////////////////////////////////////
// 3. Plane_z_relaxation
//////////////////////////////////////////////////////////////


template <class DTyp>
template <class A>
void Variable<DTyp>::Plane_z_relaxation(int iter_plane, 
					const Expr_Restriction<A>& a) {
  int Nx, Ny, Nz, i,j,k;

  stencil_typ sten_typ;

  sten_typ = a.Give_stencil_typ();


  
  Zordering* zordering = ug->Give_Zordering();
  for(int zind=0;zind<zordering->Give_number_indizes();++zind) {
    int id = zordering->Give_index(zind);

    // hexahedron
    if(zordering->Give_type(zind)==hexahedronEl) {
      if(sten_typ == yes_stencil) {
	a.template Update<hexahedronEl>(id);
      }
      if(ug->Give_hexahedron(id)->my_object(my_rank) &&
	 a.template Give_marker<hexahedronEl>(id)==yes_mark) {
	Nx = blockgrid->Give_Nx_hexahedron(id);
	Ny = blockgrid->Give_Ny_hexahedron(id);
	Nz = blockgrid->Give_Nz_hexahedron(id);
 
{ int iplane;
  
//test NOW  
//cout << " Furchtbar!!!!!!!!!!!!!!!!!  " << endl;
//cout << " ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ " << endl;
  
//#pragma omp parallel for private(iplane, j,i) if(UGBlocks::useOpenMP)
	for(k=1;k<Nz;++k) {	  
	  for(iplane=0;iplane<iter_plane;++iplane) {
	    for(j=1;j<Ny;j=j+2)
	      for(i=1;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
}	    
	    for(j=2;j<Ny;j=j+2)
	      for(i=1;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);

	    for(j=1;j<Ny;j=j+2)
	      for(i=2;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	    
	    for(j=2;j<Ny;j=j+2)
	      for(i=2;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	  }
	}
      }
    }
    

    
    // quadrangles
    if(zordering->Give_type(zind)==quadrangleEl) {
      if(sten_typ == yes_stencil){
 	       a.template  Update<quadrangleEl>(id);
      }
      if(ug->Give_quadrangle(id)->my_object(my_rank) &&
	 a.template Give_marker<quadrangleEl>(id)==yes_mark) {
	Nx = blockgrid->Give_Nx_quadrangle(id);
	Ny = blockgrid->Give_Ny_quadrangle(id);
	
	for(int iplane=0;iplane<iter_plane;++iplane) {
	  for(j=1;j<Ny;j=j+2) 
	    for(i=1;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	  
	  for(j=1;j<Ny;j=j+2) 
	    for(i=2;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	  
	  for(j=2;j<Ny;j=j+2) 
	    for(i=1;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	  for(j=2;j<Ny;j=j+2) 
	    for(i=2;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	}
      }
    }
    
  }

  






  // edges
  for(int id=0;id<ug->Give_number_edges();++id) {
    if(sten_typ == yes_stencil){
      a.template  Update<edgeEl>(id);
    }
    if(ug->Give_edge(id)->my_object(my_rank) &&
       a.template Give_marker<edgeEl>(id)==yes_mark) {
      Nx = blockgrid->Give_Nx_edge(id);

      for(i=1;i<Nx;++i)
	data_edges[id][i] =
	  a.template Give_data<edgeEl>(id,i,0,0,0,0);
    }
  }

  // points
  for(int id=0;id<ug->Give_number_points();++id) {
    if(sten_typ == yes_stencil){
      a.template  Update<pointEl>(id);
    }
    if(ug->Give_point(id)->my_object(my_rank) &&
       a.template Give_marker<pointEl>(id)==yes_mark) {
      data_points[id][0] =
	a.template Give_data<pointEl>(id,0,0,0,0,0);
    }
  }
}



template <class DTyp>
template <class A>
void Variable<DTyp>::Plane_minus_z_relaxation(int iter_plane, 
					const Expr_Restriction<A>& a) {
  int Nx, Ny, Nz, i,j,k;

  stencil_typ sten_typ;

  sten_typ = a.Give_stencil_typ();



  
  Zordering* zordering = ug->Give_Zordering();
  //  for(int zind=0;zind<zordering->Give_number_indizes();++zind) {
  for(int zind=zordering->Give_number_indizes()-1;zind>=0;--zind) {
    int id = zordering->Give_index(zind);

    // hexahedron
    if(zordering->Give_type(zind)==hexahedronEl) {
      if(sten_typ == yes_stencil) {
	a.template Update<hexahedronEl>(id);
      }
      if(ug->Give_hexahedron(id)->my_object(my_rank) &&
	 a.template Give_marker<hexahedronEl>(id)==yes_mark) {
	Nx = blockgrid->Give_Nx_hexahedron(id);
	Ny = blockgrid->Give_Ny_hexahedron(id);
	Nz = blockgrid->Give_Nz_hexahedron(id);

{ int iplane;
//#pragma omp parallel for private(iplane, j,i) if(UGBlocks::useOpenMP)
	for(k=Nz-1;k>=1;--k) {	  
	  for(iplane=0;iplane<iter_plane;++iplane) {
	    for(j=1;j<Ny;j=j+2)
	      for(i=1;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
}	    
	    for(j=2;j<Ny;j=j+2)
	      for(i=1;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);

	    for(j=1;j<Ny;j=j+2)
	      for(i=2;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	    
	    for(j=2;j<Ny;j=j+2)
	      for(i=2;i<Nx;i=i+2) 
		data_hexahedra[id][i+(Nx+1)*(j+(Ny+1)*k)] =
		  a.template Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	  }
	}
      }
    }
    

    
    // quadrangles
    if(zordering->Give_type(zind)==quadrangleEl) {
      if(sten_typ == yes_stencil){
	a.template  Update<quadrangleEl>(id);
      }
      if(ug->Give_quadrangle(id)->my_object(my_rank) &&
	 a.template Give_marker<quadrangleEl>(id)==yes_mark) {
	Nx = blockgrid->Give_Nx_quadrangle(id);
	Ny = blockgrid->Give_Ny_quadrangle(id);
	
	for(int iplane=0;iplane<iter_plane;++iplane) {
	  for(j=1;j<Ny;j=j+2) 
	    for(i=1;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	  
	  for(j=1;j<Ny;j=j+2) 
	    for(i=2;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	  
	  for(j=2;j<Ny;j=j+2) 
	    for(i=1;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	  for(j=2;j<Ny;j=j+2) 
	    for(i=2;i<Nx;i=i+2) {
	      data_quadrangles[id][i+(Nx+1)*j] =
		a.template Give_data<quadrangleEl>(id,i,j,0,Nx,0);
	    }
	}
      }
    }
    
  }

  






  // edges
  for(int id=0;id<ug->Give_number_edges();++id) {
    if(sten_typ == yes_stencil){
      a.template  Update<edgeEl>(id);
    }
    if(ug->Give_edge(id)->my_object(my_rank) &&
       a.template Give_marker<edgeEl>(id)==yes_mark) {
      Nx = blockgrid->Give_Nx_edge(id);

      for(i=1;i<Nx;++i)
	data_edges[id][i] =
	  a.template Give_data<edgeEl>(id,i,0,0,0,0);
    }
  }

  // points
  for(int id=0;id<ug->Give_number_points();++id) {
    if(sten_typ == yes_stencil){
      a.template  Update<pointEl>(id);
    }
    if(ug->Give_point(id)->my_object(my_rank) &&
       a.template Give_marker<pointEl>(id)==yes_mark) {
      data_points[id][0] =
	a.template Give_data<pointEl>(id,0,0,0,0,0);
    }
  }
}


#endif // VARI_ZORDERING_H

