/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variable.cc
//
// ------------------------------------------------------------

#include <string>
#include <stdexcept>
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "parallel.h"
#include "extemp.h"
#include "variable.h"
#include "variable_cc.h"
#include "print_var_vtk_cc.h"
#include "phaseshift_cc.h"



template <>
void Variable<double>::Update_back(int id_hex) const {
    Update_back_hex<double>(id_hex, blockgrid, 
			      data_hexahedra, data_quadrangles,
			      data_edges, data_points,
			      my_rank, comm);
}


template <>
void Variable< std::complex<double> >::Update_back(int id_hex) const {
    Update_back_hex< std::complex<double> >(id_hex, blockgrid,
                              data_hexahedra, data_quadrangles,
                              data_edges, data_points,
                              my_rank, comm);
}

template <>
void Variable<double>::AddUpFromHex(int id_hex) const {
    AddUp_back_fromHex<double>(id_hex, blockgrid,
                              data_hexahedra, data_quadrangles,
                              data_edges, data_points,
                              my_rank, comm);
}

template <>
void Variable< std::complex<double> >::AddUpFromHex(int id_hex) const {
    AddUp_back_fromHex< std::complex<double> >(id_hex, blockgrid,
                              data_hexahedra, data_quadrangles,
                              data_edges, data_points,
                              my_rank, comm);
}



template <>
template <>
void Variable<double>::Update<hexahedronEl>(int id_hex) const{
/*  if(blockgrid->Exists_phase_shift()) {
    throw(std::runtime_error(std::string("Update<hexahedronEl>: Update with shift not possible for Variable<double>")));
  }
  else {
*/
    /// TODO parallel geht nicht!!
    Update_hexahedron<double>(id_hex, blockgrid, 
			      data_hexahedra, data_quadrangles,
			      data_edges, data_points,
			      my_rank, comm);
//   }
}

template <>
template <>
void Variable<std::complex<double> >::Update<hexahedronEl>(int id_hex) const{
  if(blockgrid->Exists_phase_shift()) {
    Variable<std::complex<double> > *phase_shift  = blockgrid->Give_phase_shift();
    std::complex<double>** shift_data_hexahedra   = phase_shift->Give_pointer_data_hexahedra();
    std::complex<double>** shift_data_quadrangles = phase_shift->Give_pointer_data_quadrangle();
    
    Update_hexahedron_phase_shift<std::complex<double> >(id_hex, blockgrid, 
							 data_hexahedra, data_quadrangles, 
							 shift_data_hexahedra, 
							 shift_data_quadrangles, 
							 data_edges, 
							 data_points,
							 my_rank, comm);
  }
  else
    Update_hexahedron<std::complex<double> >(id_hex, blockgrid, 
					   data_hexahedra,
					   data_quadrangles, data_edges, 
					   data_points,
					   my_rank, comm);
}


template <>
template <>
void Variable<double>::Update<quadrangleEl>(int id_quad) const{
/*  if(blockgrid->Exists_phase_shift()) {
    throw(std::runtime_error(std::string("Update<quadrangleEl>:: Update with shift not possible for Variable<double>")));
  }
  else*/
    Update_quadrangle<double>(id_quad, blockgrid, 
			      data_hexahedra, data_quadrangles,
			      data_edges, data_points,
			      data_interior_quadrangles,
			      data_exterior_quadrangles, my_rank, comm);
}

template <>
template <>
void Variable<std::complex<double> >::Update<quadrangleEl>(int id_quad) const{
  if(blockgrid->Exists_phase_shift()) {
    Variable<std::complex<double> > *phase_shift  = blockgrid->Give_phase_shift();
    std::complex<double>** shift_data_hexahedra   = phase_shift->Give_pointer_data_hexahedra();
    std::complex<double>** shift_data_quadrangles = phase_shift->Give_pointer_data_quadrangle();
    
    Update_quadrangle_phase_shift<std::complex<double> >(id_quad, blockgrid, 
							 data_hexahedra,
							 data_quadrangles, 
							 shift_data_hexahedra, 
							 shift_data_quadrangles, 
							 data_edges, 
							 data_points,
							 data_interior_quadrangles,
							 data_exterior_quadrangles, 
							 my_rank, comm);
  }
  else  
    Update_quadrangle<std::complex<double> >(id_quad, blockgrid, 
					     data_hexahedra,
					     data_quadrangles, data_edges, 
					     data_points,
					     data_interior_quadrangles,
					     data_exterior_quadrangles, 
					     my_rank, comm);
}


template <>
template <>
void Variable<double>::Update<edgeEl>(int id_edge) const{
  Update_edge<double>(id_edge, blockgrid, data_hexahedra, data_quadrangles, data_edges,
                      data_points, data_neighbor_edges, my_rank, comm);
}


template <>
template <>
void Variable<std::complex<double> >::Update<edgeEl>(int id_edge) const{
  Update_edge<std::complex<double> >(id_edge, blockgrid, data_hexahedra,
                                     data_quadrangles, data_edges, data_points,
                                     data_neighbor_edges, my_rank, comm);
}


template <>
template <>
void Variable<double>::Update<pointEl>(int id_point) const{
  Update_point<double>(id_point, blockgrid, data_hexahedra, data_quadrangles, data_edges,
                       data_points, data_neighbor_points, my_rank, comm);
}

template <>
template <>
void Variable<std::complex<double> >::Update<pointEl>(int id_point) const{
  Update_point<std::complex<double> >(id_point, blockgrid, data_hexahedra,
                                      data_quadrangles, data_edges, data_points,
                                      data_neighbor_points, my_rank, comm);
}



template<>
bool Variable<double>::containsNaN() {
	int Nx, Ny, Nz, i, j, k;
	
	// hexahedra
	for ( int id = 0;id < ug->Give_number_hexahedra();++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			for ( k = 1;k < Nz;++k ) {
				for ( j = 1;j < Ny;++j )
					for ( i = 1;i < Nx;++i ) {
					   double x = data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ];
					   if((isInfiniteTT(x) || isIndeterminateTT(x))) return true;
					}
			}
		}
	}
	
	// quadrangles
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			
			for ( j = 1;j < Ny;++j )
				for ( i = 1;i < Nx;++i ) {
				     double x = data_quadrangles[id][i+ ( Nx+1 ) *j];
			             if((isInfiniteTT(x) || isIndeterminateTT(x))) return true;
				}
		}
		Update<quadrangleEl> ( id );
	}
	
	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {			
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			
			for ( i = 1;i < Nx;++i ) {
		             double x = data_edges[id][i];
			     if((isInfiniteTT(x) || isIndeterminateTT(x))) return true;
			}
		}
		Update<edgeEl> ( id );
	}
	
	// points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) ) {
		       double x = data_points[id][0];
	               if((isInfiniteTT(x) || isIndeterminateTT(x))) return true;
		}
		Update<pointEl> ( id );
	}
	return false;
}


template <>
Variable<double>::Variable(Blockgrid& blockgrid_, double **dataHex) :
  	Object_based_on_ug ( blockgrid_.Give_unstructured_grid()) {
	int num_total, num_neighbor;
	blockgrid = &blockgrid_;
	ug = blockgrid->Give_unstructured_grid();

        totalNumberData = 0;
        dataTotal = NULL;		
	specialVariableForOneBlock = true;

	// Kopie der UG Daten zum Löschen
	ug_number_hexahedra   = ug->Give_number_hexahedra();
	ug_number_quadrangles = ug->Give_number_quadrangles();
        ug_number_edges       = ug->Give_number_edges();
	ug_number_points      = ug->Give_number_points();
        ug_edge_number_neighbors = new int[ug_number_edges];
	idGrid = blockgrid->getId();
	
	for(int i=0;i<ug_number_edges;++i)
	    ug_edge_number_neighbors[i] = ug->Give_edge(i)->Give_number_neighbors();
	
	// for parallel
	my_rank = ug->Give_my_rank();
	comm    = ug->Give_MPI_comm();

	// hexahedra
	data_hexahedra = new double*[ug->Give_number_hexahedra()];
	for(int id=0; id< ug->Give_number_hexahedra();++id) {
		num_total = blockgrid->Give_N_total_hexahedron ( id );

		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
			data_hexahedra[id] = dataHex[id];
		} else data_hexahedra[id] = NULL;
	}	
	
	// quadrangles
	data_quadrangles = new double*[ug->Give_number_quadrangles() ];
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		num_total = blockgrid->Give_N_total_quadrangle ( id );

		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
			data_quadrangles[id] = new double[num_total];

			for ( int i=0; i<num_total; ++i )
				data_quadrangles[id][i] = 0.0;
		} else {
			data_quadrangles[id] = NULL;
		}
	}

	data_interior_quadrangles = new double*[ug->Give_number_quadrangles() ];
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		num_total = blockgrid->Give_N_total_quadrangle ( id );

		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
			data_interior_quadrangles[id] = new double[num_total];

			for ( int i=0; i<num_total; ++i )
				data_interior_quadrangles[id][i] = 0.0;
		} else data_interior_quadrangles[id] = NULL;
	}

	data_exterior_quadrangles = new double*[ug->Give_number_quadrangles() ];
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		if ( ug->Give_quadrangle ( id )->Give_exists_exterior() &&
		     ug->Give_quadrangle ( id )->my_object ( my_rank ) )        {
			num_total = blockgrid->Give_N_total_quadrangle ( id );
			data_exterior_quadrangles[id] = new double[num_total];

			for ( int i=0; i<num_total; ++i )
				data_exterior_quadrangles[id][i] = 0.0;
		} else data_exterior_quadrangles[id] = NULL;
	}

	// edges
	data_edges = new double*[ug->Give_number_edges() ];
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		num_total = blockgrid->Give_Nx_edge ( id ) +1;
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			data_edges[id] = new double[num_total];

			for ( int i=0; i<num_total; ++i )
				data_edges[id][i] = 0.0;
		} else data_edges[id] = NULL;
	}


	data_neighbor_edges = new double**[ug->Give_number_edges() ];
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		num_neighbor = ug->Give_edge ( id )->Give_number_neighbors();

		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			data_neighbor_edges[id] = new double*[num_neighbor];
			for ( int i=0; i<num_neighbor; ++i ) {
				num_total = blockgrid->Give_Nx_edge ( id ) +1;
				data_neighbor_edges[id][i] = new double[num_total];

				for ( int j=0; j<num_total; ++j )
					data_neighbor_edges[id][i][j] = 0.0;
			}
		} else data_neighbor_edges[id] = NULL;
	}

	// points
	data_points = new double*[ug->Give_number_points() ];
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		num_total = 1;

		if ( ug->Give_point ( id )->my_object ( my_rank ) ) {
			data_points[id] = new double[num_total];

			for ( int i=0; i<num_total; ++i )
				data_points[id][i] = 0.0;
		} else data_points[id] = NULL;
	}
	data_neighbor_points = new double*[ug->Give_number_points() ];
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		num_neighbor = ug->Give_point ( id )->Give_number_neighbors();

		if ( ug->Give_point ( id )->my_object ( my_rank ) ) {
			data_neighbor_points[id] = new double[num_neighbor];

			for ( int i=0; i<num_neighbor; ++i )
				data_neighbor_points[id][i] = 0.0;
		} else data_neighbor_points[id] = NULL;
	}
	
	for(int id=0; id< ug->Give_number_hexahedra();++id)
	    Update_back(id);
}



