/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variable.h
//
// ------------------------------------------------------------

#include <limits>
//#include <math.h>


#ifndef VARIABLE_H
#define VARIABLE_H
#ifdef _OPENMP
#include <omp.h>
#else
#include "../myopenmp.h"
#endif


#include "../ugopenmp.h"

#include <complex>
using std::complex;
//////////////////////////////////////////////////////////////
// 1. Restriction to subgrid
// 2. Variable for points
// 3. Other funktions like max, infty ...
// 4. Implementation of inline functions, operator=, ...
// 5. Nconst , ... funtions for copy of data
//////////////////////////////////////////////////////////////

// Gitter:
// * boundary point
// + interior point

/*
Ny    *  *  *  *  *  *
Ny-1  *  +  +  +  +  *
3     *  +  +  +  +  *
2     *  +  +  +  +  *
1     *  +  +  +  +  *
0     *  *  *  *  *  *
      0  1  2  3 Nx-1 Nx
*/
// k=0...Nz            for(k=0;k<=Nz;++k)
//   j=0...Ny            for(j=0;j<=Ny;++j)
//     i=0...Nx            for(i=0;i<=Nx;++i)
//       index = i + (Nx+1)*(j+k*(Ny+1))


///////////////////////////
// 0. Predefinitions
///////////////////////////

class IteratorZDirection;

template <typename TYPE>  class ExpressionRestriction;

template <typename TYPE>  class ExpressionProlongation;

//template <typename TYPE>  class Expr_interpolant_cell_to_point;

template <class A, class DTyp>  class Expr_Equation_of_Array;

class QString;

class IteratorZDirection;

template <class DTyp> class Variable2D;
template <class A> struct Expr2D;


template <class Typ> class vtkSmartPointer;
class vtkUnstructuredGrid;

template <class A, class B>
_TypeOf2_ ( A, B ) product_on ( Expr<A>& ao, Expr<B>& bo, Marker& marker );

///////////////////////////
// 1.a Restriction to subgrid
///////////////////////////

template <class A>
class Expr_Restriction {
		const A& a_;
		const Marker& marker;
		
	public:
		Expr_Restriction ( const Expr<A>& a, const Marker& mark ) : a_ ( a ), marker ( mark ) {}
		
		Blockgrid* Give_blockgrid() const { return a_.Give_blockgrid(); };
		
		stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ(); }
		
		typedef typename A::Result Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out );
		}
		
		inline Result Give_cell_hexahedra ( params_in_cell ) const  {
                       return a_.Give_cell_hexahedra ( params_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
		
		template <elementTyp TYP_EL>
		marker_typ Give_marker ( int i ) const{ return marker.Give_marker<TYP_EL> ( i ); }
};

template <class A>
inline Expr_Restriction<A> operator| ( const Expr<A>& a, const Marker& sg ) {
	return Expr_Restriction<A> ( a, sg );
}

template <typename DTyp>
class DTyp_Restriction {
		DTyp x_;
		const Marker& marker;
		
	public:
		DTyp_Restriction ( DTyp x, const Marker& mark ) : x_ ( x ), marker ( mark ) {}
		
		~DTyp_Restriction() {}
		
		template <elementTyp TYP_EL>
		marker_typ Give_marker ( int i ) const{ return marker.Give_marker<TYP_EL> ( i ); }
		
		inline DTyp Give_x() const { return x_; }
};

inline DTyp_Restriction<double> operator| ( double x, Marker& sg ) {
	return DTyp_Restriction<double> ( x, sg );
}

inline DTyp_Restriction<std::complex<double> > operator| ( std::complex<double> x, Marker& sg ) {
	return DTyp_Restriction<std::complex<double> > ( x, sg );
}


///////////////////////////
// 1.b Restriction to subgrid and iteration description
///////////////////////////

class Iteration_description {

	public:
		Iteration_description ( Marker& m ) : marker ( &m ) { red_black = false; }
		
		Marker* marker;
		bool red_black;
};


template <class A>
class Expr_Restriction_Iterdes {
		const A& a_;
		
	public:
		Expr_Restriction_Iterdes ( const Expr<A>& a, const Iteration_description& ides ) : a_ ( a ), iterdes ( ides ) {}
		
		const Iteration_description& iterdes;
		
		Blockgrid* Give_blockgrid() const { return a_.Give_blockgrid(); };
		
		stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ(); }
		
		typedef typename A::Result Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const {
			return a_.template Give_data<TYP_EL> ( params_out );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
		
		template <elementTyp TYP_EL>
		marker_typ Give_marker ( int i ) const{ return iterdes.marker->Give_marker<TYP_EL> ( i ); }
};

template <class A>
inline Expr_Restriction_Iterdes<A> operator| ( const Expr<A>& a, const Iteration_description& iterdes ) {
	return Expr_Restriction_Iterdes<A> ( a, iterdes );
}


///////////////////////////
// 2. Variable for points
///////////////////////////

class Assign_System;

template <class DTyp, class Vari>
class VariableVector;

/** \addtogroup ExpressionTemplates **/
/* @{ */ 
template <class DTyp>
class Variable : public Expr<Variable< DTyp > >, public Object_based_on_ug {
		template <class A, class ADTyp>
		
		friend class Expr_Equation_of_Array;		
		class Assign_System;
		
                friend  VariableVector<DTyp, Variable<DTyp> >;		
		
	public:
		int checkThreshold(double threshold, int kz );
		
		typedef DTyp Result;
		
		//  template <class DTyp_>
		//  friend DTyp_ product(Variable<DTyp_>& a, Variable<DTyp_>& b, Marker& marker);
		
		Variable(Blockgrid& blockgrid_ );
		
		/**
		 * Ganz spezieller Konstruktor fuer vtKReader
		 * äusseren Pointer (dataHex) darf man nach Aufruf löschen
		 * Daten zu innerem Pointer (*dataHex) werden verwendet und darf man nach Aufrud nicht löschen
		 **/
		Variable(Blockgrid& blockgrid_, double **dataHex);		
		
		
		~Variable();
		void Delete_data();

		int getGridId() { return idGrid; }
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		Unstructured_grid * Give_Ug() const { return ug;};
		
		void interpolateSlizeZ(const Variable2D<DTyp>* U, IteratorZDirection* iterZ);
		
		template <class A>
		void interpolateSlizeZ(const Expr2D<A>& a_, IteratorZDirection* iterZ);

		bool containsNaN();
		
		void Print_AVS ( std::ofstream *Datei );
		void Print_Dx ( std::ofstream *Datei );
		
		// QT-VTK  
		void QPrint_VTK(QString DateiName, double stretch_z = 1., QString title = QString("myfile"),  Unstructured_grid_Marker * marker = NULL );
		void QPrint_VTK(QString DateiName, double ( *convert ) ( DTyp x ), double stretch_z = 1., QString title = "myData", Unstructured_grid_Marker * marker = NULL, bool cutSmallData = true);
		void QPrint_VTK(QString DateiName, complex<double> ( *convert ) ( DTyp x ), double stretch_z = 1., QString title = "myData", Unstructured_grid_Marker * marker = NULL );
		void QPrint_VTK(QString DateiName, int z_cut, QString title = "myData" );
		
		// VTK output  
		vtkUnstructuredGrid* GetVtkOutput(double stretch_z = 1., QString title = QString("myfile"),  Unstructured_grid_Marker * marker = NULL );
		
		// z_cut = Nz, falls z_cut zu gross gewaehlt
		void Print_VTK ( std::ostream& Datei, int z_cut, std::string title = "myData" );
		void Print_VTK ( std::ostream& Datei, double ( *convert ) ( DTyp x ), double stretch_z = 1., std::string title = "myData", Unstructured_grid_Marker * marker = NULL );
		
		/// Prints absolute value in case of complex variable
		void Print_VTK ( std::ostream& Datei, double stretch_z = 1., std::string title = "myData",  Unstructured_grid_Marker * marker = NULL );
		
		
		///Binary VTK
		void Print_VTK_BIN ( std::ostream& Datei, double stretch_z = 1., std::string title = "myData",  Unstructured_grid_Marker * marker = NULL );
		
		void Print_AVS ( Variable<double>& ux, Variable<double>& uy, Variable<double>& uz, std::ofstream *Datei );
		void Print_Dx ( Variable<double>& ux, Variable<double>& uy, Variable<double>& uz, std::ofstream *Datei );
		
		// copy of data from hex_a to hex_b
		void Copy_invert_z ( int hex_b, Variable<DTyp>& a, int hex_a );
		
		template <elementTyp TYP_EL>
		inline DTyp Give_data ( params_in ) const;

        template <elementTyp TYP_EL>
        inline void Set_data ( params_in, DTyp value );
		
		
		// fuer debug-zwecke mit positionsrueckgabe
		template <elementTyp TYP_EL>
		inline DTyp Give_data ( params_in, D3vector &posVector ) const ;
		
		void UpdateHexahedra(); ///> Updated alle Randdaten in allen Hexahedra
                void UpdateHexahedraBack();
		/// todo update nur falls notwendig
		template <elementTyp TYP_EL>
		void Update ( int id ) const;
		
		void Update_back(int id) const; ///> fuer interpolation slice 2D -> 3D
				
				
                void AddUpFromHex();             ///> fuer Interpolation cell -> point
                void AddUpFromHex(int id) const; ///> fuer Interpolation cell -> point	
                void SetAllZero();               ///> fuer Interpolation cell -> point	                
                void AddFromOne();               ///> fuer Interpolation cell -> point	
	        template <class Zell>
	        void AddFromZell(Zell& ZellVar); ///> fuer Interpolation cell -> point	
				
		
		inline DTyp Give_data_quadrangle_middle ( int id, int i, int j, int Nx ) const;
		inline DTyp Give_data_quadrangle_interior ( int id, int i, int j, int Nx ) const;
		inline DTyp Give_data_quadrangle_exterior ( int id, int i, int j, int Nx ) const;
		
		inline DTyp Give_data_edge ( int id, int i ) const;
		inline DTyp Give_data_neighbor_edge ( int id, int num, int i ) const;
		
		inline DTyp Give_data_point ( int id ) const;
		inline DTyp Give_data_neighbor_point ( int id, int num ) const;
		
		inline DTyp** Give_pointer_data_points() const { return data_points; };		
		inline DTyp** Give_pointer_data_edges() const { return data_edges; };
		inline DTyp** Give_pointer_data_hexahedra() const { return data_hexahedra; };		
		inline DTyp** Give_pointer_data_quadrangle() const { return data_quadrangles; };
		
		void operator= ( const DTyp_Restriction<DTyp>& a );
		
		template <class A>
		void operator= ( const Expr_Restriction<A>& a );

		// wie oben nur Iteration Rückwärts
		template <class A>
		void operator== ( const Expr_Restriction<A>& a );		
		
		template <class A>
		void Plane_z_relaxation ( int iter_plane, const Expr_Restriction<A>& a );
		
		template <class A>
		void Plane_minus_z_relaxation ( int iter_plane, const Expr_Restriction<A>& a );
		
		template <class A>
		void operator= ( const Expr_Restriction_Iterdes<A>& a );
		
		template <class A>
		void operator= ( const Expr<A>& a );
		
		void operator= ( const Variable<DTyp>& v );
		void operator= ( DTyp x );
		
		void operator= ( const ExpressionRestriction<DTyp>& Expr_res );
		void operator= ( const ExpressionProlongation<DTyp>& Expr_prol );
//                void operator= ( const Expr_interpolant_cell_to_point<DTyp>& Expr_cell );
		
		
		static DTyp product(Variable& a, Variable& b) { return product_on(a,b,a.Give_blockgrid()->Give_unstructured_grid()->Give_all_points()); }

		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		// copies data from a in hex_a to data in hex_b
		void Copy_invert_z ( int hex_b, Variable<DTyp> a, int hex_a );
		
                int  getTotalNumberData()   const { return totalNumberData; }		
	private:
	        inline DTyp* give_startTotal() const  { assert(dataTotal!=NULL);  return dataTotal; }  	
	        
		Blockgrid*        blockgrid;
		Unstructured_grid*       ug;
		
        // own data
		DTyp** data_hexahedra;        // num_hexahedra
		DTyp** data_quadrangles;      // num_quadrangles
		DTyp** data_edges;            // num_edges
		DTyp** data_points;           // num_points, first own data
		// rest unstructured
		
		// neighbor unstructured data
		DTyp**  data_interior_quadrangles;       // num_quadrangles
		DTyp**  data_exterior_quadrangles;       // num_quadrangles
		
		DTyp*** data_neighbor_edges;             // num_edges
		
		DTyp**  data_neighbor_points;            // num_points
		
		//  DTyp**  data_triangles;                // num_triangles
		//  DTyp**  data_interior_triangles;       // num_triangles
		//  DTyp**  data_exterior_triangles;       // num_triangles
		

		
		// for parallel
		int my_rank;
		MPI_Comm comm;
		
		// zum Löschen von Daten Kopie der UG_Blocks Daten
		int  ug_number_hexahedra;
		int  ug_number_quadrangles;
		int  ug_number_edges;
		int* ug_edge_number_neighbors;  // array of size ug_number_edges
		int  ug_number_points;
		int  idGrid;
		
		bool isIndeterminateTT(const double pV) { return (pV != pV);  } 
                bool isInfiniteTT(const double pV) { return (fabs(pV) == std::numeric_limits<double>::infinity());  }
                
		// own data
                int totalNumberData;
                DTyp* dataTotal;
                bool specialVariableForOneBlock; //> dann ist dataTotal nicht verwendbar
};
/* @} */	


template <class DTyp>
void Variable<DTyp>::UpdateHexahedra() {
     for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
         Update<hexahedronEl>(id_hex);
     }
}

template <class DTyp>
void Variable<DTyp>::UpdateHexahedraBack() {
     for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
         Update_back(id_hex);
     }
}

template <class DTyp>
void Variable<DTyp>::AddUpFromHex() {
     for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
         AddUpFromHex(id_hex);
     }
}



template <class DTyp>
int Variable<DTyp>::checkThreshold ( double threshold, int kz ) {
			int id_hex = 1;

			int Ny = blockgrid->Give_Ny_hexahedron ( id_hex );
			int Nx = blockgrid->Give_Nx_hexahedron ( id_hex );
			D3vector v;
			DTyp temp;

			Update<hexahedronEl> ( id_hex );
			
			int thresholdReached = 0;

			int k = kz;
			
			for ( int j = 0;j <= Ny;++j ) {
				for ( int i = 0;i <= Nx;++i ) {
					temp = ( data_hexahedra[id_hex][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] );
					
					if ( ABS ( temp ) > threshold ) {
						thresholdReached++;
						std::cout << "Variable: real(value) = " << real ( temp ) <<std::endl;
						v = blockgrid->Give_coord_hexahedron ( id_hex, i, j, k );
						std::cout << " at coordinate (" << v.x
						<< "," << v.y
						<< "," << v.z
						<< ")." <<std::endl;
						std::cout << "with indizes (" << i
						<< "," << j
						<< "," << k
						<< ")." <<std::endl;;
					}
				}
			}
			return thresholdReached;
}


		
template <>
template <>
inline double Variable<double>::Give_data<pointEl> ( params_in ) const {
	return data_points[id][0];
}

template <>
template <>
inline std::complex<double> Variable<std::complex<double> >::Give_data<pointEl> ( params_in ) const {
	return data_points[id][0];
}

template <>
template <>
inline double Variable<double>::Give_data<edgeEl> ( params_in ) const {
	return data_edges[id][i];
}

template <>
template <>
inline std::complex<double> Variable<std::complex<double> >::Give_data<edgeEl> ( params_in ) const {
	return data_edges[id][i];
}

template <>
template <>
inline double Variable<double>::Give_data<quadrangleEl> ( params_in ) const {
	return data_quadrangles[id][i+ ( Nx+1 ) *j];
}

template <>
template <>
inline std::complex<double> Variable<std::complex<double> >::Give_data<quadrangleEl> ( params_in ) const {
	return data_quadrangles[id][i+ ( Nx+1 ) *j];
}

template <>
template <>
inline double Variable<double>::Give_data<hexahedronEl> ( params_in ) const {
	return data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ];
}

template <>
template <>
inline void Variable<double>::Set_data<hexahedronEl> ( params_in, double value )  {
     data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] = value;
}

// fuer debug zwecke koordinatenrueckgabe
template <>
template <>
inline std::complex<double> Variable<std::complex<double> >::Give_data<hexahedronEl> ( params_in, D3vector &posVector ) const {
	posVector = blockgrid->Give_coord_hexahedron ( id, i, j, k );
	return data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ];
	;
}

// fuer debug zwecke koordinatenrueckgabe
template <>
template <>
inline double Variable<double >::Give_data<hexahedronEl> ( params_in, D3vector &posVector ) const {
	posVector = blockgrid->Give_coord_hexahedron ( id, i, j, k );
	return data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ];
	;
}


template <>
template <>
inline std::complex<double> Variable<std::complex<double> >::Give_data<hexahedronEl> ( params_in ) const {
	return data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ];
}


///////////////////////////
// 3. Other functions, max, ...
///////////////////////////


/** \addtogroup ExpressionTemplates **/
/* @{ */  
template <class A, class B>
_TypeOf2_ ( A, B ) product ( Expr<A>& a, Expr<B>& b, Marker& marker );

template <class A, class B>
_TypeOf2_ ( A, B ) product ( Expr<A>& a, Expr<B>& b );



template <class A>
double L_infty ( const Expr<A>& a, Marker& marker );

template <class A>
double L_infty ( const Expr<A>& a );


template <class A>
double Maximum ( const Expr<A>& a, Marker& marker );

template <class A>
double Maximum ( const Expr<A>& a );


template <class A>
double Minimum ( const Expr<A>& a, Marker& marker );

template <class A>
double Minimum ( const Expr<A>& a );
/* @} */


//////////////////////////////////////////
// and it's implementation

template <class A, class B>
_TypeOf2_ ( A, B ) product ( Expr<A>& a, Expr<B>& b ) {
	const A& ao ( a );
	
	return product_on ( a, b, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}

template <class A>
double  L_infty ( const Expr<A>& a ) {
	const A& ao ( a );
	
	return L_infty ( a, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}


template <class A>
double  Minimum ( const Expr<A>& a ) {
	const A& ao ( a );
	
	return Minimum ( a, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}


template <class A>
double  Maximum ( const Expr<A>& a ) {
	const A& ao ( a );
	
	return Maximum ( a, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}

template <class A, class B>
_TypeOf2_ ( A, B ) product ( Expr<A>& ao, Expr<B>& bo, Marker& marker ) {
  return product_on(ao,bo,marker);
}


template <class A, class B>
_TypeOf2_ ( A, B ) product_on ( Expr<A>& ao, Expr<B>& bo, Marker& marker ) {
        int i, j, k, id;
        int Nx, Ny, Nz;

        typedef _TypeOf2_ ( A, B ) DTyp;
        const A& a ( ao );
        const B& b ( bo );

        DTyp sum;
        DTyp sum_total;

        sum = ( DTyp ) 0;

        if ( developer_version )
                if ( a.Give_blockgrid() != b.Give_blockgrid() ) std::cout << " error2 in product!" << std::endl;


        Unstructured_grid* ug = a.Give_blockgrid()->Give_unstructured_grid();

        int my_rank   = ug->Give_my_rank();

        MPI_Comm comm = ug->Give_MPI_comm();

        for ( id = 0;id < ug->Give_number_hexahedra();++id ) {
                if ( ug->Give_hexahedron ( id )->my_object ( my_rank )   &&
                        marker.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_hexahedron ( id );
                        Ny = a.Give_blockgrid()->Give_Ny_hexahedron ( id );
                        Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );

/// @todo: problem mit kombination aus reduction (+:sum) und MyConj
//#pragma omp parallel for private(k,j,i) reduction(+ : sum) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

                        for ( k = 1;k < Nz;++k )
                                for ( j = 1;j < Ny;++j )
                                        for ( i = 1;i < Nx;++i )
                                                sum += a.template Give_data<hexahedronEl> ( params_out )
                                                       * MyConj<DTyp>::Give ( b.template Give_data<hexahedronEl> ( params_out ) );
                }
        }

        for ( id = 0;id < ug->Give_number_quadrangles();++id ) {
                if ( ug->Give_quadrangle ( id )->my_object ( my_rank )   &&
                        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
                        Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );

                        for ( j = 1;j < Ny;++j )
                                for ( i = 1;i < Nx;++i )
                                        sum += a.template Give_data<quadrangleEl> ( params_out )
                                               * MyConj<DTyp>::Give ( b.template Give_data<quadrangleEl> ( params_out ) );
                }
        }

        for ( id = 0;id < ug->Give_number_edges();++id ) {
                if ( ug->Give_edge ( id )->my_object ( my_rank )    &&
                        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

                        for ( i = 1;i < Nx;++i )
                                sum += a.template Give_data<edgeEl> ( params_out )
                                       * MyConj<DTyp>::Give ( b.template Give_data<edgeEl> ( params_out ) );
                }
        }

        for ( id = 0;id < ug->Give_number_points();++id ) {
                if ( ug->Give_point ( id )->my_object ( my_rank )    &&
                        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
                        sum += a.template Give_data<pointEl> ( params_out )
                               * MyConj<DTyp>::Give ( b.template Give_data<pointEl> ( params_out ) );
                }
        }



        sum_total = Make_MPI_All_Sum<DTyp>::Do ( sum, comm );

        return sum_total;
}

template <class A, class B>
_TypeOf2_ ( A, B ) product ( const Expr<A>& ao, const Expr<B>& bo, Marker& marker ) {
	int i, j, k, id;
	int Nx, Ny, Nz;
	
	typedef _TypeOf2_ ( A, B ) DTyp;
	const A& a ( ao );
	const B& b ( bo );
	
	DTyp sum;
	DTyp sum_total;
	
	sum = ( DTyp ) 0;
	
	if ( developer_version )
		if ( a.Give_blockgrid() != b.Give_blockgrid() ) std::cout << " error2 in product!" <<std::endl;
		
		
	Unstructured_grid* ug = a.Give_blockgrid()->Give_unstructured_grid();
	
	int my_rank   = ug->Give_my_rank();
	
	MPI_Comm comm = ug->Give_MPI_comm();
	
	for ( id = 0;id < ug->Give_number_hexahedra();++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank )   &&
		        marker.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_hexahedron ( id );
			Ny = a.Give_blockgrid()->Give_Ny_hexahedron ( id );
			Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );
			
/// @todo: problem mit kombination aus reduction (+:sum) und MyConj
//#pragma omp parallel for private(k,j,i) reduction(+ : sum) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

			for ( k = 1;k < Nz;++k )
				for ( j = 1;j < Ny;++j )
					for ( i = 1;i < Nx;++i )
						sum += a.template Give_data<hexahedronEl> ( params_out )
						       * MyConj<DTyp>::Give ( b.template Give_data<hexahedronEl> ( params_out ) );
		}
	}
	
	for ( id = 0;id < ug->Give_number_quadrangles();++id ) {
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank )   &&
		        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
			Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );
			
			for ( j = 1;j < Ny;++j )
				for ( i = 1;i < Nx;++i )
					sum += a.template Give_data<quadrangleEl> ( params_out )
					       * MyConj<DTyp>::Give ( b.template Give_data<quadrangleEl> ( params_out ) );
		}
	}
	
	for ( id = 0;id < ug->Give_number_edges();++id ) {
		if ( ug->Give_edge ( id )->my_object ( my_rank )    &&
		        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_edge ( id );
			
			for ( i = 1;i < Nx;++i )
				sum += a.template Give_data<edgeEl> ( params_out )
				       * MyConj<DTyp>::Give ( b.template Give_data<edgeEl> ( params_out ) );
		}
	}
	
	for ( id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank )    &&
		        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
			sum += a.template Give_data<pointEl> ( params_out )
			       * MyConj<DTyp>::Give ( b.template Give_data<pointEl> ( params_out ) );
		}
	}
	
	
	
	sum_total = Make_MPI_All_Sum<DTyp>::Do ( sum, comm );
	
	return sum_total;
}


template <class A>
double L_infty ( const Expr<A>& ao, Marker& marker ) {
	int i, j, k, id;
	int Nx, Ny, Nz;
	
	double infty;
	double infty_total;
	
	const A& a ( ao );
	
	infty = 0.0;
	
	Unstructured_grid* ug = a.Give_blockgrid()->Give_unstructured_grid();
	int my_rank   = ug->Give_my_rank();
	MPI_Comm comm = ug->Give_MPI_comm();
	
	for ( id = 0; id < ug->Give_number_hexahedra(); ++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_hexahedron ( id );
			Ny = a.Give_blockgrid()->Give_Ny_hexahedron ( id );
			Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );
			
			for ( k = 1;k < Nz;++k )
				for ( j = 1;j < Ny;++j )
					for ( i = 1;i < Nx;++i )
						if ( infty < ABS ( a.template Give_data<hexahedronEl> ( params_out ) ) )
							infty = ABS ( a.template Give_data<hexahedronEl> ( params_out ) );
		}
	}
	
	for ( id = 0; id < ug->Give_number_quadrangles(); ++id ) {
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
			Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );
			
			for ( j = 1;j < Ny;++j )
				for ( i = 1;i < Nx;++i )
					if ( infty < ABS ( a.template Give_data<quadrangleEl> ( params_out ) ) )
						infty = ABS ( a.template Give_data<quadrangleEl> ( params_out ) );
						
		}
	}
	
	for ( id = 0; id < ug->Give_number_edges(); ++id ) {
		if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_edge ( id );
			
			for ( i = 1;i < Nx;++i )
				if ( infty < ABS ( a.template Give_data<edgeEl> ( params_out ) ) )
					infty = ABS ( a.template Give_data<edgeEl> ( params_out ) );
		}
	}
	
	for ( id = 0; id < ug->Give_number_points(); ++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
			if ( infty < ABS ( a.template Give_data<pointEl> ( params_out ) ) )
				infty = ABS ( a.template Give_data<pointEl> ( params_out ) );
		}
	}
	
	MPI_Allreduce ( &infty, &infty_total, 1, MPI_DOUBLE, MPI_MAX, comm );
	
	return infty_total;
}

template <class A>
double Minimum ( const Expr<A>& ao, Marker& marker ) {
	int i, j, k, id;
	int Nx, Ny, Nz;
	
	double minimum = 1e60;
	double minimum_total;
	
	const A& a ( ao );
	
	Unstructured_grid* ug = a.Give_blockgrid()->Give_unstructured_grid();
	int my_rank   = ug->Give_my_rank();
	MPI_Comm comm = ug->Give_MPI_comm();
	
	int set       = false;
	int set_total = false;

	for ( id = 0; id < ug->Give_number_hexahedra(); ++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_hexahedron ( id );
			Ny = a.Give_blockgrid()->Give_Ny_hexahedron ( id );
			Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );
			
			for ( k = 1;k < Nz;++k )
				for ( j = 1;j < Ny;++j )
					for ( i = 1;i < Nx;++i ) {
						if ( minimum > myREAL ( a.template Give_data<hexahedronEl> ( params_out ) ) ||
						        set == false ) {
							minimum = myREAL ( a.template Give_data<hexahedronEl> ( params_out ) );
							set = true;
						}
					}
		}
	}


	for ( id = 0; id < ug->Give_number_quadrangles(); ++id ) {
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
			Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );
			
			for ( j = 1;j < Ny;++j )
				for ( i = 1;i < Nx;++i ) {
					if ( minimum > myREAL ( a.template Give_data<quadrangleEl> ( params_out ) ) ||
					        set == false ) {
						minimum = myREAL ( a.template Give_data<quadrangleEl> ( params_out ) );
						set = true;
					}
				}
		}
	}
	
	for ( id = 0; id < ug->Give_number_edges(); ++id ) {
		if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_edge ( id );
			
			for ( i = 1;i < Nx;++i ) {
				if ( minimum > myREAL ( a.template Give_data<edgeEl> ( params_out ) ) ||
				        set == false ) {
					minimum = myREAL ( a.template Give_data<edgeEl> ( params_out ) );
					set = true;
				}
			}
		}
	}
	
	for ( id = 0; id < ug->Give_number_points(); ++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
			if ( minimum > myREAL ( a.template Give_data<pointEl> ( params_out ) ) ||
			        set == false ) {
				minimum = myREAL ( a.template Give_data<pointEl> ( params_out ) );
				set = true;
			}
		}
	}
	
	MPI_Allreduce ( &minimum, &minimum_total, 1, MPI_DOUBLE, MPI_MIN, comm );
	
	int p;
	
	MPI_Comm_size ( comm, &p );
	
	MPI_Allreduce ( &set, &set_total, 1, MPI_INT, MPI_SUM, comm );
	
	if ( set_total != p && minimum_total > 1e59 ) std::cout << "error in Minimum" << std::endl;
	
	return minimum_total;
}

template <class A>
double Maximum ( const Expr<A>& ao, Marker& marker ) {
	int i, j, k, id;
	int Nx, Ny, Nz;
	
	double maximum = -1e60;
	double maximum_total;
	
	const A& a ( ao );
	
	Unstructured_grid* ug = a.Give_blockgrid()->Give_unstructured_grid();
	int my_rank   = ug->Give_my_rank();
	MPI_Comm comm = ug->Give_MPI_comm();
	
	int set = false;
	int set_total = false;
	
	for ( id = 0; id < ug->Give_number_hexahedra(); ++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
		        marker.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_hexahedron ( id );
			Ny = a.Give_blockgrid()->Give_Ny_hexahedron ( id );
			Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );
			
			for ( k = 1;k < Nz;++k )
				for ( j = 1;j < Ny;++j )
					for ( i = 1;i < Nx;++i )
						if ( maximum < a.template Give_data<hexahedronEl> ( params_out ) ||
						        set == false ) {
							maximum = myREAL ( a.template Give_data<hexahedronEl> ( params_out ) );
							set = true;
						}
		}
	}
	
	for ( id = 0; id < ug->Give_number_quadrangles(); ++id ) {
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank )  &&
		        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
			Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );
			
			for ( j = 1;j < Ny;++j )
				for ( i = 1;i < Nx;++i )
					if ( maximum < a.template Give_data<quadrangleEl> ( params_out ) ||
					        set == false ) {
						maximum = myREAL ( a.template Give_data<quadrangleEl> ( params_out ) );
						set = true;
					}
		}
	}
	
	for ( id = 0; id < ug->Give_number_edges(); ++id ) {
		if ( ug->Give_edge ( id )->my_object ( my_rank )  &&
		        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = a.Give_blockgrid()->Give_Nx_edge ( id );
			
			for ( i = 1;i < Nx;++i )
				if ( maximum < a.template Give_data<edgeEl> ( params_out ) ||
				        set == false ) {
					maximum = myREAL ( a.template Give_data<edgeEl> ( params_out ) );
					set = true;
				}
		}
	}
	
	for ( id = 0; id < ug->Give_number_points(); ++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank )  &&
		        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
			if ( maximum < a.template Give_data<pointEl> ( params_out ) ||
			        set == false ) {
				maximum = myREAL ( a.template Give_data<pointEl> ( params_out ) );
				set = true;
			}
		}
	}
	
	MPI_Allreduce ( &maximum, &maximum_total, 1, MPI_DOUBLE, MPI_MAX, comm );
	
	int p;
	MPI_Comm_size ( comm, &p );
	
	MPI_Allreduce ( &set, &set_total, 1, MPI_INT, MPI_SUM, comm );
	
	if ( set_total != p && maximum_total < -1e59 ) std::cout << "error in Maximum! " <<std::endl;
	
	//max_total = " << maximum_total << ".\t set_total = " << set_total << ".\t p = " << p  <<std::endl;
	
	
	return maximum_total;
}

//////////////////////////////////////////
// 4. Implementation of inline functions
//////////////////////////////////////////

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_quadrangle_middle ( int id, int i, int j, int Nx ) const {
	return data_quadrangles[id][i+ ( Nx+1 ) *j];
}

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_quadrangle_interior ( int id, int i, int j, int Nx ) const {
	return data_interior_quadrangles[id][i+ ( Nx+1 ) *j];
}

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_quadrangle_exterior ( int id, int i, int j, int Nx ) const {
	return data_exterior_quadrangles[id][i+ ( Nx+1 ) *j];
}

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_edge ( int id, int i ) const {
	return data_edges[id][i];
}

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_neighbor_edge ( int id, int num, int i ) const {
	return data_neighbor_edges[id][num][i];
}

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_point ( int id ) const {
	return data_points[id][0];
}

template <class DTyp>
inline DTyp Variable<DTyp>::Give_data_neighbor_point ( int id, int num ) const {
	return data_neighbor_points[id][num];
}

template <class DTyp>
template <class A>
void Variable<DTyp>::operator= ( const Expr_Restriction<A>& a ) {
	int Nx, Ny, Nz, i, j, k;
	
	stencil_typ sten_typ = a.Give_stencil_typ();
	
	// hexahedra
	
	for ( int id = 0;id < ug->Give_number_hexahedra();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<hexahedronEl> ( id );
		}
		
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
#pragma omp parallel for private(j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
			for ( k = 1;k < Nz;++k ) {
				for ( j = 1;j < Ny;++j )
					for ( i = 1;i < Nx;++i )
						data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
						    a.template Give_data<hexahedronEl> ( id, i, j, k, Nx, Ny );
			}
		}
	}
	
	// quadrangles
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<quadrangleEl> ( id );
		}
		
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			
			for ( j = 1;j < Ny;++j )
				for ( i = 1;i < Nx;++i ) {
					data_quadrangles[id][i+ ( Nx+1 ) *j] =
					    a.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
				}
		}
	}
	
	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<edgeEl> ( id );
		}
		
		if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			
			for ( i = 1;i < Nx;++i )
				data_edges[id][i] =
				    a.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
		}
	}
	
	// points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<pointEl> ( id );
		}
		
		if ( ug->Give_point ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<pointEl> ( id ) == yes_mark ) {
			data_points[id][0] =
			    a.template Give_data<pointEl> ( id, 0, 0, 0, 0, 0 );
		}
	}
}

template <class DTyp>
template <class A>
void Variable<DTyp>::operator== ( const Expr_Restriction<A>& a ) {
	int Nx, Ny, Nz, i, j, k;
	
	stencil_typ sten_typ = a.Give_stencil_typ();

	// points
	for ( int id = ug->Give_number_points()-1;id >=0;--id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<pointEl> ( id );
		}
		
		if ( ug->Give_point ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<pointEl> ( id ) == yes_mark ) {
			data_points[id][0] =
			    a.template Give_data<pointEl> ( id, 0, 0, 0, 0, 0 );
		}
	}

	// edges
	for ( int id = ug->Give_number_edges()-1;id >=0;--id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<edgeEl> ( id );
		}
		
		if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			
			for(i=Nx-1;i>=1;--i )
				data_edges[id][i] =
				    a.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
		}
	}
	

	// quadrangles
	for ( int id = ug->Give_number_quadrangles()-1;id>=0;--id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<quadrangleEl> ( id );
		}
		
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			
			for(j=Ny-1;j>=1;--j )
				for(i=Nx-1;i>=1;--i ) {
					data_quadrangles[id][i+ ( Nx+1 ) *j] =
					    a.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
				}
		}
	}
	
	
	// hexahedra	
	for ( int id = ug->Give_number_hexahedra()-1;id>=0;--id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<hexahedronEl> ( id );
		}
		
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
#pragma omp parallel for private(j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
			for(k = Nz-1;k>=1;--k ) {
				for(j = Ny-1;j>=1;--j )
					for(i = Nx-1;i>=1;--i )
						data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
						    a.template Give_data<hexahedronEl> ( id, i, j, k, Nx, Ny );
			}
		}
	}
	
	
}


template <class DTyp>
template <class A>
void Variable<DTyp>::operator= ( const Expr_Restriction_Iterdes<A>& a ) {
	int Nx, Ny, Nz;
	
	stencil_typ sten_typ;
	
	sten_typ = a.Give_stencil_typ();
	
	// hexahedra
	
	for ( int id = 0;id < ug->Give_number_hexahedra();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<hexahedronEl> ( id );
		}
		
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<hexahedronEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			if(a.iterdes.red_black) {
			   for(int kk = 0;kk < 2;++kk ) for(int jj = 0;jj < 2;++jj ) for (int ii = 0;ii < 2;++ii) {

				    #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				    for(int k = 1 + kk;k < Nz;k = k + 2 ) {
					for(int j = 1 + jj;j < Ny;j = j + 2 ) {
					    for(int i = 1 + ii;i < Nx;i = i + 2 ) {
						data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
							    a.template Give_data<hexahedronEl> ( id, i, j, k, Nx, Ny );					      
					    }
					}
				    }
			   }		  
			  
			  /*
				for ( kk = 0;kk < 2;++kk ) {

				    #pragma omp parallel for private(jj,j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				    for ( k = 1 + kk;k < Nz;k = k + 2 ) {
						for ( jj = 0;jj < 2;++jj )
							for ( j = 1 + jj;j < Ny;j = j + 2 ) {
								data_hexahedra[id][1+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
								    a.template Give_data<hexahedronEl> ( id, 1, j, k, Nx, Ny );
								    
								for ( i = 2;i < Nx;i = i + 2 ) {
									//for(i=2;i<Nx;i=i+1) {  // so kann man etwas line_blocking machen
									data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
									    a.template Give_data<hexahedronEl> ( id, i, j, k, Nx, Ny );
									data_hexahedra[id][i-1+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
									    a.template Give_data<hexahedronEl> ( id, i - 1, j, k, Nx, Ny );
								}
								
								if ( Nx >= 2 )
									data_hexahedra[id][Nx-1+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
									    a.template Give_data<hexahedronEl> ( id, Nx - 1, j, k, Nx, Ny );
							}
					}
				}
				*/
			}
			else {

			  #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 1;k < Nz;++k ) {
					for(int j = 1;j < Ny;++j ) {
						for(int i = 1;i < Nx;i = i + 1 ) {
							data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
							    a.template Give_data<hexahedronEl> ( id, i, j, k, Nx, Ny );
						}
					}
				}
			}
		}
	}
	
	
	// quadrangles
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<quadrangleEl> ( id );
		}
		
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
		        
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			
			if ( a.iterdes.red_black ) {
				for(int jj = 0;jj < 2;++jj ) {
					for(int j = 1 + jj;j < Ny;j = j + 2 ) { // das ist eigentlich kein red-black!!!!!
						data_quadrangles[id][1+ ( Nx+1 ) *j] =
						    a.template Give_data<quadrangleEl> ( id, 1, j, 0, Nx, 0 );
						    
						for(int i = 2;i < Nx;i = i + 2 ) {
							data_quadrangles[id][i+ ( Nx+1 ) *j] =
							    a.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
							data_quadrangles[id][i-1+ ( Nx+1 ) *j] =
							    a.template Give_data<quadrangleEl> ( id, i - 1, j, 0, Nx, 0 );
						}
						
						if ( Nx >= 2 )
							data_quadrangles[id][Nx-1+ ( Nx+1 ) *j] =
							    a.template Give_data<quadrangleEl> ( id, Nx - 1, j, 0, Nx, 0 );
					}
				}
			} else {
				for(int j = 1;j < Ny;++j )
					for(int i = 1;i < Nx;++i ) {
						data_quadrangles[id][i+ ( Nx+1 ) *j] =
						    a.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
					}
			}
		}
	}
	
	
	// edges
	for(int id = 0;id < ug->Give_number_edges();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<edgeEl> ( id );
		}
		
		if(ug->Give_edge ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<edgeEl> ( id ) == yes_mark ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			
			if(a.iterdes.red_black ) {
				for(int ii = 0;ii < 2;++ii ) {
					for(int i = 1 + ii;i < Nx;i = i + 2 )
						data_edges[id][i] =
						    a.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
				}
			} else {
				for(int i = 1;i < Nx;++i )
					data_edges[id][i] =
					    a.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
			}
		}
	}
	
	// points
	for(int id = 0;id < ug->Give_number_points();++id ) {
		if ( sten_typ == yes_stencil ) {
			a.template Update<pointEl> ( id );
		}
		
		if(ug->Give_point ( id )->my_object ( my_rank ) &&
		        a.template Give_marker<pointEl> ( id ) == yes_mark ) {
			data_points[id][0] =
			    a.template Give_data<pointEl> ( id, 0, 0, 0, 0, 0 );
		}
	}
}



template <class DTyp>
template <class A>
void Variable<DTyp>::operator= ( const Expr<A>& a ) {
	int Nx, Ny, Nz;
	
	const A& ao ( a );
	
	stencil_typ sten_typ = ao.Give_stencil_typ();
	control_typ cont_typ = ao.Give_control_typ();
	
	
	bool useOpenMP_here = UGBlocks::useOpenMP && (cont_typ == thread_save);
	
	
	// hexahedra
	
	for(int id = 0;id < ug->Give_number_hexahedra();++id ) {
		if ( sten_typ == yes_stencil ) {
			ao.template Update<hexahedronEl> ( id );
		}
		
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );

//test
//if(cont_typ != thread_save) 
//  std::cout << " not thread_save " << std::endl;		
			
                        #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(useOpenMP_here)
			for(int k = 1;k < Nz;++k ) {
				for(int j = 1;j < Ny;++j )
				    for(int i = 1;i < Nx;++i )
					data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
							    ao.template Give_data<hexahedronEl> ( id, i, j, k, Nx, Ny );
			}
		}
	}
	
	// quadrangles
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
		if ( sten_typ == yes_stencil )
			ao.template Update<quadrangleEl> ( id );
			
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );

                        // #pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(useOpenMP_here)			
			for(int j = 1;j < Ny;++j )
				for (int i = 1;i < Nx;++i ) {
					data_quadrangles[id][i+ ( Nx+1 ) *j] =
                        ao.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
				}
		}
		Update<quadrangleEl> ( id );
	}
	
	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
		if ( sten_typ == yes_stencil )
			ao.template Update<edgeEl> ( id );
			
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			
			for(int i = 1;i < Nx;++i )
				data_edges[id][i] =
				    ao.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
		}
		Update<edgeEl> ( id );
	}
	
	// points
	for(int id = 0;id < ug->Give_number_points();++id ) {
		if ( sten_typ == yes_stencil )
			ao.template Update<pointEl> ( id );
			
		if(ug->Give_point ( id )->my_object ( my_rank ) ) {
			data_points[id][0] =
			    ao.template Give_data<pointEl> ( id, 0, 0, 0, 0, 0 );
		}
		Update<pointEl> ( id );
	}
//
//
//        for ( int id = 0;id < ug->Give_number_hexahedra();++id )
//          Update<hexahedronEl> ( id );
//        for ( int id = 0;id < ug->Give_number_quadrangles();++id )
//          Update<quadrangleEl> ( id );
//        for ( int id = 0;id < ug->Give_number_edges();++id )
//          Update<edgeEl> ( id );
//        for ( int id = 0;id < ug->Give_number_points();++id )
//          Update<pointEl> ( id );
}


template <class DTyp>
void Variable<DTyp>::SetAllZero() {
	int Nx, Ny, Nz;
	
	bool useOpenMP_here = UGBlocks::useOpenMP;	
	
	// hexahedra	
	for(int id = 0;id < ug->Give_number_hexahedra();++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );

                        #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(useOpenMP_here)
			for(int k = 0;k <= Nz;++k ) {
				for(int j = 0;j <= Ny;++j )
				    for(int i = 0;i <= Nx;++i )
					data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] = 0.0;
			}
		}
	}
	
	// quadrangles
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
			
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {		
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );

                        // #pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(useOpenMP_here)			
			for(int j = 0;j <= Ny;++j )
				for (int i = 0;i <= Nx;++i ) {
					data_quadrangles[id][i+ ( Nx+1 ) *j] = 0.0;
				}
		}
	}
	
	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			
			for(int i = 0;i <= Nx;++i )
				data_edges[id][i] = 0.0;
		}
	}
	
	// points
	for(int id = 0;id < ug->Give_number_points();++id ) {
		if(ug->Give_point ( id )->my_object ( my_rank ) ) {
			data_points[id][0] = 0.0;
		}
	}
}




template <class DTyp>
template <class Zell>
void Variable<DTyp>::AddFromZell(Zell& ZellVar) {
	int Nx, Ny, Nz;
		
	// hexahedra	
	for(int id = 0;id < ug->Give_number_hexahedra();++id ) {
	    if(ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			for(int k = 0;k < Nz;++k ) {
				for(int j = 0;j < Ny;++j )
				    for(int i = 0;i < Nx;++i ) {
				        DTyp value = ZellVar.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i,j,k,Nx,Ny);
					data_hexahedra[id][i   + (Nx+1) * (j   + (Ny+1)*k     ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j   + (Ny+1)*k     ) ] += value;
					data_hexahedra[id][i   + (Nx+1) * (j+1 + (Ny+1)*k     ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j+1 + (Ny+1)*k     ) ] += value;
					data_hexahedra[id][i   + (Nx+1) * (j   + (Ny+1)*(k+1) ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j   + (Ny+1)*(k+1) ) ] += value;
					data_hexahedra[id][i   + (Nx+1) * (j+1 + (Ny+1)*(k+1) ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j+1 + (Ny+1)*(k+1) ) ] += value;
				    }
			}
	    }
	}
}

template <class DTyp>
void Variable<DTyp>::AddFromOne() {
	int Nx, Ny, Nz;
		
	// hexahedra	
	for(int id = 0;id < ug->Give_number_hexahedra();++id ) {
	    if(ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			
			for(int k = 0;k < Nz;++k ) {
				for(int j = 0;j < Ny;++j )
				    for(int i = 0;i < Nx;++i ) {
				        DTyp value = 1.0;
					data_hexahedra[id][i   + (Nx+1) * (j   + (Ny+1)*k   ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j   + (Ny+1)*k   ) ] += value;
					data_hexahedra[id][i   + (Nx+1) * (j+1 + (Ny+1)*k   ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j+1 + (Ny+1)*k   ) ] += value;
					data_hexahedra[id][i   + (Nx+1) * (j   + (Ny+1)*(k+1) ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j   + (Ny+1)*(k+1) ) ] += value;
					data_hexahedra[id][i   + (Nx+1) * (j+1 + (Ny+1)*(k+1) ) ] += value;
					data_hexahedra[id][i+1 + (Nx+1) * (j+1 + (Ny+1)*(k+1) ) ] += value;
				    }
			}
	    }
	}
}

/////////////////////////////////////////////////////////////////////////
// 5. Nconst , ... funtions for copy of data
//////////////////////////////////////////////////////////////




inline int Give_Nconst ( dir3D dir, int Nxhex, int Nyhex, int Nzhex ) {
	if ( dir == Wdir3D ) return Nxhex;
	
	if ( dir == Sdir3D ) return ( Nxhex + 1 ) *Nyhex;
	
	if ( dir == Ddir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) *Nzhex;
	
	return 0;
}

inline int Give_Ni ( dir3D dir, int Nxhex, int Nyhex, int Nzhex ) {
	if ( dir == Edir3D ) return   1;
	
	if ( dir == Wdir3D ) return  -1;
	
	if ( dir == Ndir3D ) return   Nxhex + 1;
	
	if ( dir == Sdir3D ) return - ( Nxhex + 1 );
	
	if ( dir == Tdir3D ) return  ( Nxhex + 1 ) * ( Nyhex + 1 );
	
	if ( developer_version )
		if ( dir != Ddir3D ) std::cout << "error in Give_Ni !" <<std::endl;
		
	return - ( Nxhex + 1 ) * ( Nyhex + 1 );
}


// dir3D_sons { WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D };
inline int Give_Nconst ( dir3D_sons Left, dir3D_sons Right, int Nxhex, int Nyhex, int Nzhex ) {
	if ( Left == WSDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + Nxhex + 2;
	
	if ( Left == ESDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + 2*Nxhex;
	
	if ( Left == WNDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + ( Nxhex + 1 ) * ( Nyhex - 1 ) + 1;
	
	if ( Left == ENDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + ( Nxhex + 1 ) *Nyhex - 2;
	
	if ( Left == WSTdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + Nxhex + 2;
	
	if ( Left == ESTdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + 2*Nxhex;;
	
	if ( Left == WNTdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + ( Nxhex + 1 ) * ( Nyhex - 1 ) + 1;
	
	if ( developer_version )
		if ( Left != ENTdir3D ) std::cout << "error1 in Give_Nconst !" <<std::endl;
		
	return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + ( Nxhex + 1 ) *Nyhex - 2;
}

inline int Give_Nj ( dir3D_sons Left, dir3D_sons Right, int Nxhex, int Nyhex, int Nzhex ) {
	if ( Left == WSDdir3D ) {
		if ( Right ==  ESDdir3D ) return 1;
		
		if ( Right ==  WNDdir3D ) return ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  WSTdir3D ) std::cout << "error1 in Give_Nj !" <<std::endl;
			
		return ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( Left == ESDdir3D ) {
		if ( Right ==  WSDdir3D ) return -1;
		
		if ( Right ==  ENDdir3D ) return ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  ESTdir3D ) std::cout << "error2 in Give_Nj !" <<std::endl;
			
		return ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( Left == WNDdir3D ) {
		if ( Right ==  ENDdir3D ) return 1;
		
		if ( Right ==  WSDdir3D ) return - ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  WNTdir3D ) std::cout << "error3 in Give_Nj !" <<std::endl;
			
		return ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( Left == ENDdir3D ) {
		if ( Right ==  WNDdir3D ) return -1;
		
		if ( Right ==  ESDdir3D ) return - ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  ENTdir3D ) std::cout << "error4 in Give_Nj !" <<std::endl;
			
		return ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( Left == WSTdir3D ) {
		if ( Right ==  ESTdir3D ) return 1;
		
		if ( Right ==  WNTdir3D ) return ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  WSDdir3D ) std::cout << "error5 in Give_Nj !" <<std::endl;
			
		return - ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( Left == ESTdir3D ) {
		if ( Right ==  WSTdir3D ) return -1;
		
		if ( Right ==  ENTdir3D ) return ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  ESDdir3D ) std::cout << "error6 in Give_Nj !" <<std::endl;
			
		return - ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( Left == WNTdir3D ) {
		if ( Right ==  ENTdir3D ) return 1;
		
		if ( Right ==  WSTdir3D ) return - ( Nxhex + 1 );
		
		if ( developer_version )
			if ( Right !=  WNDdir3D ) std::cout << "error7 in Give_Nj !" <<std::endl;
			
		return - ( Nxhex + 1 ) * ( Nyhex + 1 );
	}
	
	if ( developer_version )
		if ( Left != ENTdir3D ) std::cout << "error8 in Give_Nj !" <<std::endl;
		
	if ( Right ==  WNTdir3D ) return -1;
	
	if ( Right ==  ESTdir3D ) return - ( Nxhex + 1 );
	
	if ( developer_version )
		if ( Right !=  ENDdir3D ) std::cout << "error9 in Give_Nj !" <<std::endl;
		
	return - ( Nxhex + 1 ) * ( Nyhex + 1 );
}

// dir2D_sons { SWdir2D, SEdir2D, NWdir2D, NEdir2D };
inline int Give_Nconst ( dir2D_sons Left, dir2D_sons Right, int Nxquad, int Nyquad ) {
	if ( Left == SWdir2D )
		return Nxquad + 2;
		
	if ( Left == SEdir2D )
		return 2*Nxquad;
		
	if ( Left == NWdir2D )
		return ( Nxquad + 1 ) * ( Nyquad - 1 ) + 1;
		
	if ( developer_version )
		if ( Left != NEdir2D ) std::cout << "error2 in Give_Nconst !" <<std::endl;
		
	return ( Nxquad + 1 ) *Nyquad - 2;
}

inline int Give_Nj ( dir2D_sons Left, dir2D_sons Right, int Nxquad, int Nyquad ) {
	if ( Left == SWdir2D ) {
		if ( Right ==  SEdir2D ) return 1;
		
		if ( developer_version )
			if ( Right !=  NWdir2D ) std::cout << "error10 in Give_Nj !" <<std::endl;
			
		return ( Nxquad + 1 );
	}
	
	if ( Left == SEdir2D ) {
		if ( Right ==  SWdir2D ) return -1;
		
		if ( developer_version )
			if ( Right !=  NEdir2D ) std::cout << "error11 in Give_Nj !" <<std::endl;
			
		return ( Nxquad + 1 );
	}
	
	if ( Left == NWdir2D ) {
		if ( Right ==  NEdir2D ) return 1;
		
		if ( developer_version )
			if ( Right !=  SWdir2D ) std::cout << "error12 in Give_Nj !" <<std::endl;
			
		return - ( Nxquad + 1 );
	}
	
	if ( developer_version )
		if ( Left != NEdir2D ) std::cout << "error13 in Give_Nj !" <<std::endl;
		
	if ( Right ==  NWdir2D ) return -1;
	
	if ( developer_version )
		if ( Right !=  SEdir2D ) std::cout << "error14 in Give_Nj !" <<std::endl;
		
	return - ( Nxquad + 1 );
}

// dir3D_sons { WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D };
inline int Give_Nconst ( dir3D_sons son, int Nxhex, int Nyhex, int Nzhex ) {
	if ( son == WSDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + Nxhex + 2;
	
	if ( son == ESDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + 2*Nxhex;
	
	if ( son == WNDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + ( Nxhex + 1 ) * ( Nyhex - 1 ) + 1;
	
	if ( son == ENDdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) + ( Nxhex + 1 ) *Nyhex - 2;
	
	if ( son == WSTdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + Nxhex + 2;
	
	if ( son == ESTdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + 2*Nxhex;;
	
	if ( son == WNTdir3D ) return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + ( Nxhex + 1 ) * ( Nyhex - 1 ) + 1;
	
	if ( developer_version )
		if ( son != ENTdir3D ) std::cout << "error3 in Give_Nconst !" <<std::endl;
		
	return ( Nxhex + 1 ) * ( Nyhex + 1 ) * ( Nzhex - 1 ) + ( Nxhex + 1 ) *Nyhex - 2;
}

// dir2D_sons { SWdir2D, SEdir2D, NWdir2D, NEdir2D };
inline int Give_Nconst ( dir2D_sons son, int Nxquad, int Nyquad ) {
	if ( son == SWdir2D )
		return Nxquad + 2;
		
	if ( son == SEdir2D )
		return 2*Nxquad;
		
	if ( son == NWdir2D )
		return ( Nxquad + 1 ) * ( Nyquad - 1 ) + 1;
		
	if ( developer_version )
		if ( son != NEdir2D ) std::cout << "error4 in Give_Nconst !" <<std::endl;
		
	return ( Nxquad + 1 ) *Nyquad - 2;
}

/*
template <class DTyp>
class Cell_variable;

template <class DTyp>
class Expr_interpolant_cell_to_point : public Expr<Expr_interpolant_cell_to_point< DTyp> >  {
 private:
  const Cell_variable< DTyp > & a_;

 public:
  Expr_interpolant_cell_to_point(const Cell_variable<DTyp>& a) : a_(a) {}

  template <elementTyp TYP_EL>
  inline DTyp Give_data ( params_in) const {
    switch(TYP_EL) {
      case hexahedronEl:
        Give_data_hexahedronEl(params_out);
      case edgeEl:
        Give_data_edgeEl(params_out);
      case pointEl:
        Give_data_pointEl(params_out);
    }
  }
  inline DTyp Give_data_hexahedronEl ( params_in) const;
  inline DTyp Give_data_edgeEl ( params_in) const;
  inline DTyp Give_data_pointEl ( params_in) const;

  template <elementTyp TYP_EL>
  void Update(int id) const {
    a_.template Update<TYP_EL>(id);
  }

};

template <class DTyp>
inline DTyp Expr_interpolant_cell_to_point<DTyp>::Give_data_hexahedronEl ( params_in) const {
  return 0.125 * (a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i  ,j  ,k  ,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i+1,j  ,k  ,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i  ,j+1,k  ,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i+1,j+1,k  ,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i  ,j  ,k+1,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i+1,j  ,k+1,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i  ,j+1,k+1,Nx,Ny) +
                  a_.Give_cell_hexahedra(id,Ind_loc_matrix_hexahedra(i,j,k),i+1,j+1,k+1,Nx,Ny));
}

template <class DTyp>
inline DTyp Expr_interpolant_cell_to_point<DTyp>::Give_data_edgeEl ( params_in) const {
  return 0;
//  return 0.125 * (a_.template Give_data<hexahedronEl>(id_hex,i  ,j  ,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j  ,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i  ,j+1,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j+1,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i  ,j  ,k+1,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j  ,k+1,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i  ,j+1,k+1,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j+1,k+1,Nx,Ny));
}

template <class DTyp>
inline DTyp Expr_interpolant_cell_to_point<DTyp>::Give_data_pointEl ( params_in) const {
  return 0;
//  return 0.125 * (a_.template Give_data<hexahedronEl>(id_hex,i  ,j  ,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j  ,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i  ,j+1,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j+1,k  ,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i  ,j  ,k+1,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j  ,k+1,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i  ,j+1,k+1,Nx,Ny) +
//                  a_.template Give_data<hexahedronEl>(id_hex,i+1,j+1,k+1,Nx,Ny));
}

template <class DTyp>
inline Expr_interpolant_cell_to_point<DTyp> Point_interpolation(const Cell_variable< DTyp >& a) {
  return Expr_interpolant_cell_to_point<DTyp>(a);
}
*/
#endif // VARIABLE_H
