/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include <sstream>
#include <stdexcept>

#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"
#include "blockgrid.h"
#include "compose_grid.h"



ComposeBlockGrid::ComposeBlockGrid(std::vector<Blockgrid*> grids_, std::vector<double> transZ_) {
  grids  = grids_;
  transZ = transZ_;
  
  ug = new ComposeUg(grids_, transZ_);

  number_points = new int[ug->degree_of_freedom() ];
  int sumDeg = 0;
  for(int i=0;i<grids.size();++i) {
     for(int num=0;num<grids.at(i)->Give_unstructured_grid()->degree_of_freedom();++num) {
         number_points[sumDeg + num] = grids.at(i)->getNumberPointsDegree(num);
     }
     sumDeg = sumDeg + grids.at(i)->Give_unstructured_grid()->degree_of_freedom();
  }
  
  cout << " degree A: " << ug->degree_of_freedom()
       << " degree B: " << sumDeg 
       << endl;
  
  
  assert(ug->degree_of_freedom()==sumDeg);
}
  
  
  
int ComposeUg::getStartHex(int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberHex.at(i);
    assert(false);
}

int ComposeUg::getEndHex(  int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberHex.at(i+1);
    assert(false);
}
  
int ComposeUg::getStartQuad(int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberQuad.at(i);
    assert(false);
}

int ComposeUg::getEndQuad(  int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberQuad.at(i+1);
    assert(false);
}
  
int ComposeUg::getStartEdge(int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberEdge.at(i);
    assert(false);
}

int ComposeUg::getEndEdge(int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberEdge.at(i+1);
    assert(false);
}

int ComposeUg::getStartPoint(int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberPoint.at(i);
    assert(false);
}

int ComposeUg::getEndPoint(int thisIdGrid) {
    for(int i=0;i<idGrid.size();++i)
        if(idGrid.at(i)==thisIdGrid) return startNumberPoint.at(i+1);
    assert(false);
}
  
ComposeUg::ComposeUg(std::vector<Blockgrid*> grids, std::vector<double> transZ) {
  // 1. unstructured grids
  std::vector<Unstructured_grid*> ugGrids;
  ugGrids.resize(grids.size());
  for(int i=0;i<ugGrids.size();++i) {
      ugGrids.at(i) = grids.at(i)->Give_unstructured_grid();
  }
  
  // 2.number points
  startNumberPoint.resize(ugGrids.size()+1);  
  startNumberPoint.at(0) = 0;      
  int numPoints=0;
  for(int i=0;i<ugGrids.size();++i) {
      numPoints = numPoints + ugGrids.at(i)->Give_number_points();
      startNumberPoint.at(i+1) = numPoints;          
  }
  Set_number_points(numPoints);
  
  // 3.set points
  numPoints=0;
  for(int i=0;i<ugGrids.size();++i) {
     D3vector zT(0.0,0.0,transZ.at(i));
     for(int num=0;num<ugGrids.at(i)->Give_number_points();++num) {
         Set_coordinate_point(numPoints + num,
			      ugGrids.at(i)->Give_point(num)->Give_coordinate() + zT);
     }
     numPoints = numPoints + ugGrids.at(i)->Give_number_points();
  }
  
  // 4.set number hexahedra
  int numHex = 0;
  for(int i=0;i<ugGrids.size();++i) {
      numHex = numHex + ugGrids.at(i)->Give_number_hexahedra();
  }
  Set_number_hexahedra(numHex); 
  
  // 5.set hexahedra
  startNumberHex.resize(ugGrids.size()+1);
  idGrid.resize(ugGrids.size());  
  numHex = 0;
  numPoints=0;  
  for(int i=0;i<ugGrids.size();++i) {
      startNumberHex.at(i) = numHex;
      idGrid.at(i)         = grids.at(i)->getId();
      for(int num=0;num<ugGrids.at(i)->Give_number_hexahedra();++num) {
          Hexahedron_el* hexE = ugGrids.at(i)->Give_hexahedron(num);
          Set_hexahedron(numHex    + num,
			 numPoints + hexE->Give_id_corner(WSDdir3D),
			 numPoints + hexE->Give_id_corner(ESDdir3D), 
			 numPoints + hexE->Give_id_corner(WNDdir3D),
			 numPoints + hexE->Give_id_corner(ENDdir3D), 
			 numPoints + hexE->Give_id_corner(WSTdir3D),
			 numPoints + hexE->Give_id_corner(ESTdir3D),
			 numPoints + hexE->Give_id_corner(WNTdir3D),
			 numPoints + hexE->Give_id_corner(ENTdir3D));
      }
      numHex = numHex + ugGrids.at(i)->Give_number_hexahedra();
      numPoints = numPoints + ugGrids.at(i)->Give_number_points();      
  }
  startNumberHex.at(ugGrids.size()) = numHex;
  construction_hexahedron_points_done();
  
  // 6.set transformations
  size_pointer_global_data = 0;
  for(int i=0;i<ugGrids.size();++i) {
      size_pointer_global_data = size_pointer_global_data + ugGrids.at(i)->sizeGlobalPointerData();
  }
  pointer_global_data = new double[size_pointer_global_data];

  int shift_data = 0;
  for(int i=0;i<ugGrids.size();++i) {
      double* local_global_pointer = ugGrids.at(i)->Get_pointer_global_data();
      for(int num=0;num < ugGrids.at(i)->sizeGlobalPointerData();++num) {
          pointer_global_data[shift_data + num] = local_global_pointer[num];	  
      }
      shift_data = shift_data + ugGrids.at(i)->sizeGlobalPointerData();
  }
 
  shift_data = 0;
  numPoints=0;  
  for(int i=0;i<ugGrids.size();++i) {
      for(int num=0;num<ugGrids.at(i)->Give_number_edges();++num) {
          Edge_el* edEl = ugGrids.at(i)->Give_edge(num);
	  int idLeft  =  numPoints + edEl->Give_id_corner(Ldir1D);
	  int idRight =  numPoints + edEl->Give_id_corner(Rdir1D);

	  Set_transformation_edge(idLeft, idRight, edEl->transform,shift_data);
      }
      numPoints = numPoints + ugGrids.at(i)->Give_number_points(); 
      shift_data = shift_data + ugGrids.at(i)->sizeGlobalPointerData();
  }  

  construction_done();  
  
  // 7.number quad
  startNumberQuad.resize(ugGrids.size()+1);  
  startNumberQuad.at(0) = 0;      
  int numQuad=0;
  for(int i=0;i<ugGrids.size();++i) {
      numQuad = numQuad + ugGrids.at(i)->Give_number_quadrangles();
      startNumberQuad.at(i+1) = numQuad;          
  }
  
  // 8.number edges
  startNumberEdge.resize(ugGrids.size()+1);  
  startNumberEdge.at(0) = 0;      
  int numEdge=0;
  for(int i=0;i<ugGrids.size();++i) {
      numEdge = numEdge + ugGrids.at(i)->Give_number_edges();
      startNumberEdge.at(i+1) = numEdge;          
  }
  
  
//test
//  Print();
}




