/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef EXAMPLES_UG_H
#define EXAMPLES_UG_H
#include <vector>

class HexahedronBasis  {
protected: 
  void PrintLengthEdgesHere(std::vector<D3vector> pointsHere);  
};

/** \addtogroup GridGeneration **/
/* @{ */   
/**
 * Normaler Hexaeder
 * 
 * **/
class Hexahedron : public HexahedronBasis , public UG_WithZdirection{
 public:
   /**
    * [0,lx] x [0,ly] x [0,lz]
    */
  Hexahedron(double lx, double ly, double lz);
   /**
    * [lx,rx] x [ly,ry] x [lz,rz]
    */
  Hexahedron(double lx, double ly, double lz,
	     double rx, double ry, double rz);
  
  void PrintLengthEdges() { PrintLengthEdgesHere(getVectorOfPoints()); }
  
  ~Hexahedron(){};

};

/**
 * Hexaeder rotated Y
 * 
 * **/
class HexahedronRotY : public HexahedronBasis, public UG_WithZdirection {
 public:
   /**
    * [lx,rx] x [ly,ry] x [lz,rz]  und rotiert um das Zentrum ((lx+rx)/2, (ly+ry)/2, (lz+rz)/2)
    */
  HexahedronRotY(double lx, double ly, double lz,
	         double rx, double ry, double rz, double sinRotY);

  void PrintLengthEdges() { PrintLengthEdgesHere(getVectorOfPoints()); }  
  
  ~HexahedronRotY(){};
};

/**
 * Normaler Hexaeder mit Unterteilung in z-Richtung
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class HexahedronList : public UG_WithZdirection {
public:
   /**
    * [offsetX,lx+offsetX] x [offsetY,ly+offsetY] x [offsetZ,lz+offsetZ]
    */
  HexahedronList(double lx, double ly, std::vector<double>& lz, double offsetX = 0., double offsetY = 0., double offsetZ = 0, bool periodic = 0 );
  
  ~HexahedronList(){};
};


/**
 * Normaler Hexaeder mit Unterteilung in x und y-Richtung
 * hierbei ist: lx[i] die i-te Korrdinate wobei ein Hexaeder mit n Elementen n+1 Koordniaten enthält.
 *
 * **/
class HexahedronXYList : public UG_WithZdirection {
 public:
   /**
    * [lx[0],lx[n_x]] x [ly[0],ly[n_y]] x [lz[0],lz[n_z]]
    */
  HexahedronXYList(std::vector<double>& lx, std::vector<double>& ly,  std::vector<double>& lz);

  ~HexahedronXYList(){};
 private:
  int n_x;
  int n_y;
  int n_z;

  friend class UGFrom3DSlice;
  friend class IteratorZDirection;

  template <class DTyp>
  friend class Variable;
  template <class DTyp>
  friend class Variable2D;
};

class HexahedronXYListTest : public UG_WithZdirection {
 public:
   /**
    * [lx[0],lx[n_x]] x [ly[0],ly[n_y]] x [lz[0],lz[n_z]]
    */
  HexahedronXYListTest(double lx, double ly, double lz,
      double rx, double ry, double rz );

  ~HexahedronXYListTest(){};
 private:
  int n_x;
  int n_y;
  int n_z;

  friend class UGFrom3DSlice;
  friend class IteratorZDirection;

  template <class DTyp>
  friend class Variable;
  template <class DTyp>
  friend class Variable2D;
};

/**
 * Hexaeder mit schraegen Endfächen
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class HexahedronSlabList : public Unstructured_grid {
public:
  HexahedronSlabList(double lx, double ly, std::vector<double>& lz, 
		     double alphaLeft, double alphaRight, 
		     double offsetX = 0., double offsetY = 0., double offsetZ = 0);
  ~HexahedronSlabList(){};
};

/**
 * Pyramide
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class PyramidList : public Unstructured_grid {
public:
   /**
    * linke  Fläche: [-lxLeft *0.5 , lxLeft *0.5 ] x [-lyLeft *0.5 , lyLeft *0.5 ]   verschoben um offset
    * rechte Fläche: [-lxRight*0.5 , lxRight*0.5 ] x [-lyRight*0.5 , lyRight*0.5 ]   verschoben um offset
    * z-Richtung:    [offsetZ,lz+offsetZ]
    */
  PyramidList(double lxLeft, double lyLeft,
	      double lxRight, double lyRight,
	      std::vector<double>& lz,
	      double offsetX = 0., double offsetY = 0., double offsetZ = 0);
  ~PyramidList(){};
};

/**
 * Slab
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class SlabList : public Unstructured_grid {
public:
   /**
    * Fläche Mitte:               [-lx *0.5 , lx *0.5 ] x [-ly *0.5 , ly *0.5 ]   verschoben um offset
    * z-Richtung in der Mitte:    [offsetZ,lz+offsetZ]
    * alpha_left, alpha_right:    Winkel der Endflächen
    */
  SlabList(double length_x,  double length_y,
	   double alpha_left, double alpha_right,
           std::vector<double>& relativeCoordZ,												 
           double offsetX = 0., double offsetY = 0., double offsetZ = 0);
  ~SlabList(){};
};

class UGFrom3DSlice;

/**
 * Cylinder in xy-Ebene durch 5 Hexaeder beschrieben
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class CylinderList : public UG_WithZdirection {
public:
  CylinderList(double Rout, double Rin, std::vector<double>& lz, double offsetX = 0., double offsetY = 0., double offsetZ =0., bool inner_grid_arched = false);

  ~CylinderList(){};

  bool getInnerGridArched() const;
  int getNz() const;
private:
  std::vector<double> length;
  bool inner_grid_arched_;
  double Radius;
  double radius;

  friend class UGFrom3DSlice;
};

/**
 * Hexaeder der in xy-Richtung durch 5 Hexaeder beschrieben wird um im Zentrum verfeinern zu können
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class AdaptiveHexahedronList : public UG_WithZdirection {
public:
  AdaptiveHexahedronList(double Routx, double Routy,
			 double Rinx,  double Riny,
			 std::vector<double>& lz, double offsetX = 0., double offsetY = 0., double offsetZ =0.);

  ~AdaptiveHexahedronList(){};
  
  friend class UGFrom3DSlice;
};


class HexahedronListCurvedEnds : public Unstructured_grid {
public:
  HexahedronListCurvedEnds(unsigned int nHex, double offsetX, double offsetY, double offsetZ, double lx, double ly, std::vector<double>& relativeCoordZ, double curvatureLeftX, double curvatureLeftY, double curvatureRightX, double curvatureRightY);

  HexahedronListCurvedEnds(double lx, double ly, std::vector<double>& lz, double curvatureLeftX, double curvatureLeftY, double curvatureRightX, double curvatureRightY) { 
      HexahedronListCurvedEnds(lz.size()-1,0.,0.,0.,lx,ly,lz, curvatureLeftX, curvatureLeftY, curvatureRightX, curvatureRightY);
   }
  
  ~HexahedronListCurvedEnds(){};
};

class Ring: public Unstructured_grid {
 public:
  Ring(double Radius, double radius, double l);
  ~Ring(){};
};

// facetest
//class face_transform_test {
//public:
//    face_transform_test(D3vector (*Transform)(double,double, double*));
//    D3vector (*transform)(double, double, double*);
//    D3vector (*transformNorth)(double,double*);
//    D3vector (*transformSouth)(double,double*);
//    D3vector (*transformEast)(double,double*);
//    D3vector (*transformWest)(double,double*);
//};


class Sphere: public Unstructured_grid {
 public:
  Sphere(double radius);
  ~Sphere(){};
};

class Cylinder : public UG_WithZdirection {
 public:
  Cylinder(double Radius, double radius, double l, bool inner_grid_arched = false);
  ~Cylinder(){};

  double Radius;
  double radius;
  double length;

  bool getInnerGridArched() const;

  private:
  bool inner_grid_arched_;
  
  friend class UGFrom3DSlice;
};

class Periodic_Cylinder : public Unstructured_grid {
 public:
  Periodic_Cylinder(double radius, double l);
  ~Periodic_Cylinder(){};
};

class Hexahedron_periodic : public Unstructured_grid {
public:
  Hexahedron_periodic(double lx, double ly, double lz);
  ~Hexahedron_periodic(){};
};


/**
 * Spezieller Prism in xy-Ebene durch 2 Hexaeder beschrieben
 * 
 *                 xS
 *        4 |-----------  3
 *          |           \
 *          |            \
 *      yL  |             \
 *       5  |--------------| 2
 *          | yH           | yS 
 *          |              |
 *          |--------------|
 *          0       xL     1
 * 
 * hierbei ist: lz[i] relative Koordinate der i-ten top-Fläche zum Offset-Punkt
 * 
 * **/
class SpecialPrismList : public UG_WithZdirection {
public:
  SpecialPrismList(double yL_, double xL_, 
		   double yS_, double xS_, 
		   double yH_, 
		   std::vector<double>& lz, double offsetX = 0., 
		   double offsetY = 0., double offsetZ =0.);

  ~SpecialPrismList(){};

  int getNz() const;
private:
  std::vector<double> length;
  bool inner_grid_arched_;
  double yL; 
  double xL; 
  double yS;
  double xS; 
  double yH;
 
  friend class UGFrom3DSlice;
};
/* @} */






#endif // EXAMPLES_UG_H
