/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// #include <assert.h>
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"
#include "blockgrid.h"
#include "marker.h"
#include <stdexcept>
#include <vector>

Marker::Marker ( Unstructured_grid* ug_ ) {
	ug = ug_;
	marker_hexahedra = new marker_typ[ug->Give_number_hexahedra() ];    // num_hexahedra
	marker_quadrangles = new marker_typ[ug->Give_number_quadrangles() ]; // num_quadrangles
	marker_edges = new marker_typ[ug->Give_number_edges() ];            // num_edges
	marker_points = new marker_typ[ug->Give_number_points() ];          // num_points

	on();
}

void Marker::on() {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, yes_mark );
}

void Marker::off() {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker::Mark_hexahedra_only( void ) {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker::Mark_quadrangles_only( void ) {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, yes_mark );
}

void Marker::Mark_edges_only( void ) {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker::Mark_points_only( void ) {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker::Unmark_hexahedra( void ) {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
		Set_marker<hexahedronEl> ( id, no_mark );
}

void Marker::Unmark_quadrangles( void ) {
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );

}
void Marker::Unmark_edges( void ) {
	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );
}
void Marker::Unmark_points( void ) {
	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );
}


// Marks hexaeder and adjacent elements.
void Marker::selectHexBlock ( unsigned int chosenHexBlock ) {
	if ( ( int ) chosenHexBlock >= ug->Give_number_hexahedra() )
		throw ( std::runtime_error ( std::string ( "Marker::selectHexBlock(): Selected block does not exist" ) ) );
//   cout << "select Block " << chosenHexBlock << endl;
	Set_marker<hexahedronEl> ( chosenHexBlock, yes_mark );

	Hexahedron_el* pHex = ug->Give_hexahedron ( chosenHexBlock );

	for ( int i=0; i<12; ++i ) {
		Set_marker<edgeEl> ( pHex->Give_id_edge ( ( Edges_cell ) i ), yes_mark );
	}

	for ( int i=0; i<6; ++i ) {
		Set_marker<quadrangleEl> ( pHex->Give_id_quadrangle ( ( dir3D ) i ), yes_mark );
	}
	for ( int i=0; i<3; ++i ) {
		Set_marker<pointEl> ( pHex->Give_id_corner ( ( dir3D_sons ) i ), yes_mark );
	}
}

// Marks hexaeder and adjacent elements.
void Marker::selectSingleHexBlock ( unsigned int chosenHexBlock ) {
	this->off();
	selectHexBlock(chosenHexBlock);
}


void Marker::selectHexBlocks ( unsigned int firstBlock, unsigned int numBlocks) {
	this->off();
	for ( unsigned int i = 0; i < numBlocks; ++i ) {
		selectHexBlock(firstBlock+i);
    }
}


void Marker::selectHexBlocks ( std::vector<unsigned int> & chosenHexBlocks ) {
	this->off();
	for ( unsigned int count = 0;
	        count <chosenHexBlocks.size();
	        ++count ) {
		unsigned chosenBlock = chosenHexBlocks.at ( count );
		selectHexBlock(chosenBlock);
	}
}

void Marker::selectQuad(unsigned int chosenQuad)
{
    if ( ( int ) chosenQuad >= ug->Give_number_quadrangles() )
        throw ( std::runtime_error ( std::string ( "Marker::selectQuad(): Selected quad does not exist" ) ) );
    Set_marker<quadrangleEl> ( chosenQuad, yes_mark );


}

void Marker::selectSingleQuad(unsigned int chosenQuad)
{
    this->off();
    selectQuad(chosenQuad);
}

void Marker::selectQuads(unsigned int firstQuad, unsigned int numQuads)
{
    this->off();
    for ( unsigned int i = 0; i < numQuads; ++i ) {
        selectQuad(firstQuad+i);
    }
}

void Marker::operatorAnd(Marker *mark)
{
    for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
    {
        if (Give_marker<hexahedronEl> ( id ) == mark->Give_marker<hexahedronEl> ( id ))
        {
            //Set_marker<hexahedronEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<hexahedronEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_edges(); ++id )
    {
        if (Give_marker<edgeEl> ( id ) == mark->Give_marker<edgeEl> ( id ))
        {
            //Set_marker<edgeEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<edgeEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_points(); ++id )
    {
        if (Give_marker<pointEl> ( id ) == mark->Give_marker<pointEl> ( id ))
        {
            //Set_marker<pointEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<pointEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
    {
        if (Give_marker<quadrangleEl> ( id ) == mark->Give_marker<quadrangleEl> ( id ))
        {
            //Set_marker<quadrangleEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<quadrangleEl> ( id, no_mark );
        }
    }
}

void Marker::operatorOr(Marker *mark)
{

    for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
    {
        if (Give_marker<hexahedronEl> ( id ) != mark->Give_marker<hexahedronEl> ( id ) || Give_marker<hexahedronEl> ( id ) == yes_mark)
        {
            Set_marker<hexahedronEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<hexahedronEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_edges(); ++id )
    {
        if (Give_marker<edgeEl> ( id ) != mark->Give_marker<edgeEl> ( id ) || Give_marker<edgeEl> ( id ) == yes_mark)
        {
            Set_marker<edgeEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<edgeEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_points(); ++id )
    {
        if (Give_marker<pointEl> ( id ) != mark->Give_marker<pointEl> ( id ) || Give_marker<pointEl> ( id ) == yes_mark)
        {
            Set_marker<pointEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<pointEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
    {
        if (Give_marker<quadrangleEl> ( id ) != mark->Give_marker<quadrangleEl> ( id ) || Give_marker<quadrangleEl> ( id ) == yes_mark)
        {
            Set_marker<quadrangleEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<quadrangleEl> ( id, no_mark );
        }
    }
}


void Marker::operatorXOr(Marker *mark)
{

    for ( int id=0; id<ug->Give_number_hexahedra(); ++id )
    {
        if (Give_marker<hexahedronEl> ( id ) != mark->Give_marker<hexahedronEl> ( id ))
        {
            Set_marker<hexahedronEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<hexahedronEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_edges(); ++id )
    {
        if (Give_marker<edgeEl> ( id ) != mark->Give_marker<edgeEl> ( id ))
        {
            Set_marker<edgeEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<edgeEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_points(); ++id )
    {
        if (Give_marker<pointEl> ( id ) != mark->Give_marker<pointEl> ( id ))
        {
            Set_marker<pointEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<pointEl> ( id, no_mark );
        }
    }

    for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
    {
        if (Give_marker<quadrangleEl> ( id ) != mark->Give_marker<quadrangleEl> ( id ))
        {
            Set_marker<quadrangleEl> ( id, yes_mark );
        }
        else
        {
            Set_marker<quadrangleEl> ( id, no_mark );
        }
    }
}


void Marker::complement ( Marker* marker ) {
    for ( int id=0; id<ug->Give_number_hexahedra(); ++id ) {
        if ( marker->Give_marker<hexahedronEl> ( id ) == no_mark )
            Set_marker<hexahedronEl> ( id, yes_mark );
        else Set_marker<hexahedronEl> ( id, no_mark );
    }
    for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
        if ( marker->Give_marker<quadrangleEl> ( id ) == no_mark )
            Set_marker<quadrangleEl> ( id, yes_mark );
        else Set_marker<quadrangleEl> ( id, no_mark );
    }
    for ( int id=0; id<ug->Give_number_edges(); ++id ) {
        if ( marker->Give_marker<edgeEl> ( id ) == no_mark )
            Set_marker<edgeEl> ( id, yes_mark );
        else Set_marker<edgeEl> ( id, no_mark );
    }
    for ( int id=0; id<ug->Give_number_points(); ++id ) {
        if ( marker->Give_marker<pointEl> ( id ) == no_mark )
            Set_marker<pointEl> ( id, yes_mark );
        else Set_marker<pointEl> ( id, no_mark );
    }
}


void Marker::interior_boundary ( Marker* marker ) {
  // kopieren
  for ( int id=0; id<ug->Give_number_hexahedra(); ++id ) {
     if ( marker->Give_marker<hexahedronEl> ( id ) == no_mark )
    Set_marker<hexahedronEl> ( id, no_mark );
     else Set_marker<hexahedronEl> ( id, yes_mark );
  }
    for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
        if ( marker->Give_marker<quadrangleEl> ( id ) == no_mark )
            Set_marker<quadrangleEl> ( id, no_mark );
        else Set_marker<quadrangleEl> ( id, yes_mark );
    }
    for ( int id=0; id<ug->Give_number_edges(); ++id ) {
        if ( marker->Give_marker<edgeEl> ( id ) == no_mark )
            Set_marker<edgeEl> ( id, no_mark );
        else Set_marker<edgeEl> ( id, yes_mark );
    }
    for ( int id=0; id<ug->Give_number_points(); ++id ) {
        if ( marker->Give_marker<pointEl> ( id ) == no_mark )
            Set_marker<pointEl> ( id, no_mark );
        else Set_marker<pointEl> ( id, yes_mark );
    }


  //aendern
  for ( int id_quad=0; id_quad<ug->Give_number_quadrangles(); ++id_quad ) {
     if ( marker->Give_marker<quadrangleEl> ( id_quad ) == no_mark ) {

       for(int e=0;e < 4;++e) {
          Set_marker<edgeEl> (ug->Give_quadrangle ( id_quad )->Give_id_edge ( (dir2D)e ),
                   no_mark );
       }
       for(int c=0;c < 4;++c) {
          Set_marker<pointEl> (ug->Give_quadrangle ( id_quad )->Give_id_corner ( (dir2D_sons)c ),
                no_mark );
       }
     }
  }
}

Marker::~Marker() {
	if ( marker_hexahedra != NULL )
		delete [] marker_hexahedra;
	marker_hexahedra = NULL;
	if ( marker_quadrangles != NULL )
		delete [] marker_quadrangles;
	marker_quadrangles = NULL;
	if ( marker_edges != NULL )
		delete [] marker_edges;
	marker_edges = NULL;
	if ( marker_points != NULL )
		delete [] marker_points;
	marker_points = NULL;
}

Boundary_Marker::Boundary_Marker ( Unstructured_grid* ug_ ) : Marker ( ug_ )    {
	Quadrangle_el* quad;
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id )     {
		Set_marker<hexahedronEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		Set_marker<edgeEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id )    {
		Set_marker<pointEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )    {
		quad = ug->Give_quadrangle ( id );
		if ( quad->Give_exists_exterior() == true )
			Set_marker<quadrangleEl> ( id, no_mark );
		else  {
			Set_marker<quadrangleEl> ( id, yes_mark );
			for ( int i=0; i<4; ++i )
                Set_marker<edgeEl> ( quad->Give_id_edge ( ( dir2D ) i ), yes_mark );
			for ( int i=0; i<4; ++i )
				Set_marker<pointEl> ( quad->Give_id_corner ( ( dir2D_sons ) i ), yes_mark );
		}
	}
}

Boundary_Marker::~Boundary_Marker() {}

void Boundary_Marker::markBoundaryDir3D(dir3D dir, int blockRangeStart, int blockRangeEnd)
    {
    if (blockRangeStart == -1)
    {return;}
    if (blockRangeEnd < 0 || blockRangeEnd > ug->Give_number_hexahedra())
    {
      blockRangeEnd =   ug->Give_number_hexahedra();
    }
    if (blockRangeStart < 0 || blockRangeStart > ug->Give_number_hexahedra())
    {
      blockRangeStart =   0;
    }

    Quadrangle_el* quad;
    Hexahedron_el* hex ;
    for ( int id = blockRangeStart;id < blockRangeEnd;++id ) {
            hex = ug->Give_hexahedron ( id );

            if ( ug->Give_quadrangle ( hex->Give_id_quadrangle ( dir ) )->Give_exists_exterior() == false ) {

            quad = ug->Give_quadrangle ( hex->Give_id_quadrangle ( dir ) );
            Set_marker<quadrangleEl> ( hex->Give_id_quadrangle ( dir ), yes_mark );
            for ( int i=0; i<4; ++i )
                Set_marker<edgeEl> ( quad->Give_id_edge ( ( dir2D ) i ), yes_mark );
            for ( int i=0; i<4; ++i )
                Set_marker<pointEl> ( quad->Give_id_corner ( ( dir2D_sons ) i ), yes_mark );
        }
        }
    }

void Boundary_Marker::unmarkBoundaryDir3D(dir3D dir, int blockRangeStart, int blockRangeEnd)
{
    if (blockRangeStart == -1)
    {return;}
    if (blockRangeEnd < 0 || blockRangeEnd > ug->Give_number_hexahedra())
    {
      blockRangeEnd =   ug->Give_number_hexahedra();
    }
    if (blockRangeStart < 0 || blockRangeStart > ug->Give_number_hexahedra())
    {
      blockRangeStart =   0;
    }

    Quadrangle_el* quad;
    Hexahedron_el* hex ;
    for ( int id = blockRangeStart;id < blockRangeStart;++id ) {
            hex = ug->Give_hexahedron ( id );

            if ( ug->Give_quadrangle ( hex->Give_id_quadrangle ( dir ) )->Give_exists_exterior() == false ) {

            quad = ug->Give_quadrangle ( hex->Give_id_quadrangle ( dir ) );
            Set_marker<quadrangleEl> ( hex->Give_id_quadrangle ( dir ), no_mark );
            for ( int i=0; i<4; ++i )
                Set_marker<edgeEl> ( quad->Give_id_edge ( ( dir2D ) i ), no_mark );
            for ( int i=0; i<4; ++i )
                Set_marker<pointEl> ( quad->Give_id_corner ( ( dir2D_sons ) i ), no_mark );
        }
    }
}


Unstructured_grid_Marker::Unstructured_grid_Marker ( Unstructured_grid* ug_ ) : Marker ( ug_ ) {
	for ( int id=0; id<ug->Give_number_hexahedra(); ++id ) {
		Set_marker<hexahedronEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		Set_marker<edgeEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		Set_marker<pointEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		Set_marker<quadrangleEl> ( id, yes_mark );
	}
}

Unstructured_grid_Marker::~Unstructured_grid_Marker() {}

Blockgrid_Marker::Blockgrid_Marker ( Unstructured_grid* ug_ ) : Marker ( ug_ ) {
}

Blockgrid_Marker::~Blockgrid_Marker() {}



void Unstructured_grid_Marker::operator= ( bool ( *Marker_formula ) ( double x, double y, double z ) ) {
	D3vector middle;

	Hexahedron_el* hex;
	Quadrangle_el* quad;
	Edge_el*       edge;

	for ( int id=0; id<ug->Give_number_hexahedra(); ++id ) {
		hex = ug->Give_hexahedron ( id );

		/* old
		   middle = D3vector(0.0,0.0,0.0);
		   for(int i=0;i<8;++i) middle = middle + hex->Give_coord((dir3D_sons)i);
		   middle = middle / 8.0;
		*/
		middle = hex->Give_middle_coord();

		if ( Marker_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<hexahedronEl> ( id, yes_mark );
		else Set_marker<hexahedronEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		quad = ug->Give_quadrangle ( id );

		/* old
		middle = D3vector(0.0,0.0,0.0);
		for(int i=0;i<4;++i)
		  middle = middle + ug->Give_point(quad->Give_id_corner((dir2D_sons)i))->Give_coordinate();
		middle = middle / 4.0;
		*/
		middle = quad->Give_middle_coord();

		if ( Marker_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<quadrangleEl> ( id, yes_mark );
		else Set_marker<quadrangleEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		edge = ug->Give_edge ( id );

		middle = ( ug->Give_point ( edge->Give_id_corner_W() )->Give_coordinate() +
		           ug->Give_point ( edge->Give_id_corner_E() )->Give_coordinate() ) / 2.0;

		if ( Marker_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<edgeEl> ( id, yes_mark );
		else Set_marker<edgeEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		middle = ug->Give_point ( id )->Give_coordinate();

		if ( Marker_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<pointEl> ( id, yes_mark );
		else Set_marker<pointEl> ( id, no_mark );
	}
}




void Unstructured_grid_Marker::setBoundary( bool ( *Marker_formula ) ( double x, double y, double z ) )  {
	D3vector middle;


	Quadrangle_el* quad;
	Edge_el*       edge;
	
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		quad = ug->Give_quadrangle ( id );
		if ( quad->Give_exists_exterior() == false) {

			// mark quadrangles
			middle = quad->Give_middle_coord();
			if ( Marker_formula ( middle.x,middle.y,middle.z ) )
				Set_marker<quadrangleEl> ( id, yes_mark );
			else Set_marker<quadrangleEl> ( id, no_mark );

			// mark edges
			for ( int i=0; i<4; ++i ) {
				int idquad = quad->Give_id_edge ( ( dir2D ) i );
				edge = ug->Give_edge ( idquad );

				middle = ( ug->Give_point ( edge->Give_id_corner_W() )->Give_coordinate() +
				           ug->Give_point ( edge->Give_id_corner_E() )->Give_coordinate() ) / 2.0;

				if ( Marker_formula ( middle.x,middle.y,middle.z ) )
					Set_marker<edgeEl> ( idquad, yes_mark );
				else Set_marker<edgeEl> ( idquad, no_mark );
			}

			// mark points
			for ( int i=0; i<4; ++i ) {
				int idcorner =  quad->Give_id_corner ( ( dir2D_sons ) i );
				middle = ug->Give_point ( idcorner )->Give_coordinate();

				if ( Marker_formula ( middle.x,middle.y,middle.z ) )
					Set_marker<pointEl> ( idcorner, yes_mark );
				else Set_marker<pointEl> ( idcorner, no_mark );
			}
		}
	}
}
