/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef MARKER_H
#define MARKER_H

#include <vector>

class Marker {
public:
	Marker( Unstructured_grid* ug_ );
	virtual ~Marker();

    //* select everything
	void on();

    //* select nothing
	void off();

    //*  select block (inkl quad, edge and corner elements) with id chosenBlock
	void selectHexBlock(unsigned int chosenHexBlock);
    //*  select block (inkl quad, edge and corner elements) with id chosenBlock and deselect all others
	void selectSingleHexBlock ( unsigned int chosenHexBlock );
    //*  select range of blocks (inkl quad, edge and corner elements)
	void selectHexBlocks ( std::vector<unsigned int> & chosenHexBlocks );
	void selectHexBlocks ( unsigned int firstBlock, unsigned int numBlocks);

    //*  select quadrangle (excl edge and corner) with id chosenQuad
    void selectQuad(unsigned int chosenQuad);
    void selectSingleQuad(unsigned int chosenQuad);
    void selectQuads( unsigned int firstQuad, unsigned int numQuads);

	template <elementTyp TYP_EL>
	inline marker_typ Give_marker(int i) const;

	template <elementTyp TYP_EL>
	inline void Set_marker(int i, marker_typ mt);

	void operator=(bool setValue) {
		if(setValue) on();
		else off();
	}

    void operatorAnd(Marker * mark);
    void operatorOr(Marker * mark);
    void operatorXOr(Marker * mark);

    void complement (Marker* marker);
    void interior_boundary (Marker* marker);

	void Mark_hexahedra_only( void ); 		///< Marks hexahedra and unmarks everything else
	void Mark_quadrangles_only( void );	///< Marks quadrangles and unmarks everything else
	void Mark_edges_only( void );			///< Marks edges and unmarks everything else
	void Mark_points_only( void );			///< Marks points and unmarks everything else

	void Unmark_hexahedra( void ); 		///< Unmarks hexahedra, keeps everything else
	void Unmark_quadrangles( void );		///< Unmarks quadrangles, keeps everything else
	void Unmark_edges( void );			///< Unmarks edges, keeps everything else
	void Unmark_points( void );			///< Unarks points, keeps everything else
		
protected:
	Unstructured_grid*       ug;

private:
	// own data
	marker_typ* marker_hexahedra;        // num_hexahedra
	marker_typ* marker_quadrangles;      // num_quadrangles
	marker_typ* marker_edges;            // num_edges
	marker_typ* marker_points;           // num_points

	Marker ( Marker &tmp ) {
		throw;
	};
};

class Boundary_Marker : public Marker {
public:
	Boundary_Marker(Unstructured_grid* ug_);
	~Boundary_Marker();
    //* mark boundary in dirction dir3D (Ndir3D, Edir3D... )
    void markBoundaryDir3D(dir3D dir, int blockRangeStart = 0, int blockRangeEnd = -1 );
    void unmarkBoundaryDir3D(dir3D dir, int blockRangeStart = 0, int blockRangeEnd = -1 );

};

class Unstructured_grid_Marker :  public Marker {
public:
	Unstructured_grid_Marker(Unstructured_grid* ug_);
    ~Unstructured_grid_Marker();
	void operator= (bool  (*Marker_formula)(double x, double y, double z));
	void operator=(bool setValue) {
		if(setValue) on();
		else off();
	}
	void setBoundary(bool  (*Marker_formula)(double x, double y, double z));
};

class Blockgrid_Marker :  public Marker {
public:
	Blockgrid_Marker(Unstructured_grid* ug_);
	~Blockgrid_Marker();

};

template <>
inline marker_typ Marker::Give_marker<hexahedronEl>(int i) const {
	return marker_hexahedra[i];
}

template <>
inline marker_typ Marker::Give_marker<quadrangleEl>(int i) const {
	return marker_quadrangles[i];
}

template <>
inline marker_typ Marker::Give_marker<edgeEl>      (int i) const {
	return marker_edges[i];
}

template <>
inline marker_typ Marker::Give_marker<pointEl>     (int i) const {
	return marker_points[i];
}

template <>
inline void Marker::Set_marker<hexahedronEl>(int i, marker_typ mt) {
	marker_hexahedra[i] = mt;
}

template <>
inline void Marker::Set_marker<quadrangleEl>(int i, marker_typ mt) {
	marker_quadrangles[i] = mt;
}

template <>
inline void Marker::Set_marker<edgeEl>      (int i, marker_typ mt) {
	marker_edges[i] = mt;
}

template <>
inline void Marker::Set_marker<pointEl>     (int i, marker_typ mt) {
	marker_points[i] = mt;
}

#endif // MARKER_H
