/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"
#include "examples_ug.h"



/**
///////////////////////////////////////////////////////////////////////////////////////////////////////
//             left_move_z 
//                |   |
//                    -----------------------                   -----------
//                   /                      /
//                  /                      /                        ^
//                 / <- alpha left        /                         |
//                ------------------------                        width_y
//               /|                     /|                          |
//              / |                    / |                          V
//             /  |                   /  |   <- alpha right
//             -----------------------                          -----------
//                                    |  |
//                                - right_move_z
//                | <-    lenght z    -> |
//
///////////////////////////////////////////////////////////////////////////////////////////////////////
**/
SlabList::SlabList(double length_x,  double length_y,
	           double alphaLeft, double alphaRight,
		   std::vector<double>& relativeCoordZ,												 
		   double offsetX, double offsetY, double offsetZ) {
	const int numBlocks(relativeCoordZ.size()-1);
	assert (numBlocks > 0);
	
	const int numInterfaces(relativeCoordZ.size());
	
	const int numpoints(numInterfaces *4);
	Set_number_points ( numpoints );
  
	double left_z_move  = length_y * 0.5 * tan(M_PI*0.5 - alphaLeft);
        double right_z_move = length_y * 0.5 * tan(M_PI*0.5 - alphaRight);
  
/*	
cout << " left_z_move: " << 	left_z_move << endl;
cout << " right_z_move: " << 	right_z_move << endl;
cout << " length_x: " << length_x << endl;
cout << " length_y: " << length_y << endl;
*/
	
	// set points
	for ( int i = 0; i < numInterfaces; ++i )	{
	        double rx = length_x * 0.5;
	        double ry = length_y * 0.5;

		if(i==0) {		
		   Set_coordinate_point ( i*4+0,  D3vector ( offsetX-rx, offsetY-ry, offsetZ+relativeCoordZ.at(i) - left_z_move ) );
		   Set_coordinate_point ( i*4+1,  D3vector ( offsetX+rx, offsetY-ry, offsetZ+relativeCoordZ.at(i) - left_z_move ) );
		   Set_coordinate_point ( i*4+2,  D3vector ( offsetX-rx, offsetY+ry, offsetZ+relativeCoordZ.at(i) + left_z_move ) );
		   Set_coordinate_point ( i*4+3,  D3vector ( offsetX+rx, offsetY+ry, offsetZ+relativeCoordZ.at(i) + left_z_move ) );
		}
		else if(i==numInterfaces-1) {
		   Set_coordinate_point ( i*4+0,  D3vector ( offsetX-rx, offsetY-ry, offsetZ+relativeCoordZ.at(i) + right_z_move) );
		   Set_coordinate_point ( i*4+1,  D3vector ( offsetX+rx, offsetY-ry, offsetZ+relativeCoordZ.at(i) + right_z_move) );
		   Set_coordinate_point ( i*4+2,  D3vector ( offsetX-rx, offsetY+ry, offsetZ+relativeCoordZ.at(i) - right_z_move) );
		   Set_coordinate_point ( i*4+3,  D3vector ( offsetX+rx, offsetY+ry, offsetZ+relativeCoordZ.at(i) - right_z_move) );
		}
		else {
		   Set_coordinate_point ( i*4+0,  D3vector ( offsetX-rx, offsetY-ry, offsetZ+relativeCoordZ.at(i) ) );
		   Set_coordinate_point ( i*4+1,  D3vector ( offsetX+rx, offsetY-ry, offsetZ+relativeCoordZ.at(i) ) );
		   Set_coordinate_point ( i*4+2,  D3vector ( offsetX-rx, offsetY+ry, offsetZ+relativeCoordZ.at(i) ) );
		   Set_coordinate_point ( i*4+3,  D3vector ( offsetX+rx, offsetY+ry, offsetZ+relativeCoordZ.at(i) ) );
		}
	}
	
	// set hexahedra
	Set_number_hexahedra ( numBlocks );
  
	// alle bis auf letzten block
	for (int i = 0; i < numBlocks; ++i ) {
		Set_hexahedron ( i, i*4 + 0, i*4 + 1, i*4 + 2, i*4 + 3,  i*4 + 4, i*4 + 5, i*4 + 6, i*4 + 7 );
	}


	construction_hexahedron_points_done();
	construction_done();
}

