/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "../grid/parti.h"
#include "ug.h"
#include "marker.h"

Zordering::Zordering(Unstructured_grid* ug) {
  // fixed data
  number_indizes = 
    ug->Give_number_hexahedra() + ug->Give_number_quadrangles();
  indices    = new int[number_indizes];
  type_index = new elementTyp[number_indizes];

  // aux data
  bool* used_hex  = new bool[ug->Give_number_hexahedra()];
  bool* used_quad = new bool[ug->Give_number_quadrangles()];
  for(int i=0; i<ug->Give_number_hexahedra();   ++i) used_hex[i]  = false;
  for(int i=0; i<ug->Give_number_quadrangles(); ++i) used_quad[i] = false;

  // sort
  double      min_z, next_z;
  int         min_z_index;
  elementTyp min_z_type;

  for(int next_i=0;next_i<number_indizes;++next_i) {
    // find first
    bool found_first=false;
    for(int j=0; j<ug->Give_number_hexahedra() && found_first==false; ++j) {
      if(used_hex[j]==false) {
	found_first=true;
	min_z = ug->Give_hexahedron(j)->Give_middle_coord().z;
	min_z_index = j;
	min_z_type  = hexahedronEl ;
      }
    }
    for(int j=0; j<ug->Give_number_quadrangles() && found_first==false; ++j) {
      if(used_quad[j]==false) {
	found_first=true;
	min_z = ug->Give_quadrangle(j)->Give_middle_coord().z;
	min_z_index = j;
	min_z_type  = quadrangleEl ;
      }
    }
    // find best
    for(int j=0; j<ug->Give_number_hexahedra(); ++j) {
      if(used_hex[j]==false) {
	next_z = ug->Give_hexahedron(j)->Give_middle_coord().z;
	if(next_z<min_z) {
	  min_z = next_z;
	  min_z_index = j;
	  min_z_type  = hexahedronEl ;
	}
      }
    }
    for(int j=0; j<ug->Give_number_quadrangles(); ++j) {
      if(used_quad[j]==false) {
	next_z = ug->Give_quadrangle(j)->Give_middle_coord().z;
	if(next_z<min_z) {
	  min_z = next_z;
	  min_z_index = j;
	  min_z_type  = quadrangleEl ;
	}
      }
    }
    // set best

    indices[next_i]    = min_z_index;
    type_index[next_i] = min_z_type;
    if(min_z_type == hexahedronEl) used_hex[min_z_index]=true;
    if(min_z_type == quadrangleEl) used_quad[min_z_index]=true;
  }


  // delete aux
  delete[] used_hex;
  delete[] used_quad;
}

Zordering::~Zordering() {
  delete[] indices;
  delete[] type_index;
}


void Zordering::Print(Unstructured_grid* ug) {
  cout << " Printing of z-ordering! " << endl;
  cout << " number hex: "  << ug->Give_number_hexahedra()   << endl;
  cout << " number quad: " << ug->Give_number_quadrangles() << endl;

  for(int i=0;i<number_indizes;++i) {
    cout << i;
    if(type_index[i]==hexahedronEl)
      cout << " hex,  index: " << indices[i] << " coordinate: " <<
	ug->Give_hexahedron(indices[i])->Give_middle_coord().z << endl;
    if(type_index[i]==quadrangleEl)
      cout << " quad, index: " << indices[i] << " coordinate: " <<
	ug->Give_quadrangle(indices[i])->Give_middle_coord().z << endl;
  }
}
