/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// interpolation cell variable
//
// ------------------------------------------------------------


#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "../extemp/extemp.h"
#include "../extemp/parallel.h"
#include "../extemp/variable.h"
#include "../extemp/cellvar.h"
#include "../extemp/co_fu.h"
#include "../extemp/functor.h"
#include "../pde_op/diffop.h"
#include "../pde_op/stenop.h"
#include "../pde_op/stenop_cc.h"


#ifndef INTERPOLCELLVAR_H
#define INTERPOLCELLVAR_H


/** \addtogroup InterpolationOperators **/
/* @{ */  
/**
 * Interpolation from Cell_variable to Variable <br>
 * FAST VERSION <br>
 * Sometimes this version is a little bit less accurate than Interpolate_from_cell_to_point in case of smooth functions <br>
 * This version is more accurate in case of discontinuous functions 
 * But in case of a non smooth function, this version is more appropriate
**/
class FastInterpolate_from_cell_to_point {
 public:
  FastInterpolate_from_cell_to_point();

  void init(Blockgrid* blockgrid_);    
  void deleteData();
  void doInterpolation(Cell_variable<double>* varCell, Variable<double>* varPoint);
  
  ~FastInterpolate_from_cell_to_point() { deleteData(); }
private:
  Variable<double>* weight;  
};




/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////


/**
 * Interpolation by Gauss-Seidel from Cell_variable to Variable <br>
 * template might be Stencil_constant_z <br>
 * SLOW VERSION <br>
 * Sometimes this version is more accurate than FastInterpolate_from_cell_to_point <br>
 * An example showed an increase of factor 3 of accuracy in case of a smooth function 
**/
template<class StenTyp>
class Interpolate_from_cell_to_point {
 public:
  Interpolate_from_cell_to_point(int numberGSinterations_ = 5);

  void init(Blockgrid* blockgrid_);    
  void deleteData();
  void doInterpolation(Cell_variable<double>* varCell, Variable<double>* varPoint);
  
  ~Interpolate_from_cell_to_point() { deleteData(); }
private:
  Blockgrid*      blockgrid;
    
  Sten_matrix<double, StenTyp>* HelmSten;
  Local_stiffness_matrix<double>*  Helm;
  Local_stiffness_matrix<double>*  sp;

  Variable<double>* d;  
  Variable<double>* f;  
  
  int numberGSinterations;
};
/* @} */


template<class StenTyp>
void Interpolate_from_cell_to_point<StenTyp>::doInterpolation(Cell_variable<double>* varCell, Variable<double>* varPoint) {
     assert(Helm!=NULL);
     (*sp) = (*Helm) * (*varCell);
     (*varPoint) = 1.0;
     (*f) = (*sp)(*varPoint);
     
     (*varPoint) = (*f) / (*d); 
     
     for(int i=0;i<numberGSinterations;++i) {
         *varPoint = (*varPoint - 0.9 * ((*HelmSten)(*varPoint) - *f) / HelmSten->diag());
     }
}

template<class StenTyp>
Interpolate_from_cell_to_point<StenTyp>::Interpolate_from_cell_to_point(int numberGSinterations_) {
     blockgrid = NULL;
     HelmSten  = NULL;
     Helm      = NULL;
     sp        = NULL;     
     f         = NULL;  
     d         = NULL;      
     numberGSinterations = numberGSinterations_;
}



template<class StenTyp>
void Interpolate_from_cell_to_point<StenTyp>::deleteData() {
     if(blockgrid!=NULL) {
        if(HelmSten!=NULL) delete HelmSten;
        if(Helm!=NULL)     delete Helm;
        if(sp!=NULL)       delete sp;	
        if(f!=NULL)        delete f;
        if(d!=NULL)        delete d;	
	HelmSten  = NULL;
	Helm      = NULL;
	f         = NULL;
	d         = NULL;	
	blockgrid = NULL;
     }
}


#endif // INTERPOLCELLVAR_H
