/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// diffop.h
// local stiffness matrix
//
// ------------------------------------------------------------

#ifndef DIFFOP_H
#define DIFFOP_H

#include "../extemp/cellvar.h"
///////////////////////////
// 0. Predefinition of class for MG
///////////////////////////


template <typename TYPE2>

class ExpRes_local_stiffness_matrix;

///////////////////////////
// 1.
///////////////////////////

// Gitter:


/////////////////////////////////////
//
//  Orientierung:   lokales Gebiet fuer Jochens code mit
//                  _COLSAMM_::ELEMENTS::_Hexahedron_2<_COLSAMM_::Gauss2>
//
//                    6--7
//                   /| /|
//                  4--5 |
//                  | 2--3
//                  |/ |/
//                  0--1
//
/////////////////////////////////////


template <typename TYPE1, typename TYPE2>

class Expr_Local_stiffness_matrix :
			public Expr<Expr_Local_stiffness_matrix<TYPE1, TYPE2> > {
			
	public:
		Expr_Local_stiffness_matrix ( const Variable<TYPE1>& v, TYPE2**  lm_hex,
		                              TYPE2**  lm_quad_in,
		                              TYPE2**  lm_quad_ex,
		                              TYPE2*** lm_neigh_ed,
		                              TYPE2**  lm_neigh_po ) :
				v_ ( v ), loc_m_hexahedra ( lm_hex ),
				loc_m_interior_quadrangles ( lm_quad_in ),
				loc_m_exterior_quadrangles ( lm_quad_ex ),
				loc_m_neighbor_edges ( lm_neigh_ed ),
				loc_m_neighbor_points ( lm_neigh_po ) {}
				
		Blockgrid* Give_blockgrid() const { return v_.Give_blockgrid(); };
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef _TypeOf0_ ( TYPE1, TYPE2 ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
            v_.template Update<TYP_EL> ( id );
		}
		
	private:
		const Variable<TYPE1>& v_;
		TYPE2**  loc_m_hexahedra;                  // num_hexahedra
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
};

template <>
template <>
inline double Expr_Local_stiffness_matrix<double, double>::Give_data<hexahedronEl> ( params_in ) const {
	int u, v;
	dir3D_sons ce;
	double sum;
	sum = 0.0;
	
	for ( v = 0;v < 8;++v ) { // test function v and cell opposite3D(v)
		ce = opposite3D ( ( dir3D_sons ) v );
		
		for ( u = 0;u < 8;++u ) {
			sum = sum +
			      loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( x1DCoord ( ce )-1+i,
			                          y1DCoord ( ce )-1+j,
			                          z1DCoord ( ce )-1+k ) *64+8*v+u] *
			      v_.Give_data<hexahedronEl> ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                    y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j,
			                                    z1DCoord ( ce ) + z1DCoord ( ( dir3D_sons ) u ) - 1 + k, Nx, Ny );
		}
	}
	
	return sum;
}

template <>
template <>
inline std::complex<double>
Expr_Local_stiffness_matrix<std::complex<double>, std::complex<double> >::
Give_data<hexahedronEl> ( params_in ) const {
	int u, v;
	dir3D_sons ce;
	std::complex<double> sum;
	sum = 0.0;
	std::complex<double> temp;
	int mx, my, mz;
	
	for ( v = 0;v < 8;++v ) { // test function v and cell opposite3D(v)
		ce = opposite3D ( ( dir3D_sons ) v );
		
		for ( u = 0;u < 8;++u ) {
			mx = x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i;
			my = y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j;
			mz = z1DCoord ( ce ) + z1DCoord ( ( dir3D_sons ) u ) - 1 + k;
			
			temp = v_.Give_data<hexahedronEl> ( id, mx, my, mz, Nx, Ny );
			sum  = sum + temp * loc_m_hexahedra[id][
			           Ind_loc_matrix_hexahedra ( x1DCoord ( ce )-1+i,
			                                      y1DCoord ( ce )-1+j,
			                                      z1DCoord ( ce )-1+k ) *64+8*v+u];
		}
	}
	
	return sum;
}



template <>
template <>
inline double Expr_Local_stiffness_matrix<double, double>::Give_data<quadrangleEl> ( params_in ) const {
	int u, v;
	dir3D_sons ce;
	double sum;
	sum = 0.0;
	
	for ( v = 0;v < 4;++v ) { // test function v and cell opposite3D(v)
		ce = opposite3D ( ( dir3D_sons ) v );
		
		for ( u = 0;u < 4;++u )
			sum = sum +
			      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+u] *    // D
			      v_.Give_data_quadrangle_middle ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                       y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx ) +
			      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+ ( u+WSTdir3D ) ] * // T
			      v_.Give_data_quadrangle_interior ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                         y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx );
	}
	
	if ( Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle ( id )->
	        Give_exists_exterior() == false )
		return sum;
		
	for ( v = 0;v < 4;++v ) { // test function v and cell opposite3D(v)
		ce = opposite3D ( ( dir3D_sons ) v );
		
		for ( u = 0;u < 4;++u )
			sum = sum +
			      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+u] * // D
			      v_.Give_data_quadrangle_middle ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                       y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx ) +
			      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+ ( u+WSTdir3D ) ] * // T
			      v_.Give_data_quadrangle_exterior ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                         y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx );
	}
	
	/*
	   std::cout << "quad locsten: "
	   << " id: "  << id
	   << " i: " << i
	   << " j: " << j
	   << " k: " << k
	   << " Nx: " << Nx
	   << " Ny: " << Ny
	   << " val: " << sum
	   <<  std::endl;
	*/
	
	
	return sum;
}




template <>
template <>
inline std::complex<double> Expr_Local_stiffness_matrix<std::complex<double>, std::complex<double>  >
::Give_data<quadrangleEl> ( params_in ) const {
	int u, v;
	dir3D_sons ce;
	std::complex<double> sum;
	sum = 0.0;
	
	for ( v = 0;v < 4;++v ) { // test function v and cell opposite3D(v)
		ce = opposite3D ( ( dir3D_sons ) v );
		
		for ( u = 0;u < 4;++u )
			sum = sum +
			      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+u] * // D
			      v_.Give_data_quadrangle_middle ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                       y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx ) +
			      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+ ( u+WSTdir3D ) ] * // T
			      v_.Give_data_quadrangle_interior ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                         y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx );
	}
	
	if ( Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle ( id )->
	        Give_exists_exterior() == false )
		return sum;
		
	for ( v = 0;v < 4;++v ) { // test function v and cell opposite3D(v)
		ce = opposite3D ( ( dir3D_sons ) v );
		
		for ( u = 0;u < 4;++u )
			sum = sum +
			      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+u] * // D
			      v_.Give_data_quadrangle_middle ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                       y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx ) +
			      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles
			                                     ( x1DCoord ( ce )-1+i,y1DCoord ( ce )-1+j ) *32+8*v+ ( u+WSTdir3D ) ] * // T
			      v_.Give_data_quadrangle_exterior ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i,
			                                         y1DCoord ( ce ) + y1DCoord ( ( dir3D_sons ) u ) - 1 + j, Nx );
	}
	
	return sum;
}

template <>
template <>
inline double Expr_Local_stiffness_matrix<double, double>::Give_data<edgeEl> ( params_in ) const {
	int u, v;
	int num;
	int id_ND, id_ST, id_NT;
	dir3D_sons ce;
	double sum;
	sum = 0.0;
	int number_neighbor_hex =
	    Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->Give_number_neighbors_hex();
	    
	for ( num = 0;num < number_neighbor_hex;++num ) {
		id_ND = Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->
		        Give_loc_id_hex ( num, ND );
		id_ST = Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->
		        Give_loc_id_hex ( num, ST );
		id_NT = Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->
		        Give_loc_id_hex ( num, NT );
		        
		for ( v = 0;v < 2;++v ) { // test function v and cell opposite3D(v)
			ce = opposite3D ( ( dir3D_sons ) v );
			
			for ( u = 0;u < 2;++u )
				sum = sum +
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+u] *   // SD
				      v_.Give_data_edge ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i ) +
				      
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+ ( u+WNDdir3D ) ] *  //ND
				      v_.Give_data_neighbor_edge ( id, id_ND, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i ) +
				      
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+ ( u+WSTdir3D ) ] * // ST
				      v_.Give_data_neighbor_edge ( id, id_ST, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i ) +
				      
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+ ( u+WNTdir3D ) ] *  // NT
				      v_.Give_data_neighbor_edge ( id, id_NT, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i );
		}
	}
	
	return sum;
}

template <>
template <>
inline std::complex<double> Expr_Local_stiffness_matrix<std::complex<double>, std::complex<double> >::
Give_data<edgeEl> ( params_in ) const {
	int u, v;
	int num;
	int id_ND, id_ST, id_NT;
	dir3D_sons ce;
	std::complex<double> sum;
	sum = 0.0;
	int number_neighbor_hex =
	    Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->Give_number_neighbors_hex();
	    
	for ( num = 0;num < number_neighbor_hex;++num ) {
		id_ND = Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->
		        Give_loc_id_hex ( num, ND );
		id_ST = Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->
		        Give_loc_id_hex ( num, ST );
		id_NT = Give_blockgrid()->Give_unstructured_grid()->Give_edge ( id )->
		        Give_loc_id_hex ( num, NT );
		        
		for ( v = 0;v < 2;++v ) { // test function v and cell opposite3D(v)
			ce = opposite3D ( ( dir3D_sons ) v );
			
			for ( u = 0;u < 2;++u )
				sum = sum +
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+u] *   // SD
				      v_.Give_data_edge ( id, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i ) +
				      
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+ ( u+WNDdir3D ) ] *  //ND
				      v_.Give_data_neighbor_edge ( id, id_ND, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i ) +
				      
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+ ( u+WSTdir3D ) ] * // ST
				      v_.Give_data_neighbor_edge ( id, id_ST, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i ) +
				      
				      loc_m_neighbor_edges[id][num][ ( x1DCoord ( ce )-1+i ) *16+8*v+ ( u+WNTdir3D ) ] *  // NT
				      v_.Give_data_neighbor_edge ( id, id_NT, x1DCoord ( ce ) + x1DCoord ( ( dir3D_sons ) u ) - 1 + i );
		}
	}
	
	return sum;
}

template <>
template <>
inline double Expr_Local_stiffness_matrix<double, double>::Give_data<pointEl> ( params_in ) const {
	int num;
	int id_SED, id_NWD, id_NED, id_SWT, id_SET, id_NWT, id_NET;
	dir3D_sons ce;
	double sum;
	sum = 0.0;
	int number_neighbor_hex =
	    Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->Give_number_neighbors_hex();
	    
	for ( num = 0;num < number_neighbor_hex;++num ) {
		id_SED = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, SED );
		id_NWD = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NWD );
		id_NED = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NED );
		id_SWT = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, SWT );
		id_SET = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, SET );
		id_NWT = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NWT );
		id_NET = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NET );
		ce = ENTdir3D; //cell
		sum = sum +
		      loc_m_neighbor_points[id][num*8+WSDdir3D] *   // SWD
		      v_.Give_data_point ( id ) +
		      
		      loc_m_neighbor_points[id][num*8+ESDdir3D] *  // SED
		      v_.Give_data_neighbor_point ( id, id_SED ) +
		      
		      loc_m_neighbor_points[id][num*8+WNDdir3D] *  // NWD
		      v_.Give_data_neighbor_point ( id, id_NWD ) +
		      
		      loc_m_neighbor_points[id][num*8+ENDdir3D] *  // NED
		      v_.Give_data_neighbor_point ( id, id_NED ) +
		      
		      loc_m_neighbor_points[id][num*8+WSTdir3D] *  // SWT
		      v_.Give_data_neighbor_point ( id, id_SWT ) +
		      
		      loc_m_neighbor_points[id][num*8+ESTdir3D] *  // SET
		      v_.Give_data_neighbor_point ( id, id_SET ) +
		      
		      loc_m_neighbor_points[id][num*8+WNTdir3D] *  // NWT
		      v_.Give_data_neighbor_point ( id, id_NWT ) +
		      
		      loc_m_neighbor_points[id][num*8+ENTdir3D] *  // NET
		      v_.Give_data_neighbor_point ( id, id_NET );
	}
	
	return sum;
}

template <>
template <>
inline std::complex<double> Expr_Local_stiffness_matrix<std::complex<double>, std::complex<double> >::
Give_data<pointEl> ( params_in ) const {
	int num;
	int id_SED, id_NWD, id_NED, id_SWT, id_SET, id_NWT, id_NET;
	dir3D_sons ce;
	std::complex<double> sum;
	sum = 0.0;
	int number_neighbor_hex =
	    Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->Give_number_neighbors_hex();
	    
	for ( num = 0;num < number_neighbor_hex;++num ) {
		id_SED = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, SED );
		id_NWD = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NWD );
		id_NED = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NED );
		id_SWT = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, SWT );
		id_SET = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, SET );
		id_NWT = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NWT );
		id_NET = Give_blockgrid()->Give_unstructured_grid()->Give_point ( id )->
		         Give_loc_id_hex ( num, NET );
		ce = ENTdir3D; //cell
		sum = sum +
		      loc_m_neighbor_points[id][num*8+WSDdir3D] *   // SWD
		      v_.Give_data_point ( id ) +
		      
		      loc_m_neighbor_points[id][num*8+ESDdir3D] *  // SED
		      v_.Give_data_neighbor_point ( id, id_SED ) +
		      
		      loc_m_neighbor_points[id][num*8+WNDdir3D] *  // NWD
		      v_.Give_data_neighbor_point ( id, id_NWD ) +
		      
		      loc_m_neighbor_points[id][num*8+ENDdir3D] *  // NED
		      v_.Give_data_neighbor_point ( id, id_NED ) +
		      
		      loc_m_neighbor_points[id][num*8+WSTdir3D] *  // SWT
		      v_.Give_data_neighbor_point ( id, id_SWT ) +
		      
		      loc_m_neighbor_points[id][num*8+ESTdir3D] *  // SET
		      v_.Give_data_neighbor_point ( id, id_SET ) +
		      
		      loc_m_neighbor_points[id][num*8+WNTdir3D] *  // NWT
		      v_.Give_data_neighbor_point ( id, id_NWT ) +
		      
		      loc_m_neighbor_points[id][num*8+ENTdir3D] *  // NET
		      v_.Give_data_neighbor_point ( id, id_NET );
	}
	
	return sum;
}

template <typename TYPE2>

class Expr_Diag_Local_stiffness_matrix :
			public Expr<Expr_Diag_Local_stiffness_matrix<TYPE2> > {
			
	public:
		Expr_Diag_Local_stiffness_matrix ( TYPE2** lm_hex, TYPE2** lm_quad_in, TYPE2** lm_quad_ex,
		                                   TYPE2*** lm_neigh_ed, TYPE2** lm_neigh_po, Blockgrid* blockgrid_ ) :
				loc_m_hexahedra ( lm_hex ),
				loc_m_interior_quadrangles ( lm_quad_in ),
				loc_m_exterior_quadrangles ( lm_quad_ex ),
				loc_m_neighbor_edges ( lm_neigh_ed ),
				loc_m_neighbor_points ( lm_neigh_po ),
				blockgrid ( blockgrid_ ) {}
				
		Blockgrid* Give_blockgrid() const {
			return blockgrid;
		};
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		
		typedef TYPE2 Result;
		
		template <elementTyp TYP_EL>
		inline TYPE2 Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		TYPE2**  loc_m_hexahedra;                  // num_hexahedra
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid* blockgrid;
};

template <>
template <>
inline double Expr_Diag_Local_stiffness_matrix<double>::Give_data<hexahedronEl> ( params_in ) const {
	return loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j  ,k  ) *64   ]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j  ,k  ) *64+ 9]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j-1,k  ) *64+18]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j-1,k  ) *64+27]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j  ,k-1 ) *64+36]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j  ,k-1 ) *64+45]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j-1,k-1 ) *64+54]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j-1,k-1 ) *64+63];
}

template <>
template <>
inline std::complex<double> Expr_Diag_Local_stiffness_matrix<std::complex<double> >::
Give_data<hexahedronEl> ( params_in ) const {
	return loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j  ,k  ) *64   ]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j  ,k  ) *64+ 9]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j-1,k  ) *64+18]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j-1,k  ) *64+27]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j  ,k-1 ) *64+36]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j  ,k-1 ) *64+45]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i  ,j-1,k-1 ) *64+54]
	       + loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i-1,j-1,k-1 ) *64+63];
}

template <>
template <>
inline double Expr_Diag_Local_stiffness_matrix<double>::Give_data<quadrangleEl> ( params_in ) const {
	if ( Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle ( id )->Give_exists_exterior() )
		return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j  ) *32   ]
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j  ) *32+ 9]
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j-1 ) *32+18]
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j-1 ) *32+27]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j  ) *32   ]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j  ) *32+ 9]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j-1 ) *32+18]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j-1 ) *32+27];
		       
	return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j  ) *32   ]
	       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j  ) *32+ 9]
	       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j-1 ) *32+18]
	       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j-1 ) *32+27];
}

template <>
template <>
inline std::complex<double> Expr_Diag_Local_stiffness_matrix<std::complex<double> >
::Give_data<quadrangleEl> ( params_in ) const {
	if ( Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle ( id )->Give_exists_exterior() )
		return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j  ) *32   ]
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j  ) *32+ 9]
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j-1 ) *32+18]
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j-1 ) *32+27]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j  ) *32   ]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j  ) *32+ 9]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j-1 ) *32+18]
		       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j-1 ) *32+27];
		       
	return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j  ) *32   ]
	       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j  ) *32+ 9]
	       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i  ,j-1 ) *32+18]
	       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles ( i-1,j-1 ) *32+27];
}

template <>
template <>
inline double Expr_Diag_Local_stiffness_matrix<double>::Give_data<edgeEl> ( params_in ) const {
	int number_neighbor_hex =
	    blockgrid->Give_unstructured_grid()->Give_edge ( id )->Give_number_neighbors_hex();
	double sum = 0.0;
	
	for ( int num = 0;num < number_neighbor_hex;++num )
		sum += loc_m_neighbor_edges[id][num][ i   *16  ]
		       + loc_m_neighbor_edges[id][num][ ( i-1 ) *16+9];
		       
	return sum;
}

template <>
template <>
inline std::complex<double> Expr_Diag_Local_stiffness_matrix<std::complex<double> >::
Give_data<edgeEl> ( params_in ) const {
	int number_neighbor_hex =
	    blockgrid->Give_unstructured_grid()->Give_edge ( id )->Give_number_neighbors_hex();
	std::complex<double> sum ( 0.0 );
	
	for ( int num = 0;num < number_neighbor_hex;++num )
		sum += loc_m_neighbor_edges[id][num][ i   *16  ]
		       + loc_m_neighbor_edges[id][num][ ( i-1 ) *16+9];
		       
	return sum;
}

template <>
template <>
inline double Expr_Diag_Local_stiffness_matrix<double>::Give_data<pointEl> ( params_in ) const {
	int number_neighbor_hex =
	    blockgrid->Give_unstructured_grid()->Give_point ( id )->Give_number_neighbors_hex();
	double sum = 0.0;
	
	for ( int num = 0;num < number_neighbor_hex;++num )
		sum += loc_m_neighbor_points[id][num*8];
		
	return sum;
}

template <>
template <>
inline std::complex<double> Expr_Diag_Local_stiffness_matrix<std::complex<double> >::
Give_data<pointEl> ( params_in ) const {
	int number_neighbor_hex =
	    blockgrid->Give_unstructured_grid()->Give_point ( id )->Give_number_neighbors_hex();
	std::complex<double> sum ( 0.0 );
	
	for ( int num = 0;num < number_neighbor_hex;++num )
		sum += loc_m_neighbor_points[id][num*8];
		
	return sum;
}


/**
Class for local stiffness matrix
**/
template <typename TYPE2>

class Local_stiffness_matrix : public Expr<Local_stiffness_matrix<TYPE2> > {
	public:
	
		typedef TYPE2 Result;
		
		Local_stiffness_matrix ( Blockgrid& bg_ );
		~Local_stiffness_matrix();
		
		template <class B>
		void Calculate ( const B& bilinear_form );
		
		template <class B>
		void Calculate_boundary ( const B& bilinear_form );
		
        template <class B>
        void Calculate_boundary ( const B& bilinear_form, Marker & marker, bool addToLocalStiffnesMatrix );

		template <class B>
		void Calculate_all_right_boundaries ( const B& bilinear_form );
		
		template <class B>
		void Calculate_all_left_boundaries(const B& bilinear_form);

        void Print();
		void Update();
		
		void makePetrovGalerkinZ();
		
		//  void addPhaseShift(const Cell_variable<std::complex<double> > & shift);
		void addPhaseShift ( const Cell_variable<TYPE2> & shift );
		
		Expr_Diag_Local_stiffness_matrix<TYPE2> diag () const;
		
		template <typename TYPE1>
		Expr_Local_stiffness_matrix<TYPE1, TYPE2> operator() ( const Variable<TYPE1>& v ) const;
		
		template <class A>
		void operator= ( const Expr<A>& a );
		
		void operator= ( const ExpRes_local_stiffness_matrix<TYPE2>& Expr_res );
		
		void operator= ( const Local_stiffness_matrix& a );
		
		void multiplyTestfunction_v_with ( const Variable<TYPE2>& coeff );
		
		
		inline TYPE2 Give_matrix_hexahedra ( params_in_loc_matrix ) const;
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		
		TYPE2** Give_loc_m_hexahedra() const { return loc_m_hexahedra; };
		
		TYPE2*  Give_loc_m_interior_quadrangles ( int id ) const { return loc_m_interior_quadrangles[id]; }
		
		TYPE2*  Give_loc_m_exterior_quadrangles ( int id ) const { return loc_m_exterior_quadrangles[id]; }
		
		TYPE2** Give_loc_m_neighbor_edges ( int id )       const { return loc_m_neighbor_edges[id]; }
		
		TYPE2*  Give_loc_m_neighbor_points ( int id )      const { return loc_m_neighbor_points[id]; }
		
	private:
		TYPE2**  loc_m_hexahedra;                  // num_hexahedra
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid* blockgrid;
		
		// for parallel
		int my_rank;
		MPI_Comm comm;
};


//////////////////
// Implementierung von Memberfunktionen
//////////////////

template <typename TYPE2>
inline TYPE2 Local_stiffness_matrix<TYPE2>::Give_matrix_hexahedra ( params_in_loc_matrix ) const {
	return loc_m_hexahedra[id_hex][ind_loc_matrix];
}

template <typename TYPE2>
template <typename TYPE1>
Expr_Local_stiffness_matrix<TYPE1, TYPE2> Local_stiffness_matrix<TYPE2>::operator()
( const Variable<TYPE1>& v ) const {
	if ( Give_blockgrid() != v.Give_blockgrid() )
		 std::cout << " wrong grids in Local_stiffness_matrix( Variable v)! "
		<<  std::endl;
		
	return Expr_Local_stiffness_matrix<TYPE1, TYPE2> ( v, loc_m_hexahedra, loc_m_interior_quadrangles,
	        loc_m_exterior_quadrangles,
	        loc_m_neighbor_edges, loc_m_neighbor_points );
}

template <typename TYPE2>
template <class B>
void Local_stiffness_matrix<TYPE2>::Calculate ( const B& bilinear_form ) {
	int id;
	int Nx, Ny, Nz, N_total;
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	for(id = 0; id < blockgrid->Give_unstructured_grid()->Give_number_hexahedra(); ++id ) {
	    if(ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
	       Nx = blockgrid->Give_Nx_hexahedron ( id );
	       Ny = blockgrid->Give_Ny_hexahedron ( id );
	       Nz = blockgrid->Give_Nz_hexahedron ( id );
	       N_total = Nx * Ny * Nz * 64;
		
	       if(ug->isConstantCellInZ()) {
		  int k = 0;
		  #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
		  for(int j = 0;j < Ny;++j) {
	              std::vector<std::vector<TYPE2> > loc_stiff;
	              std::vector<double> points ( 24, 0. ); //Länge 24
	              _COLSAMM_::ELEMENTS::_Hexahedron_2<_COLSAMM_::Gauss2, TYPE2> G;
				

			  for(int i = 0;i < Nx;++i ) {		
			      if(developer_version)  if(Ind_loc_matrix_hexahedra ( i, j, k ) *64 >= N_total )
						        std::cout << " error in Local_stiffness_matrix::Calculate" <<  std::endl;
								
				 points[0] = blockgrid->Give_coord_hexahedron ( id, i, j, k ).x;
				 points[1] = blockgrid->Give_coord_hexahedron ( id, i, j, k ).y;
				 points[2] = blockgrid->Give_coord_hexahedron ( id, i, j, k ).z;
						
				 points[3] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k ).x;
				 points[4] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k ).y;
				 points[5] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k ).z;
						
				 points[6] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k ).x;
				 points[7] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k ).y;
				 points[8] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k ).z;
						
				 points[9] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k ).x;
				 points[10] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k ).y;
				 points[11] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k ).z;
						
				 points[12] = blockgrid->Give_coord_hexahedron ( id, i, j, k + 1 ).x;
				 points[13] = blockgrid->Give_coord_hexahedron ( id, i, j, k + 1 ).y;
				 points[14] = blockgrid->Give_coord_hexahedron ( id, i, j, k + 1 ).z;
						
				 points[15] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k + 1 ).x;
				 points[16] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k + 1 ).y;
				 points[17] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k + 1 ).z;
						
				 points[18] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k + 1 ).x;
				 points[19] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k + 1 ).y;
				 points[20] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k + 1 ).z;
						
				 points[21] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k + 1 ).x;
				 points[22] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k + 1 ).y;
				 points[23] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k + 1 ).z;
						
				 G(points);
						
				 loc_stiff = G.integrate ( bilinear_form );
						
				 for(int v = 0;v < 8;++v )
				     for(int u = 0;u < 8;++u )
				         loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u]
								= loc_stiff[u][v];
			  }
		  }
		  
                  #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)		  
		  for(int k = 1;k < Nz;++k ) {
		      for(int j = 0;j < Ny;++j) {
			  for(int i = 0;i < Nx;++i ) {	
			      for(int vu = 0;vu < 64;++vu )
		                  loc_m_hexahedra[id][Ind_loc_matrix_hexahedra(i,j,k) *64+vu] =
				      loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,0 ) *64+vu];
			  }
		      }
		  }
	       }
	       else {
		  #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
		  for(int k = 0;k < Nz;++k ) {
	              std::vector<std::vector<TYPE2> > loc_stiff;
	              std::vector<double> points ( 24, 0. ); //Länge 24
	              _COLSAMM_::ELEMENTS::_Hexahedron_2<_COLSAMM_::Gauss2, TYPE2> G;
				
		      for(int j = 0;j < Ny;++j )
			  for(int i = 0;i < Nx;++i ) {		
			      if(developer_version)  if(Ind_loc_matrix_hexahedra ( i, j, k ) *64 >= N_total )
						        std::cout << " error in Local_stiffness_matrix::Calculate" <<  std::endl;
								
				 points[0] = blockgrid->Give_coord_hexahedron ( id, i, j, k ).x;
				 points[1] = blockgrid->Give_coord_hexahedron ( id, i, j, k ).y;
				 points[2] = blockgrid->Give_coord_hexahedron ( id, i, j, k ).z;
						
				 points[3] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k ).x;
				 points[4] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k ).y;
				 points[5] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k ).z;
						
				 points[6] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k ).x;
				 points[7] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k ).y;
				 points[8] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k ).z;
						
				 points[9] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k ).x;
				 points[10] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k ).y;
				 points[11] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k ).z;
						
				 points[12] = blockgrid->Give_coord_hexahedron ( id, i, j, k + 1 ).x;
				 points[13] = blockgrid->Give_coord_hexahedron ( id, i, j, k + 1 ).y;
				 points[14] = blockgrid->Give_coord_hexahedron ( id, i, j, k + 1 ).z;
						
				 points[15] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k + 1 ).x;
				 points[16] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k + 1 ).y;
				 points[17] = blockgrid->Give_coord_hexahedron ( id, i + 1, j, k + 1 ).z;
						
				 points[18] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k + 1 ).x;
				 points[19] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k + 1 ).y;
				 points[20] = blockgrid->Give_coord_hexahedron ( id, i, j + 1, k + 1 ).z;
						
				 points[21] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k + 1 ).x;
				 points[22] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k + 1 ).y;
				 points[23] = blockgrid->Give_coord_hexahedron ( id, i + 1, j + 1, k + 1 ).z;
						
				 G(points);
						
				 loc_stiff = G.integrate ( bilinear_form );
						
				 for(int v = 0;v < 8;++v )
				     for(int u = 0;u < 8;++u )
				         loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u]
								= loc_stiff[u][v];
			  }
		  }
		 
		 
	       }
	    }
	}
	
	
	Update();
}

inline void calculate_map ( dir3D_sons map_sons[8], dir3D c_x, dir3D c_y, dir3D c_z ) {
	dir3D  map3D[6];
	map3D[Edir3D] = c_x;
	map3D[Wdir3D] = opposite3D ( c_x );
	map3D[Ndir3D] = c_y;
	map3D[Sdir3D] = opposite3D ( c_y );
	map3D[Tdir3D] = c_z;
	map3D[Ddir3D] = opposite3D ( c_z );
	
	for ( int i = 0;i < 8;++i )
		map_sons[i] = ( dir3D_sons ) ( transform_part ( map3D[xCoord ( ( dir3D_sons ) i ) ] ) +
		                             transform_part ( map3D[yCoord ( ( dir3D_sons ) i ) ] ) +
		                             transform_part ( map3D[zCoord ( ( dir3D_sons ) i ) ] ) );
		                             
	// test
	if ( developer_version ) {
		int sp = 0;
		
		for ( int i = 0;i < 8;++i )
			sp = sp + map_sons[i] - i;
			
		if ( sp != 0 ) std::cout << "error in calculate_map" << std::endl;
	}
}

template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::Update() {

	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	int i, j, u, v, ed, c, id_hex, id_quad, id_edge, id_point;
	int Ni, Nj, Nconst;
	int Nxquad, Nyquad, N_total;
	int Nxhex, Nyhex, Nzhex;
	int Nxedge;
	int id_SW, id_SE, id_NW, idsp, id_W, id_E;
	int num_neighbor, num_hex, num, number_neighbors_hex;
	dir3D_sons c_SW , c_SE, c_NW;
	dir3D   c_x, c_y, c_z;
	dir3D_sons c_W, c_E, SED_corner, NWD_corner, SWT_corner;
	dir3D_sons map_sons[8];
	TYPE2* loc_pointer;
	
	int rank_send, rank_rec;  // for parallel
	
	for ( id_hex = 0;id_hex < ug->Give_number_hexahedra();++id_hex ) {
	
		rank_send = ug->Give_hexahedron ( id_hex )->Give_rank();   // for parallel
		
		Nxhex = blockgrid->Give_Nx_hexahedron ( id_hex );
		Nyhex = blockgrid->Give_Ny_hexahedron ( id_hex );
		Nzhex = blockgrid->Give_Nz_hexahedron ( id_hex );
		
		//Update local stiffness matrix for quadrangles
		
		for ( int fc = 0;fc < 6;++fc ) {
			id_quad = ug->Give_hexahedron ( id_hex )->Give_id_quadrangle ( ( dir3D ) fc );
			
			rank_rec = ug->Give_quadrangle ( id_quad )->Give_rank();  // for parallel
			
			if ( rank_send == my_rank || rank_rec == my_rank ) {  // for parallel
			
				Nxquad = blockgrid->Give_Nx_quadrangle ( id_quad );
				Nyquad = blockgrid->Give_Ny_quadrangle ( id_quad );
				N_total = Nxquad * Nyquad * 32;
				
				// a) coordinate points of quadrangle
				id_SW = ug->Give_quadrangle ( id_quad )->Give_id_corner ( SWdir2D );
				id_SE = ug->Give_quadrangle ( id_quad )->Give_id_corner ( SEdir2D );
				id_NW = ug->Give_quadrangle ( id_quad )->Give_id_corner ( NWdir2D );
				
				// b) Find corresponding points in hexahedron
				
				for ( c = 0;c < 8;++c ) {
					idsp = ug->Give_hexahedron ( id_hex )->Give_id_corner ( ( dir3D_sons ) c );
					
					if ( idsp == id_SW ) c_SW = ( dir3D_sons ) c;
					else if ( idsp == id_SE ) c_SE = ( dir3D_sons ) c;
					else if ( idsp == id_NW ) c_NW = ( dir3D_sons ) c;
				}
				
				c_x = direction_from_to ( c_SW, c_SE );
				
				c_y = direction_from_to ( c_SW, c_NW );
				c_z = opposite3D ( ( dir3D ) fc );   // hex = top
				
				// x - direction
				Nconst = Give_Nconst ( ( dir3D ) c_x, Nxhex - 1, Nyhex - 1, Nzhex - 1 );
				Ni     = Give_Ni (    ( dir3D ) c_x, Nxhex - 1, Nyhex - 1, Nzhex - 1 );
				
				// y - direction
				Nconst = Give_Nconst ( ( dir3D ) c_y, Nxhex - 1, Nyhex - 1, Nzhex - 1 ) + Nconst;
				Nj     = Give_Ni (    ( dir3D ) c_y, Nxhex - 1, Nyhex - 1, Nzhex - 1 );
				
				// empty - direction
				Nconst = Give_Nconst ( ( dir3D ) c_z,
				                       Nxhex - 1, Nyhex - 1, Nzhex - 1 ) + Nconst;
				                       
				// c) Determine if hex is interior or exterior volume
				
				if ( ug->Give_quadrangle ( id_quad )->Give_id_interior_vol() == id_hex ) { //interior
					loc_pointer = loc_m_interior_quadrangles[id_quad];
				} else { // exterior
					loc_pointer = loc_m_exterior_quadrangles[id_quad];
				}
				
				// calculate_map
				calculate_map ( map_sons, c_x, c_y, c_z );
				
				if ( rank_send == rank_rec )     // for parallel
                                        // das sollte man nur parallel machen wenn Ni klein ist	
			                //#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)			  				  
					for ( j = 0;j < Nyquad;++j )
						for ( i = 0;i < Nxquad;++i ) {
							if ( developer_version )
								if ( ( i + Nxquad*j ) *32 >= N_total )
									 std::cout << " error 1 in Local_stiffness_matrix::Update" << std::endl;
									
							for ( v = 0;v < 4;++v )
								for ( u = 0;u < 8;++u )
									loc_pointer[ ( i+Nxquad*j ) *32+8*v+u]
									= loc_m_hexahedra[id_hex][ ( i*Ni+j*Nj+Nconst ) *64+
									                           8*map_sons[v] + map_sons[u]];
						}
				else {
					if ( my_rank == rank_send ) {
						TYPE2* data = Parallel_buffer<TYPE2>::Give_send ( Nyquad * Nxquad * 32 );  // for parallel
                                                // das sollte man nur parallel machen wenn Ni klein ist	
			                        //#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)						
						for ( j = 0;j < Nyquad;++j )
							for ( i = 0;i < Nxquad;++i ) {
								if ( developer_version )
									if ( ( i + Nxquad*j ) *32 >= N_total )
										std::cout << " error 2 in Local_stiffness_matrix::Update"
										<< std::endl;
										
								for ( v = 0;v < 4;++v )
									for ( u = 0;u < 8;++u )
										data[ ( i+Nxquad*j ) *32+8*v+u] =
										    loc_m_hexahedra[id_hex][ ( i*Ni+j*Nj+Nconst ) *64+
										                             8*map_sons[v] + map_sons[u]];
							}
							
						MPI_Send ( data, Nyquad*Nxquad*32,
						
						           Give_MPI_data_typ<TYPE2>::Give(), rank_rec, fc, comm );
					} else {
						if ( developer_version )
							if ( my_rank != rank_rec )
								std::cout << " error 3 in Local_stiffness_matrix::Update" << std::endl;
								
						MPI_Status stat;
						
						MPI_Recv ( loc_pointer, Nyquad*Nxquad*32, Give_MPI_data_typ<TYPE2>::Give(),
						           rank_send, fc, comm, &stat );
					}
				}
			}
		}
		
		//Update local stiffness matrix for edges
		for ( ed = 0;ed < 12;++ed ) {
			id_edge = ug->Give_hexahedron ( id_hex )->Give_id_edge ( ( Edges_cell ) ed );
			
			rank_rec = ug->Give_edge ( id_edge )->Give_rank();  // for parallel
			
			if ( rank_send == my_rank || rank_rec == my_rank ) {  // for parallel
			
				Nxedge = blockgrid->Give_Nx_edge ( id_edge );
				N_total = Nxedge * 16;
				
				// a) coordinate points of edge
				id_W = ug->Give_edge ( id_edge )->Give_id_corner_W();
				id_E = ug->Give_edge ( id_edge )->Give_id_corner_E();
				
				// b) Find corresponding points in hexahedron
				
				for ( c = 0;c < 8;++c ) {
					idsp = ug->Give_hexahedron ( id_hex )->Give_id_corner ( ( dir3D_sons ) c );
					
					if ( idsp == id_W ) c_W = ( dir3D_sons ) c;
					else if ( idsp == id_E ) c_E = ( dir3D_sons ) c;
				}
				
				c_x = direction_from_to ( c_W, c_E );
				
				c_y = opposite3D ( OrthoD ( ( Edges_cell ) ed ) );
				c_z = opposite3D ( OrthoC ( ( Edges_cell ) ed ) );
				
				if ( ug->Give_hexahedron ( id_hex )->Give_id_corner ( Transform ( ( Edges_cell ) ed, Ldir1D ) ) >
				        ug->Give_hexahedron ( id_hex )->Give_id_corner ( Transform ( ( Edges_cell ) ed, Rdir1D ) ) ) {
					c_y = opposite3D ( OrthoC ( ( Edges_cell ) ed ) );
					c_z = opposite3D ( OrthoD ( ( Edges_cell ) ed ) );
				}
				
				// x - direction
				Nconst = Give_Nconst ( c_x, Nxhex - 1, Nyhex - 1, Nzhex - 1 );
				
				Ni     = Give_Ni    ( c_x, Nxhex - 1, Nyhex - 1, Nzhex - 1 );
				
				// A. empty - direction
				Nconst = Give_Nconst ( c_y, Nxhex - 1, Nyhex - 1, Nzhex - 1 ) + Nconst;
				
				// B. empty - direction
				Nconst = Give_Nconst ( c_z, Nxhex - 1, Nyhex - 1, Nzhex - 1 ) + Nconst;
				
				// c) Determine which neighbor hex is
				num_neighbor = ug->Give_edge ( id_edge )->Give_number_neighbors();
				
				for ( num = 0;num < num_neighbor;++num )
					if ( ug->Give_edge ( id_edge )->Give_typ_M_el ( num ) == hexahedronEl )
						if ( ug->Give_edge ( id_edge )->Give_id_M_el ( num ) == id_hex ) {
							num_hex = ug->Give_edge ( id_edge )->Give_rel_neigh_locstiff ( num );
							calculate_map ( map_sons, c_x, c_y, c_z );
							
							if ( rank_send == rank_rec )    // for parallel
							
								for ( i = 0;i < Nxedge;++i ) {
									if ( developer_version ) {
										if ( i*16 >= N_total )
											std::cout << " error 4 in Local_stiffness_matrix::Update" << std::endl;
											
										if ( num_hex >= ug->Give_edge ( id_edge )->Give_number_neighbors_hex() )
											 std::cout << " error 5 in Local_stiffness_matrix::Update" << std::endl;
											
										if ( ( i*Ni + Nconst ) *64 + 64 > Nxhex*Nyhex*Nzhex*64 ) {
											cout << " error 6 in Local_stiffness_matrix::Update"
											<< " \n Ni: " << Ni
											<< " i: " << i
											<< " Nconst: " << Nconst
											<< " Nxhex: " << Nxhex
											<< " Nyhex: " << Nyhex
											<< " Nzhex: " << Nzhex
											<< " ed: " << ed << std::endl;
											
										}
									}
									
									for ( v = 0;v < 2;++v )
										for ( u = 0;u < 8;++u )
											loc_m_neighbor_edges[id_edge][num_hex][i*16+8*v+u]
											= loc_m_hexahedra[id_hex][ ( i*Ni+Nconst ) *64+
											                           8*map_sons[v] + map_sons[u]];
								}
							else {
								if ( my_rank == rank_send ) {
									TYPE2* data = Parallel_buffer<TYPE2>::Give_send ( Nxedge * 16 );   // for parallel
									
									for ( i = 0;i < Nxedge;++i ) {
										if ( developer_version ) {
											if ( i*16 >= N_total )
												 std::cout << " error 7 in Local_stiffness_matrix::Update"
												<<  std::endl;
												
											if ( num_hex >= ug->Give_edge ( id_edge )->Give_number_neighbors_hex() )
												cout << " error 8 in Local_stiffness_matrix::Update"
												<< std::endl;
												
											if ( ( i*Ni + Nconst ) *64 + 64 > Nxhex*Nyhex*Nzhex*64 ) {
												cout << " error 9 in Local_stiffness_matrix::Update"
												<< " \n Ni: " << Ni
												<< " i: " << i
												<< " Nconst: " << Nconst
												<< " Nxhex: " << Nxhex
												<< " Nyhex: " << Nyhex
												<< " Nzhex: " << Nzhex
												<< " ed: " << ed <<  std::endl;
											}
										}
										
										for ( v = 0;v < 2;++v )
											for ( u = 0;u < 8;++u )
												data[i*16+8*v+u]
												= loc_m_hexahedra[id_hex][ ( i*Ni+Nconst ) *64+
												                           8*map_sons[v] + map_sons[u]];
									}
									
									MPI_Send ( data, Nxedge*16,
									
									           Give_MPI_data_typ<TYPE2>::Give(), rank_rec, c + 10*num, comm );
								} else {
									if ( developer_version )
										if ( my_rank != rank_rec )
											 std::cout << " error 10 in Local_stiffness_matrix::Update" <<  std::endl;
											
									MPI_Status stat;
									
									MPI_Recv ( loc_m_neighbor_edges[id_edge][num_hex], Nxedge*16,
									           Give_MPI_data_typ<TYPE2>::Give(), rank_send, c + 10*num, comm, &stat );
								}
							}
						}
			}
		}
		
		//Update local stiffness matrix for points
		for ( c = 0;c < 8;++c ) {
			id_point = ug->Give_hexahedron ( id_hex )->Give_id_corner ( ( dir3D_sons ) c );
			
			rank_rec = ug->Give_point ( id_point )->Give_rank();  // for parallel
			
			if ( rank_send == my_rank || rank_rec == my_rank ) {  // for parallel
			
				number_neighbors_hex = ug->Give_point ( id_point )->Give_number_neighbors_hex();
				N_total = number_neighbors_hex * 8;
				
				Calculate_corresponding_corners ( ( dir3D_sons ) c, SED_corner, NWD_corner, SWT_corner );
				c_x = direction_from_to ( ( dir3D_sons ) c, SED_corner );
				c_y = direction_from_to ( ( dir3D_sons ) c, NWD_corner );
				c_z = direction_from_to ( ( dir3D_sons ) c, SWT_corner );
				
				// x- direction
				Nconst = Give_Nconst ( c_x, Nxhex - 1, Nyhex - 1, Nzhex - 1 );
				
				// y- direction
				Nconst = Give_Nconst ( c_y, Nxhex - 1, Nyhex - 1, Nzhex - 1 ) + Nconst;
				
				// z- direction
				Nconst = Give_Nconst ( c_z, Nxhex - 1, Nyhex - 1, Nzhex - 1 ) + Nconst;
				
				// c) Determine which neighbor hex is
				num_neighbor = ug->Give_point ( id_point )->Give_number_neighbors();
				
				for ( num = 0;num < num_neighbor;++num )
					if ( ug->Give_point ( id_point )->Give_typ_el_neighbor ( num ) == hexahedronEl )
						if ( ug->Give_point ( id_point )->Give_id_el_neighbor ( num ) == id_hex ) {
							num_hex = ug->Give_point ( id_point )->Give_rel_neigh_locstiff ( num );
							//calculate_map
							calculate_map ( map_sons, c_x, c_y, c_z );
							
							if ( developer_version ) {
								if ( num_hex*8 >= N_total )
									 std::cout << " error 11 in Local_stiffness_matrix::Update" <<  std::endl;
									
								if ( num_hex >= ug->Give_point ( id_point )->Give_number_neighbors_hex() )
									cout << " error 12 in Local_stiffness_matrix::Update" << std::endl;
									
								if ( ( Nconst + 1 ) *64 > Nxhex*Nyhex*Nzhex*64 )
									 std::cout << " error 13 in Local_stiffness_matrix::Update" << std::endl;
							}
							
							if ( rank_send == rank_rec )      // for parallel
								for ( u = 0;u < 8;++u )
									loc_m_neighbor_points[id_point][num_hex*8+u]
									= loc_m_hexahedra[id_hex][Nconst*64+8*map_sons[0] + map_sons[u]];
									
									
							else {
								if ( my_rank == rank_send ) {
									TYPE2* data = Parallel_buffer<TYPE2>::Give_send ( 8 );   // for parallel
									
									for ( u = 0;u < 8;++u )
										data[u] = loc_m_hexahedra[id_hex][Nconst*64+8*map_sons[0] + map_sons[u]];
										
									MPI_Send ( data, 8,
									           Give_MPI_data_typ<TYPE2>::Give(), rank_rec, c + 10*num, comm );
								} else {
									if ( developer_version )
										if ( my_rank != rank_rec )
											 std::cout << " error 14 in Local_stiffness_matrix::Update" <<  std::endl;
											
									MPI_Status stat;
									
									MPI_Recv ( &loc_m_neighbor_points[id_point][num_hex*8], 8,
									           Give_MPI_data_typ<TYPE2>::Give(), rank_send, c + 10*num, comm, &stat );
								}
							}
						}
			}
		}
	}
}

template <typename TYPE2>
template <class A>
void Local_stiffness_matrix<TYPE2>::operator= ( const Expr<A>& a ) {

	int id;
	int Nx, Ny, Nz, N_total;
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	const A& ao ( a );
	
	for ( id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			N_total = Nx * Ny * Nz * 64;

			//neu eingebaut!!!
			#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
			for(int k = 0;k < Nz;++k ) {
				for (int j = 0;j < Ny;++j )
					for(int i = 0;i < Nx;++i ) {
						if ( developer_version )
							if ( Ind_loc_matrix_hexahedra ( i, j, k ) *64 >= N_total )
								std::cout << " error1 in Local_stiffness_matrix::operator=" << std::endl;
								
						for(int uv = 0;uv < 64;++uv )
							loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+uv]
							= ao.Give_matrix_hexahedra ( id,
							                             Ind_loc_matrix_hexahedra ( i, j, k ) * 64 + uv,
							                             Ind_loc_matrix_hexahedra ( i, j, k ),
							                             i, j, k, Nx, Ny );
					}
			}
		}
	}
	
	Update();
}

template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::operator= ( const Local_stiffness_matrix & a ) {
	if ( this == &a )
		return;
		
	const Expr<Local_stiffness_matrix>& b ( a );
	
	operator= ( b );
}


template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::multiplyTestfunction_v_with ( const Variable<TYPE2>& coeffV ) {

	int Nx, Ny, Nz, N_total;
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	for ( int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			coeffV.template Update<hexahedronEl> ( id );
			
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			N_total = Nx * Ny * Nz * 64;
			
			//neu eingebaut!!!
			#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
			for ( int k = 0;k < Nz;++k ) {
				for ( int j = 0;j < Ny;++j )
					for ( int i = 0;i < Nx;++i ) {
						if ( developer_version )
							if ( Ind_loc_matrix_hexahedra ( i, j, k ) *64 >= N_total )
								 std::cout << " error1 in Local_stiffness_matrix::multiplyTestfunction_v_with" <<  std::endl;
								
						for ( int v = 0;v < 8;++v ) { // test function v
							for ( int u = 0;u < 8;++u ) {
								loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u] =
								    loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u] *
								    ( coeffV.template Give_data<hexahedronEl> ( id,
								            x1DCoord ( ( dir3D_sons ) v ) + i,
								            y1DCoord ( ( dir3D_sons ) v ) + j,
								            z1DCoord ( ( dir3D_sons ) v ) + k, Nx, Ny ) );
							}
						}
					}
			}
		}
	}
	
	Update();
}


#endif // DIFFOP_H
