/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef DIFFOP_CC_H
#define DIFFOP_CC_H

template <typename TYPE2>
Expr_Diag_Local_stiffness_matrix<TYPE2> Local_stiffness_matrix<TYPE2>::diag ()
  const {
  return Expr_Diag_Local_stiffness_matrix<TYPE2>(loc_m_hexahedra,loc_m_interior_quadrangles,
                                                                   loc_m_exterior_quadrangles,
                                                                   loc_m_neighbor_edges,
                                                                   loc_m_neighbor_points,
                                                                   blockgrid);
}

template <typename TYPE2>
Local_stiffness_matrix<TYPE2>::Local_stiffness_matrix(Blockgrid& bg_) {

  blockgrid = &bg_;
  Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
  int i, id, num;
  int Nx, Ny, Nz, N_total;

  // for parallel
  my_rank = ug->Give_my_rank();
  comm    = ug->Give_MPI_comm();

  loc_m_hexahedra = new TYPE2*[ug->Give_number_hexahedra()];
  for(id=0;id<ug->Give_number_hexahedra();++id) {
    if(ug->Give_hexahedron(id)->my_object(my_rank)) {
      Nx = blockgrid->Give_Nx_hexahedron(id);
      Ny = blockgrid->Give_Ny_hexahedron(id);
      Nz = blockgrid->Give_Nz_hexahedron(id);
      N_total = Nx * Ny * Nz * 64;
      loc_m_hexahedra[id] = new TYPE2[N_total];
      for(i=0;i<N_total;++i) loc_m_hexahedra[id][i] = (TYPE2)0;
    }
    else loc_m_hexahedra[id] = NULL;
  }

  loc_m_interior_quadrangles = new TYPE2*[ug->Give_number_quadrangles()];
  loc_m_exterior_quadrangles = new TYPE2*[ug->Give_number_quadrangles()];
  for(id=0;id<ug->Give_number_quadrangles();++id) {
    if(ug->Give_quadrangle(id)->my_object(my_rank)) {
      Nx = blockgrid->Give_Nx_quadrangle(id);
      Ny = blockgrid->Give_Ny_quadrangle(id);
      N_total = Nx * Ny * 32;
      loc_m_interior_quadrangles[id] = new TYPE2[N_total];
      for(i=0;i<N_total;++i)
           loc_m_interior_quadrangles[id][i] = (TYPE2)0;
      if(ug->Give_quadrangle(id)->Give_exists_exterior()){
         loc_m_exterior_quadrangles[id] = new TYPE2[N_total];
         for(i=0;i<N_total;++i)
           loc_m_exterior_quadrangles[id][i] = (TYPE2)0;
      }
      else loc_m_exterior_quadrangles[id] = NULL;
    }
    else{
      loc_m_interior_quadrangles[id] = NULL;
      loc_m_exterior_quadrangles[id] = NULL;
    }
  }

  loc_m_neighbor_edges = new TYPE2**[ug->Give_number_edges()];
  for(id=0;id<ug->Give_number_edges();++id) {
    if(ug->Give_edge(id)->my_object(my_rank)) {
      loc_m_neighbor_edges[id] = new TYPE2*[ug->Give_edge(id)->Give_number_neighbors_hex()];
      for(num=0;num<ug->Give_edge(id)->Give_number_neighbors_hex();++num){
          Nx = blockgrid->Give_Nx_edge(id);
          N_total = Nx  * 16;
          loc_m_neighbor_edges[id][num] = new TYPE2[N_total];
          for(i=0;i<N_total;++i)
               loc_m_neighbor_edges[id][num][i] = (TYPE2)0;
      }
    }
    else loc_m_neighbor_edges[id] = NULL;
  }

  loc_m_neighbor_points = new TYPE2*[ug->Give_number_points()];
  for(id=0;id<ug->Give_number_points();++id) {
    if(ug->Give_point(id)->my_object(my_rank)) {
      N_total = ug->Give_point(id)->Give_number_neighbors_hex() * 8;
      loc_m_neighbor_points[id] = new TYPE2[N_total];
      for(num=0;num<N_total;++num)
               loc_m_neighbor_points[id][num] = (TYPE2)0;
    }
    else loc_m_neighbor_points[id] = NULL;
  }
}

template <typename TYPE2>
Local_stiffness_matrix<TYPE2>::~Local_stiffness_matrix() {
  if(loc_m_hexahedra != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id)
       if(loc_m_hexahedra[id] != NULL)
          delete [] loc_m_hexahedra[id];
    delete [] loc_m_hexahedra;
  }
  loc_m_hexahedra = NULL;

  if(loc_m_interior_quadrangles != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id)
       if(loc_m_interior_quadrangles[id] != NULL)
          delete [] loc_m_interior_quadrangles[id];
    delete [] loc_m_interior_quadrangles;
  }
  loc_m_interior_quadrangles = NULL;

  if(loc_m_exterior_quadrangles != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id)
       if(loc_m_exterior_quadrangles[id] != NULL)
          delete [] loc_m_exterior_quadrangles[id];
    delete [] loc_m_exterior_quadrangles;
  }
  loc_m_exterior_quadrangles = NULL;

  if(loc_m_neighbor_edges != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_edges();++id)
      if(loc_m_neighbor_edges[id] != NULL){
        for(int num=0;num<blockgrid->Give_unstructured_grid()
                           ->Give_edge(id)->Give_number_neighbors_hex();++num)
          if(loc_m_neighbor_edges[id][num] != NULL)
             delete [] loc_m_neighbor_edges[id][num];
        delete [] loc_m_neighbor_edges[id];
      }
    delete [] loc_m_neighbor_edges;
  }
  loc_m_neighbor_edges = NULL;

  if(loc_m_neighbor_points != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_points();++id)
      if(loc_m_neighbor_points[id] != NULL)
         delete [] loc_m_neighbor_points[id];
    delete [] loc_m_neighbor_points;
  }
  loc_m_neighbor_points = NULL;
}

template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::Print() {
  int i,j,k,u,v,id,num;
  int Nx, Ny, Nz, N_total;

  Unstructured_grid* ug = blockgrid->Give_unstructured_grid();

  for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id) {
    if(ug->Give_hexahedron(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_hexahedron(id);
      Ny = blockgrid->Give_Ny_hexahedron(id);
      Nz = blockgrid->Give_Nz_hexahedron(id);
      N_total = Nx * Ny * Nz * 64;

      for(k=0;k<Nz;++k)
        for(j=0;j<Ny;++j)
           for(i=0;i<Nx;++i){

            if(developer_version) {
               if(Ind_loc_matrix_hexahedra(i,j,k)*64>=N_total)
                   cout << " error1 in Local_stiffness_matrix::Print()" << endl;
            }
            cout << "i: " << i << " j: " << j << "k: " << k << endl;
            for(u=0;u<8;++u) {
               for(v=0;v<8;++v)
                cout << loc_m_hexahedra[id][Ind_loc_matrix_hexahedra(i,j,k)*64+8*v+u] << ", ";
                cout << endl;
            }
          }
    }
  }
  for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id) {
    if(ug->Give_quadrangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_quadrangle(id);
      Ny = blockgrid->Give_Ny_quadrangle(id);
      N_total = Nx * Ny * 32;

      for(j=0;j<Ny;++j)
        for(i=0;i<Nx;++i){
            if(developer_version)
                if(Ind_loc_matrix_quadrangles(i,j)*32>=N_total)
                    cout << " error2 in Local_stiffness_matrix::Print()" << endl;
            cout << "i: " << i << " j: " << j << endl;
            for(u=0;u<8;++u) {
               for(v=0;v<4;++v)
                  cout << loc_m_interior_quadrangles[id]
                                                    [Ind_loc_matrix_quadrangles(i,j)*32+8*v+u]
                   << ", ";
               cout << endl;
            }
          }
    }
  }
  for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id) {
      if(ug->Give_quadrangle(id)->my_object(my_rank) &&
          blockgrid->Give_unstructured_grid()->Give_quadrangle(id)->Give_exists_exterior()){
        Nx = blockgrid->Give_Nx_quadrangle(id);
        Ny = blockgrid->Give_Ny_quadrangle(id);
        N_total = Nx * Ny * 32;

        for(j=0;j<Ny;++j)
          for(i=0;i<Nx;++i){
              if(developer_version)
                 if(Ind_loc_matrix_quadrangles(i,j)*32>=N_total)
                     cout << " error3 in Local_stiffness_matrix::Print()" << endl;
              cout << "i: " << i << " j: " << j << endl;
              for(u=0;u<8;++u) {
                  for(v=0;v<4;++v)
                     cout << loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i,j)*32+8*v+u]
                          << ", ";
                  cout << endl;
              }
            }
      }
  }
  for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_edges();++id) {
    if(ug->Give_edge(id)->my_object(my_rank)) {
      for(num=0;num<blockgrid->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();++num){
        Nx = blockgrid->Give_Nx_edge(id);
        N_total = Nx  * 16;
        for(i=0;i<Nx;++i){
            if(developer_version)
               if(i*16>=N_total)
                   cout << " error4 in Local_stiffness_matrix::Print()" << endl;
            cout << "i: " << i << endl;
            for(u=0;u<8;++u) {
                for(v=0;v<2;++v)
                   cout << loc_m_neighbor_edges[id][num][i*16+8*v+u] << ", ";
                cout << endl;
            }
        }
      }
    }
  }
  
  
  //ug = blockgrid->Give_unstructured_grid();
  for(id=0;id<ug->Give_number_points();++id) {
    if(ug->Give_point(id)->my_object(my_rank)) {
      N_total = ug->Give_point(id)->Give_number_neighbors_hex() * 8;
      for(num=0;num<N_total;++num)
          cout << loc_m_neighbor_points[id][num] << ", ";
    }
  }
    
  
  /*
  for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_points();++id) {    
    if(ug->Give_edge(id)->my_object(my_rank)) {
      N_total = blockgrid->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex() * 8;
      for(num=0;num<blockgrid->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex();
           ++num){
          if(developer_version)
              if(num*8>=N_total)
                 cout << " error5 in Local_stiffness_matrix::Print()" << endl;
          for(u=0;u<8;++u)
               cout << loc_m_neighbor_points[id_point][num_hex*8+u]      loc_m_neighbor_points[id][num][num*8+u] << ", ";
          cout << endl;
      }
    }
  }
  */
}

template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::addPhaseShift(const Cell_variable<TYPE2> & shiftVariable) {
  int i,j,k,u,v,id,num;
  int Nx, Ny, Nz, N_total, uz, vz, uxy, vxy;
  TYPE2 shift;

  Unstructured_grid* ug = blockgrid->Give_unstructured_grid();

  for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id) {
    if(ug->Give_hexahedron(id)->my_object(my_rank)) {
      Nx = blockgrid->Give_Nx_hexahedron(id);
      Ny = blockgrid->Give_Ny_hexahedron(id);
      Nz = blockgrid->Give_Nz_hexahedron(id);
      N_total = Nx * Ny * Nz * 64;

    // only for k = 0, i.e. at left end of hexaedron
      k = 0;
      for ( j=0;j<Ny;++j ) {
        for ( i=0;i<Nx;++i ) {
          for ( uxy=0;uxy<4;++uxy ) {
            for ( vxy=0;vxy<4;++vxy ) {
              shift = shiftVariable.Give_cell_hexahedra(id,
					 Ind_loc_matrix_hexahedra(i,j,k),
					 i,j,k,Nx,Ny);
             
              //1. case
              uz=Ldir1D;    vz=Ldir1D;
              u = uxy + 4 * uz ;
              v = vxy + 4 * vz;


	      // loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u] *= norm(shift); 
              loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u] *= 		
		(shift * myConj(shift)); 

              //2. case
              uz=Rdir1D;    vz=Ldir1D;
              u = uxy + 4 * uz;
              v = vxy + 4 * vz;

	      loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u] *= myConj(shift);

              //3. case
              uz=Ldir1D;     vz=Rdir1D;
              u = uxy + 4 * uz;
              v = vxy + 4 * vz;

              loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*v+u] *= shift;
            }
          }
        }
      }
    }
  }
  Update();
}


#endif // DIFFOP_CC_H

