/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


// ------------------------------------------------------------
//
// mgop.cc
//
// ------------------------------------------------------------

#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "../extemp/extemp.h"
#include "../extemp/co_fu.h"
#include "../extemp/functor.h"
#include "../extemp/parallel.h"
#include "../extemp/variable.h"
#include "../../../Source_Colsamm/Colsamm.h"
#include "diffop.h"
#include "ressten.h"
#include "mgop.h"


///////////////////////////////////
// Print test functions
///////////////////////////////////


void Print_part_loc_stencil(double* sten) {
  cout << "\n   North: " << endl;
  cout << "     " << sten[WNTdir3D] << ", " << sten[ENTdir3D] << endl;
  cout << "     " << sten[WNDdir3D] << ", " << sten[ENDdir3D] << endl;
  cout << "   South: " << endl;
  cout << "     " << sten[WSTdir3D] << ", " << sten[ESTdir3D] << endl;
  cout << "     " << sten[WSDdir3D] << ", " << sten[ESDdir3D] << endl;
}

void Print_Loc_stencil(double* sten) {
  int i;
  for(i=0;i<8;++i) {
    cout << "\n Sten bei: ";  Print((dir3D_sons)i);
    Print_part_loc_stencil(&(sten[8*i]));
  }
}

///////////////////////////////////
// Restriction Stencils
///////////////////////////////////


inline int number_coarse(dir3D_sons i) {
  return trans_cellP_27(i,i);
}

// 2D
void Set_coarse_function(int* U, dir2D_sons v) {

  if(v==SWdir2D) {
    U[6] = 0;  U[7] = 0;  U[8] = 0;
    U[3] = 2;  U[4] = 1;  U[5] = 0;
    U[0] = 4;  U[1] = 2;  U[2] = 0;
  }

  if(v==SEdir2D) {
    U[6] = 0;  U[7] = 0;  U[8] = 0;
    U[3] = 0;  U[4] = 1;  U[5] = 2;
    U[0] = 0;  U[1] = 2;  U[2] = 4;
  }

  if(v==NWdir2D) {
    U[6] = 4;  U[7] = 2;  U[8] = 0;
    U[3] = 2;  U[4] = 1;  U[5] = 0;
    U[0] = 0;  U[1] = 0;  U[2] = 0;
  }

  if(v==NEdir2D) {
    U[6] = 0;  U[7] = 2;  U[8] = 4;
    U[3] = 0;  U[4] = 1;  U[5] = 2;
    U[0] = 0;  U[1] = 0;  U[2] = 0;
  }
}


// 3D
void Set_coarse_function(int* U, dir3D_sons v) {
  int k, trans, d;
  dir3D_sons next;


  // part 0: set zero
  //--------
  for(k=0;k<27;++k) U[k]=0;

  // part 1: corners of coarse cell
  //--------
  for(k=0;k<8;++k) { // iteration over cells
    if(k==v) U[number_coarse((dir3D_sons)k)]=8;
    else     U[number_coarse((dir3D_sons)k)]=0;
  }

  // part 2: middle of coarse cell
  //--------
  U[trans27(MOrt,MOrt,MOrt)] = 1;

  // part 3: faces of coarse cell
  //--------
  // TD-surface
  U[trans27(MOrt,MOrt,LOrt)] = (U[number_coarse(ESDdir3D)] +
				U[number_coarse(WSDdir3D)] +
				U[number_coarse(ENDdir3D)] +
				U[number_coarse(WNDdir3D)]) / 4;

  U[trans27(MOrt,MOrt,ROrt)] = (U[number_coarse(ESTdir3D)] +
				U[number_coarse(WSTdir3D)] +
				U[number_coarse(ENTdir3D)] +
				U[number_coarse(WNTdir3D)]) / 4;
  // NS-surface
  U[trans27(MOrt,LOrt,MOrt)] = (U[number_coarse(WSTdir3D)] +
				U[number_coarse(ESTdir3D)] +
				U[number_coarse(WSDdir3D)] +
				U[number_coarse(ESDdir3D)]) / 4;

  U[trans27(MOrt,ROrt,MOrt)] = (U[number_coarse(WNTdir3D)] +
				U[number_coarse(ENTdir3D)] +
				U[number_coarse(WNDdir3D)] +
				U[number_coarse(ENDdir3D)]) / 4;

  // EW-surface
  U[trans27(LOrt,MOrt,MOrt)] = (U[number_coarse(WSTdir3D)] +
				U[number_coarse(WSDdir3D)] +
				U[number_coarse(WNTdir3D)] +
				U[number_coarse(WNDdir3D)]) / 4;
  U[trans27(ROrt,MOrt,MOrt)] = (U[number_coarse(ESTdir3D)] +
				U[number_coarse(ESDdir3D)] +
				U[number_coarse(ENTdir3D)] +
				U[number_coarse(ENDdir3D)]) / 4;

  // part 4: edges of coarse cell
  //--------
  for(k=0;k<8;++k) { // iteration over cells
    for(d=0;d<3;++d) {  // iteration over directions
      next = Next_corner((dir3D_sons)k,(dir3D)(2*d));
      trans = trans_cellP_27((dir3D_sons)k,next);
      U[trans] = U[trans] + U[number_coarse((dir3D_sons)k)] / 2;
    }
  }
}




///////////////////////////////////
// Constructor Multigrid_operator
///////////////////////////////////

Multigrid_operator::Multigrid_operator(Blockgrid& coarse_grid_,
				       Blockgrid& fine_grid_) {
  int number_degree_of_freedom;
  bool co_x, co_y, co_z;
  Unstructured_grid* ug;
  Quadrangle_el *quad;
  Edge_el       *edge;

  // a) check if ug o.k.
  coarse_grid = &coarse_grid_;
  fine_grid   = &fine_grid_;

  if(coarse_grid->Give_unstructured_grid()!=
     fine_grid->Give_unstructured_grid())
    cout << "Error 1 in Multigrid_operator" << endl;
  ug = coarse_grid->Give_unstructured_grid();

  // for parallel
  my_rank = ug->Give_my_rank();
  comm    = ug->Give_MPI_comm();

  // b) Calculate coarse directions
  number_degree_of_freedom = ug->degree_of_freedom();


  coarse = new bool[number_degree_of_freedom];
  for(int i=0;i<number_degree_of_freedom;++i) {
    if(coarse_grid->Give_N_color(i) == fine_grid->Give_N_color(i))
      coarse[i] = false;
    else {
      if(coarse_grid->Give_N_color(i) * 2 == fine_grid->Give_N_color(i))
	coarse[i] = true;
      else  
	cout << "Error 2 in Multigrid_operator: " 
	     << i << endl;
    }

  
  }

  // c) Calculate coarse directions of hexahedron
  coarse_hexahedra = new coarsening_direction[ug->Give_number_hexahedra()];
  for(int id=0;id<ug->Give_number_hexahedra();++id) {

    if(ug->Give_hexahedron(id)->my_object(my_rank)) {

      co_x = coarse[ug->Give_color_of_edge(
		    ug->Give_hexahedron(id)->Give_id_edge(SDed))];
      co_y = coarse[ug->Give_color_of_edge(
		    ug->Give_hexahedron(id)->Give_id_edge(WDed))];
      co_z = coarse[ug->Give_color_of_edge(
		    ug->Give_hexahedron(id)->Give_id_edge(SWed))];

      if(co_x && co_y && co_z)
        coarse_hexahedra[id] = coarse_xyz;
      else 
        if(co_x && co_y)
	  coarse_hexahedra[id] = coarse_xy;
        else 
	  if(co_x && co_z)
	    coarse_hexahedra[id] = coarse_xz;
	  else 
	    if(co_y && co_z)
	      coarse_hexahedra[id] = coarse_yz;
	    else 
	      if(co_x)
	        coarse_hexahedra[id] = coarse_x;
	      else
	        if(co_y)
		  coarse_hexahedra[id] = coarse_y;
	        else
		  if(co_z)
		    coarse_hexahedra[id] = coarse_z;
		  else 
		    coarse_hexahedra[id] = no_coarse;
    }
    else coarse_hexahedra[id] = (coarsening_direction)0;
  }

  // d) Calculate coarse directions of quadrangles
  coarse_quadrangles = new coarsening_direction[ug->Give_number_quadrangles()];
  coarse_quadrangles_exterior = new bool[ug->Give_number_quadrangles()];
  coarse_quadrangles_interior = new bool[ug->Give_number_quadrangles()];

  for(int id=0;id<ug->Give_number_quadrangles();++id) {

    if(ug->Give_quadrangle(id)->my_object(my_rank)) {

      quad = ug->Give_quadrangle(id);

      co_x = coarse[ug->Give_color_of_edge(quad->Give_id_edge(Sdir2D))];
      co_y = coarse[ug->Give_color_of_edge(quad->Give_id_edge(Wdir2D))];

      if(co_x && co_y)
        coarse_quadrangles[id] = coarse_xy;
      else 
        if(co_x)
	  coarse_quadrangles[id] = coarse_x;
        else
	  if(co_y)
	    coarse_quadrangles[id] = coarse_y;
	  else 
	    coarse_quadrangles[id] = no_coarse;


      coarse_quadrangles_interior[id] =
        coarse[ug->Give_color_of_edge(quad->Give_id_interior_edge(SWdir2D))];

      if(quad->Give_exists_exterior()) {
        coarse_quadrangles_exterior[id] =
	  coarse[ug->Give_color_of_edge(quad->Give_id_exterior_edge(SWdir2D))];
      }
      else coarse_quadrangles_exterior[id] = false;
    }
    else{
      coarse_quadrangles[id] = (coarsening_direction)0;
      coarse_quadrangles_exterior[id] = (bool)0;
      coarse_quadrangles_interior[id] = (bool)0;
    }
  }

  // e) Calculate coarse for edges
  coarse_edges        = new bool[ug->Give_number_edges()];
  coarse_weight_edges = new double*[ug->Give_number_edges()];

  for(int id=0;id<ug->Give_number_edges();++id) {

    if(ug->Give_edge(id)->my_object(my_rank)) {

      edge = ug->Give_edge(id);

      coarse_edges[id] = coarse[ug->Give_color_of_edge(id)];

      int num_pos = edge->Give_number_neighbors();
      coarse_weight_edges[id] = new double[num_pos];

      for(int pos=0;pos<num_pos;++pos) {
        elementTyp typ = edge->Give_typ_W_el(pos);
        if(typ==quadrangleEl) {
	  quad = ug->Give_quadrangle(edge->Give_id_W_el(pos));
	
	  co_x = coarse[ug->Give_color_of_edge(quad->Give_id_edge(Sdir2D))];
	  co_y = coarse[ug->Give_color_of_edge(quad->Give_id_edge(Wdir2D))];
	
	  if(co_x && co_y)
	    coarse_weight_edges[id][pos] = 0.25;
	  else 
	    if(co_x || co_y)
	      coarse_weight_edges[id][pos] = 0.5;
	    else
	      coarse_weight_edges[id][pos] = 0.0;
        }
        else if(typ==edgeEl) {
	  if(coarse[ug->Give_color_of_edge(edge->Give_id_W_el(pos))])
	    coarse_weight_edges[id][pos] = 0.5;
	  else
	    coarse_weight_edges[id][pos] = 0.0;
        }
        else cout << "error in Multigrid_operator::Multigrid_operator!" << endl;
      }
    }
    else{
      coarse_edges[id] = (bool)0;
      coarse_weight_edges[id] = NULL;
    }
  }

  // f) Calculate coarse grid test function
  coarse_grid_functions3D = new int[216];
  coarse_grid_functions2D = new int[36];
  for(int i=0;i<8;++i)
    Set_coarse_function(&(coarse_grid_functions3D[i*27]),(dir3D_sons)i);

  for(int i=0;i<4;++i)
    Set_coarse_function(&(coarse_grid_functions2D[i*9]),(dir2D_sons)i);
}


///////////////////////////////////
// Destructor Multigrid_operator
///////////////////////////////////


Multigrid_operator::~Multigrid_operator() {
  Unstructured_grid* ug = coarse_grid->Give_unstructured_grid();

  if(coarse != NULL)
      delete [] coarse;
  coarse = NULL;

  if(coarse_hexahedra != NULL)
      delete [] coarse_hexahedra;
  coarse_hexahedra = NULL;

  if(coarse_quadrangles != NULL)
      delete [] coarse_quadrangles;
  coarse_quadrangles = NULL;

  if(coarse_quadrangles_exterior != NULL)
      delete [] coarse_quadrangles_exterior;
  coarse_quadrangles_exterior = NULL;

  if(coarse_quadrangles_interior != NULL)
      delete [] coarse_quadrangles_interior;
  coarse_quadrangles_interior = NULL;

  if(coarse_edges != NULL)
      delete [] coarse_edges;
  coarse_edges = NULL;

  if(coarse_weight_edges != NULL){
      for(int i=0; i<ug->Give_number_edges(); ++i)
          if(coarse_weight_edges[i] != NULL)
              delete [] coarse_weight_edges[i];
      delete [] coarse_weight_edges;
  }
  coarse_weight_edges = NULL;

  if(coarse_grid_functions3D != NULL)
      delete [] coarse_grid_functions3D;
  coarse_grid_functions3D = NULL;

  if(coarse_grid_functions2D != NULL)
      delete [] coarse_grid_functions2D;
  coarse_grid_functions2D = NULL;
}

