/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef STENOP_CC_H
#define STENOP_CC_H

template <typename TYPE2, class STENCIL_TYP>
Expr_Diag_Sten_matrix<TYPE2, STENCIL_TYP> Sten_matrix<TYPE2, STENCIL_TYP>::diag ()
const
{
	return Expr_Diag_Sten_matrix<TYPE2, STENCIL_TYP>(sten_m_hexahedra,loc_m_interior_quadrangles,
	       loc_m_exterior_quadrangles,
	       loc_m_neighbor_edges,
	       loc_m_neighbor_points,
	       blockgrid,
	       sten_typ);
}



template <typename TYPE2, class STENCIL_TYP>
Sten_matrix<TYPE2, STENCIL_TYP>::Sten_matrix(Blockgrid& bg_)
{
	blockgrid = &bg_;

	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();

	// for parallel
	my_rank = ug->Give_my_rank();
	comm    = ug->Give_MPI_comm();

//   int num_hexahedra = blockgrid->Give_unstructured_grid()->Give_number_hexahedra();

	Allocate_data();
}

template <typename TYPE2, class STENCIL_TYP>
Sten_matrix<TYPE2, STENCIL_TYP>::Sten_matrix(Blockgrid& bg_,
    STENCIL_TYP& not_constant_directions) :
		blockgrid(&bg_), sten_typ(not_constant_directions)
{

	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();

	// for parallel
	my_rank = ug->Give_my_rank();
	comm    = ug->Give_MPI_comm();

	if (blockgrid->Give_unstructured_grid()->Give_number_hexahedra() !=
	    not_constant_directions.Give_number_hexahedra())
		cout << " error in Sten_matrix<TYPE2>::Sten_matrix(Blockgrid&, Not_constant_direction_marker&) " << endl;

	Allocate_data();
}


template <typename TYPE2, class STENCIL_TYP>
void Sten_matrix<TYPE2, STENCIL_TYP>::Allocate_data()
{
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	int i, id, num;
	int Nx, Ny, Nz, N_total;

	sten_m_hexahedra = new TYPE2*[ug->Give_number_hexahedra()];
	for (id=0;id<ug->Give_number_hexahedra();++id) {

		if (ug->Give_hexahedron(id)->my_object(my_rank)) {

			Nx = blockgrid->Give_Nx_hexahedron(id);
			Ny = blockgrid->Give_Ny_hexahedron(id);
			Nz = blockgrid->Give_Nz_hexahedron(id);

			if (sten_typ.Not_const_x(id)==false) Nx = 2;
			if (sten_typ.Not_const_y(id)==false) Ny = 2;
			if (sten_typ.Not_const_z(id)==false) Nz = 2;

            //cout << "Hex: "<<id<< "Nx: " << Nx << " Ny: " << Ny << " Nz: " << Nz << endl;

            N_total = (Nx-1) * (Ny-1) * (Nz-1) * 27;
            sten_m_hexahedra[id] = new TYPE2[N_total];
            for (i=0;i<N_total;++i)
            {sten_m_hexahedra[id][i] = (TYPE2)0;}
		} else sten_m_hexahedra[id] = NULL;
	}

	loc_m_interior_quadrangles = new TYPE2*[ug->Give_number_quadrangles()];
	loc_m_exterior_quadrangles = new TYPE2*[ug->Give_number_quadrangles()];
	for (id=0;id<ug->Give_number_quadrangles();++id) {

		if (ug->Give_quadrangle(id)->my_object(my_rank)) {

			Nx = blockgrid->Give_Nx_quadrangle(id);
			Ny = blockgrid->Give_Ny_quadrangle(id);
			N_total = Nx * Ny * 32;
			loc_m_interior_quadrangles[id] = new TYPE2[N_total];
			for (i=0;i<N_total;++i)
				loc_m_interior_quadrangles[id][i] = (TYPE2)0;
			if (ug->Give_quadrangle(id)->Give_exists_exterior()) {
				loc_m_exterior_quadrangles[id] = new TYPE2[N_total];
				for (i=0;i<N_total;++i)
					loc_m_exterior_quadrangles[id][i] = (TYPE2)0;
			} else loc_m_exterior_quadrangles[id] = NULL;
		} else {
			loc_m_interior_quadrangles[id] = NULL;
			loc_m_exterior_quadrangles[id] = NULL;
		}
	}

	loc_m_neighbor_edges = new TYPE2**[ug->Give_number_edges()];
	for (id=0;id<ug->Give_number_edges();++id) {

		if (ug->Give_edge(id)->my_object(my_rank)) {

			loc_m_neighbor_edges[id] = new TYPE2*[ug->Give_edge(id)->Give_number_neighbors_hex()];
			for (num=0;num<ug->Give_edge(id)->Give_number_neighbors_hex();++num) {
				Nx = blockgrid->Give_Nx_edge(id);
				N_total = Nx  * 16;
				loc_m_neighbor_edges[id][num] = new TYPE2[N_total];
				for (i=0;i<N_total;++i)
					loc_m_neighbor_edges[id][num][i] = (TYPE2)0;
			}
		} else loc_m_neighbor_edges[id] = NULL;
	}

	loc_m_neighbor_points = new TYPE2*[ug->Give_number_points()];
	for (id=0;id<ug->Give_number_points();++id) {

		if (ug->Give_point(id)->my_object(my_rank)) {

			N_total = ug->Give_point(id)->Give_number_neighbors_hex() * 8;
			loc_m_neighbor_points[id] = new TYPE2[N_total];
			for (num=0;num<N_total;++num)
				loc_m_neighbor_points[id][num] = (TYPE2)0;
		} else loc_m_neighbor_points[id] = NULL;
	}
}

template <typename TYPE2, class STENCIL_TYP>
Sten_matrix<TYPE2, STENCIL_TYP>::~Sten_matrix()
{
	if (sten_m_hexahedra != NULL) {
		for (int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id)
			if (sten_m_hexahedra[id] != NULL)
				delete [] sten_m_hexahedra[id];
		delete [] sten_m_hexahedra;
	}
	sten_m_hexahedra = NULL;

	if (loc_m_interior_quadrangles != NULL) {
		for (int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id)
			if (loc_m_interior_quadrangles[id] != NULL)
				delete [] loc_m_interior_quadrangles[id];
		delete [] loc_m_interior_quadrangles;
	}
	loc_m_interior_quadrangles = NULL;

	if (loc_m_exterior_quadrangles != NULL) {
		for (int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id)
			if (loc_m_exterior_quadrangles[id] != NULL)
				delete [] loc_m_exterior_quadrangles[id];
		delete [] loc_m_exterior_quadrangles;
	}
	loc_m_exterior_quadrangles = NULL;

	if (loc_m_neighbor_edges != NULL) {
		for (int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_edges();++id)
			if (loc_m_neighbor_edges[id] != NULL) {
                for (int num=0;num<blockgrid->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();++num)
					if (loc_m_neighbor_edges[id][num] != NULL)
						delete [] loc_m_neighbor_edges[id][num];
				delete [] loc_m_neighbor_edges[id];
			}
		delete [] loc_m_neighbor_edges;
	}
	loc_m_neighbor_edges = NULL;

	if (loc_m_neighbor_points != NULL) {
		for (int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_points();++id)
			if (loc_m_neighbor_points[id] != NULL)
				delete [] loc_m_neighbor_points[id];
		delete [] loc_m_neighbor_points;
	}
	loc_m_neighbor_points = NULL;
}



template <typename TYPE2, class STENCIL_TYP>
void Sten_matrix<TYPE2, STENCIL_TYP>::Print()
{
	int i,j,k,u,v,id,num;
	int Nx, Ny, Nz, N_total;

	Unstructured_grid *ug = blockgrid->Give_unstructured_grid();



	for (id=0;id<blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id) {
		cout << "Local stencil of hexahedron block: " << id << endl;
		if (ug->Give_hexahedron(id)->my_object(my_rank)) {

			Nx = blockgrid->Give_Nx_hexahedron(id);
			Ny = blockgrid->Give_Ny_hexahedron(id);
			Nz = blockgrid->Give_Nz_hexahedron(id);
			N_total = Nx * Ny * Nz * 64;

			for (k=1;k<Nz;++k)
				for (j=1;j<Ny;++j)
					for (i=1;i<Nx;++i) {
						int ig = sten_typ.stenIndex(i,j,k,Nx,Ny,id);
						cout << "ig: "  << ig << endl;

						cout << "i: " << i << " j: " << j << "k: " << k << endl;
						for (int iz=0;iz<3;++iz) {
							for (int iy=0;iy<3;++iy) {
								for (int ix=0;ix<3;++ix) {
									cout << " " << sten_m_hexahedra[id][ig+trans27((Ort1D)ix,(Ort1D)iy,(Ort1D)iz)];
								}
								cout << endl;
							}
							cout << endl;
							cout << endl;
						}
					}
		}
	}

	/*
	for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id) {

	  if(ug->Give_quadrangle(id)->my_object(my_rank)) {

	    Nx = blockgrid->Give_Nx_quadrangle(id);
	    Ny = blockgrid->Give_Ny_quadrangle(id);
	    N_total = Nx * Ny * 32;

	    for(j=0;j<Ny;++j)
	      for(i=0;i<Nx;++i){
	          if(developer_version)
	              if(Ind_loc_matrix_quadrangles(i,j)*32>=N_total)
	                  cout << " error2 in Sten_matrix::Print()" << endl;
	          cout << "i: " << i << " j: " << j << endl;
	          for(u=0;u<8;++u) {
	             for(v=0;v<4;++v)
	                cout << loc_m_interior_quadrangles[id]
	                                                  [Ind_loc_matrix_quadrangles(i,j)*32+8*v+u]
	                 << ", ";
	             cout << endl;
	          }
	        }
	  }
	}
	for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id) {
	    if(blockgrid->Give_unstructured_grid()->Give_quadrangle(id)->Give_exists_exterior()){

	  if(ug->Give_quadrangle(id)->my_object(my_rank)) {

	      Nx = blockgrid->Give_Nx_quadrangle(id);
	      Ny = blockgrid->Give_Ny_quadrangle(id);
	      N_total = Nx * Ny * 32;

	      for(j=0;j<Ny;++j)
	        for(i=0;i<Nx;++i){
	            if(developer_version)
	               if(Ind_loc_matrix_quadrangles(i,j)*32>=N_total)
	                   cout << " error3 in Sten_matrix::Print()" << endl;
	            cout << "i: " << i << " j: " << j << endl;
	            for(u=0;u<8;++u) {
	                for(v=0;v<4;++v)
	                   cout << loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i,j)*32+8*v+u]
	                        << ", ";
	                cout << endl;
	            }
	          }
	    }
	  }
	}
	for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_edges();++id) {

	  if(ug->Give_edge(id)->my_object(my_rank)) {

	    for(num=0;num<blockgrid->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();++num){
	      Nx = blockgrid->Give_Nx_edge(id);
	      N_total = Nx  * 16;
	      for(i=0;i<Nx;++i){
	          if(developer_version)
	             if(i*16>=N_total)
	                 cout << " error4 in Sten_matrix::Print()" << endl;
	          cout << "i: " << i << endl;
	          for(u=0;u<8;++u) {
	              for(v=0;v<2;++v)
	                 cout << loc_m_neighbor_edges[id][num][i*16+8*v+u] << ", ";
	              cout << endl;
	          }
	      }
	    }
	  }
	}
	for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_points();++id) {

	  if(ug->Give_point(id)->my_object(my_rank)) {

	    N_total = blockgrid->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex() * 8;
	    for(num=0;num<blockgrid->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex();
	         ++num){
	        if(developer_version)
	            if(num*8>=N_total)
	               cout << " error5 in Sten_matrix::Print()" << endl;
	        for(u=0;u<8;++u)
	             cout << loc_m_neighbor_points[id][num][num*8+u] << ", ";
	        cout << endl;
	    }
	  }
	}
	*/
}


template <typename TYPE2, class STENCIL_TYP>
void Sten_matrix<TYPE2, STENCIL_TYP>::PrintStencilsToFile() const
{
	int i,j,k,id;
	int Nx, Ny, Nz;

	Unstructured_grid *ug = blockgrid->Give_unstructured_grid();

	for (id=0;id<blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id) {

		cout << "Printing stencils of hexahedron block : " << id << " to file" <<  endl;
		if (ug->Give_hexahedron(id)->my_object(my_rank)) {

			Nx = blockgrid->Give_Nx_hexahedron(id);
			Ny = blockgrid->Give_Ny_hexahedron(id);
			Nz = blockgrid->Give_Nz_hexahedron(id);

			for (k=1;k<Nz;++k) {
				/*
				for (j=1;j<Ny;++j)
					for (i=1;i<Nx;++i) */
				std::stringstream ss;
				ss << "StencilOutputBlock"<<id<<"ZIndex"<<k << ".txt" ;
				std::string filename = ss.str();
				std::ofstream out(filename.c_str());

				j=1;
				i=1;
				int ig = sten_typ.stenIndex(i,j,k,Nx,Ny,id);
				out << "#Stencil index: "  << ig << endl;
				out << "#Block index: " << id << endl;
				out << "#ZIndex: " << k << endl;
				for (int iz=0;iz<3;++iz) {
					if (iz==0) out << "left zIndex: ";
					else if (iz==1) out << "center zIndex: ";
					else if (iz==2) out << "right zIndex: ";

					out << iz << endl;

					for (int iy=0;iy<3;++iy) {
						for (int ix=0;ix<3;++ix) {
							out << sten_m_hexahedra[id][ig+trans27((Ort1D)ix,(Ort1D)iy,(Ort1D)iz)] << "\t";
						}
						out << endl;
					}
					out << endl;
					out << endl;
				}
			}
		}

		/*
		for(id=0;id<blockgrid->Give_unstructured_grid()->Give_number_quadrangles();++id) {

		  if(ug->Give_quadrangle(id)->my_object(my_rank)) {

		    Nx = blockgrid->Give_Nx_quadrangle(id);
		    Ny = blockgrid->Give_Ny_quadrangle(id);
		    N_total = Nx * Ny * 32;

		    for(j=0;j<Ny;++j)
		      for(i=0;i<Nx;++i){
		          if(developer_version)
		              if(Ind_loc_matrix_quadrangles(i,j)*32>=N_total)
		                  cout << " error2 in Sten_matrix::Print()" << endl;
		          cout << "i: " << i << " j: " << j << endl;
		          for(u=0;u<8;++u) {
		             for(v=0;v<4;++v)
		                cout << loc_m_interior_quadrangles[id]
		                                                  [Ind_loc_matrix_quadrangles(i,j)*32+8*v+u]
		                 << ", ";
		             cout << endl;
		          }
		        }
		  }
		}
		*/

	}

}

#endif // STENOP_CC_H
