/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"
#include "examples_ug.h"
#include "vtkReader.h"


VTK_Reader::VTK_Reader(QString PfadUndDateiName, double constructionShiftOfZ) {
  //Phillips job!!
  ////////////////////////////////////
  
  
  

    QFileInfo check_file(PfadUndDateiName);
    //check if file exists and if yes: Is it really a file and no directory?
    if (!check_file.exists() && !check_file.isFile()) {
        qDebug()<< "File not found! Check filename and path!: "<< PfadUndDateiName <<endl;
    } 

  TypeOfUG typ;
  QFile inputFile(PfadUndDateiName);
  int stringElement;
  int degrees;
  int numberConstructParameters;
   
  std::vector<double> constructionParameters(0);    
  std::vector<double> relativeCoordZ(0), construction_lx(0), construction_ly(0), construction_lz(0);
  bool inner_grid_arched, periodic, relativeCoordZFlag = false;
  std::vector<int> number_points(0);
  
  
  if (inputFile.open(QIODevice::ReadOnly))
  {
    QTextStream in(&inputFile);

    QString line = in.readLine();
    QStringList splitString = line.split(" ");
    stringElement = splitString.indexOf("typeUg");
    
    typ = findUGType(splitString, stringElement);
    

      
    typGrid = typ; 
   
      
   number_points = readNumberPoints(splitString);   
   
   constructionParameters = readConstructionParameters(splitString);
   

   stringElement = splitString.indexOf("relativeCoordZ");   
   numberConstructParameters = splitString.at(stringElement+1).toInt();
   for (int i = 1;i<=numberConstructParameters;++i)
   {
     relativeCoordZ.push_back(splitString.at(stringElement+1+i).toDouble()+constructionShiftOfZ);
     relativeCoordZFlag = true;
   }
 
   stringElement = splitString.indexOf("l_xyz");   
   numberConstructParameters = splitString.at(stringElement+1).toInt();
   for (int i = 1;i<=numberConstructParameters;++i)
       construction_lx.push_back(splitString.at(stringElement+1+i).toDouble());
   for (int i = 1;i<=numberConstructParameters;++i)   
       construction_ly.push_back(splitString.at(stringElement+1+i+numberConstructParameters).toDouble());
   for (int i = 1;i<=numberConstructParameters;++i)
       construction_lz.push_back(splitString.at(stringElement+1+i+numberConstructParameters*2).toDouble());
   
   
    //indexOf returns position of "inner_grid_arched", if not in string returns -1
   stringElement = splitString.indexOf("inner_grid_arched");  
   inner_grid_arched = (splitString.at(stringElement+1).toInt() == 1? true : false);
   
   stringElement = splitString.indexOf("periodic");  
   periodic = (splitString.at(stringElement+1).toInt() == 1? true : false);
   
   if(typ == noWithZTyp) {
     ug = new SpecialPrismList(constructionParameters[0],
                         constructionParameters[1],
                         constructionParameters[2],
                         constructionParameters[3],
                         constructionParameters[4],
			 relativeCoordZ,
                         constructionParameters[5],
			 constructionParameters[6],
			 constructionParameters[7]);

     blockgrid = new Blockgrid(ug, number_points);
     

   }
   
   
   
   if(typ == hexahedronUgTyp) {
     
     if (!relativeCoordZFlag)
     {     
     ug = new Hexahedron(constructionParameters[0],
                         constructionParameters[1],
                         constructionParameters[2],
                         constructionParameters[3],
                         constructionParameters[4],
                         constructionParameters[5]);
       
     }
     else
     {
     
     ug = new HexahedronList(constructionParameters[0],
                         constructionParameters[1],
			 relativeCoordZ,
                         constructionParameters[2],
                         constructionParameters[3],
                         constructionParameters[4],
			 periodic);
     }
     
     


     blockgrid = new Blockgrid(ug, number_points);

     
     
   }
   
   if(typ == cylinderUgTyp) {
     ug = new Cylinder(constructionParameters[0],
                         constructionParameters[1],
                         constructionParameters[2],
			 inner_grid_arched);

     blockgrid = new Blockgrid(ug, number_points);

   } 
   
   if(typ == cylinderListUgTyp) {
     ug = new CylinderList(constructionParameters[0],
                         constructionParameters[1],
			 relativeCoordZ,
                         constructionParameters[2],
                         constructionParameters[3],
                         constructionParameters[4],
                         inner_grid_arched);

     blockgrid = new Blockgrid(ug, number_points);

   }  
  
   if(typ == hexahedronAdaptiveUgTyp) {
     ug = new AdaptiveHexahedronList(constructionParameters[0],
                         constructionParameters[1],
                         constructionParameters[2],
                         constructionParameters[3],
			 relativeCoordZ,
                         constructionParameters[4],
			 constructionParameters[5],
			 constructionParameters[6]);

     blockgrid = new Blockgrid(ug, number_points);
    

   }  
  
   if(typ == hexahedronListUgTyp) {
     ug = new HexahedronXYList(construction_lx,
                         construction_ly,
                         construction_lz);
     


     blockgrid = new Blockgrid(ug, number_points);

   }
   
   if(typ == hexahedronListTestUgTyp) {
     ug = new HexahedronXYListTest(constructionParameters[0],
                         constructionParameters[1],
                         constructionParameters[2],
                         constructionParameters[3],
                         constructionParameters[4],
                         constructionParameters[5]);

     blockgrid = new Blockgrid(ug, number_points);

   }   
      
   
   while ("LOOKUP_TABLE" != line.left(12))
   {
     line = in.readLine();
   }




  
  ///////////////////////////////////
  

  //qDebug()<<ug->Give_number_points();
  
  // Nach  String  LOOKUP_TABLE suchen
  
  double** dataHex = new double*[ug->Give_number_hexahedra()];
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) 
  {
      int Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      int Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      int Nz = blockgrid->Give_Nz_hexahedron(id_hex);

      int size = (Nz+1) * (Ny+1) * (Nx+1);
      
      dataHex[id_hex] = new double[size];

      for(int k=0;k<=Nz;++k) {
          for(int j=0;j<=Ny;++j){
	    for(int i=0;i<=Nx;++i) {
		// dataHex[id_hex][i + (Nx + 1)*(j + (Ny + 1)*k)] = Datei....???
	        line = in.readLine();
	        dataHex[id_hex][i + (Nx + 1)*(j + (Ny + 1)*k)] = line.toDouble();
				   }
				}
			     }
  }

   
   inputFile.close();

 
  		
    variable = new Variable<double>(*blockgrid, dataHex);
    delete[]  dataHex;
  }
  else variable = NULL;
  
  //delete[] number_points;
}

  
  
Variable<double>* VTK_Reader::readNext(QString PfadUndDateiName) {
  
    //Phillips job!!
  ////////////////////////////////////
  
  TypeOfUG typ;
  QFile inputFile(PfadUndDateiName);
  int stringElement;
  int degrees;
  int numberConstructParameters;
   
  std::vector<double> constructionParameters(0);    
  std::vector<double> relativeCoordZ(0), construction_lx(0), construction_ly(0), construction_lz(0);
  bool inner_grid_arched, periodic, relativeCoordZFlag = false;
  std::vector<int> number_points(0);
  
  
  if (inputFile.open(QIODevice::ReadOnly))
  {
    QTextStream in(&inputFile);

    QString line = in.readLine();
    QStringList splitString = line.split(" ");
    stringElement = splitString.indexOf("typeUg");
    
    typ = findUGType(splitString, stringElement);

      
   if(typGrid != typ) {
     inputFile.close();
     return NULL;
   }
   
   stringElement = splitString.indexOf("degrees");  
   degrees = splitString.at(stringElement+1).toInt();
   
   if(degrees!= degreesRead) {
     inputFile.close();
     return NULL;
   }   
   
   for (int i = 1;i<=degrees;++i)
      {
      number_points.push_back(splitString.at(stringElement+1+i).toInt());
      }
   
   

   
   stringElement = splitString.indexOf("constructParameters");   
   numberConstructParameters = splitString.at(stringElement+1).toInt();
   
   if(numberConstructParameters != numberConstructionParametersRead) {
     inputFile.close();
     return NULL;
   }
   
   
   for (int i = 1;i<=numberConstructParameters;++i)
   {
     constructionParameters.push_back(splitString.at(stringElement+1+i).toDouble());
   }


   stringElement = splitString.indexOf("relativeCoordZ");   
   numberConstructParameters = splitString.at(stringElement+1).toInt();
   for (int i = 1;i<=numberConstructParameters;++i)
   {
     relativeCoordZ.push_back(splitString.at(stringElement+1+i).toDouble());
     relativeCoordZFlag = true;
   }
   
   stringElement = splitString.indexOf("l_xyz");   
   numberConstructParameters = splitString.at(stringElement+1).toInt();
   for (int i = 1;i<=numberConstructParameters;++i)
       construction_lx.push_back(splitString.at(stringElement+1+i).toDouble());
   for (int i = 1;i<=numberConstructParameters;++i)   
       construction_ly.push_back(splitString.at(stringElement+1+i+numberConstructParameters).toDouble());
   for (int i = 1;i<=numberConstructParameters;++i)
       construction_lz.push_back(splitString.at(stringElement+1+i+numberConstructParameters*2).toDouble());
   
   
    //indexOf returns position of "inner_grid_arched", if not in string returns -1
   stringElement = splitString.indexOf("inner_grid_arched");  
   inner_grid_arched = (splitString.at(stringElement+1).toInt() == 1? true : false);
   
   stringElement = splitString.indexOf("periodic");  
   periodic = (splitString.at(stringElement+1).toInt() == 1? true : false);
   
   
   while ("LOOKUP_TABLE" != line.left(12))
   {
     line = in.readLine();
   }




  
  ///////////////////////////////////
  

  //qDebug()<<ug->Give_number_points();
  
  // Nach  String  LOOKUP_TABLE suchen
  
  double** dataHex = new double*[ug->Give_number_hexahedra()];
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) 
  {
      int Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      int Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      int Nz = blockgrid->Give_Nz_hexahedron(id_hex);

      int size = (Nz+1) * (Ny+1) * (Nx+1);
      
      dataHex[id_hex] = new double[size];

      for(int k=0;k<=Nz;++k) {
          for(int j=0;j<=Ny;++j){
	    for(int i=0;i<=Nx;++i) {
		// dataHex[id_hex][i + (Nx + 1)*(j + (Ny + 1)*k)] = Datei....???
	        line = in.readLine();
	        dataHex[id_hex][i + (Nx + 1)*(j + (Ny + 1)*k)] = line.toDouble();
				   }
				}
			     }
  }

   inputFile.close();
 
  		
    Variable<double>* variableNext = new Variable<double>(*blockgrid, dataHex);
    delete[]  dataHex;
    return variableNext;
  }
  else 
    return NULL;  
}


TypeOfUG VTK_Reader::findUGType(QStringList splitString, int stringElement)
{
  TypeOfUG typ;
     switch(splitString.at(stringElement+1).toInt())
      {
	case 0: {typ = noWithZTyp;
	  break;
	}
	case 1: {typ = hexahedronUgTyp;
	  break;
	}
	case 2: {typ = cylinderUgTyp;
	  break;
	}
	case 3: {typ = cylinderListUgTyp;
	  break;
	}
	case 4: {typ = hexahedronAdaptiveUgTyp;
	  break;
	}
	case 5: {typ = hexahedronListUgTyp;
	  break;
	}
	case 6: {typ = hexahedronListTestUgTyp;
	  break;
	}    
      };   
      return typ;
}

std::vector<int> VTK_Reader::readNumberPoints(QStringList splitString)
{
   int stringElement = splitString.indexOf("degrees");  
   int degrees = splitString.at(stringElement+1).toInt();
   degreesRead = degrees;
   std::vector<int> number_points(0);
   for (int i = 1;i<=degrees;++i)
      {
      number_points.push_back(splitString.at(stringElement+1+i).toInt());
      }
      return number_points;
}

std::vector<double> VTK_Reader::readConstructionParameters(QStringList splitString)
{
   int stringElement = splitString.indexOf("constructParameters");  
   int numberConstruction = splitString.at(stringElement+1).toInt();
   numberConstructionParametersRead = numberConstruction;
   std::vector<double> construction_parameters(0);
   for (int i = 1;i<=numberConstruction;++i)
      {
      construction_parameters.push_back(splitString.at(stringElement+1+i).toDouble());
      }
      return construction_parameters;
}
