//    GUI for solid state laser simulation:
//    Copyright (C) 2016  Christoph Pflaum
//    This program is not a free software;
//    do not redistribute it and/or modify it
//    To use this software you need
//    a permission by Christoph Pflaum
/////////////////////////////////////////////////


#ifndef ITERATIVESOLVER_H
#define ITERATIVESOLVER_H

#include "../source/extemp/variable2D.h"
#include "../source/pde_op/diffop2D.h"
// siehe include unten!!!!!!!!!!!!!


/**
 Abstrake Klasse zur Beschreibung eines iterativen Lösers zur Lösung von \f$ A x = b \f$
 **/
template<class DTyp>
class IterativeSolverAbstract  {
   public:
        /** 
	 * @param blockgrid Gitter auf dem die Gleichung gelösst werden soll
	 **/
        IterativeSolverAbstract(Blockgrid2D& blockgrid_) : blockgrid(&blockgrid_) {  
	       inner_points = NULL;    eps = 0.01;    stopByEps = true;    max_iter = 50;
	       numberIterationsPerformed = 0;
	}
	virtual ~IterativeSolverAbstract() {
	  //test NOW
//cout << " ~IterativeSolverAbstract() " << endl;
	  
	}
	 
	/**
	 * @param stopByEps falls true, endet die iteration durch Abbruchkriterium mit eps
	 * @param eps       Abbruchkriterium
	 * @param max_iter  maximale Anzahl der Iterationen
	 * */
	void setBasicParameter(double eps_, bool stopByEps_ = true, int max_iter_ = 500) { eps = eps_; stopByEps = stopByEps_; max_iter = max_iter_; }
	
	/**
	 * @param inner_points linear equation system is solved only for unknowns on this grid
	 */
	void setGridMarker(Unstructured2DGrid_Marker2D& unknownGridPoints_) { inner_points = &unknownGridPoints_; }
	
	/**
	 * solves the linear equation system \f$ A x = b \f$
	 **/    
	virtual void solve(Local_stiffness_matrix2D<DTyp>& A, Variable2D<DTyp>& b, Variable2D<DTyp>& x) = 0;
	
	
	/**
	 * @return number of iterations which were performed
	 **/
	int numberOfPerformedIterations() { return numberIterationsPerformed; }
	
	/**
	 * @return residuum after iteration. This is not calculated by every iteration method
	 **/
	double getResiduum() { return residuum; }
	
         /**
	 * @return true when iteration method calculates residuum
	 **/
	virtual bool isResiduumCalculated() = 0;
	
protected:
     Blockgrid2D* blockgrid;
     Unstructured2DGrid_Marker2D* inner_points;  // NULL means all points
     double eps;
     bool stopByEps;
     int max_iter;
     
     int numberIterationsPerformed;
     double residuum;
};

#include "GaussSeidel/gaussSeidel.h"
#include "gmres/gmres.h"
#include "gmres/gmres_cc.h"

#endif //ITERATIVESOLVER_H
