/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
 
// ------------------------------------------------------------
//
// calcArg.h
//
// ------------------------------------------------------------

#include <string>
#include <stdexcept>
#include <math.h>

#include "calcArg.h"
#include "../abbrevi2D.h"

#include "../math_lib/math_lib.h"
#include "../grid/elements2D.h"
#include "../grid/parti2D.h"
#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"
#include "../grid/marker2D.h"
#include "extemp2D.h"
#include "variable2D.h"
#include "variable2D_cc.h"
#include "update2D.h"
#include "print_var2D_vtk_cc.h"
#include "co_fu2D.h"


PointForPhase::PointForPhase(int i_, int j_, double r_, int iSmall_, int jSmall_) {
    i = i_;
    j = j_;
    r = r_;
    iSmall = iSmall_;
    jSmall = jSmall_;
}

bool compare_Radius(PointForPhase& pA, PointForPhase& pB) {
   return  pA.getRadius() < pB.getRadius();
}

bool compare_Radius_blocks(BlockForPhase& pA, BlockForPhase& pB) {
   return  pA.getRadius() < pB.getRadius();
}

CalcContinuousArg::CalcContinuousArg(Blockgrid2D& blockgrid_) {  
    blockgrid = &blockgrid_;
    ug = blockgrid->Give_unstructured_grid();
    
    Variable2D<double> norm(blockgrid_);
    X_coordinate2d X(blockgrid_);
    Y_coordinate2d Y(blockgrid_);
    
    norm = X*X+Y*Y;
        
    numberBlocks = ug->Give_number_rectangle();
    
    orderedPoints = new std::list<PointForPhase>[numberBlocks];
        
    enum woSmaller { nichts, Norden, Sueden, Osten, Westen, NOst, SOst, NWest, SWest };
        
    idOrigin = 0;
    iOrigin  = 0;
    jOrigin  = 0;
    NxOrigin = 0;
    
    bool nothingFound    = true;
    double smallesRadius = 0.0;
    
    for(int id=0;id<numberBlocks;++id) {
        int Nx = blockgrid->Give_Nx_rectangle(id);
        int Ny = blockgrid->Give_Ny_rectangle(id);
        for(int i=0;i<=Nx;++i) for(int j=0;j<=Ny;++j) { 
	    norm.Update<rectangleEl>(id);
	    woSmaller woNext = nichts;
	    double radiusM = norm.template Give_data<rectangleEl>(id,i,j,Nx);
	    
/*	    
//Version nächsten der kleineren Radius hat	
	    double radiusNext = -1.0;
	    //Westen
            if(i>0) { 
	       double r = norm.template Give_data<rectangleEl>(id,i-1,j,Nx);
               if(r < radiusM && r > radiusNext) {
		  radiusNext = r;
		  woNext = Westen;
	       }
	    }
	    //Osten
            if(i<Nx) { 
	       double r = norm.template Give_data<rectangleEl>(id,i+1,j,Nx);
               if(r < radiusM && r > radiusNext) {
		  radiusNext = r;
		  woNext = Osten;
	       }
	    }
	    //Sueden
            if(j>0) { 
	       double r = norm.template Give_data<rectangleEl>(id,i,j-1,Nx);
               if(r < radiusM && r > radiusNext) {
		  radiusNext = r;
		  woNext = Sueden;
	       }
	    }
	    //Norden
            if(j<Ny) { 
	       double r = norm.template Give_data<rectangleEl>(id,i,j+1,Nx);
               if(r < radiusM && r > radiusNext) {
		  radiusNext = r;
		  woNext = Norden;
	       }
	    }
*/	    

//Version mit kleinstem Radius
	    double radiusNext = radiusM;
	    //Westen
            if(i>0) { 
	       double r = norm.template Give_data<rectangleEl>(id,i-1,j,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = Westen;
	       }
	    }
	    //Osten
            if(i<Nx) { 
	       double r = norm.template Give_data<rectangleEl>(id,i+1,j,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = Osten;
	       }
	    }
	    //Sueden
            if(j>0) { 
	       double r = norm.template Give_data<rectangleEl>(id,i,j-1,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = Sueden;
	       }
	    }
	    //Norden
            if(j<Ny) { 
	       double r = norm.template Give_data<rectangleEl>(id,i,j+1,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = Norden;
	       }
	    }
	        
	    //Nord-Westen
            if(i>0 && j<Ny) { 
	       double r = norm.template Give_data<rectangleEl>(id,i-1,j+1,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = NWest;
	       }
	    }
	    //Nord-Osten
            if(i<Nx && j<Ny) { 
	       double r = norm.template Give_data<rectangleEl>(id,i+1,j+1,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = NOst;
	       }
	    }
	    //Sued-Westen
            if(i>0 && j>0) { 
	       double r = norm.template Give_data<rectangleEl>(id,i-1,j-1,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = SWest;
	       }
	    }
	    //Sued-Osten
            if(i<Nx && j>0) { 
	       double r = norm.template Give_data<rectangleEl>(id,i+1,j-1,Nx);
               if(r < radiusNext) {
		  radiusNext = r;
		  woNext = SOst;
	       }
	    }
	    
	    if(woNext == nichts) {
	       if(nothingFound) {
		  nothingFound = false;
		  idOrigin = id;
                  iOrigin  = i;
                  jOrigin  = j;
		  NxOrigin = Nx;
                  smallesRadius = radiusM;
	       }
	       else {
		  if(smallesRadius > radiusM) {
		     idOrigin = id;
                     iOrigin  = i;
                     jOrigin  = j;
  		     NxOrigin = Nx;		     
                     smallesRadius = radiusM;		    
		  }
	       }
	    }


	    if(woNext == Westen) orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i-1,j));
	    if(woNext == Osten)  orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i+1,j));
	    if(woNext == Sueden) orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i,j-1));
	    if(woNext == Norden) orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i,j+1));
	    
	    if(woNext == NWest) orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i-1,j+1));
	    if(woNext == NOst)  orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i+1,j+1));	    
	    if(woNext == SWest) orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i-1,j-1));
	    if(woNext == SOst)  orderedPoints[id].push_back(PointForPhase(i,j,radiusM,i+1,j-1));	    	    
	}
	orderedPoints[id].sort(compare_Radius);    
	
	orderedBlocks.push_back(BlockForPhase(id,orderedPoints[id].front().getRadius()));
    }
    
    orderedBlocks.sort(compare_Radius_blocks);    

/*    
     for(std::list<BlockForPhase>::iterator it=orderedBlocks.begin(); it != orderedBlocks.end(); ++it) {
         std::cout << " id: " << it->getId() << " R: " << it->getRadius() << std::endl;
     }
     */
}

double phaseComp(std::complex<double> value) {
  return atan2(value.imag(),value.real());
}

void CalcContinuousArg::calcArg(Variable2D<double>& phase, Variable2D<std::complex<double> >& Efield, bool zeroAtOrigin) {  
   double phaseOrigin = 0.0;
   
   if(zeroAtOrigin) {
      std::complex<double> EField_origin = Efield.template Give_data<rectangleEl>(idOrigin,iOrigin,jOrigin,NxOrigin);   
      phaseOrigin = atan2(EField_origin.imag(),EField_origin.real());        
   }
   
   Function2d1<double, std::complex<double> >  Phase(phaseComp);
   phase = Phase(Efield) - phaseOrigin;      
   
   for(std::list<BlockForPhase>::iterator it=orderedBlocks.begin(); it != orderedBlocks.end(); ++it) {
       int id = it->getId();
       phase.Update<rectangleEl>(id);       
	   
       for(std::list<PointForPhase>::iterator itP=orderedPoints[id].begin(); itP != orderedPoints[id].end(); ++itP) {
           int Nx = blockgrid->Give_Nx_rectangle(id);       
	 
	   int i = itP->getI();
	   int j = itP->getJ();
	   
	   std::complex<double> EField_M = Efield.template Give_data<rectangleEl>(id,i,j,Nx); // muss man das echt nochmal holen???
           double phaseM = atan2(EField_M.imag(),EField_M.real()) - phaseOrigin;
	   	   
	   int iSmall = itP->getIsmall();
	   int jSmall = itP->getJsmall();
           
           double phaseNext = phase.template Give_data<rectangleEl>(id,iSmall,jSmall,Nx);
 	   

	   int numPi = (phaseM-phaseNext)/(2.0*M_PI);
	   phaseM = phaseM - 2.0*M_PI * numPi;

	   if(fabs(phaseM-phaseNext)>M_PI) {
	      if(phaseM < phaseNext) {
		while(phaseM < phaseNext && fabs(phaseM-phaseNext)>M_PI) phaseM = phaseM + 2.0*M_PI;
	      }
	      else {
		while(phaseM > phaseNext && fabs(phaseM-phaseNext)>M_PI) phaseM = phaseM - 2.0*M_PI;
	      }
              phase.setValueAtRectangularPoint(phaseM,id,i,j,Nx);	      
	   }
       }
       phase.Update_back(id);
    }
}






