/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// co_fu.h
//
// ------------------------------------------------------------

#include "extemp2D.h"

#ifndef CO_FU2D_H
#define CO_FU2D_H

#ifdef CPP11
#include <tuple>
using std::tuple;
#endif
//////////////////////////////////////////////////////////////
// 1. Coordinate
// 2. Meshsize x,y,z
// 3. Functions
//////////////////////////////////////////////////////////////

// Gitter:
// * boundary point
// + interior point

/*
Ny    *  *  *  *  *  *
Ny-1  *  +  +  +  +  *
3     *  +  +  +  +  *
2     *  +  +  +  +  *
1     *  +  +  +  +  *
0     *  *  *  *  *  *
      0  1  2  3 Nx-1 Nx
*/

///////////////////////////
// 1. Coordinate
///////////////////////////

/** \defgroup ExpressionTemplates2D  ''Expression Template Operators in 2D (excluding +,-,... ) ''    
 **/
/* @{  */ 
class X_coordinate2d : public Expr2D<X_coordinate2d> {

	public:
		X_coordinate2d ( Blockgrid2D& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid2D* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		
		typedef double Result;
		
		template <elementTyp TYP_EL>
		inline double Give_data ( params2D_in ) const;
		
		inline double Give_cell_rectangle ( params2D_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid2D*     blockgrid;
};
/* @} */

inline double X_coordinate2d::Give_cell_rectangle ( params2D_in_cell ) const {
	//  return  blockgrid->Give_coord_rectangle(id_rec, i, j, k).x;
	
	return 0.25* ( blockgrid->Give_coord_rectangle ( id_rec, i  , j  ).x +
	               blockgrid->Give_coord_rectangle ( id_rec, i+1, j  ).x +
	               blockgrid->Give_coord_rectangle ( id_rec, i  , j+1).x +
	               blockgrid->Give_coord_rectangle ( id_rec, i+1, j+1).x);
}

template <>
inline double X_coordinate2d::Give_data<rectangleEl> ( params2D_in ) const {
	return blockgrid->Give_coord_rectangle ( id, i, j).x;
}

template <>
inline double X_coordinate2d::Give_data<edgeEl> ( params2D_in ) const {
	return blockgrid->Give_coord_edge ( id, i ).x;
}

template <>
inline double X_coordinate2d::Give_data<pointEl> ( params2D_in ) const {
	return blockgrid->Give_coord_point ( id ).x;
}

/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
class Y_coordinate2d : public Expr2D<Y_coordinate2d> {

	public:
		Y_coordinate2d ( Blockgrid2D& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid2D* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		
		typedef double Result;
		
		template <elementTyp TYP_EL>
		inline double Give_data ( params2D_in ) const;
		
		inline double Give_cell_rectangle ( params2D_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid2D*     blockgrid;
};
/* @} */

inline double Y_coordinate2d::Give_cell_rectangle ( params2D_in_cell ) const {
	return 0.25* (blockgrid->Give_coord_rectangle ( id_rec, i  , j  ).y +
	              blockgrid->Give_coord_rectangle ( id_rec, i+1, j  ).y +
	              blockgrid->Give_coord_rectangle ( id_rec, i  , j+1).y +
	              blockgrid->Give_coord_rectangle ( id_rec, i+1, j+1).y );
}

template <>
inline double Y_coordinate2d::Give_data<rectangleEl> ( params2D_in ) const {
	return blockgrid->Give_coord_rectangle ( id, i, j).y;
}


template <>
inline double Y_coordinate2d::Give_data<edgeEl> ( params2D_in ) const {
	return blockgrid->Give_coord_edge ( id, i ).y;
}

template <>
inline double Y_coordinate2d::Give_data<pointEl> ( params2D_in ) const {
	return blockgrid->Give_coord_point ( id ).y;
}

//////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////

///////////////////////////
// 2. Meshsize_x,y,z
///////////////////////////
/*
///////// x-direction
class X_meshsize : public Expr2D<X_meshsize> {

	public:
		X_meshsize ( Blockgrid2D& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid2D* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		
		typedef double Result;
		
		
		//  template <elementTyp TYP_EL>
		//  inline double Give_data(params2D_in) const;
		
		inline double Give_cell_rectangle ( params2D_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid2D*     blockgrid;
};

inline double X_meshsize::Give_cell_rectangle ( params2D_in_cell ) const {
	return 0.25 * ABS ( ( blockgrid->Give_coord_rectangle ( id_rec, i + 1 , j     , k ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1 , j + 1 , k ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1 , j     , k + 1 ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1 , j + 1 , k + 1 ).z ) -
	                    ( blockgrid->Give_coord_rectangle ( id_rec, i,      j     , k ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i,      j + 1 , k ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i,      j     , k + 1 ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i,      j + 1 , k + 1).z ) );
}

///////// y-direction
class Y_meshsize : public Expr2D<Y_meshsize> {

	public:
		Y_meshsize ( Blockgrid2D& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid2D* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		
		typedef double Result;
		
		
		//  template <elementTyp TYP_EL>
		//  inline double Give_data(params2D_in) const;
		
		inline double Give_cell_rectangle ( params2D_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid2D*     blockgrid;
};

inline double Y_meshsize::Give_cell_rectangle ( params2D_in_cell ) const {
  	return 0.25 * ABS ( ( blockgrid->Give_coord_rectangle ( id_rec, i    , j + 1, k     ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1, j + 1, k     ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i    , j + 1, k + 1 ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1, j + 1, k + 1 ).z ) -
	                    ( blockgrid->Give_coord_rectangle ( id_rec, i    , j    , k     ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1, j    , k     ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i    , j    , k + 1 ).z +
	                      blockgrid->Give_coord_rectangle ( id_rec, i + 1, j    , k + 1 ).z ) );
}









*/

///////////////////////////
// 3. Functions
///////////////////////////

template<class A, typename DTyp, typename DTyp_Result>
  class Exp2D_Function1 : public Expr2D<Exp2D_Function1<A, DTyp, DTyp_Result > >
  {

    const A& a_;

    DTyp_Result
    (*Formula_)(DTyp x);

  public:

    inline
    Exp2D_Function1(const A& a, DTyp_Result (*Formula)(DTyp x))
        : a_(a), Formula_(Formula)
    {
    }

    stencil_typ
    Give_stencil_typ() const
    {
      return no_stencil;
    }

    Blockgrid2D* Give_blockgrid() const { return a_.Give_blockgrid(); };

    typedef DTyp_Result Result;

    template<elementTyp TYP_EL>
     inline DTyp_Result
      Give_data(params2D_in) const;

    inline DTyp_Result
    Give_cell_rectangle(params2D_in_cell) const;

    inline DTyp_Result
    Give_matrix_rectangles(params2D_in_loc_matrix) const;

    inline DTyp_Result
    Give_matrix_rectangle(params2D_in_loc_matrix) const;    
    
    template<elementTyp TYP_EL>
      void
      Update(int id) const
      {
        a_.template  Update<TYP_EL>(id);
      }

    inline Unstructured2DGrid * Give_Ug() const {
      return a_.Give_Ug();
    }
  };



template<class A, class DTyp, class DTyp_Result>
  template<elementTyp TYP_EL>
    inline DTyp_Result
    Exp2D_Function1<A, DTyp, DTyp_Result>::Give_data(params2D_in) const
    {
      return Formula_ ( a_.template Give_data<TYP_EL> ( params2D_out ) );
    }


template<class A, class DTyp, class DTyp_Result>
  inline DTyp_Result
  Exp2D_Function1<A, DTyp, DTyp_Result>::Give_cell_rectangle(params2D_in_cell) const
  {
    return Formula_ ( a_.Give_cell_rectangle ( params2D_out_cell ) );
  }

template<class A, class DTyp, class DTyp_Result>
  inline DTyp_Result
  Exp2D_Function1<A, DTyp, DTyp_Result>::Give_matrix_rectangles(params2D_in_loc_matrix) const
  {
    return Formula_ ( a_.Give_matrix_rectangles ( params2D_out_loc_matrix) );
  }

template<class A, class DTyp, class DTyp_Result>  
    inline DTyp_Result
    Exp2D_Function1<A, DTyp, DTyp_Result>::Give_matrix_rectangle(params2D_in_loc_matrix) const
    {
       return Formula_ ( a_.Give_matrix_rectangle(params2D_out_loc_matrix) );
    }
    
  
/** \addtogroup ExpressionTemplates2D **/
/* @{ */    
template <class DTyp_Result, class DTyp>
class Function2d1 {

        public:
                typedef DTyp_Result Result;
                Function2d1 ( DTyp_Result ( *formula ) ( DTyp x )) : Formula ( formula ) {}

                template <class A>
                inline Exp2D_Function1<A, DTyp, DTyp_Result> operator() ( const Expr2D<A>& a) const
                { return Exp2D_Function1<A, DTyp, DTyp_Result> ( a, Formula); }

        private:
                DTyp_Result ( *Formula) ( DTyp x );
};
/* @} */

template<class A, class B, typename DTyp, typename DTyp_Result>
class Exp2D_Function2 : public Expr2D<Exp2D_Function2<A, B, DTyp, DTyp_Result> > {
                const A& a_;
                const B& b_;

                DTyp_Result
                (*Formula_)(DTyp x, DTyp y);

                template<elementTyp TYP_EL>
                   inline DTyp_Result
                   Give_data_private(params2D_in) const;

        public:
                typedef DTyp_Result Result;

                inline Exp2D_Function2 ( const A& a, const B& b, DTyp_Result ( *Formula ) ( DTyp x, DTyp y) ) :
                                a_ ( a ), b_ ( b ), Formula_ ( Formula ) {}

                Blockgrid2D* Give_blockgrid() const { return a_.Give_blockgrid(); };

                stencil_typ Give_stencil_typ() const { return no_stencil; }


                template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params2D_in ) const;

                inline DTyp_Result Give_cell_rectangle ( params2D_in_cell ) const;

                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                        b_.template  Update<TYP_EL> ( id );
                }

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
};

template<class A, class B, class DTyp, class DTyp_Result>
  template<elementTyp TYP_EL>
    inline DTyp_Result
    Exp2D_Function2<A,B, DTyp, DTyp_Result>::Give_data_private(params2D_in) const
    {
      return Formula_ ( a_.template Give_data<TYP_EL> ( params2D_out ),b_.template Give_data<TYP_EL> ( params2D_out ));
    }

template<class A, class B, typename DTyp, typename DTyp_Result>
template <elementTyp TYP_EL>
inline DTyp_Result Exp2D_Function2<A, B, DTyp, DTyp_Result>::Give_data ( params2D_in ) const {
  return Give_data_private<TYP_EL>( params2D_out );
}

template<class A, class B, typename DTyp, typename DTyp_Result>
inline DTyp_Result Exp2D_Function2<A, B, DTyp, DTyp_Result>::Give_cell_rectangle ( params2D_in_cell ) const {
        return Formula_ ( a_.Give_cell_rectangle ( params2D_out_cell ),
                          b_.Give_cell_rectangle ( params2D_out_cell ) );
}

/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
template <class DTyp_Result, class DTyp>
class Function2d2 {

        public:
                typedef DTyp_Result Result;
                Function2d2 ( DTyp_Result ( *formula ) ( DTyp x, DTyp y )) : Formula ( formula ) {};

                template <class A, class B>
                inline Exp2D_Function2<A, B, DTyp, DTyp_Result> operator() ( const Expr2D<A>& a,
                        const Expr2D<B>& b ) const
                { return Exp2D_Function2<A, B, DTyp, DTyp_Result> ( a, b, Formula); }


        private:
                DTyp_Result ( *Formula ) ( DTyp x, DTyp y  );
};
/* @} */

//////////////////////////////////////////////////////////////////////////
//  3 arguments
//////////////////////////////////////////////////////////////////////////
template <class A, class B, class C, class DTyp>

class Exp2D_Function3 : public Expr2D<Exp2D_Function3<A, B, C, DTyp> > {
                const A& a_;
                const B& b_;
                const C& c_;
                DTyp ( *Formula_ ) ( DTyp x, DTyp y, DTyp z );

        public:
                inline Exp2D_Function3 ( const A& a, const B& b, const C& c, DTyp ( *Formula ) ( DTyp x, DTyp y, DTyp z )) :
                                a_ ( a ), b_ ( b ), c_ ( c ), Formula_ ( Formula ) {}

                stencil_typ Give_stencil_typ() const { return no_stencil; }

                Blockgrid2D* Give_blockgrid() const { return a_.Give_blockgrid(); };

                typedef DTyp Result;

                template <elementTyp TYP_EL>
                inline DTyp Give_data ( params2D_in ) const;

                inline DTyp Give_cell_rectangle ( params2D_in_cell ) const;


                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                        b_.template  Update<TYP_EL> ( id );
                        c_.template  Update<TYP_EL> ( id );
                }

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
};

template <class A, class B, class C, class DTyp>
template <elementTyp TYP_EL>
inline DTyp Exp2D_Function3<A, B, C, DTyp>::Give_data ( params2D_in ) const {
        return Formula_ ( a_.template Give_data<TYP_EL> ( params2D_out ),
                          b_.template Give_data<TYP_EL> ( params2D_out ),
                          c_.template Give_data<TYP_EL> ( params2D_out ) );
}

template <class A, class B, class C, class DTyp>
inline DTyp Exp2D_Function3<A, B, C, DTyp>::Give_cell_rectangle ( params2D_in_cell ) const {
        return Formula_ ( a_.Give_cell_rectangle ( params2D_out_cell ),
                          b_.Give_cell_rectangle ( params2D_out_cell ),
                          c_.Give_cell_rectangle ( params2D_out_cell ) );
}

/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
template <class DTyp>
class Function2d3 {

        public:
                typedef DTyp Result;

                Function2d3 ( DTyp ( *formula ) ( DTyp x, DTyp y, DTyp z )) : Formula ( formula ) {};

                template <class A, class B, class C>
                inline Exp2D_Function3<A, B, C, DTyp> operator() ( const Expr2D<A>& a,
                        const Expr2D<B>& b,
                        const Expr2D<C>& c ) const
                { return Exp2D_Function3<A, B, C, DTyp> ( a, b, c, Formula); }



        private:
                DTyp ( *Formula ) ( DTyp x, DTyp y, DTyp z );
};
/* @} */


//////////////////////////////////////////////////////////////////////////
//  4 arguments
//////////////////////////////////////////////////////////////////////////
template <class A, class B, class C, class D, class DTyp>
class Exp2D_Function4 : public Expr2D<Exp2D_Function4<A, B, C, D, DTyp> > {
                const A& a_;
                const B& b_;
                const C& c_;
                const D& d_;

                DTyp ( *Formula_ ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4);

        public:
                inline Exp2D_Function4(const A& a,
                                       const B& b,
                                       const C& c,
                                       const D& d,
                                       DTyp ( *Formula ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4 )) :
                                a_(a), b_(b), c_(c), d_(d), Formula_ ( Formula ) {}

                stencil_typ Give_stencil_typ() const { return no_stencil; }

                typedef DTyp Result;

                template <elementTyp TYP_EL>
                inline DTyp Give_data ( params2D_in ) const;

                inline DTyp Give_cell_rectangle ( params2D_in_cell ) const;


                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                        b_.template  Update<TYP_EL> ( id );
                        c_.template  Update<TYP_EL> ( id );
                        d_.template  Update<TYP_EL> ( id );
                }

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
};

template <class A, class B, class C, class D, class DTyp>
template <elementTyp TYP_EL>
inline DTyp Exp2D_Function4<A, B, C, D, DTyp>::Give_data ( params2D_in ) const {
        return Formula_ ( a_.template Give_data<TYP_EL> ( params2D_out ),
                          b_.template Give_data<TYP_EL> ( params2D_out ),
                          c_.template Give_data<TYP_EL> ( params2D_out ),
                          d_.template Give_data<TYP_EL> ( params2D_out )
                        );
}

template <class A, class B, class C, class D, class DTyp>
inline DTyp Exp2D_Function4<A, B, C, D, DTyp>::Give_cell_rectangle ( params2D_in_cell ) const {
        return Formula_ ( a_.Give_cell_rectangle ( params2D_out_cell ),
                          b_.Give_cell_rectangle ( params2D_out_cell ),
                          c_.Give_cell_rectangle ( params2D_out_cell ),
                          d_.Give_cell_rectangle ( params2D_out_cell )
                        );
}

/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
template <class DTyp>
class Function2d4 {
        public:
                typedef DTyp Result;

                Function2d4( DTyp ( *formula ) (DTyp x1, DTyp x2, DTyp x3, DTyp x4)) : Formula ( formula ) {};

                template <class A, class B, class C, class D>
                inline Exp2D_Function4<A, B, C, D, DTyp> operator() ( const Expr2D<A>& a,
                        const Expr2D<B>& b,
                        const Expr2D<C>& c,
                        const Expr2D<D>& d) const
                { return Exp2D_Function4<A, B, C, D, DTyp> ( a, b, c, d, Formula); }



        private:
                DTyp ( *Formula ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4);
};
/* @} */


//////////////////////////////////////////////////////////////////////////
//  5 arguments
//////////////////////////////////////////////////////////////////////////
template <class A, class B, class C, class D, class E, class DTyp>
class Exp2D_Function5 : public Expr2D<Exp2D_Function5<A, B, C, D, E, DTyp> > {
                const A& a_;
                const B& b_;
                const C& c_;
                const D& d_;
                const E& e_;

                DTyp ( *Formula_ ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5);

        public:
                inline Exp2D_Function5(const A& a,
                                       const B& b,
                                       const C& c,
                                       const D& d,
                                       const E& e,
                                       DTyp ( *Formula ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5 )) :
                                a_(a), b_(b), c_(c), d_(d), e_(e), Formula_ ( Formula ) {}

                stencil_typ Give_stencil_typ() const { return no_stencil; }

                typedef DTyp Result;

                template <elementTyp TYP_EL>
                inline DTyp Give_data ( params2D_in ) const;

                inline DTyp Give_cell_rectangle ( params2D_in_cell ) const;


                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                        b_.template  Update<TYP_EL> ( id );
                        c_.template  Update<TYP_EL> ( id );
                        d_.template  Update<TYP_EL> ( id );
                        e_.template  Update<TYP_EL> ( id );
                }

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
};

template <class A, class B, class C, class D, class E, class DTyp>
template <elementTyp TYP_EL>
inline DTyp Exp2D_Function5<A, B, C, D, E, DTyp>::Give_data ( params2D_in ) const {
        return Formula_ ( a_.template Give_data<TYP_EL> ( params2D_out ),
                          b_.template Give_data<TYP_EL> ( params2D_out ),
                          c_.template Give_data<TYP_EL> ( params2D_out ),
                          d_.template Give_data<TYP_EL> ( params2D_out ),
                          e_.template Give_data<TYP_EL> ( params2D_out )
                        );
}

template <class A, class B, class C, class D, class E, class DTyp>
inline DTyp Exp2D_Function5<A, B, C, D, E, DTyp>::Give_cell_rectangle ( params2D_in_cell ) const {
        return Formula_ ( a_.Give_cell_rectangle ( params2D_out_cell ),
                          b_.Give_cell_rectangle ( params2D_out_cell ),
                          c_.Give_cell_rectangle ( params2D_out_cell ),
                          d_.Give_cell_rectangle ( params2D_out_cell ),
                          e_.Give_cell_rectangle ( params2D_out_cell )
                        );
}

/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
template <class DTyp>
class Function2d5 {

        public:
                typedef DTyp Result;

                Function2d5( DTyp ( *formula ) (DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5)) : Formula ( formula ) {};

                template <class A, class B, class C, class D, class E>
                inline Exp2D_Function5<A, B, C, D, E, DTyp> operator() ( const Expr2D<A>& a,
                        const Expr2D<B>& b,
                        const Expr2D<C>& c,
                        const Expr2D<D>& d,
                        const Expr2D<E>& e) const
                { return Exp2D_Function5<A, B, C, D, E, DTyp> ( a, b, c, d, e, Formula); }



        private:
                DTyp ( *Formula ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5 );
};
/* @} */

//////////////////////////////////////////////////////////////////////////
//  6 arguments
//////////////////////////////////////////////////////////////////////////
template <class A, class B, class C, class D, class E, class F, class DTyp>
class Exp2D_Function6 : public Expr2D<Exp2D_Function6<A, B, C, D, E, F, DTyp> > {
                const A& a_;
                const B& b_;
                const C& c_;
                const D& d_;
                const E& e_;
                const F& f_;

                DTyp ( *Formula_ ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5, DTyp x6);

        public:
                inline Exp2D_Function6(const A& a,
                                       const B& b,
                                       const C& c,
                                       const D& d,
                                       const E& e,
                                       const F& f,
                                       DTyp ( *Formula ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5, DTyp x6 )) :
                                a_(a), b_(b), c_(c), d_(d), e_(e), f_(f), Formula_ ( Formula ) {}

                stencil_typ Give_stencil_typ() const { return no_stencil; }

                typedef DTyp Result;

                template <elementTyp TYP_EL>
                inline DTyp Give_data ( params2D_in ) const;

                inline DTyp Give_cell_rectangle ( params2D_in_cell ) const;


                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template  Update<TYP_EL> ( id );
                        b_.template  Update<TYP_EL> ( id );
                        c_.template  Update<TYP_EL> ( id );
                        d_.template  Update<TYP_EL> ( id );
                        e_.template  Update<TYP_EL> ( id );
                        f_.template  Update<TYP_EL> ( id );
                }

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
};

template <class A, class B, class C, class D, class E, class F, class DTyp>
template <elementTyp TYP_EL>
inline DTyp Exp2D_Function6<A, B, C, D, E, F, DTyp>::Give_data ( params2D_in ) const {
        return Formula_ ( a_.template Give_data<TYP_EL> ( params2D_out ),
                          b_.template Give_data<TYP_EL> ( params2D_out ),
                          c_.template Give_data<TYP_EL> ( params2D_out ),
                          d_.template Give_data<TYP_EL> ( params2D_out ),
                          e_.template Give_data<TYP_EL> ( params2D_out ),
                          f_.template Give_data<TYP_EL> ( params2D_out )
                        );
}

template <class A, class B, class C, class D, class E, class F, class DTyp>
inline DTyp Exp2D_Function6<A, B, C, D, E, F, DTyp>::Give_cell_rectangle ( params2D_in_cell ) const {
        return Formula_ ( a_.Give_cell_rectangle ( params2D_out_cell ),
                          b_.Give_cell_rectangle ( params2D_out_cell ),
                          c_.Give_cell_rectangle ( params2D_out_cell ),
                          d_.Give_cell_rectangle ( params2D_out_cell ),
                          e_.Give_cell_rectangle ( params2D_out_cell ),
                          f_.Give_cell_rectangle ( params2D_out_cell )
                        );
}

/** \addtogroup ExpressionTemplates2D **/
/* @{ */ 
template <class DTyp>
class Function2d6 {

        public:
                typedef DTyp Result;

                Function2d6( DTyp ( *formula ) (DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5, DTyp x6)) : Formula ( formula ) {};

                template <class A, class B, class C, class D, class E, class F>
                inline Exp2D_Function6<A, B, C, D, E, F, DTyp> operator() ( const Expr2D<A>& a,
                        const Expr2D<B>& b,
                        const Expr2D<C>& c,
                        const Expr2D<D>& d,
                        const Expr2D<E>& e,
                        const Expr2D<F>& f) const
                { return Exp2D_Function6<A, B, C, D, E, F, DTyp> ( a, b, c, d, e, f, Formula); }



        private:
                DTyp ( *Formula ) ( DTyp x1, DTyp x2, DTyp x3, DTyp x4, DTyp x5, DTyp x6 );
};
/* @} */


#endif // CO_FU_H
