/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// interpolateFrom3D.h
//
// ------------------------------------------------------------

#ifndef INTERPOLFROM3D_H
#define INTERPOLFROM3D_H
#ifdef _OPENMP
#include <omp.h>
#include "../ugopenmp.h"
#endif

//////////////////////////////////////////////////////////////
// 1. z-Iterator
// 2. Interpolation 2D->2D
// 3. Interpolation 3D->2D
// 4. Intergrate 3D->2D
//////////////////////////////////////////////////////////////

// Gitter:
// * boundary point
// + interior point

/*
Ny    *  *  *  *  *  *
Ny-1  *  +  +  +  +  *
3     *  +  +  +  +  *
2     *  +  +  +  +  *
1     *  +  +  +  +  *
0     *  *  *  *  *  *
      0  1  2  3 Nx-1 Nx
*/
// k=0...Nz            for(k=0;k<=Nz;++k)
//   j=0...Ny            for(j=0;j<=Ny;++j)
//     i=0...Nx            for(i=0;i<=Nx;++i)
//       index = i + (Nx+1)*(j+k*(Ny+1))


/////////////////////////////////////////////////////////
// 1. z-Iterator
/////////////////////////////////////////////////////////

class IteratorZDirection {
public:
   IteratorZDirection(Blockgrid* blockgrid);
   
   bool next(); ///> nächster slice in z-Richtung
   bool has_next(); ///> true falls es nächsten slice in z-Richtung gibt   
   bool previous(); ///> vorheriger slice in z-Richtung
   bool nextDirection(bool right); ///> nächster oder vorheriger slice in z-Richtung, nächster falls right==true
   void gotoFront(); ///> geh zum ersten slice in z-Richtung, dies ist default beim constructor
   void gotoEnd(); ///> geh zum letzten slice in z-Richtung
   void gotoStartDirection(bool right); ///> geh zum start slice in z-Richtung in Richtung right   
   
   int get_k() { return index; } ///> globaler NummerierungsIndex: 0, ... , getSize()-1
   int getSize()  { return totalNumberSlices; } ///> maximale Anzahl der indizes
      
   int get_k_intern() { return k; } ///> nur nummer in einem Block
   int get_nummerBlock() { return nummerBlock; }
   double get_z() { return z_Wert; }
   double get_hz() { return hz; }
   TypeOfUG giveTypeUg() {return typeUg;}
   
private:
  void startBlock(int numBlock);
  void endBlock(int numBlock);
  
  //bool iterate_Interfaces_Two_Times;
  
  // Koordinaten informationen
  double z_Wert;
  double hz;
  double zLeft;  // start z-Koordinate des Blocks
  
  // fuer id in einem Block
  int k;
  int Nz;

  // globaler NummerierungsIndex
  int index;
  int totalNumberSlices;
  
  // fuer nummer des Blocks in z-Richtung
  int nummerBlock;
  int nummernBloecke;
  
  // pointer auf das ug
  UG_WithZdirection* ug_withZ;
  Blockgrid* blockgrid;
  TypeOfUG typeUg;
};

/////////////////////////////////////////////////////////
// 2. Interpolation 2D->2D
/////////////////////////////////////////////////////////

template <class DTyp>
void Variable<DTyp>::interpolateSlizeZ(const Variable2D<DTyp>* U, IteratorZDirection* iterZ) {
   Blockgrid2D* blockgrid2D     = U->Give_blockgrid();
   Unstructured2DGrid* ug2Dp  = blockgrid2D->Give_unstructured_grid();
   TypeOfUG typeUg = ug->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug);
   
   int numberCorners2D = ug2Dp->Give_number_points();

   // Rectangle -> Hexaeder
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int Nx = blockgrid2D->Give_Nx_rectangle(0);
      int Ny = blockgrid2D->Give_Ny_rectangle(0);
      U->template  Update<rectangleEl>(0);
      
      int k  = iterZ->get_k_intern();
      int id = iterZ->get_nummerBlock();

      Update<hexahedronEl>(id);
      U->template  Update<rectangleEl>(0);
      for(int j = 0;j <= Ny;++j)
          for(int i = 0;i <= Nx;++i) { 
              data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template Give_data<rectangleEl>(0, i, j, Nx);
	  }
      Update_back(id);
   }
   else if(typeUg==cylinderUgTyp || typeUg == hexahedronAdaptiveUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = iterZ->get_nummerBlock();
      int num_blocks = ug3DZ->getNummernBloecke();
      int k  = iterZ->get_k_intern();
	  
      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

	  Update<hexahedronEl>(id);
          U->template  Update<rectangleEl>(id_rec);
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i) { 
                  data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template Give_data<rectangleEl>(id_rec, i, j, Nx);
	      }
          Update_back(id);
      }
   }
   else if(typeUg==cylinderListUgTyp) { // List Cylinder -> disc // neu Christoph: 29.10.14
      int id_block = iterZ->get_nummerBlock();
      int num_blocks = ug3DZ->getNummernBloecke();
      int k  = iterZ->get_k_intern();
	  
      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

	  Update<hexahedronEl>(id);
          U->template  Update<rectangleEl>(id_rec);
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i) { 
                  data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template  Give_data<rectangleEl>(id_rec, i, j, Nx);
	      }
          Update_back(id);
      }
   }   
   else if(typeUg == hexahedronListUgTyp) {
     HexahedronXYList* hexahedron = (HexahedronXYList*)ug;
     int n_x = numberCorners2D==hexahedron->n_x;
     int n_y = numberCorners2D==hexahedron->n_y;
     assert((n_x+1) * (n_y+1));

     int id_block = iterZ->get_nummerBlock();
     int num_blocks = ug3DZ->getNummernBloecke();
     int k  = iterZ->get_k_intern();

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         Update<hexahedronEl>(id);
         U->template  Update<rectangleEl>(id_rec);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template  Give_data<rectangleEl>(id_rec, i, j, Nx);
             }
         Update_back(id);
     }
   }
   else if(typeUg == hexahedronListTestUgTyp) {
     HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)ug;
     int n_x = numberCorners2D==hexahedron->n_x;
     int n_y = numberCorners2D==hexahedron->n_y;
     assert((n_x+1) * (n_y+1));

     int id_block = iterZ->get_nummerBlock();
     int num_blocks = ug3DZ->getNummernBloecke();
     int k  = iterZ->get_k_intern();

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         Update<hexahedronEl>(id);
         U->template  Update<rectangleEl>(id_rec);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template  Give_data<rectangleEl>(id_rec, i, j, Nx);
             }
         Update_back(id);
     }
   }
   else {
     std::cout << " noch nicht implementiert!! " << std::endl;
     assert(false); 
   }
}


template <class DTyp>
template <class A>
void Variable<DTyp>::interpolateSlizeZ(const Expr2D<A>& a_, IteratorZDirection* iterZ) {
   const A& a(a_);
   Blockgrid2D* blockgrid2D     = a.Give_blockgrid();
   Unstructured2DGrid* ug2Dp  = blockgrid2D->Give_unstructured_grid();
   TypeOfUG typeUg = ug->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug);

   int numberCorners2D = ug2Dp->Give_number_points();

   // Rectangle -> Hexaeder
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int Nx = blockgrid2D->Give_Nx_rectangle(0);
      int Ny = blockgrid2D->Give_Ny_rectangle(0);
      a.template Update<rectangleEl>(0);

      int k  = iterZ->get_k_intern();
      int id = iterZ->get_nummerBlock();

      Update<hexahedronEl>(id);
      a.template Update<rectangleEl>(0);
      for(int j = 0;j <= Ny;++j)
          for(int i = 0;i <= Nx;++i) {
              data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = a.template Give_data<rectangleEl>(0, i, j, Nx);
          }
      Update_back(id);
   }
   else if(typeUg==cylinderUgTyp || typeUg == hexahedronAdaptiveUgTyp || typeUg==cylinderListUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = iterZ->get_nummerBlock();
      int num_blocks = ug3DZ->getNummernBloecke();
      int k  = iterZ->get_k_intern();

      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          Update<hexahedronEl>(id);
          a.template Update<rectangleEl>(id_rec);
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i) {
                  data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = a.template Give_data<rectangleEl>(id_rec, i, j, Nx);
              }
          Update_back(id);
      }
   }
   else if(typeUg == hexahedronListUgTyp) {
     HexahedronXYList* hexahedron = (HexahedronXYList*)ug;
     int n_x = hexahedron->n_x;
     int n_y = hexahedron->n_y;
     assert((n_x+1) * (n_y+1));

     int id_block = iterZ->get_nummerBlock();
     int num_blocks = ug3DZ->getNummernBloecke();
     int k  = iterZ->get_k_intern();

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         Update<hexahedronEl>(id);
         a.template Update<rectangleEl>(id_rec);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = a.template Give_data<rectangleEl>(id_rec, i, j, Nx);
             }
         Update_back(id);
     }
   }
   else if(typeUg == hexahedronListTestUgTyp) {
     HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)ug;
     int n_x = numberCorners2D==hexahedron->n_x;
     int n_y = numberCorners2D==hexahedron->n_y;
     assert((n_x+1) * (n_y+1));

     int id_block = iterZ->get_nummerBlock();
     int num_blocks = ug3DZ->getNummernBloecke();
     int k  = iterZ->get_k_intern();

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid2D->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid2D->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         Update<hexahedronEl>(id);
         a.template Update<rectangleEl>(id_rec);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = a.template Give_data<rectangleEl>(id_rec, i, j, Nx);
             }
         Update_back(id);
     }
   }
   else {
     std::cout << " noch nicht implementiert!! " << std::endl;
     assert(false);
   }
}

/////////////////////////////////////////////////////////
// 3. Interpolation 3D->2D
// 3.1. mit double z
/////////////////////////////////////////////////////////



int find_k(double zLeft, double zRight, double z, int Nz);

template <class DTyp>
template <class DTyp1>
void Variable2D<DTyp>::interpolateSlizeZ(const Variable<DTyp1>* U, double z) {
//  assert(-1);
//}
//
//template <>
//template <class DTyp1>
//void Variable2D<DTyp1>::interpolateSlizeZ(Variable<DTyp1>* U, double z) {
   Blockgrid* blockgrid3D    = U->Give_blockgrid();
   Unstructured_grid* ug3Dp  = blockgrid3D->Give_unstructured_grid();
   TypeOfUG typeUg = ug3Dp->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3Dp);
   
   int numberCorners2D = ug->Give_number_points();
   // Hexaeder -> Rectangle 
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int id = ug3DZ->getNummerBlockZ(z);
            
      int Nx = blockgrid->Give_Nx_rectangle(0);
      int Ny = blockgrid->Give_Ny_rectangle(0);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id);
            
      double zLeft  = ug3DZ->zLeft(id);                           //ug3Dp->Give_point(0)->Give_coordinate().z;
      double zRight = ug3DZ->zRight(id);                          //ug3Dp->Give_point(4)->Give_coordinate().z;
      int kf = find_k(zLeft,zRight,z,Nz);
      if(kf==0 || kf==Nz) { //links und rechts
         int k=kf;
	 U->template  Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) { 
                 data_rectangles[0][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );
	     }
	 Update_back(0);
      }
      else { //mitte
         int k;
	 double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
	 double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

	 U->template  Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
		 k=kf-1;
                 double Ul = U->template  Give_data<hexahedronEl> ( params_out );
		 k=kf;
                 double Ur = U->template  Give_data<hexahedronEl> ( params_out );
		 
		 data_rectangles[0][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl); 
	     }
	  Update_back(0);
      }
   }
   else if(typeUg==cylinderUgTyp || typeUg==cylinderListUgTyp || typeUg==hexahedronAdaptiveUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();
	  
      double zLeft  = ug3DZ->zLeft(id_block);  
      double zRight = ug3DZ->zRight(id_block);  
	  
      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;
	  
          int kf = find_k(zLeft,zRight,z,Nz);
      
          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
	     U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) { 
                     data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out ); 
	         }
	     Update_back(id_rec);
	  }
          else { //mitte
             int k;
	     double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
	     double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

	     U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
		     k=kf-1;
                     double Ul = U->template  Give_data<hexahedronEl> ( params_out );
		     k=kf;
                     double Ur = U->template  Give_data<hexahedronEl> ( params_out );
		 
		     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);	     
	         }
	      Update_back(id_rec);
	  }
      }
   }
   else if(typeUg==hexahedronListUgTyp) { //adaptiveHex -> adaptiveRec
      HexahedronXYList* hexahedron = (HexahedronXYList*)(ug3Dp);
      int n_x = hexahedron->n_x;
      int n_y = hexahedron->n_y;
      assert((n_x) * (n_y));

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<n_x*n_y;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );
                 }
             Update_back(id_rec);
          }
          else { //mitte
             int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     k=kf-1;
                     double Ul = U->template  Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     double Ur = U->template  Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                 }
              Update_back(id_rec);
          }
      }
   }
   else if(typeUg==hexahedronListTestUgTyp) { //adaptiveHex -> adaptiveRec
      HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)(ug3Dp);
      int n_x = hexahedron->n_x;
      int n_y = hexahedron->n_y;
      assert((n_x) * (n_y));

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<n_x*n_y;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );
                 }
             Update_back(id_rec);
          }
          else { //mitte
             int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     k=kf-1;
                     double Ul = U->template  Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     double Ur = U->template  Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                 }
              Update_back(id_rec);
          }
      }
   }
   else {
     assert(false); 
   }
}



template <class DTyp>
template <class DTyp1, class A>
void Variable2D<DTyp>::interpolateSlizeZ(const Exp_Function1<A, DTyp1, DTyp>&  U, double z) {
   Blockgrid* blockgrid3D    = U.a_.Give_blockgrid();
   Unstructured_grid* ug3Dp  = blockgrid3D->Give_unstructured_grid();
   TypeOfUG typeUg = ug3Dp->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3Dp);

   int numberCorners2D = ug->Give_number_points();

   // Hexaeder -> Rectangle
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int id = ug3DZ->getNummerBlockZ(z);

      int Nx = blockgrid->Give_Nx_rectangle(0);
      int Ny = blockgrid->Give_Ny_rectangle(0);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id);

      double zLeft  = ug3DZ->zLeft(id);                           //ug3Dp->Give_point(0)->Give_coordinate().z;
      double zRight = ug3DZ->zRight(id);                          //ug3Dp->Give_point(4)->Give_coordinate().z;
      int kf = find_k(zLeft,zRight,z,Nz);

      if(kf==0 || kf==Nz) { //links und rechts
         int k=kf;
         U.Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_rectangles[0][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
             }
         Update_back(0);
      }
      else { //mitte
         int k;
         double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
         double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

         U.Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 k=kf-1;
                 DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                 k=kf;
                 DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                 data_rectangles[0][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
             }
          Update_back(0);
      }
   }
   else if(typeUg==cylinderUgTyp || typeUg==hexahedronAdaptiveUgTyp || typeUg==cylinderListUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
                 }
             Update_back(id_rec);
          }
          else { //mitte
             int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     k=kf-1;
                     DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                 }
              Update_back(id_rec);
          }
      }
   }
   else if(typeUg==hexahedronListUgTyp) { //adaptiveHex -> adaptiveRec
     HexahedronXYList* hexahedron = (HexahedronXYList*)(ug3Dp);
     int n_x = hexahedron->n_x;
     int n_y = hexahedron->n_y;
      assert((n_x+1) * (n_y+1));

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<n_x*n_y;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;
          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
                 }
             Update_back(id_rec);
          }
          else { //mitte
             int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     k=kf-1;
                     DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                 }
              Update_back(id_rec);
          }
      }
   }
   else if(typeUg==hexahedronListTestUgTyp) { //adaptiveHex -> adaptiveRec
     HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)(ug3Dp);
     int n_x = hexahedron->n_x;
     int n_y = hexahedron->n_y;
      assert((n_x+1) * (n_y+1));

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<n_x*n_y;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;
          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
                 }
             Update_back(id_rec);
          }
          else { //mitte
             int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     k=kf-1;
                     DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                 }
              Update_back(id_rec);
          }
      }
   }
   else {
     assert(false);
   }
}

template <class DTyp>
template <class A, class B, class C, class D, class DTyp1>
void Variable2D<DTyp>::interpolateSlizeZ(const Exp_Function4<A, B, C, D, DTyp1,DTyp>&  U, double z) {
   Blockgrid* blockgrid3D    = U.a_.Give_blockgrid();
   Unstructured_grid* ug3Dp  = blockgrid3D->Give_unstructured_grid();
   TypeOfUG typeUg = ug3Dp->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3Dp);

   int numberCorners2D = ug->Give_number_points();

   // Hexaeder -> Rectangle
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int id = ug3DZ->getNummerBlockZ(z);

      int Nx = blockgrid->Give_Nx_rectangle(0);
      int Ny = blockgrid->Give_Ny_rectangle(0);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id);

      double zLeft  = ug3DZ->zLeft(id);                           //ug3Dp->Give_point(0)->Give_coordinate().z;
      double zRight = ug3DZ->zRight(id);                          //ug3Dp->Give_point(4)->Give_coordinate().z;
      int kf = find_k(zLeft,zRight,z,Nz);

      if(kf==0 || kf==Nz) { //links und rechts
         int k=kf;
         U.Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_rectangles[0][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
             }
         Update_back(0);
      }
      else { //mitte
         int k;
         double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
         double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

         U.Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 k=kf-1;
                 DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                 k=kf;
                 DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                 data_rectangles[0][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
             }
          Update_back(0);
      }
   }
   else if(typeUg==cylinderUgTyp || typeUg==hexahedronAdaptiveUgTyp || typeUg==cylinderListUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
                 }
             Update_back(id_rec);
          }
          else { //mitte
             int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U.Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     k=kf-1;
                     DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                 }
              Update_back(id_rec);
          }
      }
   }
   else if(typeUg==hexahedronListUgTyp) { //adaptiveHex -> adaptiveRec
     HexahedronXYList* hexahedron = (HexahedronXYList*)(ug3Dp);
     int n_x = hexahedron->n_x;
     int n_y = hexahedron->n_y;

     int id_block = ug3DZ->getNummerBlockZ(z);
     int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
     int num_blocks = ug3DZ->getNummernBloecke();
     double zLeft  = ug3DZ->zLeft(id_block);
     double zRight = ug3DZ->zRight(id_block);

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         int kf = find_k(zLeft,zRight,z,Nz);
         if(kf==0 || kf==Nz) { //links und rechts
            int k=kf;
            U.Update<hexahedronEl>(id);
            for(int j = 0;j <= Ny;++j)
                for(int i = 0;i <= Nx;++i) {
                    data_rectangles[id_rec][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
                }
            Update_back(id_rec);
         }
         else { //mitte
            int k;
            double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
            double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

            U.Update<hexahedronEl>(id);
            for(int j = 0;j <= Ny;++j)
                for(int i = 0;i <= Nx;++i) {
                    k=kf-1;
                    DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                    k=kf;
                    DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                    data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                }
             Update_back(id_rec);
         }
     }
  }
   else if(typeUg==hexahedronListTestUgTyp) { //adaptiveHex -> adaptiveRec
     HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)(ug3Dp);
     int n_x = hexahedron->n_x;
     int n_y = hexahedron->n_y;

     int id_block = ug3DZ->getNummerBlockZ(z);
     int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
     int num_blocks = ug3DZ->getNummernBloecke();
     double zLeft  = ug3DZ->zLeft(id_block);
     double zRight = ug3DZ->zRight(id_block);

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         int kf = find_k(zLeft,zRight,z,Nz);
         if(kf==0 || kf==Nz) { //links und rechts
            int k=kf;
            U.Update<hexahedronEl>(id);
            for(int j = 0;j <= Ny;++j)
                for(int i = 0;i <= Nx;++i) {
                    data_rectangles[id_rec][i+(Nx+1)*j] = U.Give_data<hexahedronEl> ( params_out );
                }
            Update_back(id_rec);
         }
         else { //mitte
            int k;
            double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
            double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

            U.Update<hexahedronEl>(id);
            for(int j = 0;j <= Ny;++j)
                for(int i = 0;i <= Nx;++i) {
                    k=kf-1;
                    DTyp Ul = U.Give_data<hexahedronEl> ( params_out );
                    k=kf;
                    DTyp Ur = U.Give_data<hexahedronEl> ( params_out );

                    data_rectangles[id_rec][i+(Nx+1)*j] = Ul + (Ur-Ul) * (z-zl)/(zr-zl);
                }
             Update_back(id_rec);
         }
     }
  }
   else {
     assert(false);
   }
}


/////////////////////////////////////////////////////////
// 3. Interpolation 3D->2D
// 3.2. mit Iterator
/////////////////////////////////////////////////////////



template <class DTyp>
template <class DTyp1>
void Variable2D<DTyp>::interpolateSlizeZ(const Variable<DTyp1>* U, IteratorZDirection* iterZ) {
   Blockgrid* blockgrid3D  = U->Give_blockgrid();
   Unstructured_grid* ug3D  = blockgrid3D->Give_unstructured_grid();
   TypeOfUG typeUg = ug3D->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3D);
   
   int numberCorners2D = ug->Give_number_points();
//   U->UpdateHexahedra();

   // Rectangle -> Hexaeder
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int Nx = blockgrid->Give_Nx_rectangle(0);
      int Ny = blockgrid->Give_Ny_rectangle(0);
//      U->template  Update<rectangleEl>(0);
      
      int k  = iterZ->get_k_intern();
      int id = iterZ->get_nummerBlock();

      U->template  Update<hexahedronEl>(id);
      for(int j = 0;j <= Ny;++j)
          for(int i = 0;i <= Nx;++i) { 
              data_rectangles[0][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );    
//              data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template Give_data<rectangleEl>(0, i, j, Nx);
	  }
      Update_back(0);
   }
   else if(typeUg==cylinderUgTyp || typeUg == hexahedronAdaptiveUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = iterZ->get_nummerBlock();
      int num_blocks = ug3DZ->getNummernBloecke();
      int k  = iterZ->get_k_intern();
	  
      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          U->template  Update<hexahedronEl>(id);
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i) { 
                  data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );    		
//                  data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template Give_data<rectangleEl>(id_rec, i, j, Nx);
	      }
          Update_back(id_rec);
      }
   }
   else if(typeUg==cylinderListUgTyp) { // List Cylinder -> disc // neu Christoph: 29.10.14
      int id_block = iterZ->get_nummerBlock();
      int num_blocks = ug3DZ->getNummernBloecke();
      int k  = iterZ->get_k_intern();
	  
      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          U->template  Update<hexahedronEl>(id);
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i) { 
                  data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );   		
//                  data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template  Give_data<rectangleEl>(id_rec, i, j, Nx);
	      }
          Update_back(id_rec);
      }
   }   
   else if(typeUg == hexahedronListUgTyp) {
     HexahedronXYList* hexahedron = (HexahedronXYList*)ug;
     int n_x = numberCorners2D==hexahedron->n_x;
     int n_y = numberCorners2D==hexahedron->n_y;
     assert((n_x+1) * (n_y+1));

     int id_block = iterZ->get_nummerBlock();
     int num_blocks = ug3DZ->getNummernBloecke();
     int k  = iterZ->get_k_intern();

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         U->template  Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );   	       
//                 data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template  Give_data<rectangleEl>(id_rec, i, j, Nx);
             }
         Update_back(id_rec);
     }
   }
   else if(typeUg == hexahedronListTestUgTyp) {
     HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)ug;
     int n_x = numberCorners2D==hexahedron->n_x;
     int n_y = numberCorners2D==hexahedron->n_y;
     assert((n_x+1) * (n_y+1));

     int id_block = iterZ->get_nummerBlock();
     int num_blocks = ug3DZ->getNummernBloecke();
     int k  = iterZ->get_k_intern();

     for(int id_rec = 0;id_rec< n_x*n_y;++id_rec) {
         int Nx = blockgrid->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid->Give_Ny_rectangle(id_rec);
         int id = id_rec*num_blocks + id_block;

         U->template  Update<hexahedronEl>(id);	 
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_rectangles[id_rec][i+(Nx+1)*j] = U->template  Give_data<hexahedronEl> ( params_out );   	       
//                 data_hexahedra[id][(i)+(Nx+1)*((j)+(Ny+1)*(k))] = U->template  Give_data<rectangleEl>(id_rec, i, j, Nx);
             }
         Update_back(id_rec);
     }
   }
   else {
     std::cout << " noch nicht implementiert!! " << std::endl;
     assert(false); 
   }
}




/////////////////////////////////////////////////////////
// 4. Intergrate 3D->2D
/////////////////////////////////////////////////////////



template <class DTyp>
void Variable2D<DTyp>::integrateZ(const Variable<DTyp>* U) {
   Blockgrid* blockgrid3D    = U->Give_blockgrid();
   Unstructured_grid* ug3Dp  = blockgrid3D->Give_unstructured_grid();
   TypeOfUG typeUg = ug3Dp->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3Dp);  
   
   int numberCorners2D = ug->Give_number_points();

   // Hexaeder -> Rectangle 
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
        
      int Nx = blockgrid->Give_Nx_rectangle(0);
      int Ny = blockgrid->Give_Ny_rectangle(0);

      //start NULL setzen
      for(int j = 0;j <= Ny;++j)
          for(int i = 0;i <= Nx;++i)
              data_rectangles[0][i+(Nx+1)*j] = 0.0;
            
      for(int id = 0;id<ug3DZ->getNummernBloecke();++id) {
          int Nz = blockgrid3D->Give_Nz_hexahedron(id);
	 
         // aufaddieren in Zellen    
         double hz = 0.0;
         for(int ksp = 0;ksp < Nz;++ksp) {
	     hz = blockgrid3D->Give_coord_hexahedron( id,0,0,ksp+1).z -
	          blockgrid3D->Give_coord_hexahedron( id,0,0,ksp  ).z;
	     U->template  Update<hexahedronEl>(id);
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) { 
		        int k = ksp;
		        double mittel   = U->template Give_data<hexahedronEl>(params_out);
		        k = ksp+1;
		        mittel = mittel + U->template Give_data<hexahedronEl>(params_out);
                        data_rectangles[0][i+(Nx+1)*j] =
                           data_rectangles[0][i+(Nx+1)*j] + mittel * 0.5 * hz;
	         }
	 }
      }
      Update_back(0);      
   }
   else if(typeUg==cylinderUgTyp || typeUg == cylinderListUgTyp || typeUg==hexahedronAdaptiveUgTyp) { // Cylinder -> disc
     assert(numberCorners2D%8 == 0);
      
      int num_blocks = ug3DZ->getNummernBloecke();

      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);

    	  //start NULL setzen
	  for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i)
                  data_rectangles[id_rec][i+(Nx+1)*j] = 0.0;

	  // aufaddieren in Zellen
	  for(int id_block = 0; id_block < num_blocks; ++id_block) {
              int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
	      int id = id_rec*num_blocks + id_block;
			
	      double hz = 0.0;
	      for(int ksp = 0;ksp < Nz;++ksp) {
	          hz = blockgrid3D->Give_coord_hexahedron( id,0,0,ksp+1).z -
	               blockgrid3D->Give_coord_hexahedron( id,0,0,ksp  ).z;
	          U->template  Update<hexahedronEl>(id);
                  for(int j = 0;j <= Ny;++j)
                  for(int i = 0;i <= Nx;++i) { 
		      int k = ksp;
		      double mittel   = U->template  Give_data<hexahedronEl>(params_out);
		      k = ksp+1;
		      mittel = mittel + U->template  Give_data<hexahedronEl>(params_out);
                      data_rectangles[id_rec][i+(Nx+1)*j] =
                           data_rectangles[id_rec][i+(Nx+1)*j] + mittel * 0.5 * hz;
		  }
	      }
	  }
	  Update_back(id_rec);
      }
   }
   else if(typeUg==hexahedronListUgTyp) { //adaptiveHex -> adaptiveRec
      HexahedronXYList* hexahedron = (HexahedronXYList*)ug;
      int n_x = numberCorners2D==hexahedron->n_x;
      int n_y = numberCorners2D==hexahedron->n_y;
      assert((n_x+1) * (n_y+1));
      int num_blocks = ug3DZ->getNummernBloecke();

      for(int id_rec = 0;id_rec<n_x*n_y;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);

          //start NULL setzen
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i)
                  data_rectangles[id_rec][i+(Nx+1)*j] = 0.0;

          // aufaddieren in Zellen
          for(int id_block = 0; id_block < num_blocks; ++id_block) {
              int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
              int id = id_rec*num_blocks + id_block;

              double hz = 0.0;
              for(int ksp = 0;ksp < Nz;++ksp) {
                  hz = blockgrid3D->Give_coord_hexahedron( id,0,0,ksp+1).z -
                       blockgrid3D->Give_coord_hexahedron( id,0,0,ksp  ).z;
                  U->template  Update<hexahedronEl>(id);
                  for(int j = 0;j <= Ny;++j)
                  for(int i = 0;i <= Nx;++i) {
                      int k = ksp;
                      double mittel   = U->template  Give_data<hexahedronEl>(params_out);
                      k = ksp+1;
                      mittel = mittel + U->template  Give_data<hexahedronEl>(params_out);
                      data_rectangles[id_rec][i+(Nx+1)*j] =
                           data_rectangles[id_rec][i+(Nx+1)*j] + mittel * 0.5 * hz;
                  }
              }
          }
          Update_back(id_rec);
      }
   }
   else if(typeUg==hexahedronListTestUgTyp) { //adaptiveHex -> adaptiveRec
      HexahedronXYListTest* hexahedron = (HexahedronXYListTest*)ug;
      int n_x = numberCorners2D==hexahedron->n_x;
      int n_y = numberCorners2D==hexahedron->n_y;
      assert((n_x+1) * (n_y+1));
      int num_blocks = ug3DZ->getNummernBloecke();

      for(int id_rec = 0;id_rec<n_x*n_y;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);

          //start NULL setzen
          for(int j = 0;j <= Ny;++j)
              for(int i = 0;i <= Nx;++i)
                  data_rectangles[id_rec][i+(Nx+1)*j] = 0.0;

          // aufaddieren in Zellen
          for(int id_block = 0; id_block < num_blocks; ++id_block) {
              int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
              int id = id_rec*num_blocks + id_block;

              double hz = 0.0;
              for(int ksp = 0;ksp < Nz;++ksp) {
                  hz = blockgrid3D->Give_coord_hexahedron( id,0,0,ksp+1).z -
                       blockgrid3D->Give_coord_hexahedron( id,0,0,ksp  ).z;
                  U->template  Update<hexahedronEl>(id);
                  for(int j = 0;j <= Ny;++j)
                  for(int i = 0;i <= Nx;++i) {
                      int k = ksp;
                      double mittel   = U->template  Give_data<hexahedronEl>(params_out);
                      k = ksp+1;
                      mittel = mittel + U->template  Give_data<hexahedronEl>(params_out);
                      data_rectangles[id_rec][i+(Nx+1)*j] =
                           data_rectangles[id_rec][i+(Nx+1)*j] + mittel * 0.5 * hz;
                  }
              }
          }
          Update_back(id_rec);
      }
   }
   else {
     std::cout << "typeUg = " << typeUg << " has no interpolation defined!" << std::endl;
     assert(false); 
   }
}


#endif // INTERPOLFROM3D_H
