/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include <sstream>
#include <stdexcept>

#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "elements2D.h"
#include "parti2D.h"
#include "ug2D.h"
#include "blockgrid2D.h"

#define HWS rec->Give_coord(SWdir2D)
#define HES rec->Give_coord(SEdir2D)
#define HWN rec->Give_coord(NWdir2D)
#define HEN rec->Give_coord(NEdir2D)

int Blockgrid2D::id_count_grid = 0;

Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_) {
  strechDomainX =  strechDomainY = 1.0;
  id_of_grid = id_count_grid; ++id_count_grid;  
  ug = ug_;
  number_points.resize(ug->degree_of_freedom());
  for ( int i=0;i<ug->degree_of_freedom();++i ) {
    number_points.at(i) = 10;
  }
  variable_set = false;
}

Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, int N) {
  strechDomainX =  strechDomainY = 1.0;  
  id_of_grid = id_count_grid; ++id_count_grid;  
  ug = ug_;
  number_points.resize(ug->degree_of_freedom());
  for ( int i=0;i<ug->degree_of_freedom();++i ) {
    number_points.at(i) = N;
  }
  variable_set = false;
}

Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, int Nx, int Ny) {
  strechDomainX =  strechDomainY = 1.0;
  id_of_grid = id_count_grid; ++id_count_grid;  
  ug = ug_;
  if ( ug->degree_of_freedom() <2 )
    cout << " error in Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, int Nx, int Ny) " << std::endl;
  number_points.resize(ug->degree_of_freedom());

  number_points[0] = Nx;
  for ( int i = 1; i < ug->degree_of_freedom();++i) {
    number_points.at(i) = Ny;
  }

  variable_set = false;
}


Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, const std::vector<int>& N) {
  strechDomainX =  strechDomainY = 1.0;  
  id_of_grid = id_count_grid; ++id_count_grid;  
  int size = N.size();
  ug = ug_;
  if ( ug->degree_of_freedom() != size)
    cout << " error in Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, vector<int>& Nx,  vector<int>&Ny) "
    << endl;

  number_points.resize(size);
  for(int i = 0; i < size; ++i)
  {
    number_points[i] = N[i];
  }
  variable_set = false;
}

Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, const std::vector<int>& Nx,  const std::vector<int>&Ny) {
  strechDomainX =  strechDomainY = 1.0;  
  id_of_grid = id_count_grid; ++id_count_grid;  
  int size = Nx.size() + Ny.size();
  ug = ug_;
  if ( ug->degree_of_freedom() != size)
    cout << " error in Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_, vector<int>& Nx,  vector<int>&Ny) "
    << endl;

  number_points.resize(size);
  number_points[0] = Nx[0];
  number_points[1] = Ny[0];
  for(int i = 1; i < Nx.size(); ++i)
  {
    number_points[i+1] = Nx[i];
  }
  for(int i = 1; i < Ny.size(); ++i)
  {
    number_points[Nx.size() + i] = Ny[i];
  }
  variable_set = false;
}

Blockgrid2D::Blockgrid2D ( Unstructured2DGrid *ug_, int Nx, int Ny, int Nr) {
  strechDomainX =  strechDomainY = 1.0; 
  id_of_grid = id_count_grid; ++id_count_grid;  
  ug = ug_;
  if ( ug->degree_of_freedom() !=3 )
    cout << " error in Blockgrid2D::Blockgrid2D(Unstructured2DGrid *ug_,int Nx,int Ny, int Nr) "
    << endl;

  number_points.resize(3);

  number_points[0] = Nx;
  number_points[1] = Ny;
  number_points[2] = Nr;

  variable_set = false;
}

Blockgrid2D::Blockgrid2D(const Blockgrid2D& copy) {
  strechDomainX = copy.strechDomainX;    strechDomainY = copy.strechDomainY;
  
  id_of_grid = id_count_grid; ++id_count_grid;  
  number_points = copy.number_points;
  variable_set = copy.variable_set;
  ug = copy.ug;
}

Blockgrid2D::Blockgrid2D(const Blockgrid2D& copy, const std::vector<int>& number_points_) {
  strechDomainX = copy.strechDomainX;    strechDomainY = copy.strechDomainY;  
  id_of_grid = id_count_grid; ++id_count_grid;  
  assert(copy.number_points.size() == number_points_.size());
  ug = copy.ug;
  number_points = number_points_;
  variable_set = false;
}

Blockgrid2D::~Blockgrid2D() {
}

const std::vector<int>& Blockgrid2D::Give_N() const {
  return number_points;
}

Blockgrid2D Blockgrid2D::CopyRefinement(std::vector<int>& N) const {
  Blockgrid2D copy(*this);
  copy.number_points = N;
  return copy;
}

void Blockgrid2D::Refine(std::vector<int>& N) {
  number_points = N;
}

int Blockgrid2D::Give_Nx_rectangle (int id) const {
  return number_points.at(ug->Give_edge((ug->Give_rectangle(id)->
                                         Give_id_edge(Sdir2D)))->Give_color_of_edge());
}

int Blockgrid2D::Give_Ny_rectangle (int id) const {
  return number_points.at(ug->Give_edge((ug->Give_rectangle(id)->
                                         Give_id_edge(Wdir2D)))->Give_color_of_edge());
}

int Blockgrid2D::Give_N_total_rectangle(int id) const {
  return
    ( Give_Nx_rectangle(id) +1 ) *
    ( Give_Ny_rectangle(id) +1 );
}




int Blockgrid2D::Give_Nx_edge(int id) const {
  return number_points.at(ug->Give_edge(id)->Give_color_of_edge());
}

int Blockgrid2D::Give_N_color(int color) const {
  assert(color<ug->degree_of_freedom());
  assert(color>=0);

  return number_points.at(color);
}

inline double change(bool ori, double q) {
  if(ori) return q;
  return 1.0-q;
}

/*
inline double find_p ( Edges_cell ed, double eta, double xi,double phi )
{
  if ( ed<=NTed ) return eta;
  if ( ed<=ETed ) return xi;
  if ( developer_version )
    if ( ed>NEed ) cout << " error in find_p!" << endl;
  return phi;
}
*/


D2vector Blockgrid2D::Give_coord_rectangle(int id, int i, int j) const  {
  double eta, xi;
  double phiN, phiS, phiE, phiW;

  D2vector PSW, PSE, PNW, PNE, PW, PE, PN, PS;
  RectangleEl* quad;
  EdgeEl* edge;

  quad = ug->Give_rectangle(id);

  eta = ( double ) i/ ( double ) Give_Nx_rectangle( id );
  xi = ( double ) j/ ( double ) Give_Ny_rectangle( id );

  if(quad->getOrientationEdge(Wdir2D)) phiW = 1.0 - xi;
  else                                 phiW = xi;

  if(quad->getOrientationEdge(Edir2D)) phiE = 1.0 - xi;
  else                                 phiE = xi;

  if(quad->getOrientationEdge(Ndir2D)) phiN = 1.0 - eta;
  else                                 phiN = eta;

  if(quad->getOrientationEdge(Sdir2D)) phiS = 1.0 - eta;
  else                                 phiS = eta;
  
  
  // @todo : so sollte es sein
  PSW  = quad->Give_coord ( SWdir2D );
  PSE  = quad->Give_coord ( SEdir2D );
  PNW  = quad->Give_coord ( NWdir2D );
  PNE  = quad->Give_coord ( NEdir2D );

  
  
  
  // aber ist noch so
  //PSW  = ug->Give_point(quad->Give_id_corner(SWdir2D))->Give_coordinate();
  //PSE  = ug->Give_point(quad->Give_id_corner(SEdir2D))->Give_coordinate();
  //PNW  = ug->Give_point(quad->Give_id_corner(NWdir2D))->Give_coordinate();
  //PNE  = ug->Give_point(quad->Give_id_corner(NEdir2D))->Give_coordinate();
  
/*  fuer was ist das gut????
  assert(quad->Give_id_corner ( SWdir2D ) < quad->Give_id_corner ( SEdir2D )  &&
         quad->Give_id_corner ( SEdir2D ) < quad->Give_id_corner ( NWdir2D )  &&
         quad->Give_id_corner ( SWdir2D ) < quad->Give_id_corner ( NEdir2D )     );
*/

  edge = ug->Give_edge(quad->Give_id_edge(Wdir2D));
  PW = edge->transformer->transform( phiW );
  PW = PW + PSW + ( PNW-PSW ) * xi;

  edge = ug->Give_edge(quad->Give_id_edge ( Sdir2D ));
  PS = edge->transformer->transform( phiS );
  PS = PS + PSW + ( PSE-PSW ) * eta;

  edge = ug->Give_edge( quad->Give_id_edge ( Edir2D ));
  PE = edge->transformer->transform( phiE );
  PE = PE + PSE + ( PNE-PSE ) * xi;

  edge = ug->Give_edge ( quad->Give_id_edge ( Ndir2D ) );
  PN = edge->transformer->transform( phiN );
  PN = PN + PNW + ( PNE-PNW ) * eta;

  return
    D2vector(strechDomainX,strechDomainY) * (
    PW + eta* ( PE-PW ) +
    PS + xi* ( PN-PS )
    - ( PSW +
        ( PSE - PSW ) * eta +
        ( PNW - PSW ) * xi +
        ( PNE - PSE - PNW + PSW ) * eta * xi )
    );
}


D2vector Blockgrid2D::Give_coord_edge ( int id, int i ) const {
  double eta;
  D2vector Psi, PL, PR;
  EdgeEl* edge;

  edge = ug->Give_edge ( id );

  eta = ( double ) i/ ( double ) Give_Nx_edge ( id );

  Psi = ug->Give_edge ( id )->transformer->transform( eta );

  PL  = edge->Give_coord ( Wd );
  PR  = edge->Give_coord ( Ed );
  /*
  if(edge->Give_id_corner_W() < edge->Give_id_corner_E()) {
     PL  = edge->Give_coord ( Wd );
     PR  = edge->Give_coord ( Ed );
  }
  else {
     PR  = edge->Give_coord ( Wd );
     PL  = edge->Give_coord ( Ed );
  }
*/

  return  D2vector(strechDomainX,strechDomainY) * (Psi + PL + ( PR-PL ) * eta);
}

D2vector Blockgrid2D::Give_coord_point ( int id_point ) const {
  return  D2vector(strechDomainX,strechDomainY) * (ug->Give_point(id_point)->Give_coordinate());
}



int Blockgrid2D::Total_number_of_points() {
  using namespace std;
  
  int num_points, num_points_rec, num_points_edges;

  int Nx, Ny;

  num_points = num_points_rec = num_points_edges = 0;
  // rectangles
  for ( int id=0;id<ug->Give_number_rectangle();++id )
  {
    Nx = Give_Nx_rectangle ( id );
    Ny = Give_Ny_rectangle ( id );

    num_points_rec += ( Nx-1 ) * ( Ny-1 );
  }

  // edges
  for ( int id=0;id<ug->Give_number_edges();++id )
  {
    Nx = Give_Nx_edge ( id );

    num_points_edges += ( Nx-1 );
  }

  
  // points
  num_points = ug->Give_number_points();
  // alltogether
  num_points = num_points + num_points_rec + num_points_edges;

  return num_points;
}


