/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "elements2D.h"
#include "parti2D.h"
#include "ug2D.h"
#include "blockgrid2D.h"
#include "examples_ug2D.h"

#define Ind_SectionedRectangle_point(i,k) ((k)*(v_x.size())+(i))
#define Ind_SectionedRectangle_rectangle(i,k) ((k)*(v_x.size()-1)+(i))
#define Ind_SectionedRectangle_SWdir(i,k) ((k)*(v_x.size())+(i))
#define Ind_SectionedRectangle_SEdir(i,k) ((k)*(v_x.size())+(i+1))
#define Ind_SectionedRectangle_NWdir(i,k) ((k+1)*(v_x.size())+(i))
#define Ind_SectionedRectangle_NEdir(i,k) ((k+1)*(v_x.size())+(i+1))
#define Coord_SectionedRectangle_point(i,k) D2vector(v_x[i], v_y[k])
#define Coord_SectionedRectangle_rectangle(i,k) D2vector(v_x[i], v_y[k])

#ifdef WIN32
  inline double round(double x) { return (floor(x + 0.5)); }
#endif
	

	
SectionedRectangle::SectionedRectangle(std::vector<double>& v_x, std::vector<double>& v_y) {
  assert((v_x.size()-1)*(v_y.size()-1));
  
  Set_number_points ( v_x.size() * v_y.size() );

  for(int k = 0; k < v_y.size(); ++k)
  {
	  for(int i = 0; i < v_x.size(); ++i)
	  {
		  Set_coordinate_point(Ind_SectionedRectangle_point(i,k), Coord_SectionedRectangle_point(i,k));
	  }
  }

  Set_number_rectanglEs ( (v_x.size()-1)*(v_y.size()-1) );
  for(int k = 0; k < v_y.size()-1; ++k)
  {
	  for(int i = 0; i < v_x.size()-1; ++i)
	  {
		  assert(v_x[i] < v_x[i+1]);
		  assert(v_y[k] < v_y[k+1]);
		  Set_rectangle(Ind_SectionedRectangle_rectangle(i,k), Ind_SectionedRectangle_SWdir(i,k), Ind_SectionedRectangle_SEdir(i,k),
				  Ind_SectionedRectangle_NWdir(i,k), Ind_SectionedRectangle_NEdir(i,k));
	  }
  }

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();
}

TestRectangle::TestRectangle(double l_x, double l_y,
    double r_x, double r_y) {
  Set_number_points ( 4*4 );

  for(int k = 0; k < 4; ++k)
  {
          for(int i = 0; i < 4; ++i)
          {
              double x;
              double y;

              if(i == 0)
                x = l_x;
              else if(i == 1)
                x = -0.2*fabs(1.5-k)*(r_y-l_y);
              else if(i == 2)
                x =  0.2*fabs(1.5-k)*(r_y-l_y);
              else if(i == 3)
                x = r_x;

              if(k == 0)
                y = l_y;
              else if(k == 1)
                y = -0.2*fabs(1.5-i)*(r_x-l_x);
              else if(k == 2)
                y =  0.2*fabs(1.5-i)*(r_x-l_x);
              else if(k == 3)
                y = r_y;
              Set_coordinate_point(4*k+i, D2vector(x, y));
          }
  }

  Set_number_rectanglEs ( 3*3 );
  for(int k = 0; k < 3; ++k)
  {
          for(int i = 0; i < 3; ++i)
          {
                  Set_rectangle(3*k+i, 4*(k)+i, 4*(k)+i+1,
                      4*(k+1)+i, 4*(k+1)+i+1);
          }
  }

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();
}


Rectangle::Rectangle(double rx_, double ry_) : lx(0.0), rx(rx_), ly(0.0), ry(ry_) {
  Set_number_points ( 4 ); // 4 an den Ecken

  Set_coordinate_point ( SWdir2D, D2vector ( lx,ly) );
  Set_coordinate_point ( SEdir2D, D2vector ( rx,ly) );
  Set_coordinate_point ( NWdir2D, D2vector ( lx,ry) );
  Set_coordinate_point ( NEdir2D, D2vector ( rx,ry) );

  Set_number_rectanglEs ( 1 );
  Set_rectangle ( 0, SWdir2D, SEdir2D, NWdir2D, NEdir2D );

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();
}


Rectangle::Rectangle(double lx_, double ly_,
                     double rx_, double ry_): lx(lx_), ly(ly_), rx(rx_), ry(ry_) {
  Set_number_points ( 4 ); // 4 an den Ecken
  Set_coordinate_point ( SWdir2D, D2vector ( lx,ly) );
  Set_coordinate_point ( SEdir2D, D2vector ( rx,ly) );
  Set_coordinate_point ( NWdir2D, D2vector ( lx,ry) );
  Set_coordinate_point ( NEdir2D, D2vector ( rx,ry) );

  Set_number_rectanglEs ( 1 );
  Set_rectangle ( 0, SWdir2D, SEdir2D, NWdir2D, NEdir2D );

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();
}



Rechteck::Rechteck(double rx_, double ry_) : lx(0.0), rx(rx_), ly(0.0), ry(ry_) {
	Set_number_points(4); // 4 an den Ecken

	Set_coordinate_point(SWdir2D, D2vector(lx, ly));
	Set_coordinate_point(SEdir2D, D2vector(rx, ly));
	Set_coordinate_point(NWdir2D, D2vector(lx, ry));
	Set_coordinate_point(NEdir2D, D2vector(rx, ry));

	Set_number_rectanglEs(1);
	Set_rectangle(0, SWdir2D, SEdir2D, NWdir2D, NEdir2D);

	construction_rectangle_points_done();

	// no transformations of the edges and faces
	construction_done();
}


Rechteck::Rechteck(double lx_, double ly_,
	double rx_, double ry_) : lx(lx_), ly(ly_), rx(rx_), ry(ry_) {
	Set_number_points(4); // 4 an den Ecken
	Set_coordinate_point(SWdir2D, D2vector(lx, ly));
	Set_coordinate_point(SEdir2D, D2vector(rx, ly));
	Set_coordinate_point(NWdir2D, D2vector(lx, ry));
	Set_coordinate_point(NEdir2D, D2vector(rx, ry));

	Set_number_rectanglEs(1);
	Set_rectangle(0, SWdir2D, SEdir2D, NWdir2D, NEdir2D);

	construction_rectangle_points_done();

	// no transformations of the edges and faces
	construction_done();
}





////////////////////////////////////////////////////////////////////
// disc
////////////////////////////////////////////////////////////////////

// Fuer DEformation der 4 aaeusseren Kanten
//D2vector transform_disc_NE ( double t, double* pointer_global_data) {
//  double R_global_data = pointer_global_data[0];
//  return R_global_data*D2vector ( cos ( t*0.5*M_PI )- ( 1-t ), sin ( t*0.5*M_PI )-t);
//}
//
//D2vector transform_disc_NW ( double t, double* pointer_global_data ) {
//  double R_global_data = pointer_global_data[0];
//  return R_global_data*D2vector ( -sin ( t*0.5*M_PI ) +t, cos ( t*0.5*M_PI )- ( 1-t ));
//}
//
//D2vector transform_disc_NW_inner ( double t, double* pointer_global_data ) {
//  //return D2vector(0.0,0.0); //Keine ANEdirerung
//  double r = pointer_global_data[1];
//  double rho = pointer_global_data[2];
//  return D2vector (rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t, rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t));
//}
//
//D2vector transform_disc_SW ( double t, double* pointer_global_data ) {
//  double R_global_data = pointer_global_data[0];
//  return R_global_data*D2vector ( -cos ( t*0.5*M_PI ) + ( 1-t ), -sin ( t*0.5*M_PI ) +t);
//}
//
//D2vector transform_disc_SE ( double t, double* pointer_global_data ) {
//  double R_global_data = pointer_global_data[0];
//  return R_global_data*D2vector ( cos ( t*0.5*M_PI )- ( 1-t ), -sin ( t*0.5*M_PI ) +t);
//}
//
//// Fuer DEformation der 4 inneren Kanten t=0,...,1
//D2vector transform_disc_NE_inner ( double t, double* pointer_global_data) {
//  //return D2vector(0.0,0.0); //Keine ANEdirerung
//  double r = pointer_global_data[1];
//  double rho = pointer_global_data[2];
//  return D2vector (rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t), rho*(sin(M_PI/6.0*(1.0+t))-0.5)-r*t);
//}
//
//D2vector transform_disc_SW_inner ( double t, double* pointer_global_data ) {
//  //return D2vector(0.0,0.0); //Keine ANEdirerung
//  double r = pointer_global_data[1];
//  double rho = pointer_global_data[2];
//  return D2vector (-rho*(cos(M_PI/6.0*(1.0+t))-0.5)+r*(1.0-t), rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t);
//}
//
//D2vector transform_disc_SE_inner ( double t, double* pointer_global_data ) {
//  //return D2vector(0.0,0.0); //Keine ANEdirerung
//  double r = pointer_global_data[1];
//  double rho = pointer_global_data[2];
//  return D2vector (rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t), rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t);
//}

bool SectionedDisc::inDomain(const D2vector& point_coord) {
  return D2VectorNorm(point_coord) <= MaximumX();
}


SectionedDisc::SectionedDisc(std::vector<double>& radius_vec, bool inner_grid_arched) {
  double offsetX = 0.0;
  double offsetY = 0.0;

  int num_points_per_line = radius_vec.size();

  outerR = radius_vec[num_points_per_line-1];

  // new version
  pointer_global_data = new double[3];
  pointer_global_data[1] = radius_vec[0];
  pointer_global_data[0] = radius_vec[num_points_per_line-1];
  pointer_global_data[2] = 2.0*radius_vec[0]*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));


  //1.Schritt Punkte setzen
  //------------------------
  Set_number_points( 4*num_points_per_line );
  for(int i = 0; i < num_points_per_line; ++i)
    Set_coordinate_point(                      i, D2vector ( offsetX+radius_vec[i],offsetY ) );

  for(int i = 0; i < num_points_per_line; ++i)
    Set_coordinate_point(  num_points_per_line+i, D2vector ( offsetX,offsetY+radius_vec[i] ) );

  for(int i = 0; i < num_points_per_line; ++i)
    Set_coordinate_point(2*num_points_per_line+i, D2vector ( offsetX-radius_vec[i],offsetY ) );

  for(int i = 0; i < num_points_per_line; ++i)
    Set_coordinate_point(3*num_points_per_line+i, D2vector ( offsetX,offsetY-radius_vec[i] ) );

  //2.Schritt Rechtecke setzen
  //------------------------
  Set_number_rectanglEs(1+4*(radius_vec.size()-1));
  Set_rectangle( 0, 0, radius_vec.size(), 3*radius_vec.size(), 2*radius_vec.size());

  for(int i = 0; i < num_points_per_line-1; ++i) {
    for(int j = 0; j < 4; ++j) {
      int myIndex = 1+4*i+j;
      Set_rectangle(myIndex,
          num_points_per_line*j+i,
          num_points_per_line*((j+1)%4)+i,
          num_points_per_line*j+i+1,
          num_points_per_line*((j+1)%4)+i+1);
    }
  }


  construction_rectangle_points_done();
  transform_disc_NE = new NETransform2D[num_points_per_line];
  transform_disc_NW = new NWTransform2D[num_points_per_line];
  transform_disc_SE = new SETransform2D[num_points_per_line];
  transform_disc_SW = new SWTransform2D[num_points_per_line];

  //3.Schritt Kanten deformieren
  //------------------------------
  //3.1 Kanten aussen

  if (inner_grid_arched == true) {
    double R_innen = 2.0*radius_vec[0]*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
    transform_disc_NE[0].setRadius(radius_vec[0],R_innen);
    transform_disc_NW[0].setRadius(radius_vec[0],R_innen);
    transform_disc_SE[0].setRadius(radius_vec[0],R_innen);
    transform_disc_SW[0].setRadius(radius_vec[0],R_innen);

    Set_transformation_edge ( 0, num_points_per_line, transform_disc_NE );
    Set_transformation_edge ( num_points_per_line, 2*num_points_per_line, transform_disc_NW );
    Set_transformation_edge ( 2*num_points_per_line, 3*num_points_per_line, transform_disc_SW );
    Set_transformation_edge ( 0, 3*num_points_per_line, transform_disc_SE );
  }

  for (int i = 1; i < num_points_per_line; ++i) {
    transform_disc_NE[i].setRadius(radius_vec[i]);
    transform_disc_NW[i].setRadius(radius_vec[i]);
    transform_disc_SE[i].setRadius(radius_vec[i]);
    transform_disc_SW[i].setRadius(radius_vec[i]);

    Set_transformation_edge(i, num_points_per_line + i, transform_disc_NE + i);
    Set_transformation_edge(num_points_per_line + i, 2 * num_points_per_line + i, transform_disc_NW + i);
    Set_transformation_edge(2 * num_points_per_line + i, 3 * num_points_per_line + i, transform_disc_SW + i);
    Set_transformation_edge(i, 3 * num_points_per_line + i, transform_disc_SE + i);
  }

  construction_done();
}

SectionedDisc::~SectionedDisc(){
  delete transform_disc_NE;
  delete transform_disc_NW;
  delete transform_disc_SE;
  delete transform_disc_SW;
}

bool Disc::inDomain(const D2vector& point_coord) {
  return D2VectorNorm(point_coord) <= MaximumX();
}

Disc::Disc ( double R, double r, bool inner_grid_arched_) : Radius(R), radius(r), inner_grid_arched(inner_grid_arched_) {
  double offsetX = 0.0;
  double offsetY = 0.0;

  // new version
  pointer_global_data = new double[3];
  pointer_global_data[0] = R;
  pointer_global_data[1] = r;
  
  //Thomas: Diesen Wert aus anderen daten berechnen
  double R_innen = 2.0*r*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));//10.0*R;
  pointer_global_data[2] = R_innen;
  //double l_y = 2.0*r/(sqrt(6.0)-sqrt(2.0));
  //pointer_global_data[3] = l_y;

  //TEST:
  double rho_test = 2.0*r*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  //double l_y = 2.0*R_global_data/(sqrt(6.0)-sqrt(2.0));
  double l_y_test = 2.0*r/(sqrt(6.0)-sqrt(2.0));
  
  ////int num_blocks = Lz.size() -1;
  ////assert(num_blocks>0);
  ////int num_blocks_total = 5* ( Lz.size()-1 );
  int num_points_per_face = 8;
  ////int numPoints = num_points_per_face*(num_blocks+1) ;

  //1.Schritt Punkte setzen
  //------------------------
  Set_number_points( 8 );


  Set_coordinate_point(0, D2vector ( offsetX+r,offsetY ) );
  Set_coordinate_point(1, D2vector ( offsetX+R,offsetY ) );

  Set_coordinate_point(2, D2vector ( offsetX,offsetY+r ) );
  Set_coordinate_point(3, D2vector ( offsetX,offsetY+R ) );

  Set_coordinate_point(4, D2vector ( offsetX-r,offsetY ) );
  Set_coordinate_point(5, D2vector ( offsetX-R,offsetY ) );

  Set_coordinate_point(6, D2vector ( offsetX,offsetY-r ) );
  Set_coordinate_point(7, D2vector ( offsetX,offsetY-R ) );

  //2.Schritt Rechtecke setzen
  //------------------------
  Set_number_rectanglEs(5);
  Set_rectangle( 0, 0, 2, 6, 4);

  for(int j = 0; j < num_points_per_face; j=j+2 ) {
      int myIndex = j/2+1;
      Set_rectangle(myIndex,
                     j+0,
                     j+1,
                     ((j+2) % num_points_per_face),
                     ((j+3) % num_points_per_face));
  }

  
  construction_rectangle_points_done();
  

  //3.Schritt Kanten deformieren
  //------------------------------
  //3.1 Kanten aussen
  transform_disc_NE.setRadius(R);
  Set_transformation_edge ( 1, 3, &transform_disc_NE );
  transform_disc_NW.setRadius(R);
  Set_transformation_edge ( 3, 5, &transform_disc_NW );
  transform_disc_SE.setRadius(R);
  Set_transformation_edge ( 5, 7, &transform_disc_SW );
  transform_disc_SW.setRadius(R);
  Set_transformation_edge ( 1, 7, &transform_disc_SE );

  if (inner_grid_arched == true) {
    //3.2 Kanten innen
    transform_disc_NE_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 0, 2, &transform_disc_NE_inner );
    transform_disc_NW_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 2, 4, &transform_disc_NW_inner );
    transform_disc_SE_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 4, 6, &transform_disc_SW_inner );
    transform_disc_SW_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 0, 6, &transform_disc_SE_inner );
  }
 
  
  construction_done();
}

void Disc::inverseTransformation(const D2vector& point_coord, double& i, double& j) const {
  i = 0.5;
  j = 0.0;
  int n = 0;

  while (n<5) {
    const double gx= (sqrt(0.5)*Radius*i + (1-i)*sqrt(0.5)*radius)*j-point_coord.x;
    const double gy= (i*Radius*cos(M_PI/4*j) + (1-i)*sqrt(0.5)*radius)-point_coord.y;
    const double a = (sqrt(2.0)*Radius*i)/2.0 - sqrt(0.5)*radius*(i - 1);
    const double b = -j*(sqrt(0.5)*radius - (sqrt(2.0)*Radius)/2.0);
    const double c = -(M_PI*Radius*i*sin((M_PI*j)/4))/4;
    const double d = Radius*cos((M_PI*j)/4.0) - sqrt(0.5)*radius;

    j = j - (d*gx-b*gy)/(a*d-b*c);
    i = i + (c*gx-a*gy)/(a*d-b*c);

    n=n+1;
  }
  if(i>= 1.0) i=1.0;
}

void Disc::Test() const {
  std::cout << "Disc\n";
}

void Rectangle::Test() const {
  std::cout << "Rectangle\n";
}

void Rechteck::Test() const {
	std::cout << "Rechteck\n";
}


void Disc::getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const {
  const double r_Radius = 2.0*radius*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  const double angle = fmod(0.5*M_PI + atan(point_coord.y / point_coord.x),0.5*M_PI);
  const double a = r_Radius - radius;
  const double shift = sqrt(Radius*Radius - radius*radius);
  const double phi = M_PI/2 - atan(shift/radius);
  const double bend = Radius*cos(angle)-shift + radius;

  
  if((fabs(point_coord.x) + fabs(point_coord.y))<radius) {
    id=0;
    const D2matrix rotMat(1.,1.,-1.,1.);
    D2vector rotated = D2vector(0.5,0.5)+rotMat*(point_coord/(2*radius));
    i = round(blockgrid.Give_Nx_rectangle(id)*rotated.y);
    j = round(blockgrid.Give_Ny_rectangle(id)*(1.-rotated.x));
  } else if( point_coord.x >= 0. && point_coord.y >= 0. ) {
    id = 1;
    const D2matrix rotMat(sqrt(0.5),-sqrt(0.5),sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x >= 0 && point_coord.y < 0 ) {
    id = 4;
    const D2matrix rotMat(-sqrt(0.5),-sqrt(0.5),sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x < 0 && point_coord.y < 0 ) {
    id = 3;
    const D2matrix rotMat(-sqrt(0.5),sqrt(0.5),-sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else {
    id = 2;
    const D2matrix rotMat(sqrt(0.5),sqrt(0.5),-sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  }
}

bool Disc::getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const {
  const double r_Radius = 2.0*radius*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  const double angle = fmod(0.5*M_PI + atan(point_coord.y / point_coord.x),0.5*M_PI);
  const double a = r_Radius - radius;
  const double shift = sqrt(Radius*Radius - radius*radius);
  const double phi = M_PI/2 - atan(shift/radius);
  const double bend = Radius*cos(angle)-shift + radius;

  if((fabs(point_coord.x) + fabs(point_coord.y))<radius) {
    id=0;
    const D2matrix rotMat(1.,1.,-1.,1.);
    D2vector rotated = D2vector(0.5,0.5)+rotMat*(point_coord/(2*radius));
    i = (blockgrid.Give_Nx_rectangle(id)*rotated.y);
    j = (blockgrid.Give_Ny_rectangle(id)*(1.-rotated.x));
  } else if( point_coord.x >= 0. && point_coord.y >= 0. ) {
    id = 1;
    const D2matrix rotMat(sqrt(0.5),-sqrt(0.5),sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x >= 0 && point_coord.y < 0 ) {
    id = 4;
    const D2matrix rotMat(-sqrt(0.5),-sqrt(0.5),sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x < 0 && point_coord.y < 0 ) {
    id = 3;
    const D2matrix rotMat(-sqrt(0.5),sqrt(0.5),-sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else {
    id = 2;
    const D2matrix rotMat(sqrt(0.5),sqrt(0.5),-sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  }
  return (point_coord.radius() <= Radius);
}

void Rechteck::getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const {
  id = 0;
  const int Nx = blockgrid.Give_Nx_rectangle(0);
  const int Ny = blockgrid.Give_Ny_rectangle(0);
  i = (point_coord.x - lx)/(rx-lx)*Nx;
  j = (point_coord.y - ly)/(ry-ly)*Ny;
  if(i < 0) i=0;
  if(i > Nx) i = Nx;
  if(j < 0) j=0;
  if(j > Ny) j = Ny;
}


bool Rechteck::getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const {
  id = 0;
  const int Nx = blockgrid.Give_Nx_rectangle(0);
  const int Ny = blockgrid.Give_Ny_rectangle(0);
  i = (point_coord.x - lx)/(rx-lx)*Nx;
  j = (point_coord.y - ly)/(ry-ly)*Ny;
  if(i >=0.0 && j >=0.0 && i < Nx && j < Ny) return true;
  if(i < 0)  i=0;
  if(i > Nx) i = Nx;
  if(j < 0)  j=0;
  if(j > Ny) j = Ny;
  return false;
}


void Rectangle::getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const {
	id = 0;
	const int Nx = blockgrid.Give_Nx_rectangle(0);
	const int Ny = blockgrid.Give_Ny_rectangle(0);
	i = (point_coord.x - lx) / (rx - lx)*Nx;
	j = (point_coord.y - ly) / (ry - ly)*Ny;
	if (i < 0) i = 0;
	if (i > Nx) i = Nx;
	if (j < 0) j = 0;
	if (j > Ny) j = Ny;
}


bool Rectangle::getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const {
	id = 0;
	const int Nx = blockgrid.Give_Nx_rectangle(0);
	const int Ny = blockgrid.Give_Ny_rectangle(0);
	i = (point_coord.x - lx) / (rx - lx)*Nx;
	j = (point_coord.y - ly) / (ry - ly)*Ny;
	if (i >= 0.0 && j >= 0.0 && i < Nx && j < Ny) return true;
	if (i < 0)  i = 0;
	if (i > Nx) i = Nx;
	if (j < 0)  j = 0;
	if (j > Ny) j = Ny;
	return false;
}
