/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef EXAMPLES_UG2D_H
#define EXAMPLES_UG2D_H
#include <vector>

class Blockgrid2D;


/**
 * Normal Rectangle
 *
 * **/
class Rechteck : public Unstructured2DGrid {
public:
	/**
	 * [0,rx] x [0,ry]
	 */
	Rechteck(double rx, double ry);
	/**
	 * [lx,rx] x [ly,ry]
	 */
	Rechteck(double lx, double ly,
		double rx, double ry);
	~Rechteck() {};

	void getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const;
	/**
	 * inverse computation
	 * @param point_coord input getCoordinate
	 * @param id id des rectangle
	 * @param i,j indizes der Koordinaten
	 * @return true falls Punkt im Gebiet, sonst false
	 **/
	bool getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const;
	void Test() const;

    double getSizeX() { return rx-lx; }
    double getSizeY() { return ry-ly; }

private:
	double lx, ly, rx, ry;
};


/**
 * Normal Rectangle
 * 
 * **/
class Rectangle : public Unstructured2DGrid {
 public:
   /**
    * [0,rx] x [0,ry]
    */
  Rectangle(double rx, double ry);
   /**
    * [lx,rx] x [ly,ry]
    */
  Rectangle(double lx, double ly,
	     double rx, double ry);
  ~Rectangle(){};

  void getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const ;
  /**
   * inverse computation
   * @param point_coord input getCoordinate
   * @param id id des rectangle
   * @param i,j indizes der Koordinaten
   * @return true falls Punkt im Gebiet, sonst false  
   **/ 
  bool getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const ;
  void Test() const;
  
 private:
  double lx,ly,rx,ry;
};


class NETransform2D: public EdgeTransformator2D{
private:
 double r;
 double rho;
public:
 void setRadius(double r_) {
   r = r_;
   rho = -1.;
 }

 void setRadius(double r_, double rho_) {
   r = r_;
   rho = rho_;
 }

 D2vector transform(double t) {
   if(rho < 1)
     return r*D2vector ( cos ( t*0.5*M_PI )- ( 1-t ), sin ( t*0.5*M_PI )-t);
   else
     return D2vector (rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t), rho*(sin(M_PI/6.0*(1.0+t))-0.5)-r*t);
 }
};

class NWTransform2D: public EdgeTransformator2D{
private:
 double r;
 double rho;
public:
 void setRadius(double r_) {
   r = r_;
   rho = -1.;
 }

 void setRadius(double r_, double rho_) {
   r = r_;
   rho = rho_;
 }

  D2vector transform(double t) {
    if(rho < 1)
      return r*D2vector ( -sin ( t*0.5*M_PI ) +t, cos ( t*0.5*M_PI )- ( 1-t ));
    else
      return D2vector (rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t, rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t));
  }
};

class SETransform2D: public EdgeTransformator2D{
private:
 double r;
 double rho;
public:
 void setRadius(double r_) {
   r = r_;
   rho = -1.;
 }

 void setRadius(double r_, double rho_) {
   r = r_;
   rho = rho_;
 }

 D2vector transform(double t) {
   if(rho < 1)
     return r*D2vector ( cos ( t*0.5*M_PI )- ( 1-t ), -sin ( t*0.5*M_PI ) +t);
   else
     return D2vector (rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t), rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t);
 }
};

class SWTransform2D: public EdgeTransformator2D{
private:
 double r;
 double rho;
public:
 void setRadius(double r_) {
   r = r_;
   rho = -1.;
 }

 void setRadius(double r_, double rho_) {
   r = r_;
   rho = rho_;
 }

 D2vector transform(double t) {
   if(rho < 1) // 0?
     return r*D2vector ( -cos ( t*0.5*M_PI ) + ( 1-t ), -sin ( t*0.5*M_PI ) +t);
   else
     return D2vector (-rho*(cos(M_PI/6.0*(1.0+t))-0.5)+r*(1.0-t), rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t);
 }
};


class Disc : public Unstructured2DGrid {
 public:
  Disc(double Radius, double radius, bool inner_grid_arched = false);

  bool inDomain(const D2vector& point_coord);

  /**
   * inverse computation
   * @param point_coord input getCoordinate
   * @param id id des rectangle
   * @param i,j indizes der Koordinaten
   **/
  void getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const ;
  /**
   * inverse computation
   * @param point_coord input getCoordinate
   * @param id id des rectangle
   * @param i,j indizes der Koordinaten
   * @return true falls Punkt im Gebiet, sonst false  
   **/  
  bool getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const ;

  void Test() const;

  ~Disc(){};
  
 private:
  NETransform2D transform_disc_NE;
  NETransform2D transform_disc_NE_inner;
  NWTransform2D transform_disc_NW;
  NWTransform2D transform_disc_NW_inner;
  SETransform2D transform_disc_SE;
  SETransform2D transform_disc_SE_inner;
  SWTransform2D transform_disc_SW;
  SWTransform2D transform_disc_SW_inner;
  double Radius;
  double radius;
  bool inner_grid_arched;
  void inverseTransformation(const D2vector& point_coord, double& i, double& j) const;  
};

class SectionedDisc : public Unstructured2DGrid {
 public:
  SectionedDisc(std::vector<double>& radius_vec, bool inner_grid_arched = false);
  
  bool inDomain(const D2vector& point_coord); 

  ~SectionedDisc();
  
private:
  NETransform2D* transform_disc_NE;
  NWTransform2D* transform_disc_NW;
  SETransform2D* transform_disc_SE;
  SWTransform2D* transform_disc_SW;
  double outerR;
};

/**
 * Normal Rectangle
 *
 * **/
class SectionedRectangle : public Unstructured2DGrid {
 public:
   /**
    * v_x x v_y
    */
	SectionedRectangle(std::vector<double>& v_x, std::vector<double>& v_y);

	
	~SectionedRectangle(){};
private:
  double xMin, xMax, yMin, yMax;	
};

class TestRectangle : public Unstructured2DGrid {
 public:
   /**
    * v_x x v_y
    */
  TestRectangle(double lx, double ly,
            double rx, double ry);

	
        ~TestRectangle(){};
};

#endif // EXAMPLES_UG_H
