/**********************************************************************************
 * Copyright 2010 Christoph Pflaum
 *              Department Informatik Lehrstuhl 10 - Systemsimulation
 *              Friedrich-Alexander Universität Erlangen-Nürnberg
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "ug2D.h"
#include "blockgrid2D.h"
#include "../extemp/extemp2D.h"
#include "../../../common_source/abbrevi_common.h"
#include "../abbrevi2D.h"

#ifndef FLEXBLEINTERPOLATION2D_H
#define FLEXBLEINTERPOLATION2D_H

template<class UGclass, class A>
class ExpressionFlexibleInterpolation2D;

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
struct FlexibleInterpolation2D_helper_base {
  static Result getNeighborEdgeValue(const ExpressionFlexibleInterpolation2D<UGclass,A>& ref, int NeighborId, int edgeID, int index1);

  static int getNeighborX(const ExpressionFlexibleInterpolation2D<UGclass,A>& ref, int id, int x_index);

  static int getNeighborY(const ExpressionFlexibleInterpolation2D<UGclass,A>& ref, int id, int y_index);

  static Result f(params2D_in, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref);

  static void update(int id, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref);

  static Result findPointInUG(D2vector& point_coord, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref);

  static Result QubicQuadInterpolate(Result data[4][4], double x, double y);
};

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
struct FlexibleInterpolation2D_helper : public FlexibleInterpolation2D_helper_base<A,Result, UGclass, TYP_EL> {
  template<class UGclass_, class A_>
  friend class ExpressionFlexibleInterpolation2D;
};


/***
 * Objekt zum Interpolieren von Daten von einem Gitter auf das andere
 * Beispiel:
 *        FlexibleInterpolation2D interpolator(grid_from, grid_to);
 *        Variable2D<double> uFrom(grid_from);
 *        Variable2D<double> uTo(grid_to);
 *            ...
 *        uTo = interpolator(uFrom);
 * 
 * grid_from sollte Kreis sein!!!
 * grid_to   beliebige Geometrie
 * Interpolation durch 4 Punkte an denen Wert + Steigung bekannt ist.
 * 
 ***/
template<class UGclass>
class FlexibleInterpolation2D
{
  const Blockgrid2D& grid_from;
  const UGclass& ug_from;

  const Blockgrid2D& grid_to;
  //const UGclass& ug_to;


  // for parallel
  int my_rank;
  MPI_Comm comm;

 public:
  FlexibleInterpolation2D(Blockgrid2D& grid_from_, Blockgrid2D& grid_to_);

  template<class A>
  ExpressionFlexibleInterpolation2D<UGclass,A> operator() (const Expr2D<A>& a_) const;

  template<class UGclass_, class A>
  friend class ExpressionFlexibleInterpolation2D;

  template<class A_,typename Result, class UGclass_, elementTyp TYP_EL>
  friend struct FlexibleInterpolation2D_helper_base;

  template<class A_,typename Result, class UGclass_, elementTyp TYP_EL>
  friend struct FlexibleInterpolation2D_helper;
};

template<class UGclass, class A>
class ExpressionFlexibleInterpolation2D: public Expr2D< ExpressionFlexibleInterpolation2D<UGclass,A> >
{
private:
  const FlexibleInterpolation2D<UGclass>& interpol;
  const A& a;

  ExpressionFlexibleInterpolation2D(const FlexibleInterpolation2D<UGclass>& interpol_, const A& a_);
public:
  typedef typename A::Result Result;

  stencil_typ Give_stencil_typ() const {return yes_stencil;}

  template <elementTyp TYP_EL>
  void Update(int id) const {
    FlexibleInterpolation2D_helper<A,Result,UGclass, TYP_EL>::update(id,*this);
  }

  template <elementTyp TYP_EL>
  Result Give_data( params2D_in ) const {
    return FlexibleInterpolation2D_helper<A,Result,UGclass, TYP_EL>::f(params2D_out,*this);
  }

  Unstructured2DGrid * Give_Ug() const {
    return a.Give_Ug();
  }

  template<class A_,typename Result, class UGclass_, elementTyp TYP_EL>
  friend struct FlexibleInterpolation2D_helper;

  Blockgrid2D* Give_blockgrid() const {
    return &interpol.grid_to;
  }

  template<class UGclass_>
  friend class FlexibleInterpolation2D;

  template<class A_,typename Result, class UGclass_, elementTyp TYP_EL>
  friend struct FlexibleInterpolation2D_helper_base;

  template<class A_,typename Result, class UGclass_, elementTyp TYP_EL>
  friend struct FlexibleInterpolation2D_helper;
};

#endif
