/**********************************************************************************
 * Copyright 2010 Christoph Pflaum
 *              Department Informatik Lehrstuhl 10 - Systemsimulation
 *              Friedrich-Alexander Universität Erlangen-Nürnberg
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "ug2D.h"
#include "blockgrid2D.h"
#include "elements2D.h"
#include "flexibleInterpolation2D.h"
#include "../../../common_source/abbrevi_common.h"

#ifndef _NOARRAYLIB

#ifdef CPP11
  #include <array>
  using std::array;
#else
  #ifdef WIN32
    #include <array>
  #else
    #include <tr1/array>
  #endif
#endif

#endif //_NOARRAYLIB

#include <numeric>

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
inline Result FlexibleInterpolation2D_helper_base<A,Result, UGclass, TYP_EL>::getNeighborEdgeValue(const ExpressionFlexibleInterpolation2D<UGclass,A>& ref, int NeighborId, int edgeID, int index1) {
  RectangleEl& rectangle = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(NeighborId));
  if(rectangle.Give_id_edge(Wdir2D) == edgeID) {
    const int index2 = 1;
    return ref.a.template Give_data < rectangleEl > (NeighborId,index2,index1,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
  } else if(rectangle.Give_id_edge(Edir2D) == edgeID) {
    const int index2 = (ref.interpol.grid_from.Give_Nx_rectangle(NeighborId))-1;
    return ref.a.template Give_data < rectangleEl > (NeighborId,index2,index1,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
  } else if(rectangle.Give_id_edge(Sdir2D) == edgeID) {
    const int index2 = 1;
    return ref.a.template Give_data < rectangleEl > (NeighborId,index1,index2,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
  } else if(rectangle.Give_id_edge(Ndir2D) == edgeID) {
    const int index2 = (ref.interpol.grid_from.Give_Ny_rectangle(NeighborId)-1);
    return ref.a.template Give_data < rectangleEl > (NeighborId,index1,index2,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
  }
  else {
    std::cout << "Fehler" << std::endl;
  }
}

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
inline int FlexibleInterpolation2D_helper_base<A,Result, UGclass, TYP_EL>::getNeighborX(const ExpressionFlexibleInterpolation2D<UGclass,A>& ref, int id, int x_index) {

  RectangleEl& element = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(id));
  if(x_index == -1) {
    const int edgeId = element.Give_id_edge(Wdir2D);
    EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
    if(edge.isAtBoundary()) return -1;
    else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
  }
  if(x_index == -2) {
    const int edgeId = element.Give_id_edge(Edir2D);
    EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
    if(edge.isAtBoundary()) return -1;
    else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
  }
  assert(-1);
}

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
inline int FlexibleInterpolation2D_helper_base<A,Result, UGclass, TYP_EL>::getNeighborY(const ExpressionFlexibleInterpolation2D<UGclass,A>& ref, int id, int y_index) {

  RectangleEl& element = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(id));
  if(y_index == -1) {
    const int edgeId = element.Give_id_edge(Sdir2D);
    EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
    if(edge.isAtBoundary()) return -1;
    else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
  }
  if(y_index == -2) {
    const int edgeId = element.Give_id_edge(Ndir2D);
    EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
    if(edge.isAtBoundary()) return -1;
    else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
  }
  assert(-1);
}

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
Result FlexibleInterpolation2D_helper_base<A,Result, UGclass, TYP_EL>::findPointInUG(D2vector& point_coord, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
  const UGclass& ug_from(ref.interpol.ug_from);
  int id;
  double i=0.0;
  double j=0.0;

  ug_from.getCoordinate(point_coord, id, i, j, ref.interpol.grid_from);

  const int Nx_from = ref.interpol.grid_from.Give_Nx_rectangle(id);
  const int Ny_from = ref.interpol.grid_from.Give_Ny_rectangle(id);

  if(i >= Nx_from) i = Nx_from;
  if(j >= Ny_from) j = Ny_from;
  ref.a.template Update < rectangleEl > (id);

  int ii[4];
  ii[1] = (floor(i)<Nx_from)?floor(i):floor(i)-1;
  ii[2] = ii[1] + 1;
  ii[0] = (ii[1] > 0) ? ii[1] - 1 : -1;
  ii[3] = (ii[2] < Nx_from) ? ii[2] + 1 : -2;
  double weight_i =i-ii[1];
  int jj[4];
  jj[1] = (floor(j)<Ny_from)?floor(j):floor(j)-1;
  jj[2] = jj[1] + 1;
  jj[0] = (jj[1] > 0) ? jj[1] - 1 : -1;
  jj[3] = (jj[2] < Ny_from) ? jj[2] + 1 : -2;
  double weight_j = j-jj[1];
  RectangleEl& me = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(id));

  Result data[4][4];
  int print = 0;
  for (int ii_i = 0; ii_i < 4; ++ii_i) {
    for (int jj_i = 0; jj_i < 4; ++jj_i) {
      if(ii[ii_i] >= 0 && jj[jj_i] >= 0) {
        data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id, ii[ii_i], jj[jj_i], Nx_from);

      } else if(ii[ii_i] < 0 && jj[jj_i] >= 0) {
        int neighbor_x = getNeighborX(ref,id,ii[ii_i]);

        int edgeID;
        if(neighbor_x >=0) {
          //ref.a.Update<rectangleEl>(neighbor_x);
          if(ii[ii_i]==-1) edgeID = me.Give_id_edge(Wdir2D);
          else edgeID = me.Give_id_edge(Edir2D);
          data[jj_i][ii_i] = getNeighborEdgeValue(ref,neighbor_x,edgeID,jj[jj_i]);
        }
        else {
          //Boundary point, no neighbor exists
          if(ii[ii_i]==-1)
            data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id,0, jj[jj_i], Nx_from);
          else
            data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id, (ref.interpol.grid_from.Give_Nx_rectangle(id)) , jj[jj_i], Nx_from);
        }
      } else if(ii[ii_i] >= 0 && jj[jj_i] < 0) {
        int neighbor_y = getNeighborY(ref,id,jj[jj_i]);

        int edgeID;
        if(neighbor_y >=0) {
//            ref.a.Update<rectangleEl>(neighbor_y);
          if(jj[jj_i]==-1) edgeID = me.Give_id_edge(Sdir2D);
          else edgeID = me.Give_id_edge(Ndir2D);
          data[jj_i][ii_i] = getNeighborEdgeValue(ref,neighbor_y,edgeID,ii[ii_i]);
        }
        else {
          //Boundary point, no neighbor exists
          if(ii[ii_i]==-1)
            data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id,0, jj[jj_i], Nx_from);
          else
            data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id, (ref.interpol.grid_from.Give_Ny_rectangle(id)) , jj[jj_i], Nx_from);
        }
      } else {
        int neighbor_x = getNeighborX(ref,id,ii[ii_i]);
        int neighbor_y = getNeighborY(ref,id,jj[jj_i]);
        Result data_x;
        Result data_y;
        int index_x = (ii[ii_i] == -1)?Nx_from:0;
        int index_y = (jj[jj_i] == -1)?Ny_from:0;
        if(neighbor_x >=0) {
//            ref.a.Update<rectangleEl>(neighbor_x);
          int edgeID;

          if(ii[ii_i]==-1) edgeID = me.Give_id_edge(Wdir2D);
          else edgeID = me.Give_id_edge(Edir2D);
          data_x = getNeighborEdgeValue(ref,neighbor_x,edgeID,index_y);
        } else {
          if(ii[ii_i]==-1)
            data_x = ref.a.template Give_data < rectangleEl > (id,Nx_from, index_y, Nx_from);
          else
            data_x = ref.a.template Give_data < rectangleEl > (id,0, index_y, Nx_from);
        }
        if(neighbor_y >=0) {
//            ref.a.Update<rectangleEl>(neighbor_y);
          int edgeID;

          if(jj[jj_i]==-1) edgeID = me.Give_id_edge(Sdir2D);
          else edgeID = me.Give_id_edge(Ndir2D);
          data_y = getNeighborEdgeValue(ref,neighbor_y,edgeID,index_x);
        } else {
          if(jj[jj_i]==-1)
            data_y = ref.a.template Give_data < rectangleEl > (id,Nx_from, index_x, Nx_from);
          else
            data_y = ref.a.template Give_data < rectangleEl > (id,0, index_x, Nx_from);
        }
        data[jj_i][ii_i] = 0.5*(data_x+data_y);
        print=1;
      }
    }
  }

  return QubicQuadInterpolate(data, weight_i, weight_j);
}

template<class A,typename Result, class UGclass, elementTyp TYP_EL>
inline Result FlexibleInterpolation2D_helper_base<A,Result, UGclass, TYP_EL>::QubicQuadInterpolate(Result data[4][4], double x, double y) {
#ifdef _NOARRAYLIB
    return Result(0.0);       
#else  
  for(int i = 0; i < 4; ++i) {
    data[i][0] = 0.5*(data[i][2]-data[i][0]);
    data[i][3] = 0.5*(data[i][3]-data[i][1]);
  }

  Result data_x[4];
  array< array<double,4>,4 > A_inv = {{{2.,-2.,1.,1.},{-3.,3.,0-2.,-1.},{0.,0.,1.,0.},{1.,0.,0.,0.}}};
  for(int i = 0; i < 4; ++i) {

    array<Result,4> vec_f = {data[i][1],data[i][2],data[i][0],data[i][3]};
    array<double,4> vec_x = {x*x*x,x*x,x,1};
    array<Result,4> par;
    for(int k = 0; k < 4; ++k) {
      par[k] = std::inner_product(vec_f.begin(),vec_f.end(),A_inv[k].begin(),Result(0.0));
    }
    data_x[i] = std::inner_product(vec_x.begin(),vec_x.end(),par.begin(),Result(0.0));
  }

  data_x[0] = 0.5*(data_x[2]-data_x[0]);
  data_x[3] = 0.5*(data_x[3]-data_x[1]);

  array<Result,4> vec_f = {data_x[1],data_x[2],data_x[0],data_x[3]};
  array<double,4> vec_y = {y*y*y,y*y,y,1};
  array<Result,4> par;
  for(int k = 0; k < 4; ++k) {
    par[k] = std::inner_product(vec_f.begin(),vec_f.end(),A_inv[k].begin(),Result(0.0));
  }
  return std::inner_product(vec_y.begin(),vec_y.end(),par.begin(),Result(0.0));
#endif // _NOARRAYLIB
}

template<class A,typename Result, class UGclass>
struct FlexibleInterpolation2D_helper<A,Result,UGclass,rectangleEl> : public FlexibleInterpolation2D_helper_base<A,Result, UGclass,rectangleEl> {
  static Result f(params2D_in, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
    D2vector point_coord = ref.interpol.grid_to.Give_coord_rectangle(id,i,j);
    return FlexibleInterpolation2D_helper_base<A,Result,UGclass,rectangleEl>::findPointInUG(point_coord,ref);
  }

  static void update(int id, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
    if(id < ref.a.Give_Ug()->Give_number_rectangle())
      ref.a.template Update<rectangleEl>(id);
  }
};

template<class A,typename Result, class UGclass>
struct FlexibleInterpolation2D_helper<A,Result, UGclass,edgeEl> : public FlexibleInterpolation2D_helper_base<A,Result, UGclass,rectangleEl> {
  static Result f(params2D_in, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
    D2vector point_coord = ref.interpol.grid_to.Give_coord_edge(id,i);
    return FlexibleInterpolation2D_helper_base<A,Result,UGclass,edgeEl>::findPointInUG(point_coord,ref);
  }

  static void update(int id, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
    if(id < ref.a.Give_Ug()->Give_number_edges())
          ref.a.template Update<edgeEl>(id);
  }
};

template<class A,typename Result, class UGclass>
struct FlexibleInterpolation2D_helper<A,Result, UGclass,pointEl> : public FlexibleInterpolation2D_helper_base<A,Result, UGclass,rectangleEl> {
  static Result f(params2D_in, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
    D2vector point_coord = ref.interpol.grid_to.Give_coord_point(id);
    return FlexibleInterpolation2D_helper_base<A,Result,UGclass,pointEl>::findPointInUG(point_coord,ref);
  }

  static void update(int id, const ExpressionFlexibleInterpolation2D<UGclass,A>& ref) {
    if(id < ref.a.Give_Ug()->Give_number_points())
      ref.a.template Update<pointEl>(id);
  }
};

template<class UGclass>
FlexibleInterpolation2D<UGclass>::FlexibleInterpolation2D(Blockgrid2D& grid_from_, Blockgrid2D& grid_to_)
    : grid_from(grid_from_), ug_from(*static_cast<UGclass*>(grid_from.Give_unstructured_grid())),
      grid_to(grid_to_)
{
  /// TODO @Rainer assert
//  assert(typeof(ug_from) == UGclass);
//  // for parallel
//  my_rank = ug_from.Give_my_rank();
//  comm = ug_from.Give_MPI_comm();
}

template<class UGclass>
template<class A>
ExpressionFlexibleInterpolation2D<UGclass, A>
  FlexibleInterpolation2D<UGclass>::operator()(const Expr2D<A>& a_) const
  {
    const A& a(a_);
    return ExpressionFlexibleInterpolation2D < UGclass,A > (*this, a);
  }

template<class UGclass, class A>
ExpressionFlexibleInterpolation2D<UGclass,A>::ExpressionFlexibleInterpolation2D(const FlexibleInterpolation2D<UGclass>& interpol_, const A& a_)
      : interpol(interpol_), a(a_)
  {
  assert(a_.Give_blockgrid() == &interpol_.grid_from);
  }
