/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "../extemp/co_fu2D.h"
#include "../extemp/variable2D.h"
#include "../extemp/variable2D_cc.h"
#include "elements2D.h"
#include "parti2D.h"
#include "ug2D.h"
#include "gridFrom3D.h"
#include "transformations_ug2D.h"
#include "../../../program/source/grid/examples_ug.h"
#include <typeinfo>
#include <cmath>

#ifdef WIN32
  inline double round(double x) { return (floor(x + 0.5)); }
#endif

UGFrom3DSlice::UGFrom3DSlice(Unstructured_grid* ug3D, double PMLwidth) : PMLwidth(PMLwidth), PMLfunctor(this){
  assert(ug3D->getUgTyp()!=noWithZTyp);
  ug3Dp = ug3D;
  TypeOfUG typ3D = ug3D->getUgTyp();
  if(typ3D==hexahedronUgTyp) constructRectangle();
  else if(typ3D==cylinderUgTyp)  PMLwidth>0.0?constructPMLDisc(PMLwidth):constructDisc();
  else if(typ3D==cylinderListUgTyp) PMLwidth>0.0?constructPMLDisc(PMLwidth):constructDisc();
  else if(typ3D==hexahedronListUgTyp) constructSectionedRectangle();
  else if(typ3D==hexahedronListTestUgTyp) constructTestRectangle();
  else if(typ3D==hexahedronAdaptiveUgTyp) constructAdaptiveRectangle();
  else {}
}

bool UGFrom3DSlice::inDomain(const D2vector& point_coord) {
  if(ug3Dtype == HexaederUG) {
     if((point_coord.x < MinimumX()) || (point_coord.x > MaximumX())) return false;
     if((point_coord.y < MinimumY()) || (point_coord.y > MaximumY())) return false;
     return true;    
  }
  return D2VectorNorm(point_coord) <= MaximumX();
}

////////////////////////////////////////////////////////////////////
// rectangle
////////////////////////////////////////////////////////////////////


void UGFrom3DSlice::constructRectangle() {
  ug3Dtype = HexaederUG;
  Set_number_points ( 4 ); // 4 an den Ecken

  D3vector vSWdir = ug3Dp->Give_point(SWdir2D)->Give_coordinate();
  D3vector vSEdir = ug3Dp->Give_point(SEdir2D)->Give_coordinate();
  D3vector vWND = ug3Dp->Give_point(NWdir2D)->Give_coordinate();
  D3vector vNEdir = ug3Dp->Give_point(NEdir2D)->Give_coordinate();
  
  Set_coordinate_point( SWdir2D, D2vector(vSWdir.x,vSWdir.y) );
  Set_coordinate_point( SEdir2D, D2vector(vSEdir.x,vSEdir.y) );
  Set_coordinate_point( NWdir2D, D2vector(vWND.x,vWND.y) );
  Set_coordinate_point( NEdir2D, D2vector(vNEdir.x,vNEdir.y) );

  Set_number_rectanglEs ( 1 );
  Set_rectangle ( 0, SWdir2D, SEdir2D, NWdir2D, NEdir2D );

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();  
}

void UGFrom3DSlice::constructSectionedRectangle() {
  ug3Dtype = HexaederUG;
  
  int n_x = ((HexahedronXYList*)ug3Dp)->n_x;
  int n_y = ((HexahedronXYList*)ug3Dp)->n_y;

  
  assert(n_x*n_y);
  Set_number_points ( (n_x+1) * (n_y+1) );
  for(int i_y = 0; i_y <= n_y; ++i_y ) {
    for(int i_x = 0; i_x <= n_x; ++i_x ) {
      D3vector v = ug3Dp->Give_point(i_y*(n_x+1) + i_x)->Give_coordinate();
      Set_coordinate_point( i_y*(n_x+1) + i_x, D2vector(v.x,v.y) );
    }
  }

  Set_number_rectanglEs ( n_x*n_y );

  for(int i_y = 0; i_y < n_y; ++i_y ) {
    for(int i_x = 0; i_x < n_x; ++i_x ) {
      Set_rectangle((i_y * n_x + i_x),
        (i_y)   * (n_x+1) + (i_x),   // SWdir2D
        (i_y)   * (n_x+1) + (i_x+1), // SEdir2D
        (i_y+1) * (n_x+1) + (i_x),   // NWdir2D
        (i_y+1) * (n_x+1) + (i_x+1));// NEdir2D
    }
  }

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();
}

void UGFrom3DSlice::constructTestRectangle() {
  ug3Dtype = HexaederUG;
  
  int n_x = 3;
  int n_y = 3;
  Set_number_points ( (n_x+1) * (n_y+1) );
  for(int i_y = 0; i_y <= n_y; ++i_y ) {
    for(int i_x = 0; i_x <= n_x; ++i_x ) {
      D3vector v = ug3Dp->Give_point(i_y*(n_x+1) + i_x)->Give_coordinate();
      Set_coordinate_point( i_y*(n_x+1) + i_x, D2vector(v.x,v.y) );
    }
  }

  Set_number_rectanglEs ( n_x*n_y );

  for(int i_y = 0; i_y < n_y; ++i_y ) {
    for(int i_x = 0; i_x < n_x; ++i_x ) {
      Set_rectangle((i_y * n_x + i_x),
        (i_y)   * (n_x+1) + (i_x),   // SWdir2D
        (i_y)   * (n_x+1) + (i_x+1), // SEdir2D
        (i_y+1) * (n_x+1) + (i_x),   // NWdir2D
        (i_y+1) * (n_x+1) + (i_x+1));// NEdir2D
    }
  }

  construction_rectangle_points_done();

  // no transformations of the edges and faces
  construction_done();
}

void UGFrom3DSlice::getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const {
  if(ug3Dtype == CylinderUG)
    getDiscCoordinate(point_coord, id, i, j, blockgrid);
  else if(ug3Dtype == HexaederUG)
    getRectangleCoordinate(point_coord, id, i, j, blockgrid);
  else
    assert(0);
}

bool UGFrom3DSlice::getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const {
  if(ug3Dtype == CylinderUG)
    return getDiscCoordinate(point_coord, id, i, j, blockgrid);
  else if(ug3Dtype == HexaederUG)
    return getRectangleCoordinate(point_coord, id, i, j, blockgrid);
  else
    assert(0);
}

void UGFrom3DSlice::Test() const {
  std::cout << "UGFrom3DSlice\n";
}

////////////////////////////////////////////////////////////////////
// disc
////////////////////////////////////////////////////////////////////

void UGFrom3DSlice::constructDisc() {
  ug3Dtype = CylinderUG;
  double R = ug3Dp->Give_point(1)->Give_coordinate().x;
  double r = ug3Dp->Give_point(0)->Give_coordinate().x;

  double offsetX = 0.0;
  double offsetY = 0.0;

  // new version
  pointer_global_data = new double[3];
  pointer_global_data[0] = R;
  pointer_global_data[1] = r;
  double R_innen = 2.0*r*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));//10.0*R;
  pointer_global_data[2] = R_innen;
  ////int num_blocks = Lz.size() -1;
  ////assert(num_blocks>0);
  ////int num_blocks_total = 5* ( Lz.size()-1 );
  int num_points_per_face = 8;
  ////int numPoints = num_points_per_face*(num_blocks+1) ;

  Set_number_points( 8 );


  Set_coordinate_point(0, D2vector ( offsetX+r,offsetY ) );
  Set_coordinate_point(1, D2vector ( offsetX+R,offsetY ) );

  Set_coordinate_point(2, D2vector ( offsetX,offsetY+r ) );
  Set_coordinate_point(3, D2vector ( offsetX,offsetY+R ) );

  Set_coordinate_point(4, D2vector ( offsetX-r,offsetY ) );
  Set_coordinate_point(5, D2vector ( offsetX-R,offsetY ) );

  Set_coordinate_point(6, D2vector ( offsetX,offsetY-r ) );
  Set_coordinate_point(7, D2vector ( offsetX,offsetY-R ) );

  Set_number_rectanglEs(5);
  Set_rectangle( 0, 0, 2, 6, 4);

  for(int j = 0; j < num_points_per_face; j=j+2 ) {
      int myIndex = j/2+1;
      Set_rectangle(myIndex,
                     j+0,
                     j+1,
                     ((j+2) % num_points_per_face),
                     ((j+3) % num_points_per_face));
  }

  
  construction_rectangle_points_done();

  transform_disc_NE.setRadius(R);
  Set_transformation_edge ( 1, 3, &transform_disc_NE );
  transform_disc_NW.setRadius(R);
  Set_transformation_edge ( 3, 5, &transform_disc_NW );
  transform_disc_SE.setRadius(R);
  Set_transformation_edge ( 5, 7, &transform_disc_SW );
  transform_disc_SW.setRadius(R);
  Set_transformation_edge ( 1, 7, &transform_disc_SE );

  TypeOfUG typ3D = ug3Dp->getUgTyp();
  bool inner_grid_arched = false;
  if(typ3D==cylinderUgTyp) {
    inner_grid_arched = ((Cylinder*)ug3Dp)->getInnerGridArched();
  }
  else if(typ3D==cylinderListUgTyp) {
    inner_grid_arched = ((CylinderList*)ug3Dp)->getInnerGridArched();
  }

  if (inner_grid_arched) {
    //3.2 Kanten innen
    transform_disc_NE_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 0, 2, &transform_disc_NE_inner );
    transform_disc_NW_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 2, 4, &transform_disc_NW_inner );
    transform_disc_SE_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 4, 6, &transform_disc_SW_inner );
    transform_disc_SW_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 0, 6, &transform_disc_SE_inner );
  }

  construction_done();
}

void UGFrom3DSlice::constructPMLDisc(double PMLwidth) {
  ug3Dtype = CylinderUG;

  double R_PML = ug3Dp->Give_point(1)->Give_coordinate().x + PMLwidth;
  double R = ug3Dp->Give_point(1)->Give_coordinate().x;
  double r = ug3Dp->Give_point(0)->Give_coordinate().x;

  double offsetX = 0.0;
  double offsetY = 0.0;

  // new version
  pointer_global_data = new double[3];
  pointer_global_data[0] = R;
  pointer_global_data[1] = r;
  double R_innen = 2.0*r*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));//10.0*R;
  pointer_global_data[2] = R_innen;
  ////int num_blocks = Lz.size() -1;
  ////assert(num_blocks>0);
  ////int num_blocks_total = 5* ( Lz.size()-1 );
  int num_points_per_face = 8;
  ////int numPoints = num_points_per_face*(num_blocks+1) ;

  Set_number_points( 12 );


  Set_coordinate_point(0, D2vector ( offsetX+r,offsetY ) );
  Set_coordinate_point(1, D2vector ( offsetX+R,offsetY ) );
  Set_coordinate_point(8, D2vector ( offsetX+R_PML,offsetY ) );

  Set_coordinate_point(2, D2vector ( offsetX,offsetY+r ) );
  Set_coordinate_point(3, D2vector ( offsetX,offsetY+R ) );
  Set_coordinate_point(9, D2vector ( offsetX,offsetY+R_PML ) );

  Set_coordinate_point(4, D2vector ( offsetX-r,offsetY ) );
  Set_coordinate_point(5, D2vector ( offsetX-R,offsetY ) );
  Set_coordinate_point(10, D2vector ( offsetX-R_PML,offsetY ) );

  Set_coordinate_point(6 , D2vector ( offsetX,offsetY-r ) );
  Set_coordinate_point(7, D2vector ( offsetX,offsetY-R ) );
  Set_coordinate_point(11, D2vector ( offsetX,offsetY-R_PML ) );

  Set_number_rectanglEs(9);
  Set_rectangle( 0, 0, 2, 6, 4);

  for(int j = 0; j < num_points_per_face; j=j+2 ) {
      int myIndex = j/2+1;
      Set_rectangle(myIndex,
                     j+0,
                     j+1,
                     ((j+2) % num_points_per_face),
                     ((j+3) % num_points_per_face));
  }

  for(int j = 0; j <= 3; ++j) {
    Set_rectangle(5+j,2*j+1,j+8,(2*j+3)%num_points_per_face,(j+1)%4+8);
  }


  construction_rectangle_points_done();

  transform_disc_NE.setRadius(R);
  Set_transformation_edge ( 1, 3, &transform_disc_NE );
  transform_disc_NW.setRadius(R);
  Set_transformation_edge ( 3, 5, &transform_disc_NW );
  transform_disc_SE.setRadius(R);
  Set_transformation_edge ( 5, 7, &transform_disc_SW );
  transform_disc_SW.setRadius(R);
  Set_transformation_edge ( 1, 7, &transform_disc_SE );

  transform_disc_NE_PML.setRadius(R_PML);
  Set_transformation_edge ( 8, 9, &transform_disc_NE_PML );
  transform_disc_NW_PML.setRadius(R_PML);
  Set_transformation_edge ( 9, 10, &transform_disc_NW_PML );
  transform_disc_SE_PML.setRadius(R_PML);
  Set_transformation_edge ( 10, 11, &transform_disc_SW_PML );
  transform_disc_SW_PML.setRadius(R_PML);
  Set_transformation_edge ( 8, 11, &transform_disc_SE_PML );

  TypeOfUG typ3D = ug3Dp->getUgTyp();
  bool inner_grid_arched = false;
  if(typ3D==cylinderUgTyp) {
    inner_grid_arched = ((Cylinder*)ug3Dp)->getInnerGridArched();
  }
  else if(typ3D==cylinderListUgTyp) {
    inner_grid_arched = ((CylinderList*)ug3Dp)->getInnerGridArched();
  }

  if (inner_grid_arched) {
    //3.2 Kanten innen
    transform_disc_NE_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 0, 2, &transform_disc_NE_inner );
    transform_disc_NW_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 2, 4, &transform_disc_NW_inner );
    transform_disc_SE_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 4, 6, &transform_disc_SW_inner );
    transform_disc_SW_inner.setRadius(r,R_innen);
    Set_transformation_edge ( 0, 6, &transform_disc_SE_inner );
  }

  construction_done();
}

void UGFrom3DSlice::inverseTransformation(const D2vector& point_coord, double& i, double& j) const {
  i = 0.5;
  j = 0.0;
  int n = 0;

  const double radius = ug3Dp->Give_point(0)->Give_coordinate().x;
  const double Radius = ug3Dp->Give_point(1)->Give_coordinate().x;

  while (n<5) {
    const double gx= (sqrt(0.5)*Radius*i + (1-i)*sqrt(0.5)*radius)*j-point_coord.x;
    const double gy= (i*Radius*cos(M_PI/4*j) + (1-i)*sqrt(0.5)*radius)-point_coord.y;
    const double a = (sqrt(2.0)*Radius*i)/2.0 - sqrt(0.5)*radius*(i - 1);
    const double b = -j*(sqrt(0.5)*radius - (sqrt(2.0)*Radius)/2.0);
    const double c = -(M_PI*Radius*i*sin((M_PI*j)/4.0))/4.0;
    const double d = Radius*cos((M_PI*j)/4.0) - sqrt(0.5)*radius;

    j = j - (d*gx-b*gy)/(a*d-b*c);
    i = i + (c*gx-a*gy)/(a*d-b*c);

    n=n+1;
  }
  if(i>= 1.0) i=1.0;
}

void UGFrom3DSlice::getRectangleCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const {
  id = 0;
  double lx = ug3Dp->Give_point(0)->Give_coordinate().x;
  double ly = ug3Dp->Give_point(0)->Give_coordinate().y;
  double rx = ug3Dp->Give_point(3)->Give_coordinate().x;
  double ry = ug3Dp->Give_point(3)->Give_coordinate().y;

  const int Nx = blockgrid.Give_Nx_rectangle(0);
  const int Ny = blockgrid.Give_Ny_rectangle(0);
  i = (point_coord.x - lx)/(rx-lx)*Nx;
  j = (point_coord.y - ly)/(ry-ly)*Ny;
  if(i < 0) i=0;
  if(i > Nx) i = Nx;
  if(j < 0) j=0;
  if(j > Ny) j = Ny;
}


bool UGFrom3DSlice::getRectangleCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const {
  id = 0;
  const int Nx = blockgrid.Give_Nx_rectangle(0);
  const int Ny = blockgrid.Give_Ny_rectangle(0);
  double lx = ug3Dp->Give_point(0)->Give_coordinate().x;
  double ly = ug3Dp->Give_point(0)->Give_coordinate().y;
  double rx = ug3Dp->Give_point(3)->Give_coordinate().x;
  double ry = ug3Dp->Give_point(3)->Give_coordinate().y;
  i = (point_coord.x - lx)/(rx-lx)*Nx;
  j = (point_coord.y - ly)/(ry-ly)*Ny;
  if(i >=0.0 && j >=0.0 && i < Nx && j < Ny) return true;  
  if(i < 0) i=0;
  if(i > Nx) i = Nx;
  if(j < 0) j=0;
  if(j > Nx) j = Nx;
  return false;  
}

void UGFrom3DSlice::getDiscCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const  {
  i = j = 0;
  const double radius = ug3Dp->Give_point(0)->Give_coordinate().x; // inner
  const double Radius = ug3Dp->Give_point(1)->Give_coordinate().x; // outer
  const double r_Radius = 2.0*radius*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  const double angle = fmod(0.5*M_PI + atan(point_coord.y / point_coord.x),0.5*M_PI);
  const double a = r_Radius - radius;
  const double shift = sqrt(Radius*Radius - radius*radius);
  const double phi = M_PI/2 - atan(shift/radius);
  const double bend = Radius*cos(angle)-shift + radius;

  if((fabs(point_coord.x) + fabs(point_coord.y))<radius) {
    id=0;
    const D2matrix rotMat(1.,1.,-1.,1.);
    D2vector rotated = D2vector(0.5,0.5)+rotMat*(point_coord/(2*radius));
    i = round(blockgrid.Give_Nx_rectangle(id)*rotated.y);
    j = round(blockgrid.Give_Ny_rectangle(id)*(1.-rotated.x));
  } else if( point_coord.x >= 0. && point_coord.y >= 0. ) {
    id = 1;
    const D2matrix rotMat(sqrt(0.5),-sqrt(0.5),sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;

    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x >= 0 && point_coord.y < 0 ) {
    id = 4;
    const D2matrix rotMat(-sqrt(0.5),-sqrt(0.5),sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x < 0 && point_coord.y < 0 ) {
    id = 3;
    const D2matrix rotMat(-sqrt(0.5),sqrt(0.5),-sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else {
    id = 2;
    const D2matrix rotMat(sqrt(0.5),sqrt(0.5),-sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = round((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = round((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  }
}

bool UGFrom3DSlice::getDiscCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const {
  i = j = 0;
  const double radius = ug3Dp->Give_point(0)->Give_coordinate().x; // inner
  const double Radius = ug3Dp->Give_point(1)->Give_coordinate().x; // outer
  const double r_Radius = 2.0*radius*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  const double angle = fmod(0.5*M_PI + atan(point_coord.y / point_coord.x),0.5*M_PI);
  const double a = r_Radius - radius;
  const double shift = sqrt(Radius*Radius - radius*radius);
  const double phi = M_PI/2 - atan(shift/radius);
  const double bend = Radius*cos(angle)-shift + radius;

  if((fabs(point_coord.x) + fabs(point_coord.y))<radius) {
    id=0;
    const D2matrix rotMat(1.,1.,-1.,1.);
    D2vector rotated = D2vector(0.5,0.5)+rotMat*(point_coord/(2*radius));
    i = (blockgrid.Give_Nx_rectangle(id)*rotated.y);
    j = (blockgrid.Give_Ny_rectangle(id)*(1.-rotated.x));
  } else if( point_coord.x >= 0. && point_coord.y >= 0. ) {
    id = 1;
    const D2matrix rotMat(sqrt(0.5),-sqrt(0.5),sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x >= 0 && point_coord.y < 0 ) {
    id = 4;
    const D2matrix rotMat(-sqrt(0.5),-sqrt(0.5),sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else if( point_coord.x < 0 && point_coord.y < 0 ) {
    id = 3;
    const D2matrix rotMat(-sqrt(0.5),sqrt(0.5),-sqrt(0.5),-sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  } else {
    id = 2;
    const D2matrix rotMat(sqrt(0.5),sqrt(0.5),-sqrt(0.5),sqrt(0.5));
    D2vector rotated = rotMat*point_coord;
    double j_tmp,i_tmp;
    inverseTransformation(rotated,i_tmp,j_tmp);
    i = ((i_tmp)*blockgrid.Give_Nx_rectangle(id));
    j = ((0.5-(0.5*j_tmp))*blockgrid.Give_Ny_rectangle(id));
  }
  return (point_coord.radius() <= Radius);
}

////////////////////////////////////////////////////////////////////
// adaptive rectangle
////////////////////////////////////////////////////////////////////

void UGFrom3DSlice::constructAdaptiveRectangle() {
  ug3Dtype = HexaederUG;  
  
  double Rx = ug3Dp->Give_point(1)->Give_coordinate().x;
  double rx = ug3Dp->Give_point(0)->Give_coordinate().x;
  double Ry = ug3Dp->Give_point(1)->Give_coordinate().y;
  double ry = ug3Dp->Give_point(0)->Give_coordinate().y;

  double offsetX = 0.0;
  double offsetY = 0.0;

  int num_points_per_face = 8;

  Set_number_points( 8 );


  Set_coordinate_point(0, D2vector ( offsetX+rx,offsetY+ry ) );
  Set_coordinate_point(1, D2vector ( offsetX+Rx,offsetY+Ry ) );

  Set_coordinate_point(2, D2vector ( offsetX-rx,offsetY+ry ) );
  Set_coordinate_point(3, D2vector ( offsetX-Rx,offsetY+Ry ) );

  Set_coordinate_point(4, D2vector ( offsetX-rx,offsetY-ry ) );
  Set_coordinate_point(5, D2vector ( offsetX-Rx,offsetY-Ry ) );

  Set_coordinate_point(6, D2vector ( offsetX+rx,offsetY-ry ) );
  Set_coordinate_point(7, D2vector ( offsetX+Rx,offsetY-Ry ) );

  Set_number_rectanglEs(5);
  Set_rectangle( 0, 0, 2, 6, 4);

  for(int j = 0; j < num_points_per_face; j=j+2 ) {
      int myIndex = j/2+1;
      Set_rectangle(myIndex,
                     j+0,
                     j+1,
                     ((j+2) % num_points_per_face),
                     ((j+3) % num_points_per_face));
  }

  construction_rectangle_points_done();
  construction_done();  
}

std::complex<double> UGFrom3DSlice::evaluate(const double x, const double y)
{
  const double r = std::sqrt(x*x+y*y);
  return std::complex<double>(1.0,  lambda/(2.0*M_PI)*(r > r_inner-1e-12 && r <= r_outer+1e-12)? (r-r_inner)/(r_outer-r_inner):0.0);
}

//Function2d2< std::complex<double>,double,double, double, double> UGFrom3DSlice::Set_pml(set_pml);

Variable2D< std::complex<double> > UGFrom3DSlice::getPMLlayer(Blockgrid2D& blockgrid, const double lambda) {
  assert(blockgrid.Give_unstructured_grid() == this);
  Variable2D< std::complex<double> > pml(blockgrid);
  X_coordinate2d X(blockgrid);
  Y_coordinate2d Y(blockgrid);

  r_inner = ug3Dp->Give_point(1)->Give_coordinate().x;
  r_outer = r_inner + PMLwidth;
  this->lambda = lambda;

  pml = PMLfunctor(X,Y);//Set_pml(X,Y,r_inner, r_outer, lambda);



  return pml;

}

PMLGrid_Marker2D_Inner UGFrom3DSlice::getInnerMarker() {
  return PMLGrid_Marker2D_Inner(*this);
}

PMLGrid_Marker2D_PML UGFrom3DSlice::getPMLMarker() {
  return PMLGrid_Marker2D_PML(*this);
}

////////////////////////////////////////////////////////////////////
// blockgrid
////////////////////////////////////////////////////////////////////



Blockgrid2DFrom3D::Blockgrid2DFrom3D(UGFrom3DSlice* ug2D, Blockgrid* bg3D) : Blockgrid2D(ug2D) {
  Unstructured_grid* ug3D = ug2D->Give_unstructured_grid();
  assert(bg3D->Give_unstructured_grid()==ug3D);
  
  int numberColors = ug2D->degree_of_freedom();
  assert(numberColors<=ug3D->degree_of_freedom());
  number_points.at(0) = bg3D->Give_N_color(1);  // muss hier vertauscht werden!!
  number_points.at(1) = bg3D->Give_N_color(0);
  

  if(numberColors == 3) {
     number_points.at(2) = bg3D->Give_N_color(ug3D->degree_of_freedom()-1);
  }
  else if(numberColors>3) {
    for(int i = 2; i < numberColors;++i)
      number_points.at(i) = bg3D->Give_N_color(i+1);
  }
  else {
     assert(numberColors==2);
  }  
};

Blockgrid2DFrom3D::Blockgrid2DFrom3D(UGFrom3DSlice* ug2D, const std::vector<int>& N_fine) : Blockgrid2D(ug2D) {
  Unstructured_grid* ug3D = ug2D->Give_unstructured_grid();

  int numberColors = ug2D->degree_of_freedom();

  for(int i = 0; i < numberColors; ++i)
  {
    if(i >= N_fine.size()) {
      number_points.at(i) = N_fine[N_fine.size()-1];
    } else {
      number_points.at(i) = N_fine[i];
    }
  }
};
