/**********************************************************************************
 * Copyright 2010 Christoph Pflaum
 *              Department Informatik Lehrstuhl 10 - Systemsimulation
 *              Friedrich-Alexander Universität Erlangen-Nürnberg
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "ug2D.h"
#include "blockgrid2D.h"
#include "gridFrom3D.h"
#include "../extemp/extemp2D.h"
#include "../../../common_source/abbrevi_common.h"
#include "../abbrevi2D.h"

#ifndef INTERPOLATION2D_H
#define INTERPOLATION2D_H

template<class A>
class ExpressionInterpolation2D;


template<class A,typename Result, elementTyp TYP_EL>
struct Interpolation2D_helper {
  static Result f(params2D_in, const ExpressionInterpolation2D<A>& ref);

  template<class A_>
  friend class ExpressionInterpolation2D;
};

/***
 * Objekt zum Interpolieren von Daten von einem Gitter auf das andere
 * Beispiel:
 *        Interpolation2D interpolator(grid_from, grid_to);
 *        Variable2D<double> uFrom(grid_from);
 *        Variable2D<double> uTo(grid_to);
 *            ...
 *        uTo = interpolator(uFrom);
 * 
 * grid_from und grid_to müssen gleiches UG haben ( gleiche Geometrie)
 * Interpolation durch 4 Punkte an denen Wert + Steigung bekannt ist.
 * 
 ***/
class Interpolation2D
{
  Blockgrid2D& grid_from;
  Blockgrid2D& grid_to;

  // for parallel
  int my_rank;
  MPI_Comm comm;

 public:
  Interpolation2D(Blockgrid2D& grid_from_, Blockgrid2D& grid_to_);

  Interpolation2D(Blockgrid2DFrom3D& grid_from_, Blockgrid2DFrom3D& grid_to_);

  template<class A>
  ExpressionInterpolation2D<A> operator() (const Expr2D<A>& a_) const;

  template<class A>
  friend class ExpressionInterpolation2D;

  template<class A_,typename Result, elementTyp TYP_EL>
  friend struct Interpolation2D_helper;

  bool isSlice;
};

template<class A>
class ExpressionInterpolation2D: public Expr2D< ExpressionInterpolation2D<A> >
{
private:
  const Interpolation2D& interpol;
  const A& a;

  ExpressionInterpolation2D(const Interpolation2D& interpol_, const A& a_);
public:
  typedef typename A::Result Result;

  stencil_typ Give_stencil_typ() const {return yes_stencil;}

  template <elementTyp TYP_EL>
  void Update(int id) const {
    Interpolation2D_helper<A,Result, TYP_EL>::update(*this,id);
  }

  template <elementTyp TYP_EL>
  Result Give_data( params2D_in ) const {
    return Interpolation2D_helper<A,Result, TYP_EL>::f(params2D_out,*this);
  }

  template<class A_,typename Result, elementTyp TYP_EL>
  friend struct Interpolation2D_helper;

  Blockgrid2D* Give_blockgrid() const {
    return &interpol.grid_to;
  }

  friend class Interpolation2D;
};

#endif
