/**********************************************************************************
 * Copyright 2010 Christoph Pflaum
 *              Department Informatik Lehrstuhl 10 - Systemsimulation
 *              Friedrich-Alexander Universität Erlangen-Nürnberg
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "ug2D.h"
#include "blockgrid2D.h"
#include "elements2D.h"
#include "../extemp/extemp2D.h"
#include "../../../common_source/abbrevi_common.h"

#ifndef _NOARRAYLIB

#ifdef CPP11
  #include <array>
  using std::array;
#else
  #ifdef WIN32
    #include <array>
    using std::array;
  #else
    #include <tr1/array>
using std::tr1::array;
  #endif
#endif
  
#endif // _NOARRAYLIB  
  
#include <numeric>

template<class A,typename Result>
struct Interpolation2D_helper<A,Result,rectangleEl> {

  static int getNeighborX(const ExpressionInterpolation2D<A>& ref, int id, int x_index) {
    RectangleEl& element = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(id));
    if(x_index == -1) {
      const int edgeId = element.Give_id_edge(Wdir2D);
      EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
      if(edge.isAtBoundary()) return -1;
      else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
    }
    if(x_index == -2) {
      const int edgeId = element.Give_id_edge(Edir2D);
      EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
      if(edge.isAtBoundary()) return -1;
      else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
    }
    assert(-1);
  }

  static int getNeighborY(const ExpressionInterpolation2D<A>& ref, int id, int y_index) {
    RectangleEl& element = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(id));
    if(y_index == -1) {
      const int edgeId = element.Give_id_edge(Sdir2D);
      EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
      if(edge.isAtBoundary()) return -1;
      else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
    }
    if(y_index == -2) {
      const int edgeId = element.Give_id_edge(Ndir2D);
      EdgeEl edge = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_edge(edgeId));
      if(edge.isAtBoundary()) return -1;
      else return edge.getIdNextRectangle(0)!=id?edge.getIdNextRectangle(0):edge.getIdNextRectangle(1);
    }
    assert(-1);
  }

  static inline void update(const ExpressionInterpolation2D<A>& ref, int id) {
    if(id < ref.interpol.grid_from.Give_unstructured_grid()->Give_number_rectangle())
      ref.a.template Update<rectangleEl>(id);
  }

  static inline Result f(int id, int i_to, int j_to, int Nx, const ExpressionInterpolation2D<A>& ref) {
    if(id >= ref.interpol.grid_from.Give_unstructured_grid()->Give_number_rectangle())
    {
      id-=4;
      i_to =   ref.interpol.grid_to.Give_Nx_rectangle(id);
    }
    const int Nx_from = ref.interpol.grid_from.Give_Nx_rectangle(id);
    const int Ny_from = ref.interpol.grid_from.Give_Ny_rectangle(id);
    const int Nx_to =   ref.interpol.grid_to.Give_Nx_rectangle(id);
    const int Ny_to =   ref.interpol.grid_to.Give_Ny_rectangle(id);

    const double i =  double(i_to) * double(Nx_from)/double(Nx_to);
    const double j =  double(j_to) * double(Ny_from)/double(Ny_to);

    ref.a.template Update < rectangleEl > (id);

    int ii[4];
    ii[1] = (floor(i)<Nx_from)?floor(i):(floor(i)-1);
    ii[2] = ii[1] + 1;
    ii[0] = (ii[1] > 0) ? ii[1] - 1 : -1;
    ii[3] = (ii[2] < Nx_from) ? ii[2] + 1 : -2;
    double weight_i =i-ii[1];
    int jj[4];
    jj[1] = (floor(j)<Ny_from)?floor(j):(floor(j)-1);
    jj[2] = jj[1] + 1;
    jj[0] = (jj[1] > 0) ? jj[1] - 1 : -1;
    jj[3] = (jj[2] < Ny_from) ? jj[2] + 1 : -2;
    double weight_j = j-jj[1];
    D2vector coord_t = ref.interpol.grid_to.Give_coord_rectangle(id,i_to,j_to);

    RectangleEl& me = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(id));

    Result data[4][4];
    D2vector coord[4][4];
    for (int ii_i = 0; ii_i < 4; ++ii_i) {
      for (int jj_i = 0; jj_i < 4; ++jj_i) {
        if(ii[ii_i] >= 0 && jj[jj_i] >= 0) {
          data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id, ii[ii_i], jj[jj_i], Nx_from);
          coord[jj_i][ii_i] = ref.interpol.grid_from.Give_coord_rectangle(id, ii[ii_i], jj[jj_i]);
        } else if(ii[ii_i] < 0 && jj[jj_i] >= 0) {
          int neighbor_x = Interpolation2D_helper<A,Result,rectangleEl>::getNeighborX(ref,id,ii[ii_i]);

          int edgeID;
          if(neighbor_x >=0) {
            //ref.a.Update<rectangleEl>(neighbor_x);
            int index_y = jj[jj_i];
            if(ii[ii_i]==-1) {
              edgeID = me.Give_id_edge(Wdir2D);
              index_y = Ny_from - index_y;
            } else {
              edgeID = me.Give_id_edge(Edir2D);
            }
            //if(me.getOrientationEdge(edgeID))
            data[jj_i][ii_i] = getNeighborEdgeValue(ref,neighbor_x,edgeID,index_y);
            coord[jj_i][ii_i] = getNeighborEdgeCoordinate(ref,neighbor_x,edgeID,index_y);
          }
          else {
            //Boundary point, no neighbor exists
            if(ii[ii_i]==-1) {
              data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id,0, jj[jj_i], Nx_from);
              coord[jj_i][ii_i] = ref.interpol.grid_from.Give_coord_rectangle(id,0, jj[jj_i]);

            } else {
              data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id, (ref.interpol.grid_from.Give_Nx_rectangle(id)) , jj[jj_i], Nx_from);
              coord[jj_i][ii_i] = ref.interpol.grid_from.Give_coord_rectangle(id, (ref.interpol.grid_from.Give_Nx_rectangle(id)) , jj[jj_i]);
            }
          }
        } else if(ii[ii_i] >= 0 && jj[jj_i] < 0) {
          int neighbor_y = Interpolation2D_helper<A,Result,rectangleEl>::getNeighborY(ref,id,jj[jj_i]);

          int edgeID;
          if(neighbor_y >=0) {
            int index_x = ii[ii_i];
  //            ref.a.Update<rectangleEl>(neighbor_y);
            if(jj[jj_i]==-1) {
              edgeID = me.Give_id_edge(Sdir2D);
            } else {
              edgeID = me.Give_id_edge(Ndir2D);
              index_x = Nx_from - index_x;
            }

            data[jj_i][ii_i] = getNeighborEdgeValue(ref,neighbor_y,edgeID,index_x);
            coord[jj_i][ii_i] = getNeighborEdgeCoordinate(ref,neighbor_y,edgeID,index_x);

          }
          else {
            //Boundary point, no neighbor exists
            if(ii[ii_i]==-1) {
              data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id,0, jj[jj_i], Nx_from);
              coord[jj_i][ii_i] = ref.interpol.grid_from.Give_coord_rectangle(id,0, jj[jj_i]);
            } else {
              data[jj_i][ii_i] = ref.a.template Give_data < rectangleEl > (id, (ref.interpol.grid_from.Give_Ny_rectangle(id)) , jj[jj_i], Nx_from);
              coord[jj_i][ii_i] = ref.interpol.grid_from.Give_coord_rectangle(id, (ref.interpol.grid_from.Give_Ny_rectangle(id)) , jj[jj_i]);
            }
          }
        } else {
          int neighbor_x = getNeighborX(ref,id,ii[ii_i]);
          int neighbor_y = getNeighborY(ref,id,jj[jj_i]);
          Result data_x;
          Result data_y;
          D2vector coord_x;
          D2vector coord_y;

          int index_x = (ii[ii_i] == -2)?Nx_from:0;
          int index_y = (jj[jj_i] == -2)?Ny_from:0;
          if(neighbor_x >=0) {
  //            ref.a.Update<rectangleEl>(neighbor_x);
            int edgeID;

            if(ii[ii_i]==-1) {
              edgeID = me.Give_id_edge(Wdir2D);
              index_y = Ny_from - index_y;
            }
            else {
              edgeID = me.Give_id_edge(Edir2D);
            }
            data_x = getNeighborEdgeValue(ref,neighbor_x,edgeID,index_y);
            coord_x = getNeighborEdgeCoordinate(ref,neighbor_x,edgeID,index_y);
          } else {
            if(ii[ii_i]==-1) {
              data_x = ref.a.template Give_data < rectangleEl > (id,0, index_y, Nx_from);
              coord_x = ref.interpol.grid_from.Give_coord_rectangle(id,0, index_y);
            } else {
              data_x = ref.a.template Give_data < rectangleEl > (id,Nx_from, index_y, Nx_from);
              coord_x = ref.interpol.grid_from.Give_coord_rectangle(id,Nx_from, index_y);
            }
          }
          if(neighbor_y >=0) {
  //            ref.a.Update<rectangleEl>(neighbor_y);
            int edgeID;

            if(jj[jj_i]==-1) {
              edgeID = me.Give_id_edge(Sdir2D);
            } else {
              edgeID = me.Give_id_edge(Ndir2D);
              index_x = Nx_from - index_x;
            }
            data_y = getNeighborEdgeValue(ref,neighbor_y,edgeID,index_x);
            coord_y = getNeighborEdgeCoordinate(ref,neighbor_y,edgeID,index_x);
          } else {
            if(jj[jj_i]==-1) {
              data_y = ref.a.template Give_data < rectangleEl > (id,0, index_x, Nx_from);
              coord_y = ref.interpol.grid_from.Give_coord_rectangle(id,0, index_x);
            } else {
              data_y = ref.a.template Give_data < rectangleEl > (id,Nx_from, index_x, Nx_from);
              coord_y = ref.interpol.grid_from.Give_coord_rectangle(id,Nx_from, index_x);
            }
          }
          data[jj_i][ii_i] = 0.5*(data_x+data_y);
          coord[jj_i][ii_i] = 0.5*(coord_x+coord_y);
        }
      }
    }



    return QubicQuadInterpolate(data, coord, weight_i, weight_j);
  }

  static inline Result getNeighborEdgeValue(const ExpressionInterpolation2D<A>& ref, int NeighborId, int edgeID, int index1) {
    RectangleEl& rectangle = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(NeighborId));

    if(rectangle.Give_id_edge(Wdir2D) == edgeID) {
      const int index2 = 1;
      return ref.a.template Give_data < rectangleEl > (NeighborId,index2,index1,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
    } else if(rectangle.Give_id_edge(Edir2D) == edgeID) {
      const int index2 = (ref.interpol.grid_from.Give_Nx_rectangle(NeighborId))-1;
      index1 = ref.interpol.grid_from.Give_Ny_rectangle(NeighborId) - index1;
      return ref.a.template Give_data < rectangleEl > (NeighborId,index2,index1,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
    } else if(rectangle.Give_id_edge(Sdir2D) == edgeID) {
      const int index2 = 1;
      index1 = ref.interpol.grid_from.Give_Nx_rectangle(NeighborId) - index1;
      return ref.a.template Give_data < rectangleEl > (NeighborId,index1,index2,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
    } else if(rectangle.Give_id_edge(Ndir2D) == edgeID) {
      const int index2 = (ref.interpol.grid_from.Give_Ny_rectangle(NeighborId)-1);
      return ref.a.template Give_data < rectangleEl > (NeighborId,index1,index2,ref.interpol.grid_from.Give_Nx_rectangle(NeighborId));
    }
    else {
      std::cout << "Fehler" << std::endl;
    }
  }

  static inline D2vector getNeighborEdgeCoordinate(const ExpressionInterpolation2D<A>& ref, int NeighborId, int edgeID, int index1) {
    RectangleEl& rectangle = *(ref.interpol.grid_from.Give_unstructured_grid()->Give_rectangle(NeighborId));

    if(rectangle.Give_id_edge(Edir2D) == edgeID) {
      const int index2 = (ref.interpol.grid_from.Give_Nx_rectangle(NeighborId))-1;
      index1 = ref.interpol.grid_from.Give_Ny_rectangle(NeighborId) - index1;
      return ref.interpol.grid_from.Give_coord_rectangle(NeighborId,index2,index1);
    } else if(rectangle.Give_id_edge(Wdir2D) == edgeID) {
      const int index2 = 1;
      return ref.interpol.grid_from.Give_coord_rectangle(NeighborId,index2,index1);
    } else if(rectangle.Give_id_edge(Sdir2D) == edgeID) {
      const int index2 = 1;
      index1 = ref.interpol.grid_from.Give_Nx_rectangle(NeighborId) - index1;
      return ref.interpol.grid_from.Give_coord_rectangle(NeighborId,index1,index2);
    } else if(rectangle.Give_id_edge(Ndir2D) == edgeID) {
      const int index2 = (ref.interpol.grid_from.Give_Ny_rectangle(NeighborId)-1);
      return ref.interpol.grid_from.Give_coord_rectangle(NeighborId,index1,index2);
    }
    else {
      std::cout << "Fehler" << std::endl;
    }
  }

  static inline Result QubicQuadInterpolate(Result data[4][4], D2vector coord[4][4], double x, double y) {
#ifdef _NOARRAYLIB
    return Result(0.0);       
#else
    for(int i = 0; i < 4; ++i) {
      if(ABS(coord[i][1]-coord[i][0]) > 1e-8)
        data[i][0] = 0.5*(data[i][1]-data[i][0])/ABS(coord[i][1]-coord[i][0])*ABS(coord[i][2]-coord[i][1])+0.5*(data[i][2]-data[i][1]);
      else
        data[i][0] = (data[i][2]-data[i][1])/ABS(coord[i][2]-coord[i][1]);

      if(ABS(coord[i][3]-coord[i][2]) > 1e-8)
        data[i][3] = 0.5*(data[i][2]-data[i][1])+0.5*(data[i][3]-data[i][2])/ABS(coord[i][3]-coord[i][2])*ABS(coord[i][2]-coord[i][1]);
      else
        data[i][3] = (data[i][2]-data[i][1])/ABS(coord[i][2]-coord[i][1]);
    }

    Result data_x[4];
    D2vector coord_x[4];
    array< array<double,4>,4 > A_inv = {{{2.,-2.,1.,1.},{-3.,3.,0-2.,-1.},{0.,0.,1.,0.},{1.,0.,0.,0.}}};
    for(int i = 0; i < 4; ++i) {
      array<Result,4> vec_f = {data[i][1],data[i][2],data[i][0],data[i][3]};
      array<double,4> vec_x = {x*x*x,x*x,x,1};
      array<Result,4> par;
      for(int k = 0; k < 4; ++k) {
        par[k] = std::inner_product(vec_f.begin(),vec_f.end(),A_inv[k].begin(),Result(0.0));
      }
      data_x[i] = std::inner_product(vec_x.begin(),vec_x.end(),par.begin(),Result(0.0));
      coord_x[i] = (1.0-x)*coord[i][1] + x*coord[i][2];
    }


    if(ABS(coord_x[1]-coord_x[0]) > 1e-8)
      data_x[0] = 0.5*(data_x[1]-data_x[0])/ABS(coord_x[1]-coord_x[0])*ABS(coord_x[2]-coord_x[1])+0.5*(data_x[2]-data_x[1]);
    else
      data_x[0] = 0.0; //(data_x[2]-data_x[1])/ABS(coord_x[2]-coord_x[1]);

    if(ABS(coord_x[3]-coord_x[2]) > 1e-8)
      data_x[3] = 0.5*(data_x[3]-data_x[2])/ABS(coord_x[3]-coord_x[2])*ABS(coord_x[2]-coord_x[1])+0.5*(data_x[2]-data_x[1]);
    else
      data_x[3] = (data_x[2]-data_x[1])/ABS(coord_x[2]-coord_x[1]);

    array<Result,4> vec_f = {data_x[1],data_x[2],data_x[0],data_x[3]};
    array<double,4> vec_y = {y*y*y,y*y,y,1};
    array<Result,4> par;
    for(int k = 0; k < 4; ++k) {
      par[k] = std::inner_product(vec_f.begin(),vec_f.end(),A_inv[k].begin(),Result(0.0));
    }
    return std::inner_product(vec_y.begin(),vec_y.end(),par.begin(),Result(0.0));
#endif // _NOARRAYLIB
  }
};


template<class A,typename Result>
struct Interpolation2D_helper<A,Result,edgeEl> {
  static inline void update(const ExpressionInterpolation2D<A>& ref, int id) {
    if(id < ref.interpol.grid_from.Give_unstructured_grid()->Give_number_edges())
      ref.a.template Update<edgeEl>(id);
  }

  static Result f(params2D_in, const ExpressionInterpolation2D<A>& ref) {
    if(id >= ref.interpol.grid_from.Give_unstructured_grid()->Give_number_edges())
    {
      //std::cout << "Id=" << id << ", i=" << i << "-->";
      EdgeEl& edge = *ref.interpol.grid_to.Give_unstructured_grid()->Give_edge(id);
      int pointW = edge.Give_id_corner_W();
      int pointE = edge.Give_id_corner_E();

      int pointWinner = pointW>7?(pointW-8)*2+1:pointW-1;
      int pointEinner = pointE>7?(pointE-8)*2+1:pointE-1;

      //std::cout << "pointW=" << pointW << ", pointE=" << pointE << "-->pointW=" << pointWinner << ", pointE=" << pointEinner << std::endl;
      bool test;
      id = ref.interpol.grid_to.Give_unstructured_grid()->getEdge(pointWinner,pointEinner, &test);
      const int Nx = ref.interpol.grid_to.Give_Nx_edge(id);
      if(pointWinner%2 == 0 || pointEinner%2 == 0) {
        return ref.a.template Give_data<pointEl>(pointEinner,0,0,0);
        //i = Nx-1;
      } else {
        const int Nx_from = ref.interpol.grid_from.Give_Nx_edge(id);
        const int Nx_to = ref.interpol.grid_to.Give_Nx_edge(id);

        double ii = fabs((i/(double)Nx_to) * Nx_from);
        double weight_i = ii-floor(ii);
        return (1.-weight_i)*(ref.a.template Give_data<edgeEl>(id,(int)floor(ii),j,Nx_from)) + (weight_i)*(ref.a.template Give_data<edgeEl>(id,(int)ceil(ii),j,Nx_from));
      }
//      EdgeEl& newEdge = *ref.interpol.grid_to.Give_unstructured_grid()->Give_edge(id);
//      pointW = newEdge.Give_id_corner_W();
//      pointE = newEdge.Give_id_corner_E();
      //std::cout << "pointW=" << pointW << ", pointE=" << pointE << std::endl;

      //std::cout << "Id=" << id << ", i=" << i << std::endl;


    }
    else
    {
      const int Nx_from = ref.interpol.grid_from.Give_Nx_edge(id);
      const int Nx_to = ref.interpol.grid_to.Give_Nx_edge(id);
      double ii = fabs((i/(double)Nx_to) * Nx_from);
      double weight_i = ii-floor(ii);
      return (1.-weight_i)*(ref.a.template Give_data<edgeEl>(id,(int)floor(ii),j,Nx_from)) + (weight_i)*(ref.a.template Give_data<edgeEl>(id,(int)ceil(ii),j,Nx_from));
    }
  }
};
template<class A,typename Result>
struct Interpolation2D_helper<A,Result,pointEl> {

  static inline void update(const ExpressionInterpolation2D<A>& ref, int id) {
  if(id < ref.interpol.grid_from.Give_unstructured_grid()->Give_number_points())
    ref.a.template Update<pointEl>(id);
  }

  static Result f(params2D_in, const ExpressionInterpolation2D<A>& ref) {
    if(id >= ref.interpol.grid_from.Give_unstructured_grid()->Give_number_points())
    {
      id = (id-8)*2+1;
    }
    return ref.a.template Give_data<pointEl>(id,i,j,Nx);
  }
};

template<class A>
ExpressionInterpolation2D<A>
  Interpolation2D::operator()(const Expr2D<A>& a_) const
  {
    const A& a(a_);
    return ExpressionInterpolation2D < A > (*this, a);
  }

template<class A>
ExpressionInterpolation2D<A>::ExpressionInterpolation2D(const Interpolation2D& interpol_, const A& a_)
      : interpol(interpol_), a(a_)
  {
  }
