/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "elements2D.h"
#include "../grid/parti2D.h"
#include "ug2D.h"
#include "marker2D.h"

#include <algorithm>
#include <iomanip>


using namespace std;

//////////////////////////////////////////////////////////////////
// other member functions
//////////////////////////////////////////////////////////////////



D2vector transform2D_nothing(double x, double* gd) {
  return D2vector(0.0,0.0);
}

D2vector transform2D_nothing(double x, double y) {
  return D2vector(0.0,0.0);
}

Unstructured2DGrid::Unstructured2DGrid() {
  construction_rectangle_points_done_yn = false;
  construction_done_yn                   = false;

  points.clear();     points.resize(0);     
  edges.clear();      edges.resize(0); 
  rectanglEs.clear(); rectanglEs.resize(0);   

  num_points = 0;
  num_edges = 0;
  num_rectanglEs  = 0;

  pointer_global_data = NULL;
  
  number_degree_of_freedom = 0;

  all_points = NULL;
}

Unstructured2DGrid::~Unstructured2DGrid() {
  //>>>>>>>>>>>ok NEdir2D
  
  delete [] pointer_global_data;
  delete all_points;
}


bool Unstructured2DGrid::inDomain(const D2vector& point_coord) {
  if((point_coord.x < minCorX) || (point_coord.x > maxCorX)) return false;
  if((point_coord.y < minCorY) || (point_coord.y > maxCorY)) return false;
  return true;
}


void Unstructured2DGrid::Print() {
    cout << " Print data UG: " << endl << " ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ " << endl;
    cout <<  points.size() << " Points: " << endl;
    for(int i=0;i<points.size();++i) {
      cout << " point: " << i << endl;
      points.at(i).Print();
    }
    
    cout << endl << edges.size() << " Edges: " << endl;
    for(int i=0;i<edges.size();++i) {
       cout << " edge: " << i << endl;
       edges.at(i).Print();
    } 
    cout << endl << rectanglEs.size() << " Rectangles: " << endl;
    for(int i=0;i<rectanglEs.size();++i) {
      cout << " rectangles: " << i << endl;
      rectanglEs.at(i).Print();
    }
}
    
    
void Unstructured2DGrid::Set_number_points(int num_points_) {
  num_points = num_points_;
  assert(num_points>0);
  points.resize(num_points);
}

void Unstructured2DGrid::Set_coordinate_point(int id, D2vector coord) {
  points.at(id).Set_coordinate(coord);
}

void Unstructured2DGrid::Set_number_rectanglEs(int num_rectanglEs_) {
  // Set informations of rectanglEs
  num_rectanglEs = num_rectanglEs_;
  assert(num_rectanglEs>0);
  rectanglEs.resize(num_rectanglEs);
}

void Unstructured2DGrid::Set_rectangle(int id,
				      int i_WS, int i_ES,
				      int i_WN, int i_EN) {
  assert(construction_rectangle_points_done_yn==false);

  // AnwNEdirer muss sehr schlau sein beim AnwNEdiren dieser
  // Funktion, da sonst Orientierung nicht stimmt.

  rectanglEs[id].Set_id_corner(SWdir2D,i_WS);
  rectanglEs[id].Set_id_corner(SEdir2D,i_ES);
  rectanglEs[id].Set_id_corner(NWdir2D,i_WN);
  rectanglEs[id].Set_id_corner(NEdir2D,i_EN);

  
  rectanglEs[id].Set_coord(SWdir2D,points[i_WS].Give_coordinate());
  rectanglEs[id].Set_coord(SEdir2D,points[i_ES].Give_coordinate());
  rectanglEs[id].Set_coord(NWdir2D,points[i_WN].Give_coordinate());
  rectanglEs[id].Set_coord(NEdir2D,points[i_EN].Give_coordinate());
}


/*

void Calculate_corresponding_edges(dir3D_sons corner, Edges_cell& SD_edge,
                                   Edges_cell& WD_edge, Edges_cell& SW_edge){
  if(corner == SWdird){
     SD_edge = SDed;
     WD_edge = WDed;
     SW_edge = SWed;
  }
  else if(corner == SEdird){
     SD_edge = EDed;
     WD_edge = SDed;
     SW_edge = SEed;
  }
  else if(corner == WNDdir){
     SD_edge = WDed;
     WD_edge = NDed;
     SW_edge = NWed;
  }
  else if(corner == NEdird){
     SD_edge = NDed;
     WD_edge = EDed;
     SW_edge = NEed;
  }
  else if(corner == WSTdir3d){
     SD_edge = WTed;
     WD_edge = STed;
     SW_edge = SWed;
  }
  else if(corner == ESTdir3d){
     SD_edge = STed;
     WD_edge = ETed;
     SW_edge = SEed;
  }
  else if(corner == WNTdir3d){
     SD_edge = NTed;
     WD_edge = WTed;
     SW_edge = NWed;
  }
  else{
     if(developer_version)
        if(corner != ENTdir3d) cout << "error in Calculate_corresponding_edges!" << endl;
     SD_edge = ETed;
     WD_edge = NTed;
     SW_edge = NEed;
  }
}
*/

int Unstructured2DGrid::getEdge(int id_left, int id_right, bool* edgeExists) {
  int id_edge = points.at(id_left).getEdge(id_right);
  if(id_edge < 0) {  // neue Kante erzeugen
     *edgeExists = false;
    ++num_edges;
    points.at(id_left).addEdge(id_right, num_edges-1);
    points.at(id_right).addEdge(id_left, num_edges-1);
    return num_edges-1;
  }
  // alte Kante ok!
  *edgeExists = true;
  return id_edge;
}

// NW   NE
// SW   SE
void Unstructured2DGrid::construction_rectangle_points_done() {
  construction_rectangle_points_done_yn = true;

  // a) Iterieren um Kanten zu zählen
  bool edgeExists;
  for(int i=0;i<rectanglEs.size();++i) {
    int idSW = rectanglEs.at(i).Give_id_corner(SWdir2D);
    int idSE = rectanglEs.at(i).Give_id_corner(SEdir2D);
    int idNW = rectanglEs.at(i).Give_id_corner(NWdir2D);
    int idNE = rectanglEs.at(i).Give_id_corner(NEdir2D);

    
    // a) Analyse Kanten
     rectanglEs.at(i).Set_id_edge(Wdir2D, getEdge(idSW,idNW,
						  &edgeExists));
     if(edgeExists) rectanglEs.at(i).edgeIsOtherOrientation(Wdir2D);

     rectanglEs.at(i).Set_id_edge(Ndir2D, getEdge(idNW,idNE,
						  &edgeExists));
     if(edgeExists) rectanglEs.at(i).edgeIsOtherOrientation(Ndir2D);

     rectanglEs.at(i).Set_id_edge(Edir2D, getEdge(idNE,idSE,
						  &edgeExists));
     if(edgeExists) rectanglEs.at(i).edgeIsOtherOrientation(Edir2D);

     rectanglEs.at(i).Set_id_edge(Sdir2D, getEdge(idSE,idSW,
						  &edgeExists));
     if(edgeExists) rectanglEs.at(i).edgeIsOtherOrientation(Sdir2D);
 
     // b) Rechteck zu Punkten hinzufügen    
     //    Form       
     //             3    4        NW  NE 
     //             1    2        SW  SE
     points.at(rectanglEs.at(i).Give_id_corner(SWdir2D)).add_rectangle(Id_of_rec2D(idSE,idNW,idNE),i);
     //             4    2        NW  NE
     //             3    1        SW  SE
     points.at(rectanglEs.at(i).Give_id_corner(SEdir2D)).add_rectangle(Id_of_rec2D(idNE,idSW,idNW),i);
     //             2    1        NW  NE
     //             4    3        SW  SE
     points.at(rectanglEs.at(i).Give_id_corner(NEdir2D)).add_rectangle(Id_of_rec2D(idNW,idSE,idSW),i);
     //             1    3        NW  NE
     //             2    4        SW  SE
     points.at(rectanglEs.at(i).Give_id_corner(NWdir2D)).add_rectangle(Id_of_rec2D(idSW,idNE,idSE),i);
  }
  
  // b) Kanten erzeugen und setzen Daten: Ecken, Rec, boundary info
  edges.resize(num_edges);
  for(int i=0;i<rectanglEs.size();++i) {
    int idSW = rectanglEs.at(i).Give_id_corner(SWdir2D);
    int idSE = rectanglEs.at(i).Give_id_corner(SEdir2D);
    int idNW = rectanglEs.at(i).Give_id_corner(NWdir2D);
    int idNE = rectanglEs.at(i).Give_id_corner(NEdir2D);

    // West-Kante
    if(rectanglEs.at(i).getOrientationEdge(Wdir2D)) { // das hiesst Orientierung umgedreht
       edges.at(rectanglEs.at(i).Give_id_edge(Wdir2D)).setSecond_Id_Rectangle(i);
    }
    else {
       edges.at(rectanglEs.at(i).Give_id_edge(Wdir2D)).setFirst_Id_Rectangle(i);
       edges.at(rectanglEs.at(i).Give_id_edge(Wdir2D)).Set_id_corners(idSW,idNW);
    }

    // Nord-Kante
    if(rectanglEs.at(i).getOrientationEdge(Ndir2D)) { // das hiesst Orientierung umgedreht
       edges.at(rectanglEs.at(i).Give_id_edge(Ndir2D)).setSecond_Id_Rectangle(i);
    }
    else {
       edges.at(rectanglEs.at(i).Give_id_edge(Ndir2D)).setFirst_Id_Rectangle(i);
       edges.at(rectanglEs.at(i).Give_id_edge(Ndir2D)).Set_id_corners(idNW,idNE);
    }
    
    // Ost-Kante
    if(rectanglEs.at(i).getOrientationEdge(Edir2D)) { // das hiesst Orientierung umgedreht
       edges.at(rectanglEs.at(i).Give_id_edge(Edir2D)).setSecond_Id_Rectangle(i);
    }
    else {
       edges.at(rectanglEs.at(i).Give_id_edge(Edir2D)).setFirst_Id_Rectangle(i);
       edges.at(rectanglEs.at(i).Give_id_edge(Edir2D)).Set_id_corners(idNE,idSE);
    }
    
    // Sued-Kante
    if(rectanglEs.at(i).getOrientationEdge(Sdir2D)) { // das hiesst Orientierung umgedreht
       edges.at(rectanglEs.at(i).Give_id_edge(Sdir2D)).setSecond_Id_Rectangle(i);
    }
    else {
       edges.at(rectanglEs.at(i).Give_id_edge(Sdir2D)).setFirst_Id_Rectangle(i);
       edges.at(rectanglEs.at(i).Give_id_edge(Sdir2D)).Set_id_corners(idSE,idSW);
    }
  }  
  // c) Setze neue Orientierung der Kanten
  for(int i=0;i<rectanglEs.size();++i) {
    int idSW = rectanglEs.at(i).Give_id_corner(SWdir2D);
    int idSE = rectanglEs.at(i).Give_id_corner(SEdir2D);
    int idNW = rectanglEs.at(i).Give_id_corner(NWdir2D);
    int idNE = rectanglEs.at(i).Give_id_corner(NEdir2D);

    // West-Kante
    if(edges.at(rectanglEs.at(i).Give_id_edge(Wdir2D)).Give_id_corner(Wd)==idSW) {
       rectanglEs.at(i).edgeIsNormalOrientation(Wdir2D);
    }
    else {
       rectanglEs.at(i).edgeIsOtherOrientation(Wdir2D);
    }
    
    // Nord-Kante
    if(edges.at(rectanglEs.at(i).Give_id_edge(Ndir2D)).Give_id_corner(Wd)==idNW) {
       rectanglEs.at(i).edgeIsNormalOrientation(Ndir2D);
    }
    else {
       rectanglEs.at(i).edgeIsOtherOrientation(Ndir2D);
    }    

    // Ost-Kante
    if(edges.at(rectanglEs.at(i).Give_id_edge(Edir2D)).Give_id_corner(Wd)==idSE) {
       rectanglEs.at(i).edgeIsNormalOrientation(Edir2D);
    }
    else {
       rectanglEs.at(i).edgeIsOtherOrientation(Edir2D);
    }    

    // Sued-Kante
    if(edges.at(rectanglEs.at(i).Give_id_edge(Sdir2D)).Give_id_corner(Wd)==idSW) {
       rectanglEs.at(i).edgeIsNormalOrientation(Sdir2D);
    }
    else {
       rectanglEs.at(i).edgeIsOtherOrientation(Sdir2D);
    }    
  }
  
  // d) Setze Koordinaten der Rectangles
  for(int i=0;i<rectanglEs.size();++i) {
    for(int c=0;c<4;++c) {
        rectanglEs.at(i).Set_coord((dir2D_sons)c,
		points.at(rectanglEs.at(i).Give_id_corner((dir2D_sons)c)).Give_coordinate());
    }
  }
  
  // e) Setze Koordinaten der Kanten
  for(int i=0;i<edges.size();++i) {
    for(int c=0;c<2;++c) {
        edges.at(i).Set_coord((dir1d)c, 
		points.at(edges.at(i).Give_id_corner((dir1d)c)).Give_coordinate());
    }
  }
  
  // f) Setze lokale Id's in Rectangles
  for(int i=0;i<points.size();++i) {
//test NOW    
      points.at(i).addLocalPointInfosToRectangles();
  }
}




void Unstructured2DGrid::construction_done() {
  int  color, num;
  int id_first_colored_edge;

  assert(construction_rectangle_points_done_yn);

  // I. color edges with respect to degree of freedom
  color = 0;
  for(id_first_colored_edge=-2;id_first_colored_edge!=-1;++color) {
    id_first_colored_edge=-1;
    // 1. Find not colored edge
    for(num=0;num<edges.size() && id_first_colored_edge==-1;++num) {
      if(edges.at(num).Give_color_of_edge()==-1) {
	edges.at(num).Set_color_of_edge(color);
	id_first_colored_edge = num;
      }
    }

    // 2. find edges with same color
    if(id_first_colored_edge!=-1)
      Set_recursive_color(id_first_colored_edge,color);
  }
  number_degree_of_freedom = color-1;

  // II. set transformation edges in rectanglEs
  for(int id_rec=0;id_rec<num_rectanglEs;++id_rec) {
    for(int i=0;i<4;++i) if(rectanglEs[id_rec].transformer[i] == NULL) {
      // set transformation function
      rectanglEs[id_rec].transformer[i] =
	      edges[rectanglEs[id_rec].Give_id_edge((dir2D)i)].transformer;
    };
  }

  // III. set all points
  all_points = new Marker2D(this);
  
  // IV. max min Coordinates for convex domain
  D2vector vv = Give_point(0)->Give_coordinate();  
    
  minCorX = vv.x;
  minCorY = vv.y;
  maxCorX = vv.x;
  maxCorY = vv.y;

  for(int nump=0; nump < Give_number_points(); ++nump) {
      D2vector v = Give_point(nump)->Give_coordinate();
      minCorX = MIN(minCorX,v.x);
      maxCorX = MAX(maxCorX,v.x);
      minCorY = MIN(minCorY,v.y);
      maxCorY = MAX(maxCorY,v.y);
  }
}

void Unstructured2DGrid::Set_recursive_color(int id_edge, int color) {
  // iterate neighbors of edge
  for(int i=0; i<edges.at(id_edge).getNumberNextRectangles();++i) {
      int id_rec = edges.at(id_edge).getIdNextRectangle(i);
      // iterate edges of rectangle with edge id_edge
      for(int ed=0;ed<4;++ed) {
	// find colored edge
	if(rectanglEs.at(id_rec).Give_id_edge((dir2D)ed)==id_edge) {
	   int id_edge_next = rectanglEs.at(id_rec).Give_id_edge(oppositeEdge((dir2D)ed));
	   if(edges.at(id_edge_next).Give_color_of_edge()==-1) {
              // set color and recursive call
              edges.at(id_edge_next).Set_color_of_edge(color);
	      Set_recursive_color(id_edge_next,color);
	   }
	}
      }
  }
}

void Unstructured2DGrid::Set_transformation_edge(int idL, int idR,  EdgeTransformator2D* transformer) {
  // transform must be zero vector at 0,1
  // transform: [0,1] ->  D2vector
  assert(construction_rectangle_points_done_yn==true);
  assert(construction_done_yn==false);
  assert(idL < idR);

  for(int i=0;i<edges.size();++i) {
    int idLed = edges.at(i).Give_id_corner(Wd);
    int idRed = edges.at(i).Give_id_corner(Ed);

    if((idL==idLed && idR==idRed) || (idR==idLed && idL==idRed)) {
        edges[i].transformer = transformer;
    }
  }
}


