/**********************************************************************************
 * Copyright 2013 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#ifndef UG2D_H
#define UG2D_H
// next two includes for exception handling
#include <stdexcept>
#include <string>
//#include <list>
#include <vector>

#include "parti2D.h"
#include "elements2D.h"


////////////////////////////////////////////////////////////////
//
//
//  1. class Unstructured2DGrid and other classes
//           main class describing an unstructured grid
//
//  2. some inline functions
//
////////////////////////////////////////////////////////////////


class Marker2D;
class Unstructured2DGrid;
class Blockgrid2D;

//////////////////////////////////////////////////////////////////
//
//  1. class Unstructured2DGrid and other classes
//           main class describing an unstructured grid
//////////////////////////////////////////////////////////////////


class Unstructured2DGrid : public Partitioning2D {
  public:
    Unstructured2DGrid();
    ~Unstructured2DGrid();

    void Print();
    
    virtual void Test() const = 0;

    /**
   * inverse computation
   * @param point_coord input getCoordinate
   * @param id id des rectangle
   * @param i,j indizes der Koordinaten
   * @return true falls Punkt im Gebiet, sonst false  
   **/ 
    virtual bool getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) 
                 const { std::cout << " not implemented" << std::endl; return false; } 
    
    virtual bool inDomain(const D2vector& point_coord); 
    
    double MinimumX() { return minCorX; };
    double MinimumY() { return minCorY; };
    double MaximumX() { return maxCorX; };
    double MaximumY() { return maxCorY; };
    
    
    // for parallel
    void Make_Partitioning2D(MPI_Comm com);

    void Set_number_points(int num_points_);
    int  Give_number_points() { return num_points; }
    void Set_coordinate_point(int id, D2vector coord);

    void Set_number_rectanglEs(int num_rectangle_);
    int  Give_number_rectangle() { return num_rectanglEs; }

    int  Give_number_edges() { return num_edges; }

//    bool Adjacent(int num_hex, int num_edge);

    RectangleEl* Give_rectangle(int i) { return &(rectanglEs.at(i)); }
    EdgeEl* Give_edge(int i) { return &(edges.at(i)); }
    PointEl* Give_point(int i) { return &(points.at(i)); }

    void Set_rectangle(int id,
			int i_WSD, int i_ESD, int i_WND, int i_END);


    void construction_rectangle_points_done(); // now calculate edges

	/**
	 // ``transform'' must be zero vector at 0,1
     // transform: [0,1] ->  D2vector
     //  transform(t,pointer_global_data)
     //  es muss sein:  idL < idR
	**/
    void Set_transformation_edge(int idL, int idR,   EdgeTransformator2D* transformer);
                                    // ``transform'' must be zero vector at 0,1
                                    // transform: [0,1] ->  D2vector
									// 
  
    void construction_done();

    int  degree_of_freedom()  { return number_degree_of_freedom; };
    inline int Give_color_of_edge(int id_edge);

    void PrintVtk(ofstream *Datei);

    Marker2D& Give_all_points() { return *all_points; };

    double* Get_pointer_global_data() { return pointer_global_data; }
    
    /**
     * @return true: kante im Rechteck id_rec hat umgedrehte Orientierung
     */
    bool getOrientationEdge(int id_rec, dir2D kante) { return rectanglEs.at(id_rec).getOrientationEdge(kante); }
    
  protected:
    // pointer for global objects used in coordinate transformations
    double* pointer_global_data;
    
  private:
    double minCorX;
    double minCorY;
    double maxCorX;
    double maxCorY;     
    
    bool construction_rectangle_points_done_yn;
    bool construction_done_yn;

    int getEdge(int id_left, int id_right, bool* edgeExists);

    // Vektoren der Daten
    std::vector<PointEl>      points;     
    std::vector<EdgeEl>       edges;    
    std::vector<RectangleEl>  rectanglEs;   

    //zugehörige Längen der Vektoren
    int num_points;
    int num_edges;
    int num_rectanglEs;

    int number_degree_of_freedom;

    void Set_recursive_color(int id_edge, int color);

    Marker2D* all_points;
    
    // function called in construction_rectangle_points_done()
    void Set_coordinates_edges();

    template<class A_,typename Result, elementTyp TYP_EL>
    friend struct Interpolation2D_helper;  
};


//////////////////////////////////////////////////////////////////
// 4. some inline functions
//////////////////////////////////////////////////////////////////


inline int Unstructured2DGrid::Give_color_of_edge(int id_edge) {
  return edges[id_edge].Give_color_of_edge();
}



#endif // UG2D_H
