/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// interpolTwoD.h
//
// ------------------------------------------------------------

#include <QtGui>

#ifndef INTERPOLTwoD_H
#define INTERPOLTwoD_H
#include <vector>

/////////////////////////////////////////////////////////////
// 1. Interpolate from  blockgrid to rectangular blockgrid
/////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////
// 1. Interpolate from  blockgrid to rectangular blockgrid
/////////////////////////////////////////////////////////////

/** \defgroup InterpolationOperators  ''Interpolation operators ''    
 * Different interpolation operators.
 * Observe different data structures in 2D and 3D.
 **/
/* @{  */ 
/**
 * Interpolation 2D:
data structure on structured grid
 \verbatim
ny-1  *  *  *  *  *  *
      *  +  +  +  +  *
3     *  +  +  +  +  *
2     *  +  +  +  +  *
1     *  +  +  +  +  *
0     *  *  *  *  *  *
      0  1  2  3     nx-1
  \endverbatim
 * interpolates data from blockgrid_ to rectangular block [pWS, pEN]
**/
class Interpolate_on_structured_2Dgrid {
 public:
   ~Interpolate_on_structured_2Dgrid();
   
/**
 * preparation for interpolation
 @param  nx_ number of structured grid points x-direction
 @param  ny_ number of structured grid points y-direction
 @param  pWS Corner WSD of structured grid
 @param  pEN Corner WSD of structured grid
 @param  blockgrid_ of unstructured grid
**/
  Interpolate_on_structured_2Dgrid(int nx_, int ny_,
				     D2vector pWS, D2vector pEN,
				     Blockgrid2D& blockgrid_);

/**
 * preparation for interpolation
 @param  blockgrid_ of unstructured grid
**/
  Interpolate_on_structured_2Dgrid(Blockgrid2D* blockgrid_, double factorIncrease = 4);  
  
  
/**
 * interpolates from Variable u(of unstructured blockgrid) to structured data
 @param  u:  Variable on unstructured blockgrid
 @param  data: pointer to structured grid data i+nx*j
**/
  template <class DTyp>
    void interpolate(Variable2D<DTyp>& u, DTyp* data, DTyp defaultInterpolation);

/**
 * interpolates ABS of Variable u(of unstructured blockgrid) to structured data
 @param  u:  Variable on unstructured blockgrid
 @param  data: pointer to structured grid data i+nx*j
**/
void interpolateAbs(Variable2D< std::complex<double> >& u, double* data, double defaultInterpolation);    
    
protected:
  int nx, ny;
  double hx, hy;
  
  D2vector pWS;
  D2vector pEN;  
  
     
  static double absExpression(std::complex<double> value) {
        return sqrt(value.real()*value.real() + value.imag()*value.imag());
  }    
  
private:
  void initialize();
  
  int* ids_quad;
  int *ids_i, *ids_j;
  int *typ_tri;

  D2vector *lambda;

  Blockgrid2D*       blockgrid;
  Unstructured2DGrid*       ug;  
};


/**
 * interpolates data from blockgrid_ to rectangular block [pWSD, pENT] with own data
**/
template <class DTyp>
class Interpolate_on_structured_2Dgrid_with_data : protected Interpolate_on_structured_2Dgrid {
  public:
    Interpolate_on_structured_2Dgrid_with_data(int nx_, int ny_,
				               D2vector pWS, D2vector pEN,
				               Blockgrid2D& blockgrid_);
    Interpolate_on_structured_2Dgrid_with_data(Blockgrid2D& blockgrid_, double factorIncrease = 4);    
    
    void setData(Variable2D<DTyp>& u, DTyp defaultInterpolation);
    
    DTyp evaluate(double coord_x, double coord_y);
    
  private:
    DTyp* data;
};
/* @} */

class ColorMap {
  public:
    static QImage getImageColorMap(QSize resultSize);    
    static void makeColorMap();

    /**
     @wave Wellenlaenge in nm von 380nm bis 780nm 
     **/
    static uint rgbFromWaveLength(double wave);
    enum { ColormapSize = 512 };
    static uint colormap[ColormapSize];    
};

/** \addtogroup InterpolationOperators **/
/* @{ */  
/**
 * this class is used für visualization using Qt
 **/
/*
class QImageMakerAbstract : public ColorMap, public Interpolate_on_structured_2Dgrid {
  public:
    QImageMakerAbstract(Blockgrid2D* bg_, int thicknessContourLines_ = 4);
    ~QImageMakerAbstract();
    
    QImage makeImage(QSize resultSize, unsigned int numberContours, bool showData, bool origionalSize = true);
    

  protected:

    double* data;
    //Variable2D<double>* u;
    //Unstructured2DGrid* ug;
    Blockgrid2D* bg;
    double maxU, minU;
    int valueBackground;
    int contourColor;
    int valueEmpty;  
    
    int thicknessContourLines;
};

class QImageMaker : public QImageMakerAbstract {
  public:
      QImageMaker(Variable2D<double>& u_, int thicknessContourLines_ = 4);
};
*/

//test NOW
class QImageMaker : public ColorMap, public Interpolate_on_structured_2Dgrid {
  public:
    QImageMaker(Variable2D<double>& u_, int thicknessContourLines_ = 4);
    QImageMaker(Variable2D<std::complex<double> >& u, int thicknessContourLines_ = 4);    
    ~QImageMaker();
    
    QImage makeImage(QSize resultSize, unsigned int numberContours, bool showData, bool origionalSize = true);
    

  protected:
    double* data;
    //Variable2D<double>* u;
    //Unstructured2DGrid* ug;
    Blockgrid2D* bg;
    double maxU, minU;
    int valueBackground;
    int contourColor;
    int valueEmpty;  
    
    int thicknessContourLines;
};



/*
class QImageMakerAbs : public QImageMakerAbstract {
  public:
      QImageMakerAbs(Variable2D< std::complex<double> >& u_, int thicknessContourLines_ = 4);
};
*/

/**
 * this class is used für visualization using Qt. It creates vector of arrows of a vector field
 **/
class VectorOfArrowsCreator : public ColorMap, public Interpolate_on_structured_2Dgrid {
  public:
    VectorOfArrowsCreator(Variable2D<double>& ux_, Variable2D<double>& uy_);
    ~VectorOfArrowsCreator();
    
    void createVectorList(QVector<QPointF>& startP, QVector<QPointF>& endP,
                   QSize resultSize, unsigned int numberArrowX_, unsigned int numberArrowY_, 
		   bool origionalSize = true);
  private:
    double* dataX;
    double* dataY;    
    //Unstructured2DGrid* ug;
    Blockgrid2D* bg;
    int valueBackground;
    int contourColor;
    int valueEmpty;  

};

/**
 * this class is used für visualization using Qt. 
 * It creates vector of arrows of a vector field for complex data
 **/
class VectorOfArrowsCreatorComplex : public ColorMap, public Interpolate_on_structured_2Dgrid {
  public:
    VectorOfArrowsCreatorComplex(Variable2D< std::complex<double> >& ux_, 
                                 Variable2D< std::complex<double> >& uy_);
    ~VectorOfArrowsCreatorComplex();
    
    void createVectorList(QVector<QPointF>& startP, QVector<QPointF>& endP,
                   QSize resultSize, unsigned int numberArrowX_, unsigned int numberArrowY_, 
		   bool origionalSize = true);
  private:
    std::complex<double>* dataX;
    std::complex<double>* dataY;    
    Blockgrid2D* bg;
    int valueBackground;
    int contourColor;
    int valueEmpty;    
};
/* @} */

/////////////////////////////////////////////////////////////////
// Implementierung von Methoden
/////////////////////////////////////////////////////////////////

template <class DTyp>
class Give_corner_data_of_quad {
 public:
  Give_corner_data_of_quad(Variable2D<DTyp>& u_, int quad_, int i_, int j_);
  
  DTyp WS() { return u->template Give_data<rectangleEl>(id, i,   j,   Nx); }
  DTyp ES() { return u->template Give_data<rectangleEl>(id, i+1, j,   Nx); }
  DTyp WN() { return u->template Give_data<rectangleEl>(id, i,   j+1, Nx); }
  DTyp EN() { return u->template Give_data<rectangleEl>(id, i+1, j+1, Nx); }

 private:
  int id, i, j;
  int Nx;
  Variable2D<DTyp>* u;
};

template <class DTyp>
Give_corner_data_of_quad<DTyp>::Give_corner_data_of_quad(Variable2D<DTyp>& u_,
							 int quad_, int i_, int j_) {
  Blockgrid2D*        blockgrid;

  u = &u_;
  blockgrid = u->Give_blockgrid();
  id = quad_; i = i_;  j = j_;
  
  Nx = blockgrid->Give_Nx_rectangle(id);
}


template <class DTyp>
void Interpolate_on_structured_2Dgrid::interpolate(Variable2D<DTyp>& u, DTyp* data, DTyp defaultInterpolation) {
  int i,j, id_quad, typ;
  int ind_global;


  for(int id=0;id<ug->Give_number_rectangle();++id)
    u.template Update<rectangleEl>(id);

  for(int js = 0; js < ny;++js)
      for(int is = 0; is < nx;++is) {
	ind_global = is+nx*js;
	i = ids_i[ind_global];
	j = ids_j[ind_global];
	id_quad = ids_quad[ind_global];

	/*
	// test : das geht nicht::
	if(id_hex < 0) {
	  ind_global = is+nx*((ny-1-js)+ny*ks);
	  i = ids_i[ind_global];
	  j = ids_j[ind_global];
	  k = ids_k[ind_global];
	  id_hex = ids_hex[ind_global];
	}
*/		
	if(id_quad < 0) data[ind_global] = defaultInterpolation;
	else {
	  Give_corner_data_of_quad<DTyp> du(u, id_quad, i,j);

	  typ = typ_tri[ind_global];

	
	  /*
	  if( ind_global == 3)
	    cout << " \n ind_global: " << is + nx * (js + ny * ks)
	         << " typ: " << typ
	         << " id_hex " << id_hex
	         << " i: " << i
	         << " j: " << j
	         << " k: " << k
	         << endl;
  	  */

	  if(typ==0) data[ind_global] = interpolate_in_tri(lambda[ind_global],
							   du.WN(),du.WS(),du.ES());
	  if(typ==1) data[ind_global] = interpolate_in_tri(lambda[ind_global],
							   du.EN(),du.WN(),du.ES());
	}
      }
}



template <class DTyp>
Interpolate_on_structured_2Dgrid_with_data<DTyp>::Interpolate_on_structured_2Dgrid_with_data(int nx_, int ny_,
				                                                       D2vector pWS, D2vector pEN,
				                                                       Blockgrid2D& blockgrid_) 
   : Interpolate_on_structured_2Dgrid(nx_,ny_,pWS,pEN,blockgrid_) {
   data = new DTyp[nx*ny];
   for(int i=0;i<nx*ny;++i) data[i] = 0.0; 
}

template <class DTyp>
Interpolate_on_structured_2Dgrid_with_data<DTyp>::Interpolate_on_structured_2Dgrid_with_data(Blockgrid2D& blockgrid_, double factorIncrease) 
   : Interpolate_on_structured_2Dgrid(&blockgrid_, factorIncrease) {
   data = new DTyp[nx*ny];
   for(int i=0;i<nx*ny;++i) data[i] = 0.0; 
}

template <class DTyp>
void Interpolate_on_structured_2Dgrid_with_data<DTyp>::setData(Variable2D<DTyp>& u, DTyp defaultInterpolation) {
   interpolate(u,data,defaultInterpolation);
}

template <class DTyp>
DTyp Interpolate_on_structured_2Dgrid_with_data<DTyp>::evaluate(double coord_x, double coord_y) {
   if(coord_x > pEN.x) return 0.0;
   if(coord_x < pWS.x) return 0.0;
   if(coord_y > pEN.y) return 0.0;
   if(coord_y < pWS.y) return 0.0;
  
  int i = (coord_x - pWS.x) / hx;   
  int j = (coord_y - pWS.y) / hy;   
    
  if(i < 0)   i=0;     if(j <0   ) j=0;
  if(i>=nx-1) i=nx-2;  if(j>=ny-1) j=ny-2;
  
    
  DTyp uWS = data[i    +nx*j    ];
  DTyp uES = data[(i+1)+nx*j    ];
  DTyp uWN = data[i    +nx*(j+1)];
  DTyp uEN = data[(i+1)+nx*(j+1)];
  
  
  // assert( (i+1)+nx*((j+1)+ny*(k+1)) < nx*ny*nz);
  
  double locX = (coord_x - pWS.x) / hx - i;
  double locY = (coord_y - pWS.y) / hy - j;
  
  //return uWSD;
  
  return uWS * (1.0 - locX) * (1.0 - locY) +
         uES *        locX  * (1.0 - locY) +
         uWN * (1.0 - locX) *        locY  +
         uEN *        locX  *        locY;
}
    


#endif // INTERPOL_H
