/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


// ------------------------------------------------------------
// math_lib.h
//
// ------------------------------------------------------------

#ifndef MATHLIB2D_H_
#define MATHLIB2D_H_

#include "../../../common_source/mathlib/math_lib_common.h"
#include <ostream>
#include <iostream>
#include <fstream>
using std::ofstream;

class D2matrix;

//////////////////////////////////////////////////////////////////////
// 1. 2D vector class
// 2. a simple 2D matrix class
// 3. geometric operators for 2D vectors
// 4. interpolate in tri
//////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////
// 1. a simple 2D vector class
//////////////////////////////////////////////////////////////////////

class D2vector {
 public:
  double x,y;
  D2vector(double cx, double cy) : x(cx), y(cy) {};
  explicit D2vector(double c) : x(c), y(c)  {};
  D2vector() : x(0), y(0) {};
  ~D2vector(){};
  void Print();
  void Print(ofstream *Datei);
  void operator=(const D2vector& v) { x=v.x; y=v.y; }
  double operator[](int i) {
    if(i==0) return x;
    return y;
  }
  double radius() const { return sqrt(x*x+y*y); }
  
  bool testZwischen(D2vector A, D2vector B); // testet obPunkt auf Gerade zwischen A und B ist.

  D2vector& operator +=(const D2vector& V);
  D2vector& operator -=(const D2vector& V);

  friend std::ostream& operator<< (std::ostream& os, const D2vector& v);
  friend
  D2vector operator*(const D2matrix& m,const D2vector& v);
};

inline D2vector& D2vector::operator +=(const D2vector& V) {
  x += V.x;
  y += V.y;
}

inline D2vector& D2vector::operator -=(const D2vector& V) {
  x -= V.x;
  y -= V.y;
}

inline double MIN(D2vector V) {
  if(V.x<V.y) return V.x;
  V.y;
}

inline double ANGLE(D2vector V) {
  return atan(V.y/V.x);
}

inline double MAX(D2vector V) {
  if(V.x>V.y) return V.x;
  return V.y;
}

inline double ABS(D2vector V) {
  return sqrt(V.x*V.x + V.y*V.y);
}

inline D2vector MAX(D2vector V1, D2vector V2) {
  return D2vector(MAX(V1.x,V2.x),MAX(V1.y,V2.y));
}

inline D2vector MIN(D2vector V1, D2vector V2) {
  return D2vector(MIN(V1.x,V2.x),MIN(V1.y,V2.y));
}

inline D2vector operator+(const D2vector& v,const D2vector& w) {
  return D2vector(v.x+w.x,v.y+w.y);
}

inline D2vector operator*(const D2vector& v,const D2vector& w) {
  return D2vector(v.x*w.x,v.y*w.y);
}

inline D2vector operator/(const D2vector& v,const D2vector& w) {
  return D2vector(v.x/w.x,v.y/w.y);
}

inline D2vector operator-(const D2vector& v,const D2vector& w) {
  return D2vector(v.x-w.x,v.y-w.y);
}

inline D2vector operator/(const D2vector& v,const double f) {
  return D2vector(v.x/f,v.y/f);
}

inline D2vector operator*(const D2vector& v,const double f) {
  return D2vector(v.x*f,v.y*f);
}

inline D2vector operator*(const double f, const D2vector& v) {
  return D2vector(v.x*f,v.y*f);
}

inline bool operator<(const D2vector& v,const D2vector& w) {
  return (v.x<w.x && v.y<w.y);
}
inline bool operator==(const D2vector& v,const D2vector& w) {
  return (v.x==w.x && v.y==w.y);
}

inline std::ostream& operator<< (std::ostream& os, const D2vector& v) {
  os << "("<<v.x<<","<<v.y<<")";
  return os;
}

//////////////////////////////////////////////////////////////////////
// 2. a simple 2D matrix class 
//////////////////////////////////////////////////////////////////////


class D2matrix {
 private:
  double x1,y1; // first  row
  double x2,y2; // second row

 public:
  D2matrix(D2vector cx, D2vector cy) : 
    x1(cx.x), y1(cy.x),
    x2(cx.y), y2(cy.y)   {};

  D2matrix(double a, double b, double c, double d) :
    x1(a), y1(c),
    x2(b), y2(d)   {};

  double Determinante() {
    return   (x1 * y2 - y1 * x2);
  }
  ~D2matrix(){};


  D2vector invert_apply(const D2vector& v) {
    double x,y;
    double det =  Determinante();


    x =  (  y2*v.x - y1*v.y) / det;
    y =  (- x2*v.x + x1*v.y) / det;

    return D2vector(x,y);
  }

  friend std::ostream& operator<< (std::ostream& os, const D2matrix& v);

  friend
  D2vector operator*(const D2matrix& m,const D2vector& v);

};

inline D2vector operator*(const D2matrix& m,const D2vector& v) {
  return D2vector(m.x1*v.x+m.x2*v.y,m.y1*v.x+m.y2*v.y);
}


inline std::ostream& operator<< (std::ostream& os, const D2matrix& m) {
  os << "("<<m.x1<<","<<m.x2<<";"<<m.y1<<","<<m.y2<<")";
  return os;
}

//////////////////////////////////////////////////////////////////////
// 3. geometric operators for 2D vectors


inline double D2VectorNorm(const D2vector& v) {
  return sqrt(v.x*v.x+v.y*v.y);
}

// scalar product
inline double product(const D2vector& v, const D2vector& w) {
  return v.x*w.x+v.y*w.y;
}

//////////////////////////////////////////////////////////////////////
// 4. interpolate in tri

D2vector lambda_of_p_in_tri(D2vector p,
			    D2vector cA, D2vector cB, D2vector cC);

template <class DTyp>
DTyp interpolate_in_tri(D2vector lambda,
			DTyp vA, DTyp vB, DTyp vC) {
  return vA + (vB-vA) * lambda.x + (vC-vA) * lambda.y;
}



//////////////////////////////////////////////
// Implementierung einiger Memberfunktionen
//////////////////////////////////////////////


// D2vector 
// ----------

inline void D2vector::Print() {
  std::cout << "Coordinate: " << x << ", " << y << ";";
}

inline void D2vector::Print(ofstream *Datei) {
  *Datei  << x << " " << y;
}


#endif // MATHLIB2D_H_


