/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


// ------------------------------------------------------------
//
// mgop.cc
//
// ------------------------------------------------------------

#include "../mympi2D.h"
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "../grid/elements2D.h"
//#include "../grid/parti2D.h"
#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"
#include "../grid/marker2D.h"
#include "../extemp/extemp2D.h"
#include "../extemp/co_fu2D.h"
#include "../extemp/variable2D.h"
#include "../../../Source_Colsamm/Colsamm.h"
#include "diffop2D.h"
#include "ressten2D.h"
#include "ressten2D_cc.h"
#include "mgop2D.h"

///////////////////////////////////
// Restriction Stencils
///////////////////////////////////


//inline int number_coarse(dir3D_sons i) {
//  return trans_cellP_27(i,i);
//}

// 2D
void Set_coarse_function2D(int* U, dir2D_sons v) {

  if(v==SWdir2D) {
    U[6] = 0;  U[7] = 0;  U[8] = 0;
    U[3] = 2;  U[4] = 1;  U[5] = 0;
    U[0] = 4;  U[1] = 2;  U[2] = 0;
  }

  if(v==SEdir2D) {
    U[6] = 0;  U[7] = 0;  U[8] = 0;
    U[3] = 0;  U[4] = 1;  U[5] = 2;
    U[0] = 0;  U[1] = 2;  U[2] = 4;
  }

  if(v==NWdir2D) {
    U[6] = 4;  U[7] = 2;  U[8] = 0;
    U[3] = 2;  U[4] = 1;  U[5] = 0;
    U[0] = 0;  U[1] = 0;  U[2] = 0;
  }

  if(v==NEdir2D) {
    U[6] = 0;  U[7] = 2;  U[8] = 4;
    U[3] = 0;  U[4] = 1;  U[5] = 2;
    U[0] = 0;  U[1] = 0;  U[2] = 0;
  }
}

///////////////////////////////////
// Constructor Multigrid2D_operator
///////////////////////////////////

Multigrid2D_operator::Multigrid2D_operator(Blockgrid2D& coarse_grid_,
				       Blockgrid2D& fine_grid_) {
  int number_degree_of_freedom;
  bool co_x, co_y, co_edge;
  Unstructured2DGrid* ug;
  EdgeEl       *edge;

  // a) check if ug o.k.
  coarse_grid = &coarse_grid_;
  fine_grid   = &fine_grid_;

  if(coarse_grid->Give_unstructured_grid()!=
     fine_grid->Give_unstructured_grid())
    cout << "Error 1 in Multigrid2D_operator" << endl;
  ug = coarse_grid->Give_unstructured_grid();

  // for parallel
  my_rank = ug->Give_my_rank();
  comm    = ug->Give_MPI_comm();

  // b) Calculate coarse directions
  number_degree_of_freedom = ug->degree_of_freedom();


  coarse = new bool[number_degree_of_freedom];
  for(int i=0;i<number_degree_of_freedom;++i) {
    if(coarse_grid->Give_N_color(i) == fine_grid->Give_N_color(i))
      coarse[i] = false;
    else {
      if(coarse_grid->Give_N_color(i) * 2 == fine_grid->Give_N_color(i))
	coarse[i] = true;
      else  
	cout << "Error 2 in Multigrid2D_operator: "
	     << i << endl;
    }

  
  }

  // c) Calculate coarse directions of rectangle
  coarse_rectangles = new coarsening2D_direction[ug->Give_number_rectangle()];
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      co_x = coarse[ug->Give_color_of_edge(
		    ug->Give_rectangle(id)->Give_id_edge(Sdir2D))];
      co_y = coarse[ug->Give_color_of_edge(
		    ug->Give_rectangle(id)->Give_id_edge(Wdir2D))];

      if(co_x && co_y)
    	  coarse_rectangles[id] = coarse2D_xy;
      else if(co_x)
    	  coarse_rectangles[id] = coarse2D_x;
	  else if(co_y)
		  coarse_rectangles[id] = coarse2D_y;
	  else 
		  coarse_rectangles[id] = no_coarse2D;
    }
    else coarse_rectangles[id] = (coarsening2D_direction)0;
  }

  // d) Calculate coarse directions of edges
  coarse_edges_exterior = new bool[ug->Give_number_edges()];
  coarse_edges_interior = new bool[ug->Give_number_edges()];

  for(int id=0;id<ug->Give_number_edges();++id) {

    if(ug->Give_edge(id)->my_object(my_rank)) {

      edge = ug->Give_edge(id);

      co_edge = coarse[edge->Give_color_of_edge()];

      coarse_edges_interior[id] = co_edge;

      if(edge->isAtBoundary()) {
        coarse_edges_exterior[id] = co_edge;
      }
      else coarse_edges_exterior[id] = false;
    }
    else{
      coarse_edges_exterior[id] = (bool)0;
      coarse_edges_interior[id] = (bool)0;
    }
  }

  // e) Calculate coarse grid test function
  coarse_grid_functions2D = new int[36];

  for(int i=0;i<4;++i)
    Set_coarse_function2D(&(coarse_grid_functions2D[i*9]),(dir2D_sons)i);
}


///////////////////////////////////
// Destructor Multigrid2D_operator
///////////////////////////////////


Multigrid2D_operator::~Multigrid2D_operator() {

  Unstructured2DGrid* ug = coarse_grid->Give_unstructured_grid();

  if(coarse != NULL)
      delete [] coarse;
  coarse = NULL;

  if(coarse_rectangles != NULL)
      delete [] coarse_rectangles;
  coarse_rectangles = NULL;

  if(coarse_edges_exterior != NULL)
      delete [] coarse_edges_exterior;
  coarse_edges_exterior = NULL;

  if(coarse_edges_interior != NULL)
      delete [] coarse_edges_interior;
  coarse_edges_interior = NULL;

  if(coarse_grid_functions2D != NULL)
      delete [] coarse_grid_functions2D;
  coarse_grid_functions2D = NULL;
}

