/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// mgop.h
//
// ------------------------------------------------------------
#ifndef MGOP2D_H_
#define MGOP2D_H_

void Print_Loc_stencil(double* sten);

enum coarsening2D_direction {
	coarse2D_xy, coarse2D_x, coarse2D_y, no_coarse2D
};

#define recWS  v.Give_data<rectangleEl>(id,i-1,j-1,Nxf)
#define recS  v.Give_data<rectangleEl>(id,i,j-1,Nxf)
#define recES  v.Give_data<rectangleEl>(id,i+1,j-1,Nxf)

#define recW  v.Give_data<rectangleEl>(id,i-1,j,Nxf)
#define recM  v.Give_data<rectangleEl>(id,i,j,Nxf)
#define recE  v.Give_data<rectangleEl>(id,i,j-1,Nxf)

#define recWN  v.Give_data<rectangleEl>(id,i+1,j-1,Nxf)
#define recN  v.Give_data<rectangleEl>(id,i,j+1,Nxf)
#define recEN  v.Give_data<rectangleEl>(id,i+1,j+1,Nxf)

///////////////////////////////////////////////////////////////
// Restriction

class Multigrid2D_operator;

template<typename TYPE2>

class ExpRes2D_local_stiffness_matrix {
public:
	ExpRes2D_local_stiffness_matrix(const Local_stiffness_matrix2D<TYPE2>& lm_,
			Multigrid2D_operator* mgop_) :
			lm(lm_), mgop(mgop_) {
	}
	;

	const Local_stiffness_matrix2D<TYPE2>& lm;
	Multigrid2D_operator* mgop;
};

template<class DTyp>

class ExpressionRestriction2D {
public:
	ExpressionRestriction2D(const Variable2D<DTyp>& v_,
			Multigrid2D_operator* mgop_) :
			v(v_), mgop(mgop_) {
	}
	;

	const Variable2D<DTyp>& v;
	Multigrid2D_operator* mgop;
};

template<class DTyp>

class ExpressionProlongation2D {
public:
	ExpressionProlongation2D(const Variable2D<DTyp>& v_,
			Multigrid2D_operator* mgop_) :
			v(v_), mgop(mgop_) {
	}
	;

	const Variable2D<DTyp>& v;
	Multigrid2D_operator* mgop;
};

class Multigrid2D_operator {
public:
	Multigrid2D_operator(Blockgrid2D& coarse_grid_, Blockgrid2D& fine_grid_);
	~Multigrid2D_operator();

	template<typename TYPE2>
	ExpRes2D_local_stiffness_matrix<TYPE2> Restrict(
			const Local_stiffness_matrix2D<TYPE2>& lm);


	template<class DTyp>
	ExpressionRestriction2D<DTyp> Restrict(const Variable2D<DTyp>& v);

	template<class DTyp>
	ExpressionProlongation2D<DTyp> Prolongate(const Variable2D<DTyp>& v);

	coarsening2D_direction Give_coarse_rect(int id) {
		return coarse_rectangles[id];
	}

	bool Give_coarse_edge_exterior(int id) {
		return coarse_edges_exterior[id];
	}

	bool Give_coarse_edge_interior(int id) {
		return coarse_edges_interior[id];
	}

	//double* Give_coarse_weight_edges ( int id ) { return coarse_weight_edges[id]; }

	Blockgrid2D* Give_coarse_grid() {
		return coarse_grid;
	}

	Blockgrid2D* Give_fine_grid() {
		return fine_grid;
	}

	int* Give_coarse_grid_functions2D() {
		return coarse_grid_functions2D;
	}

private:
	bool *coarse;    // coarse[i] means coarsening in direction i

	coarsening2D_direction* coarse_rectangles; // num_rectangles

	bool* coarse_edges_interior;               // num_edges
	bool* coarse_edges_exterior;               // num_edges

	//double**  coarse_weight_edges;                 // num_edges

	Blockgrid2D* coarse_grid;
	Blockgrid2D* fine_grid;
	// for coarsening: coarse grid test functions
	int *coarse_grid_functions2D;

	// for parallel
	int my_rank;
	MPI_Comm comm;
};

template<typename TYPE2>
ExpRes2D_local_stiffness_matrix<TYPE2> Multigrid2D_operator::Restrict(
		const Local_stiffness_matrix2D<TYPE2>& lm) {
	if (lm.Give_blockgrid() != fine_grid)
		cout << " Error in Restriction lm, wrong fine grid! " << endl;

	return ExpRes2D_local_stiffness_matrix<TYPE2>(lm, this);
}

template<class DTyp>
ExpressionRestriction2D<DTyp> Multigrid2D_operator::Restrict(
		const Variable2D<DTyp>& v) {
	if (v.Give_blockgrid() != fine_grid)
		cout << " Error in Restriction v, wrong fine grid! " << endl;

	return ExpressionRestriction2D<DTyp>(v, this);
}

template<class DTyp>
ExpressionProlongation2D<DTyp> Multigrid2D_operator::Prolongate(
		const Variable2D<DTyp>& v) {
	if (v.Give_blockgrid() != coarse_grid)
		cout << " Error in Prolongation v, wrong coarse grid! " << endl;

	return ExpressionProlongation2D<DTyp>(v, this);
}

///////////////////////////////////////////////////////////////
// Restriction stiffness matrix

template<typename TYPE2>
  void
  Local_stiffness_matrix2D<TYPE2>::operator=(
      const ExpRes2D_local_stiffness_matrix<TYPE2>& Expr_res)
  {
    int id, i, j;
    int Nx, Ny, N_total;
    int Nx_fine, Ny_fine;

    Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

    Multigrid2D_operator* mgop;
    coarsening2D_direction coarse_rectangle;

    int *coarse_grid_functions2D;

    TYPE2** loc_fine;
    TYPE2* Stencil_coarse;
    TYPE2* Stencils_fine[4];

    loc_fine = Expr_res.lm.loc_m_rectangles;

    mgop = Expr_res.mgop;

    coarse_grid_functions2D = mgop->Give_coarse_grid_functions2D();

    for (id = 0;
        id < blockgrid->Give_unstructured_grid()->Give_number_rectangle();
        ++id) {

      if (ug->Give_rectangle(id)->my_object(my_rank)) {
        Nx = blockgrid->Give_Nx_rectangle(id);
        Ny = blockgrid->Give_Ny_rectangle(id);
        N_total = Nx * Ny * 16;

        coarse_rectangle = mgop->Give_coarse_rect(id);

        if (coarse_rectangle == coarse2D_xy) {
          Nx_fine = Nx * 2;
          Ny_fine = Ny * 2;

          for (j = 0; j < Ny; ++j)
            for (i = 0; i < Nx; ++i) {
              Stencil_coarse = &loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,
                  j) * 16];

              Stencils_fine[SWdir2D] =
                  &loc_fine[id][Ind_loc_matrix_rectangle_fine(2 * i, 2 * j) * 16];
              Stencils_fine[SEdir2D] =
                  &loc_fine[id][Ind_loc_matrix_rectangle_fine(2 * i + 1, 2 * j)
                      * 16];
              Stencils_fine[NWdir2D] =
                  &loc_fine[id][Ind_loc_matrix_rectangle_fine(2 * i, 2 * j + 1)
                      * 16];
              Stencils_fine[NEdir2D] =
                  &loc_fine[id][Ind_loc_matrix_rectangle_fine(2 * i + 1,
                      2 * j + 1) * 16];

              Restrict2D_local_stiffness_matrix_2D<TYPE2, Index2D_calc_xy>(
                  Stencil_coarse, Stencils_fine, coarse_grid_functions2D);
            }
        }
        if (coarse_rectangle == coarse2D_x) {
          Nx_fine = Nx * 2;
          Ny_fine = Ny;

          for (j = 0; j < Ny; ++j)
            for (i = 0; i < Nx; ++i) {
              Stencil_coarse = &loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,
                  j) * 16];

              Stencils_fine[Wd] = &loc_fine[id][Ind_loc_matrix_rectangle_fine(
                  2 * i, j) * 16];
              Stencils_fine[Ed] = &loc_fine[id][Ind_loc_matrix_rectangle_fine(
                  2 * i + 1, j) * 16];

              Restrict2D_local_stiffness_matrix_1D<TYPE2, Index2D_calc_x>(
                  Stencil_coarse, Stencils_fine);
            }
        }

        if (coarse_rectangle == coarse2D_y) {
          Nx_fine = Nx;
          Ny_fine = Ny * 2;

          for (j = 0; j < Ny; ++j)
            for (i = 0; i < Nx; ++i) {
              Stencil_coarse = &loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,
                  j) * 16];

              Stencils_fine[Wd] = &loc_fine[id][Ind_loc_matrix_rectangle_fine(i,
                  2 * j) * 16];
              Stencils_fine[Ed] = &loc_fine[id][Ind_loc_matrix_rectangle_fine(i,
                  2 * j + 1) * 16];

              Restrict2D_local_stiffness_matrix_1D<TYPE2, Index2D_calc_y>(
                  Stencil_coarse, Stencils_fine);
            }

        }
        if (coarse_rectangle == no_coarse2D) {
          Nx_fine = Nx;
          Ny_fine = Ny;

          for (j = 0; j < Ny; ++j)
            for (i = 0; i < Nx; ++i)
              for (int l = 0; l < 16; ++l) {
                loc_m_rectangles[id][Ind_loc_matrix_rectangle(i, j) * 16 + l] =
                    loc_fine[id][Ind_loc_matrix_rectangle(i, j) * 16 + l];
              }
        }
      }
    }
    Update();
  }

///////////////////////////////////////////////////////////////
// Restriction

template<class DTyp>
void Variable2D<DTyp>::operator=(
		const ExpressionRestriction2D<DTyp>& Expr_res) {
	int Nx, Ny, Nxf, Nyf;
	Multigrid2D_operator* mgop;
	coarsening2D_direction coarse_rectangle;
	bool coarse_edge_exterior;
	bool coarse_edge_interior;

	const Variable2D<DTyp>& v(Expr_res.v);

	mgop = Expr_res.mgop;

	if (Give_blockgrid() != Expr_res.mgop->Give_coarse_grid())
		cout << " Error in Restriction, wrong coares grid! " << endl;

	// 1. rectangle
	for (int id = 0; id < ug->Give_number_rectangle(); ++id) {

		v.Update < rectangleEl > (id);

		if (ug->Give_rectangle(id)->my_object(my_rank)) {

			Nx = blockgrid->Give_Nx_rectangle(id);
			Ny = blockgrid->Give_Ny_rectangle(id);

			coarse_rectangle = mgop->Give_coarse_rect(id);

			//    cout << " co" << (int)coarse_hexahedron << endl;

			if (coarse_rectangle == coarse2D_xy) {
				Nxf = Nx * 2;
				Nyf = Ny * 2;

				for (int j = 2; j < Nyf; j = j + 2)
					for (int i = 2; i < Nxf; i = i + 2)
						data_rectangles[id][i / 2 + (Nx + 1) * (j / 2)] =
						( 1. * recWN + 2. * recN + 1. * recEN
						+ 2. * recW  + 4. * recM + 2. * recE
						+ 1. * recWS + 2. * recS + 1. * recES)/16.0;
			}

			if (coarse_rectangle == coarse2D_x) {
				Nxf = Nx * 2;
				Nyf = Ny;

				for (int j = 1; j < Nyf; j = j + 1)
					for (int i = 2; i < Nxf; i = i + 2)
						data_rectangles[id][i / 2 + (Nx + 1) * (j)] = 0.25 * recE
								+ 0.5*recM + 0.25 * recW;
			}

			if (coarse_rectangle == coarse2D_y) {
				Nxf = Nx;
				Nyf = Ny * 2;

				for (int j = 2; j < Nyf; j = j + 2)
					for (int i = 1; i < Nxf; i = i + 1)
						data_rectangles[id][i + (Nx + 1) * (j / 2)] = 0.25 * recS
								+ 0.5 * recM + 0.25 * recN;
			}

			if (coarse_rectangle == no_coarse2D) {
				Nxf = Nx;
				Nyf = Ny;

				for (int j = 1; j < Ny; ++j)
					for (int i = 1; i < Nx; ++i)
						data_rectangles[id][i + (Nx + 1) * (j)] = recM;
			}
		}
		Update<rectangleEl>(id);
	}

	// 2. edges
	int num_pos;

	double weight;
	for (int id = 0; id < ug->Give_number_edges(); ++id) {

		v.Update < edgeEl > (id);

		if (ug->Give_edge(id)->my_object(my_rank)) {

			// Update edge
			Nx = blockgrid->Give_Nx_edge(id);
			//num_pos = ug->Give_edge ( id )->Give_number_neighbors();

			coarse_edge_exterior = mgop->Give_coarse_edge_exterior(id);
			coarse_edge_interior = mgop->Give_coarse_edge_interior(id);

			if (coarse_edge_interior || coarse_edge_exterior) { // coarse edge
				Nxf = Nx * 2;

				for (int i = 2; i < Nxf; i = i + 2) {
//					if (coarse_edge_interior)
//						data_interior_edges[id][i / 2] = (
//								  0.25 * v.Give_data_edge_interior(id, i - 1)
//								+ 0.25 * v.Give_data_edge_interior(id, i + 1)
//								+ v.Give_data_edge_interior(id, i));
//					if (coarse_edge_exterior)
//						data_exterior_edges[id][i / 2] = (
//								  0.25 * v.Give_data_edge_exterior(id, i - 1)
//								+ 0.25 * v.Give_data_edge_exterior(id, i + 1)
//								+ 0.5 * v.Give_data_edge_exterior(id, i));

					data_edges[id][i / 2] = (
						 0.25 * v.Give_data<edgeEl>(id, i - 1,0,0)
						+ 0.25 * v.Give_data<edgeEl>(id, i + 1,0,0)
						+ 0.5 * v.Give_data<edgeEl>(id, i,0,0));

				}
			} else { // no coarse edge
				for (int i = 1; i < Nx; i = i + 1) {
					data_edges[id][i] = v.Give_data<edgeEl>(id,i,0,0);
				}
			}
		}
	}
	// 3. points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) )
		data_points[id][0] = v.Give_data<pointEl> ( id, 0 , 0 , 0);
	}
}

///////////////////////////////////////////////////////////////
// Prolongation

template<class DTyp>
void Variable2D<DTyp>::operator=(
	const ExpressionProlongation2D<DTyp>& Expr_prol) {
	Blockgrid2D* coarse_blockgrid;
	Multigrid2D_operator* mgop;
	mgop = Expr_prol.mgop;
	coarse_blockgrid = mgop->Give_coarse_grid();

	int Nx, Ny, Nxf, Nyf;
	coarsening2D_direction coarse_rectangle;
	bool coarse_edge_exterior;
	bool coarse_edge_interior;

	const Variable2D<DTyp>& v(Expr_prol.v);

	if (Give_blockgrid() != mgop->Give_fine_grid())
		cout << " Error in Prolongation, wrong fine grid! " << endl;

	// 1. rectangle
	for (int id = 0; id < ug->Give_number_rectangle(); ++id) {

		v.Update<rectangleEl>(id);

		if (ug->Give_rectangle(id)->my_object(my_rank)) {

			Nx = coarse_blockgrid->Give_Nx_rectangle(id);
			Ny = coarse_blockgrid->Give_Ny_rectangle(id);

			coarse_rectangle = mgop->Give_coarse_rect(id);

			//    cout << " co" << (int)coarse_hexahedron << endl;

			if (coarse_rectangle == coarse2D_xy) {
				Nxf = Nx * 2;
				Nyf = Ny * 2;
				for (int j = 0; j < Nyf; ++j) {
					for (int i = 0; i < Nxf; ++i) {
						data_rectangles[id][i + (Nxf + 1) * j] = 0.0;
					}
				}


				for (int j = 0; j <= Ny; ++j) {
					for (int i = 0; i <= Nx; ++i) {
						if(j > 0 && i > 0)  data_rectangles[id][(2*i-1) + (Nxf + 1) * (2*j-1)] += 0.25 * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(j > 0)           data_rectangles[id][(2*i)   + (Nxf + 1) * (2*j-1)] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(j > 0 && i < Nx) data_rectangles[id][(2*i+1) + (Nxf + 1) * (2*j-1)] += 0.25 * v.Give_data<rectangleEl>(id,i,j,Nx);

						if(i > 0)  data_rectangles[id][(2*i-1) + (Nxf + 1) * 2*j    ] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
						data_rectangles[id][(2*i)   + (Nxf + 1) * 2*j    ] += 1.0  * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(i < Nx) data_rectangles[id][(2*i+1) + (Nxf + 1) * 2*j    ] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);

						if(j < Ny && i > 0)  data_rectangles[id][(2*i-1) + (Nxf + 1) * (2*j+1)] += 0.25 * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(j < Ny)           data_rectangles[id][(2*i)   + (Nxf + 1) * (2*j+1)] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(j < Ny && i < Nx) data_rectangles[id][(2*i+1) + (Nxf + 1) * (2*j+1)] += 0.25 * v.Give_data<rectangleEl>(id,i,j,Nx);
					}
				}

			}

			if (coarse_rectangle == coarse2D_x) {
				Nxf = Nx * 2;
				Nyf = Ny;

				for (int j = 0; j < Nyf; ++j) {
					for (int i = 0; i < Nxf; ++i) {
						data_rectangles[id][i + (Nxf + 1) * j] = 0.0;
					}
				}


				for (int j = 1; j < Ny; ++j) {
					for (int i = 0; i <= Nx; ++i) {
						if(i > 0)  data_rectangles[id][(2*i-1) + (Nxf + 1) * j] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
						           data_rectangles[id][(2*i)   + (Nxf + 1) * j] += 1.0  * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(i < Nx) data_rectangles[id][(2*i+1) + (Nxf + 1) * j] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
					}
				}
			}

			if (coarse_rectangle == coarse2D_y) {
				Nxf = Nx * 2;
				Nyf = Ny;

				for (int j = 0; j < Nyf; ++j) {
					for (int i = 0; i < Nxf; ++i) {
						data_rectangles[id][i + (Nxf + 1) * j] = 0.0;
					}
				}


				for (int j = 0; j <= Ny; ++j) {
					for (int i = 1; i < Nx; ++i) {
						if(j > 0 )  data_rectangles[id][i + (Nxf + 1) * (2*j-1)] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
						            data_rectangles[id][i + (Nxf + 1) * (2*j)  ] += 1.0  * v.Give_data<rectangleEl>(id,i,j,Nx);
						if(j < Ny)  data_rectangles[id][i + (Nxf + 1) * (2*j+1)] += 0.5  * v.Give_data<rectangleEl>(id,i,j,Nx);
					}
				}
			}

			if (coarse_rectangle == no_coarse2D) {
				Nxf = Nx;
				Nyf = Ny;

				for (int j = 0; j < Nyf; ++j) {
					for (int i = 0; i < Nxf; ++i) {
						data_rectangles[id][i + (Nxf + 1) * j] = v.Give_data<rectangleEl>(id,i,j,Nx);
					}
				}
			}
		}
		Update<rectangleEl>(id);
	}

	// 2. edges
	int num_pos;

	double weight;
	for (int id = 0; id < ug->Give_number_edges(); ++id) {

		v.Update < edgeEl > (id);

		if (ug->Give_edge(id)->my_object(my_rank)) {

			// Update edge
			Nx = coarse_blockgrid->Give_Nx_edge(id);
			//num_pos = ug->Give_edge ( id )->Give_number_neighbors();

			coarse_edge_exterior = mgop->Give_coarse_edge_exterior(id);
			coarse_edge_interior = mgop->Give_coarse_edge_interior(id);

			if (coarse_edge_interior || coarse_edge_exterior) { // coarse edge
				Nxf = Nx * 2;
				for (int i = 1; i < Nxf; ++i) {
					data_edges[id][i] = 0;
				}

				data_edges[id][1] = 0.5 * v.Give_data<edgeEl>(id, 0,0,0);
				for (int i = 1; i < Nx; ++i) {
					data_edges[id][2*i-1] += 0.5 * v.Give_data<edgeEl>(id, i,0,0);
					data_edges[id][2*i  ] += 1.0 * v.Give_data<edgeEl>(id, i,0,0);
					data_edges[id][2*i+1] += 0.5 * v.Give_data<edgeEl>(id, i,0,0);
				}
				data_edges[id][Nxf-1] += 0.5 * v.Give_data<edgeEl>(id, Nx,0,0);
			} else { // no coarse edge
				for (int i = 1; i < Nx; i = i + 1) {
					data_edges[id][i] = v.Give_data<edgeEl>(id,i,0,0);
				}
			}
		}
		Update<edgeEl>(id);
	}
	// 3. points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) )
		data_points[id][0] = v.Give_data<pointEl> ( id, 0 , 0 , 0);
		Update<pointEl>(id);
	}
}


#endif // MGOP_H
