
/**********************************************************************************
 * Copyright 2020 Christoph Pflaum
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/



template<class DTyp, class Vari, class Mat>
GmResSystem<DTyp,Vari,Mat>::GmResSystem(VariableVector<DTyp,Vari>* res) : IterativeSolverSystemAbstract<DTyp,Vari,Mat>(res) {
   Restart = 100;
   
   Q.resize(Restart+1);
   
   H = new DTyp*[Restart+1];
   C = new DTyp[Restart];
   S = new DTyp[Restart];
   y = new DTyp[Restart];
   gamma = new DTyp[Restart+1];
   
   for(int i=0;i<Restart+1;++i) {
       H[i] = new DTyp[Restart];
       if(this->size==1) {
	  Vari* v = new Vari(*this->rXcomponent);
	  Q.at(i) = new VariableVector<DTyp,Vari>(*v);
       }
       if(this->size==2) {
	  Vari* vx = new Vari(*this->rXcomponent);
	  Vari* vy = new Vari(*this->rXcomponent);	 
	  Q.at(i) = new VariableVector<DTyp,Vari>(*vx,*vy);
       }
       if(this->size==3) {
	  Vari* vx = new Vari(*this->rXcomponent);
	  Vari* vy = new Vari(*this->rXcomponent); 
	  Vari* vz = new Vari(*this->rXcomponent);	  
	  Q.at(i) = new VariableVector<DTyp,Vari>(*vx,*vy,*vz);
       }
   }
}

template<class DTyp, class Vari, class Mat>
GmResSystem<DTyp,Vari,Mat>::~GmResSystem() {
   for(int i=0;i<Restart+1;++i) {
       delete[] H[i];
       Q.at(i)->deleteVariables();
       delete Q.at(i);
   }
   delete H;
   delete C;
   delete S;
   delete y;
   delete gamma;
} 


template<class DTyp, class Vari, class Mat>
void GmResSystem<DTyp,Vari,Mat>::setRestartNumber(int numberRestart_) {
  int oldRestart = Restart;
  Restart = numberRestart_;
  if(oldRestart==Restart) return;

  for(int i=0;i<oldRestart+1;++i) delete H[i];
  delete H;
  delete C;
  delete S;
  delete y;
  delete gamma;
    
   H = new DTyp*[Restart+1];
   C = new DTyp[Restart];
   S = new DTyp[Restart];
   y = new DTyp[Restart];
   gamma = new DTyp[Restart+1];
   
   for(int i=0;i<Restart+1;++i) {
       H[i] = new DTyp[Restart];
   }
  
  
  if(oldRestart<Restart) {
     Q.resize(Restart+1);
     for(int i=oldRestart+1;i<Restart+1;++i) {
         //Q.at(i) = new Variable2D<DTyp>(*(this->blockgrid));
         if(this->size==1) {
	    Vari* v = new Vari(*this->rXcomponent);
	    Q.at(i) = new VariableVector<DTyp,Vari>(*v);
         }
         if(this->size==2) {
	    Vari* vx = new Vari(*this->rXcomponent);
	    Vari* vy = new Vari(*this->rXcomponent);	 
	    Q.at(i) = new VariableVector<DTyp,Vari>(*vx,*vy);
         }
         if(this->size==3) {
	    Vari* vx = new Vari(*this->rXcomponent);
	    Vari* vy = new Vari(*this->rXcomponent); 
	    Vari* vz = new Vari(*this->rXcomponent);	  
	    Q.at(i) = new VariableVector<DTyp,Vari>(*vx,*vy,*vz);
         }	 
     }
  }
  else {
     for(int i=Restart+1;i<oldRestart+1;++i) {
         Q.at(i)->deleteVariables();
         delete Q.at(i);
     }
     Q.resize(Restart+1);
  }
}



template<class DTyp, class Vari, class Mat>
inline void GmResSystem<DTyp, Vari, Mat>::GivensRotation(const complex<double>& x1, complex<double>& x2, 
						    complex<double>& c, complex<double>& s)  {
    complex<double> rho = x1 / abs(x1) * sqrt(x1.real() * x1.real() + x1.imag() * x1.imag() + x2.real() * x2.real() + x2.imag() * x2.imag());
    c = x1 / rho;
    s = std::conj(x2 / rho);
  }

template<class DTyp, class Vari, class Mat>
inline void GmResSystem<DTyp, Vari, Mat>::GivensRotation(const double& x1, double& x2, double& c, double& s)  {
    double rho = x1 / fabs(x1) * sqrt(x1 * x1 + x2 * x2);
    c = x1 / rho;
    s = x2 / rho;
  }

  



template<class DTyp, class Vari, class Mat>
void GmResSystem<DTyp,Vari,Mat>::solve(Mat& A, VariableVector<DTyp,Vari>& b, VariableVector<DTyp,Vari>& x) {
    this->numberIterationsPerformed = 0;
      
//test
//cout << " max_iter: " << this->max_iter << endl;    
    
    double r_abs;
    do {
      (*this->r) = 0.0;
      A.apply(x,*this->r);
      (*this->r) = b - (*this->r);      //  r = (b - A(x));
     
      r_abs = L2Norm(*this->r);
      if(r_abs < 1e-12) break;

      gamma[0] = r_abs;
      *Q[0] = (*this->r) / gamma[0];


      
      for(int j = 1; j <= Restart; ++j) {
        this->numberIterationsPerformed++;
        *Q[j] = 0.0;
	A.apply(*Q[j-1] , *Q[j]);  // *Q[j] = (A(*Q[j-1]));

        for (int i = 0; i < j; ++i) {
          H[i][j - 1] = product(*Q[j], *Q[i]);
          *Q[j] = *Q[j] - H[i][j - 1] * *Q[i];
        }

        //H[j][j - 1] = L2NORM(*Q[j]);  // old
        H[j][j - 1] = L2Norm(*Q[j]);    
        *Q[j] = *Q[j] / H[j][j - 1];

        for (int i = 1; i < j; ++i) {
          const DTyp h_tmp = H[i - 1][j - 1];
          H[i - 1][j - 1] = (C[i - 1]) * h_tmp + S[i - 1] * H[i][j - 1];
          H[i][j - 1] = -std::conj(S[i - 1]) * h_tmp + C[i - 1] * H[i][j - 1];
        }
      
        GivensRotation(H[j - 1][j - 1], H[j][j - 1], C[j - 1], S[j - 1]);

        H[j - 1][j - 1] = (C[j - 1]) * H[j - 1][j - 1] + S[j - 1] * H[j][j - 1];

        gamma[j] = -std::conj(S[j - 1]) * gamma[j - 1];
        gamma[j - 1] = (C[j - 1]) * gamma[j - 1];
        r_abs = abs(gamma[j]);

        if (j == Restart || r_abs < this->eps || this->numberIterationsPerformed >= this->max_iter) {
          for (int i = j - 1; i >= 0; i--) {
            DTyp tmp_sum = gamma[i];
            for (int l = i + 1; l < j; ++l) {
              tmp_sum -= H[i][l] * y[l];

            }
            y[i] = tmp_sum / H[i][i];

            x = x + y[i] * *Q[i];
          }
          if (r_abs < this->eps || this->numberIterationsPerformed >= this->max_iter) {
	    this->residuum = r_abs;
	    return;
	  }
        }
      }
      
    } while (r_abs > this->eps && this->numberIterationsPerformed < this->max_iter);
    this->residuum = r_abs;
    return;
}


  
  
  
  