/**********************************************************************************
 * Copyright 2020 Christoph Pflaum
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#ifndef ITERATIVESOLVERSYSTEM_H
#define ITERATIVESOLVERSYSTEM_H


/**
 Abstrake Klasse zur Beschreibung eines iterativen Lösers zur Lösung von \f$ A x = b \f$
 **/
template<class DTyp, class Vari, class Mat>
class IterativeSolverSystemAbstract  {
   public:
        /** 
	 * @param res
	 **/
        IterativeSolverSystemAbstract(VariableVector<DTyp,Vari>* res)  {  
	       eps = 0.01;    stopByEps = true;    max_iter = 50;
	       numberIterationsPerformed = 0;
	       r    = res;
	       rXcomponent = r->getComponent(0);
	       size = r->getSize();
	}
	virtual ~IterativeSolverSystemAbstract() {
	  //test NOW
//cout << " ~IterativeSolverAbstract() " << endl;
	  
	}
	 
	/**
	 * @param stopByEps falls true, endet die iteration durch Abbruchkriterium mit eps
	 * @param eps       Abbruchkriterium
	 * @param max_iter  maximale Anzahl der Iterationen
	 * */
	void setBasicParameter(double eps_, bool stopByEps_ = true, int max_iter_ = 500) { eps = eps_; stopByEps = stopByEps_; max_iter = max_iter_; }
		
	/**
	 * solves the linear equation system \f$ A x = b \f$
	 * class Mat must contain member function   Mat::apply(x,b)
	 **/
	virtual void solve(Mat& A, VariableVector<DTyp,Vari>& b, VariableVector<DTyp,Vari>& x) = 0;
	
	
	/**
	 * @return number of iterations which were performed
	 **/
	int numberOfPerformedIterations() { return numberIterationsPerformed; }
	
	/**
	 * @return residuum after iteration. This is not calculated by every iteration method
	 **/
	double getResiduum() { return residuum; }
	
         /**
	 * @return true when iteration method calculates residuum
	 **/
	virtual bool isResiduumCalculated() = 0;
	
protected:
     double eps;
     bool stopByEps;
     int max_iter;
     
     int numberIterationsPerformed;
     double residuum;
     
     VariableVector<DTyp,Vari>* r;
     Vari* rXcomponent;     
     int size;
};

#include "gmres/gmresSystem.h"
#include "gmres/gmresSystem_cc.h"

#endif //ITERATIVESOLVERSYSTEM_H
