/**********************************************************************************
 * Copyright 2020 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#ifndef EXTEMPSYSTEM_H
#define EXTEMPSYSTEM_H



template <class A> struct ExprSystem {
	inline operator const A&() const {
		return *static_cast<const A*> ( this );
	}
};

//----------------------------------------------------------------------------
template <class A, class B>
class AddSystem : public ExprSystem<AddSystem<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline AddSystem ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
						
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) +
			       b_.template Give_data<numComponent> (i);
		}		
};

//----------------------------------------------------------------------------
template <class A, class B>
inline AddSystem<A, B> operator+ ( const ExprSystem<A>& a, const ExprSystem<B>& b ) {
	return AddSystem<A, B> ( a, b );
}




//----------------------------------------------------------------------------
template <class A, typename TYPE>
class CAddSystem : public ExprSystem<CAddSystem<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline CAddSystem ( const A& a, TYPE b ) : a_ ( a ), b_ ( b ) {}
		
		typedef _TypeOf1_ ( A, TYPE ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) + b_;
		}
};

//----------------------------------------------------------------------------
template <class A>
inline CAddSystem<A, double>
operator+ ( const ExprSystem<A>& a, double b ) {
	return CAddSystem<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddSystem<A, double>
operator+ ( double b, const ExprSystem<A>& a ) {
	return CAddSystem<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddSystem<A, double>
operator- ( const ExprSystem<A>& a, double b ) {
	return CAddSystem<A, double> ( a, -b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddSystem<A, std::complex<double> >
operator+ ( const ExprSystem<A>& a, std::complex<double>  b ) {
	return CAddSystem<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddSystem<A, std::complex<double> >
operator+ ( std::complex<double>  b, const ExprSystem<A>& a ) {
	return CAddSystem<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAddSystem<A, std::complex<double> >
operator- ( const ExprSystem<A>& a, std::complex<double>  b ) {
	return CAddSystem<A, std::complex<double> > ( a, -b );
}


//----------------------------------------------------------------------------
template <class A, class B>

class SubSystem : public ExprSystem<SubSystem<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline SubSystem ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) -
			       b_.template Give_data<numComponent> (i);
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline SubSystem<A, B> operator- ( const ExprSystem<A>& a, const ExprSystem<B>& b ) {
	return SubSystem<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <typename TYPE, class B>

class CSubSystem : public ExprSystem<CSubSystem<TYPE, B> > {
		TYPE a_;
		const B& b_;
		
	public:
		inline CSubSystem ( TYPE a, const B& b ) : a_ ( a ), b_ ( b ) {}
				
		typedef _TypeOf1_ ( B, TYPE ) Result;
				
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_ - b_.template Give_data<numComponent> (i);
		}		
};

//----------------------------------------------------------------------------
template <class B>
inline CSubSystem<double, B> operator- ( double a, const ExprSystem<B>& b ) {
	return CSubSystem<double, B> ( a, b );
}

template <class B>
inline CSubSystem<std::complex<double>, B> operator- ( std::complex<double> a, const ExprSystem<B>& b ) {
	return CSubSystem<std::complex<double>, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>

class MinusSystem : public ExprSystem<MinusSystem<A> > {
		const A& a_;
		
	public:
		inline MinusSystem ( const A& a ) : a_ ( a ) {}
				
		typedef typename A::Result Result;

		template <int numComponent>
		inline Result Give_data(int i) const {
			return - a_.template Give_data<numComponent> (i);
		}		
		
};

//----------------------------------------------------------------------------
template <class A>
inline MinusSystem<A> operator- ( const ExprSystem<A>& a ) {
	return MinusSystem<A> ( a );
}

//----------------------------------------------------------------------------
template <class A, typename TYPE>

class TimesSystem : public ExprSystem<TimesSystem<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline TimesSystem ( const A& a, const TYPE& b ) : a_ ( a ), b_ ( b ) {}
					
		typedef _TypeOf1_ ( A, TYPE ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) * b_;
		}		
		
		
};

//----------------------------------------------------------------------------
template <class A>
inline TimesSystem<A, double>
operator* ( const ExprSystem<A>& a, double b ) {
	return TimesSystem<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesSystem<A, double>
operator* ( double b, const ExprSystem<A>& a ) {
	return TimesSystem<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesSystem<A, std::complex<double> >
operator* ( const ExprSystem<A>& a, std::complex<double> b ) {
	return TimesSystem<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesSystem<A, std::complex<double> >
operator* ( std::complex<double> b, const ExprSystem<A>& a ) {
	return TimesSystem<A, std::complex<double> > ( a, b );
}


//----------------------------------------------------------------------------
// spezielle Konstruktion falls ein Operant bool
//----------------------------------------------------------------------------
template <class A>
class TimesBoolSystem : public ExprSystem<TimesBoolSystem<A> > {
		const A& a_;
		bool b_;
		
	public:
		inline TimesBoolSystem ( const A& a, const bool& b ) : a_ ( a ), b_ ( b ) {}
					
		typedef typename A::Result Result;
			
		template <int numComponent>
		inline Result Give_data(int i) const {
		       if(b_) return a_.template Give_data<numComponent>(i);
		       return 0.0;
		}				
		
};


template <class A>
inline TimesBoolSystem<A>
operator*( bool b, const ExprSystem<A>& a ) {
	return TimesBoolSystem<A> ( a, b );
}


template <class A>
inline TimesBoolSystem<A>
operator*( const ExprSystem<A>& a, bool b) {
	return TimesBoolSystem<A> ( a, b );
}



//----------------------------------------------------------------------------
template <typename TYPE, class B>

class DivSystem : public ExprSystem<DivSystem<TYPE, B> > {
		const TYPE a_;
		const B& b_;
		
	public:
		inline DivSystem ( const TYPE& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		typedef _TypeOf1_ ( B, TYPE ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_ / b_.template Give_data<numComponent> (i);
		}		
		
};

//----------------------------------------------------------------------------
template <class A>
inline TimesSystem<A, double> operator/ ( const ExprSystem<A>& a, double b ) {
	return TimesSystem<A, double> ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class A>
inline TimesSystem<A, std::complex<double> > operator/ ( const ExprSystem<A>& a, std::complex<double> b ) {
	return TimesSystem<A, std::complex<double> > ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class B>
inline DivSystem<double, B> operator/ ( double a, const ExprSystem<B>& b ) {
	return DivSystem<double, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class B>
inline DivSystem<std::complex<double>, B> operator/ ( std::complex<double> a, const ExprSystem<B>& b ) {
	return DivSystem<std::complex<double>, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class VTimesSystem : public ExprSystem<VTimesSystem<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VTimesSystem ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
				
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) *
			       b_.template Give_data<numComponent> (i);
		}		
		
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VTimesSystem<A, B> operator* ( const ExprSystem<A>& a, const ExprSystem<B>& b ) {
	return VTimesSystem<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class VDivSystem : public ExprSystem<VDivSystem<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VDivSystem ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}

		typedef _TypeOf2_ ( A, B ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) /
			       b_.template Give_data<numComponent> (i);
		}	
		
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VDivSystem<A, B> operator/ ( const ExprSystem<A>& a, const ExprSystem<B>& b ) {
	return VDivSystem<A, B> ( a, b );
}


/*
//----------------------------------------------------------------------------
template <class A, class B>

class SaveDivSystem : public ExprSystem<SaveDivSystem<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline SaveDivSystem ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <int numComponent>
		inline Result Give_data(int i) const {
			return a_.template Give_data<numComponent> (i) /
			       b_.template Give_data<numComponent> (i);
		       Result valueB = b_.template Give_data<numComponent> (i);
		       if(myABS(valueB) < saveDivisionSystem::eps) return 0.0;
                       return a_.template Give_data<numComponent> (i) / valueB;			  
		}
		                
   private:
     static double myABS(double a) {  return a < 0.0 ? -a : a; }
     static std::complex<double> myABS(std::complex<double> c) {  
         return (c.real() < 0.0 ? -c.real() : c.real()) + (c.imag() < 0.0 ? -c.imag() : c.imag());
     }
};

//----------------------------------------------------------------------------
template <class A, class B>
inline SaveDivSystem<A, B> saveDiv( const ExprSystem<A>& a, const ExprSystem<B>& b ) {
	return SaveDivSystem<A, B> ( a, b );
}
*/

//----------------------------------------------------------------------------
#endif // EXTEMPSYSTEM_H
