/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// systemVar.h
//
// ------------------------------------------------------------

#include <limits>
//#include <math.h>


#ifndef SYSTEMVAR_H
#define SYSTEMVAR_H

#include <extempSystem.h>

#include <complex>
using std::complex;




///////////////////////////
// 2. Variable for points
///////////////////////////


template <class DTyp, class Vari>
class VariableVector : public ExprSystem<VariableVector<DTyp, Vari> > {
	public:
		VariableVector(Vari& variable);
		VariableVector(Vari& variableX, Vari& variableY);		
		VariableVector(Vari& variableX, Vari& variableY, Vari& variableZ);		
		
		~VariableVector();

		typedef DTyp Result;		
		
		template <int numComponent>
		inline Result Give_data(int i) const { return dataTotal[numComponent][i]; }
		
		
		template <class A>
		void operator= ( const ExprSystem<A>& a );
		
		void operator= ( const VariableVector<DTyp, Vari>& v );

		void operator= ( DTyp x );
				
		Vari* getComponent(int comp) { assert(comp>=0 && comp<size); return variables[comp]; }
		
		int getSize() { return size; }
		
		void deleteVariables();
	private:
	        int size;
		DTyp** dataTotal;
		int*   totalNumberData;
		
		Vari**  variables;
};
	

/////////////////////////////////////////////
// implementations
/////////////////////////////////////////////


/*
template <class DTyp>
VariableVector<DTyp>::VariableVector(int size_) {
   size = size_; 
   assert(size<=3);
   dataAreSet = false;
   totalNumberData = new int[size];
   dataTotal = new DTyp*[size];
   for(int s=0;s<size;++s) dataTotal[s] = NULL;
   sizeSet = 0;
}
*/

template <class DTyp, class Vari>
VariableVector<DTyp, Vari>::~VariableVector() {
   delete[] totalNumberData;
   delete[] dataTotal;
   delete[] variables;
}

template <class DTyp, class Vari>
void VariableVector<DTyp, Vari>::deleteVariables() {
   for(int s=0;s<size;++s) delete variables[s];
}

template <class DTyp, class Vari>
VariableVector<DTyp, Vari>::VariableVector(Vari& variable) {
   size = 1; 
   totalNumberData = new int[size];
   dataTotal = new DTyp*[size];
   variables = new Vari*[size];
   
   dataTotal[0] = variable.give_startTotal();
   totalNumberData[0] = variable.getTotalNumberData();   
   variables[0] = &variable;   
   assert(dataTotal[0]!=NULL);   
}

template <class DTyp, class Vari>
VariableVector<DTyp, Vari>::VariableVector(Vari& variableX, Vari& variableY) {
   size = 2; 
   totalNumberData = new int[size];
   dataTotal = new DTyp*[size];
   variables = new Vari*[size];   
   
   dataTotal[0] = variableX.give_startTotal();
   totalNumberData[0] = variableX.getTotalNumberData();
   variables[0] = &variableX;
   assert(dataTotal[0]!=NULL);
   
   dataTotal[1] = variableY.give_startTotal();
   totalNumberData[1] = variableY.getTotalNumberData();
   variables[1] = &variableY;   
   assert(dataTotal[1]!=NULL);   
}
	
template <class DTyp, class Vari>
VariableVector<DTyp, Vari>::VariableVector(Vari& variableX, Vari& variableY, Vari& variableZ) {
   size = 3; 
   totalNumberData = new int[size];
   dataTotal = new DTyp*[size];
   variables = new Vari*[size];
   
   dataTotal[0] = variableX.give_startTotal();
   totalNumberData[0] = variableX.getTotalNumberData();
   variables[0] = &variableX;
   assert(dataTotal[0]!=NULL);
   
   dataTotal[1] = variableY.give_startTotal();
   totalNumberData[1] = variableY.getTotalNumberData();
   variables[1] = &variableY;
   assert(dataTotal[1]!=NULL);

   dataTotal[2] = variableZ.give_startTotal();
   totalNumberData[2] = variableZ.getTotalNumberData();   
   variables[2] = &variableZ;   
   assert(dataTotal[2]!=NULL);   
}	
	
	
template <class DTyp, class Vari>		
template <class A>
void VariableVector<DTyp, Vari>::operator= ( const ExprSystem<A>& a ) {
     const A& ao ( a );     
          
     
     for(int i=0;i<totalNumberData[0];++i) {
          dataTotal[0][i] = ao.template Give_data<0>(i);
     }
     
     if(size==1) return;
     for(int i=0;i<totalNumberData[1];++i) {
          dataTotal[1][i] = ao.template Give_data<1>(i);
     }     
     
     if(size==2) return;
     for(int i=0;i<totalNumberData[2];++i) {
          dataTotal[2][i] = ao.template Give_data<2>(i);
     }     
}


template <class DTyp, class Vari>		
void VariableVector<DTyp, Vari>::operator= ( const VariableVector<DTyp, Vari>& v ) {     
     for(int i=0;i<totalNumberData[0];++i) {
          dataTotal[0][i] = v.template Give_data<0>(i);
     }
     
     if(size==1) return;
     for(int i=0;i<totalNumberData[1];++i) {
          dataTotal[1][i] = v.template Give_data<1>(i);
     }     
     
     if(size==2) return;
     for(int i=0;i<totalNumberData[2];++i) {
          dataTotal[2][i] = v.template Give_data<2>(i);
     }     
}

template <class DTyp, class Vari>		
void VariableVector<DTyp, Vari>::operator= ( DTyp x ) {     
     for(int i=0;i<totalNumberData[0];++i) {
          dataTotal[0][i] = x;
     }
     
     if(size==1) return;
     for(int i=0;i<totalNumberData[1];++i) {
          dataTotal[1][i] = x;
     }     
     
     if(size==2) return;
     for(int i=0;i<totalNumberData[2];++i) {
          dataTotal[2][i] = x;
     }     
}



///////////////////////////
// 3. Other functions, max, ...
///////////////////////////




template <class DTyp, class Vari>
DTyp product ( VariableVector<DTyp, Vari>& a, VariableVector<DTyp, Vari>& b ) {
     DTyp sum = product(*(a.getComponent(0)),*(b.getComponent(0)));
     int size = a.getSize();
     if(size==1) return sum;
     
     sum = sum + product(*(a.getComponent(1)),*(b.getComponent(1)));
     if(size==2) return sum;

     sum = sum + product(*(a.getComponent(2)),*(b.getComponent(2)));
     return sum;     
}

template <class DTyp, class Vari>
double L2Norm( VariableVector<DTyp, Vari>& a) {
     double norm = L2Norm(*(a.getComponent(0)));
     int size = a.getSize();
     if(size==1) return norm;

     double sum = norm*norm;
     
     norm = L2Norm(*(a.getComponent(1)));     
     sum = sum + norm*norm;
     if(size==2) return sqrt(sum);
     
     norm = L2Norm(*(a.getComponent(2)));
     sum = sum + norm*norm;
     return sqrt(sum); 
}

template <class DTyp, class Vari>
double L_infty ( VariableVector<DTyp, Vari>& a) {
     double infty = L_infty(*(a.getComponent(0)));
     int size = a.getSize();
     if(size==1) return infty;
     
     double infty_new = L_infty(*(a.getComponent(1)));
     if(infty_new > infty) infty = infty_new;
     if(size==2) return infty;

     infty_new = L_infty(*(a.getComponent(2)));
     if(infty_new > infty) infty = infty_new;
     return infty;     
}

template <class DTyp, class Vari>
double Maximum ( VariableVector<double, Vari>& a) {
     DTyp maxi = Maximum(*(a.getComponent(0)));
     int size = a.getSize();
     if(size==1) return maxi;
     
     DTyp maxi_new = Maximum(*(a.getComponent(1)));
     if(maxi_new > maxi) maxi = maxi_new;
     if(size==2) return maxi;

     maxi_new = Maximum(*(a.getComponent(2)));
     if(maxi_new > maxi) maxi = maxi_new;
     return maxi;     
}


template <class DTyp, class Vari>
double Minimum ( VariableVector<double, Vari>& a) {
     DTyp mini = Minimum(*(a.getComponent(0)));
     int size = a.getSize();
     if(size==1) return mini;
     
     DTyp mini_new = Minimum(*(a.getComponent(1)));
     if(mini_new < mini) mini = mini_new;
     if(size==2) return mini;

     mini_new = Minimum(*(a.getComponent(2)));
     if(mini_new < mini) mini = mini_new;
     return mini;     
}



#endif // SYSTEMVAR_H
